# This file is generated by rebuild_raw_files.py
from base64 import b64decode
from dataclasses import dataclass
import gzip


@dataclass
class RawFiles:
    metadata: dict
    scripts: dict
    styles: dict
    deploy: dict

RAW_FILES = {}
RAW_FILES['global'] = RawFiles({}, {}, {}, {})
RAW_FILES['global'].scripts['global.js'] = "let snippets = document.getElementsByClassName('copyable');\nconst buttonText = \"\u00f0\u0178\u201c\u2039\";\nconsole.log(snippets);\nfor (let i = 0; i < snippets.length; i++) {\n    code = snippets[i].textContent;\n    //snippets[i].classList.add('hljs'); // append copy button to pre tag\n    snippets[i].innerHTML = '<button class=\"copy-button\">'+buttonText+'</button>' + snippets[i].innerHTML; // append copy button\n    snippets[i].getElementsByClassName(\"copy-button\")[0].addEventListener(\"click\", function () {\n        this.innerText = 'Copying..';\n        navigator.clipboard.writeText(code);\n        this.innerText = 'Copied!';\n        button = this;\n        setTimeout(function () {\n            button.innerText = buttonText;\n        }, 1000)\n    });\n}\n\nlet expandables = document.getElementsByClassName('expandable');\n// Any span with the expandable class will be turned into \"...\", and can be clicked\n// to expand the rest of the content.\nfor (let i = 0; i < expandables.length; i++) {\n    let expandable = expandables[i];\n    let content = expandable.textContent;\n    if (content.length > 100) {\n        expandable.textContent = content.slice(0, 100) + '...';\n        expandable.style.cursor = 'pointer';\n        expandable.addEventListener('click', function () {\n            if (expandable.textContent.endsWith('...')) {\n                expandable.textContent = content;\n            } else {\n                expandable.textContent = content.slice(0, 100) + '...';\n            }\n        });\n    }\n}\n\nfunction navigateToAbout() {\n    // if Sk.bottle is available, we'll use that:\n    if (window.Sk && window.Sk.bottle) {\n        window.Sk.bottle.changeLocation('/--about');\n    } else {\n        window.location.href = '--about';\n    }\n}\n\nwindow.stopInformationPageListener = function() {\n    throw new Error(\"Information page listener not initialized yet.\");\n};\ndocument.addEventListener('DOMContentLoaded', () => {\n    let lastPressTime = 0;\n    const doublePressDelay = 600; // milliseconds allowed between presses\n\n    function doubleIHandler(event) {\n        // Check for Ctrl+I (case-insensitive)\n        if (event.ctrlKey && (event.key === 'i' || event.key === 'I')) {\n            const now = Date.now();\n\n            if (now - lastPressTime <= doublePressDelay) {\n                // Detected a double press within the allowed time\n                navigateToAbout();\n            }\n\n            // Update the last press time\n            lastPressTime = now;\n\n            // Prevent default browser behavior if needed (e.g., to stop opening dev tools)\n            event.preventDefault();\n        }\n    }\n\n    document.addEventListener('keydown', doubleIHandler);\n    window.stopInformationPageListener = () => {\n        document.removeEventListener('keydown', doubleIHandler);\n    };\n\n});"
RAW_FILES['global'].styles['diff.css'] = "table.diff {\n    font-family:Courier,serif;\n    border:medium;\n    margin-bottom: 4em;\n    width: 100%;\n}\n.diff_header {background-color:#e0e0e0}\ntd.diff_header { text-align:right; }\ntable.diff td {\n    padding: 0px;\n    border-bottom: 0px solid black;\n}\n.diff_next {background-color:#c0c0c0}\n.diff_add {background-color:#aaffaa}\n.diff_chg {background-color:#ffff77}\n.diff_sub {background-color:#ffaaaa}"
RAW_FILES['global'].styles['global.css'] = "div.btlw-debug .copy-button {\n     float: right;\n     cursor: pointer;\n     background-color: white;\n     padding: 4px;\n }\n\ndiv.btlw-container {\n    padding: 1em;\n    border: 1px solid lightgrey;\n}\n\ndiv.btlw-header {\n    border: 1px solid lightgrey;\n    border-bottom: 0px;\n    background-color: #EEE;\n    padding-left: 1em;\n    font-weight: bold;\n}\n\ndiv.btlw-container img {\n    display: block;\n}\n\ndiv.btlw-header .btlw-reset,div.btlw-header .btlw-about {\n    float: right;\n    cursor: pointer;\n    background-color: white;\n    padding: 4px;\n    text-decoration: none;\n    color: black;\n}"
RAW_FILES['global'].deploy['drafter-setup.js'] = "const MAIN_FILENAME = \"main\";\n\nif (typeof Sk.environ == \"undefined\") {\n    Sk.environ = new Sk.builtin.dict();\n}\nSk.environ.set$item(new Sk.builtin.str(\"DRAFTER_SKULPT\"), Sk.builtin.bool.true$);\n\nfunction builtinRead(x) {\n    if (\n        Sk.builtinFiles === undefined ||\n        Sk.builtinFiles[\"files\"][x] === undefined\n    )\n        throw \"File not found: '\" + x + \"'\";\n    return Sk.builtinFiles[\"files\"][x];\n}\n// (Sk.TurtleGraphics || (Sk.TurtleGraphics = {})).target = \"mycanvas\";\nSk.BottleSiteTarget = \"#website\";\n\nSk.configure({ read: builtinRead, __future__: Sk.python3 });\n\nSk.inBrowser = false;\n\nif (typeof Sk.console === \"undefined\") {\n    Sk.console = {};\n}\n\nSk.console = {\n    // TODO: Move handleError into this object, and make it so that drafter is a function that creates the object\n    drafter: {},\n    printPILImage: function (img) {\n        document.body.append(img.image);\n    },\n    plot: function (chart) {\n        let container = document.createElement(\"div\");\n        document.body.append(container);\n        return {\"html\": [container]};\n    },\n    getWidth: function() {\n        return 300;\n    },\n    getHeight: function() {\n        return 300;\n    }\n};\n\n$.ajaxSetup({ cache: false });\n\nconst preStyle = `background-color: #f0f0f0; padding: 4px; border: 1px solid lightgrey; margin: 0px`;\n\nfunction startWebserver(pythonSite) {\n    Sk.console.drafter.handleError = function (code, message) {\n        document.body.innerHTML = `<h1>Error Running Site!</h1><div>There was an error running your site. Here is the error message:</div><div><pre style=\"${preStyle}\">${code}: ${message}</pre></div>`;\n    };\n    try {\n        Sk.misceval\n            .asyncToPromise(() =>\n                Sk.importMainWithBody(\n                    MAIN_FILENAME,\n                    false,\n                    pythonSite,\n                    true\n                )\n            )\n            .then((result) => {\n                console.log(result.$d);\n                handleRedirectNavigation();\n            })\n            .catch((e) => {\n                showError(e, MAIN_FILENAME+\".py\", pythonSite);\n            });\n    } catch (e) {\n        showError(e, MAIN_FILENAME+\".py\", pythonSite);\n    }\n}\n\nfunction handleRedirectNavigation() {\n    console.log(\"Handling redirect navigation if needed\");\n    const pathMatch = window.location.search.match(/\\?(\\/.+)/);\n    if (pathMatch) {\n        // Navigate to the target path using the Bottle mechanism\n        const target = pathMatch[1].replace(/~and~/g, '&');\n        Sk.bottle.changeLocation(target);\n    }\n}\n\nfunction showError(err, filenameExecuted, code) {\n    console.error(err);\n    console.error(err.args.v[0].v);\n    document.body.innerHTML = [\n        \"<h1>Error Running Site!</h1><div>There was an error running your site. Here is the error message:</div><div>\",\n        presentRunError(err, filenameExecuted, code),\n        \"</div>\"].join(\"\\n\");\n}\n\nfunction presentRunError(error, filenameExecuted, code) {\n    let label = error.tp$name;\n    let category = \"runtime\";\n    let message = convertSkulptError(error, filenameExecuted, code);\n\n    return message;\n\n    /*let linesError = [];\n    if (lineno !== undefined && lineno !== null) {\n        linesError.push(lineno);\n    }*/\n}\n\nfunction buildTraceback(error, filenameExecuted, code) {\n    return error.traceback.map(frame => {\n        if (!frame) {\n            return \"??\";\n        }\n        let lineno = frame.lineno;\n        let file = `File <code class=\"filename\">\"${frame.filename}\"</code>, `;\n        let line = `on line <code class=\"lineno\">${lineno}</code>, `;\n        let scope = (frame.scope !== \"<module>\" &&\n        frame.scope !== undefined) ? `in scope ${frame.scope}` : \"\";\n        let source = \"\";\n        console.log(filenameExecuted, frame.filename, code);\n        if (frame.source !== undefined) {\n            source = `\\n<pre style=\"${preStyle}\"><code>${frame.source}</code></pre>`;\n        } else if (filenameExecuted === frame.filename && code) {\n            const lines = code.split(\"\\n\");\n            const lineIndex = lineno - 1;\n            const lineContent = lines[lineIndex];\n            source = `\\n<pre style=\"${preStyle}\"><code>${lineContent}</code></pre>`;\n        }\n        return file + line + scope + source;\n    });\n}\n\nfunction convertSkulptError(error, filenameExecuted, code) {\n    let name = error.tp$name;\n    let args = Sk.ffi.remapToJs(error.args);\n    let top = `${name}: ${args[0]}`;\n    let traceback = \"\";\n    if (name === \"TimeoutError\") {\n        if (error.err && error.err.traceback && error.err.traceback.length) {\n            const allFrames = buildTraceback(error.err, filenameExecuted, code);\n            const result = [\"Traceback:\"];\n            if (allFrames.length > 5) {\n                result.push(...allFrames.slice(0, 3),\n                            `... Hiding ${allFrames.length - 3} other stack frames ...,`,\n                            ...allFrames.slice(-3, -2));\n            } else {\n                result.push(...allFrames);\n            }\n            traceback = result.join(\"\\n<br>\");\n        }\n    } else if (error.traceback && error.traceback.length) {\n        traceback = \"<strong>Traceback:</strong><br>\\n\" + buildTraceback(error, filenameExecuted, code).join(\"\\n<br>\");\n    }\n    return `<pre style=\"${preStyle}\">${top}</pre>\\n<br>\\n${traceback}\\n<br>\\n<div>\n    <p><strong>Advice:</strong><br>\n    Some common things to check:\n    <ul>\n    <li>Check your site to make sure it has no errors and runs fine when not deployed.</li>\n    <li>Make sure you are not using third party libraries or modules that are not supported (e.g., <code>threading</code>).</li>\n    <li>Check that you are correctly referencing any files or images you are using.</li>\n    </ul>\n    </p>\n    </div>`;\n}\n\n$(document).ready(function() {\n\tif (\n        Sk.builtinFiles !== undefined &&\n        Sk.builtinFiles[\"files\"][\"main.py\"] !== undefined\n    ) {\n\t\tstartWebserver(Sk.builtinFiles[\"files\"][\"main.py\"]);\n    } else if (document.querySelector(\"iframe\")) {\n        let iframe = document.getElementsByTagName(\"iframe\")[0];\n        iframe.onload = ev => {\n            let code = iframe.contentWindow.document.querySelector(\"pre\").textContent;\n            startWebserver(code);\n        };\n    } else {\n        $.ajax({\n            type: \"GET\",\n            url: \"website.py\",\n            success: function(contents) {\n                startWebserver(contents);\n            }\n        });\n    }\n});\n"
RAW_FILES['global'].deploy['global.js'] = "let snippets = document.getElementsByClassName('copyable');\nconst buttonText = \"\u00f0\u0178\u201c\u2039\";\nconsole.log(snippets);\nfor (let i = 0; i < snippets.length; i++) {\n    code = snippets[i].textContent;\n    //snippets[i].classList.add('hljs'); // append copy button to pre tag\n    snippets[i].innerHTML = '<button class=\"copy-button\">'+buttonText+'</button>' + snippets[i].innerHTML; // append copy button\n    snippets[i].getElementsByClassName(\"copy-button\")[0].addEventListener(\"click\", function () {\n        this.innerText = 'Copying..';\n        navigator.clipboard.writeText(code);\n        this.innerText = 'Copied!';\n        button = this;\n        setTimeout(function () {\n            button.innerText = buttonText;\n        }, 1000)\n    });\n}\n\nlet expandables = document.getElementsByClassName('expandable');\n// Any span with the expandable class will be turned into \"...\", and can be clicked\n// to expand the rest of the content.\nfor (let i = 0; i < expandables.length; i++) {\n    let expandable = expandables[i];\n    let content = expandable.textContent;\n    if (content.length > 100) {\n        expandable.textContent = content.slice(0, 100) + '...';\n        expandable.style.cursor = 'pointer';\n        expandable.addEventListener('click', function () {\n            if (expandable.textContent.endsWith('...')) {\n                expandable.textContent = content;\n            } else {\n                expandable.textContent = content.slice(0, 100) + '...';\n            }\n        });\n    }\n}\n\nfunction navigateToAbout() {\n    // if Sk.bottle is available, we'll use that:\n    if (window.Sk && window.Sk.bottle) {\n        window.Sk.bottle.changeLocation('/--about');\n    } else {\n        window.location.href = '--about';\n    }\n}\n\nwindow.stopInformationPageListener = function() {\n    throw new Error(\"Information page listener not initialized yet.\");\n};\ndocument.addEventListener('DOMContentLoaded', () => {\n    let lastPressTime = 0;\n    const doublePressDelay = 600; // milliseconds allowed between presses\n\n    function doubleIHandler(event) {\n        // Check for Ctrl+I (case-insensitive)\n        if (event.ctrlKey && (event.key === 'i' || event.key === 'I')) {\n            const now = Date.now();\n\n            if (now - lastPressTime <= doublePressDelay) {\n                // Detected a double press within the allowed time\n                navigateToAbout();\n            }\n\n            // Update the last press time\n            lastPressTime = now;\n\n            // Prevent default browser behavior if needed (e.g., to stop opening dev tools)\n            event.preventDefault();\n        }\n    }\n\n    document.addEventListener('keydown', doubleIHandler);\n    window.stopInformationPageListener = () => {\n        document.removeEventListener('keydown', doubleIHandler);\n    };\n\n});"
RAW_FILES['skeleton'] = RawFiles({"credit": "<a href='http://getskeleton.com/' target='_blank'>Skeleton</a> by Dave Gamache"}, {}, {}, {})
RAW_FILES['skeleton'].styles['skeleton.css'] = "/**\nhttps://github.com/dhg/Skeleton\nThe MIT License (MIT)\nCopyright (c) 2011-2014 Dave Gamache\nPermission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the \"Software\"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:\nThe above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.\nTHE SOFTWARE IS PROVIDED \"AS IS\", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.\n */\n\n.btlw .container {\n    position: relative;\n    width: 100%;\n    max-width: 960px;\n    margin: 0 auto;\n    padding: 0 20px;\n    box-sizing: border-box\n}\n\n.btlw .column,.btlw .columns {\n    width: 100%;\n    float: left;\n    box-sizing: border-box\n}\n\n@media (min-width: 400px) {\n    .btlw .container {\n        width:85%;\n        padding: 0\n    }\n}\n\n@media (min-width: 550px) {\n    .btlw .container {\n        width:80%\n    }\n\n    .btlw .column,.btlw .columns {\n        margin-left: 4%\n    }\n\n    .btlw .column:first-child,.btlw .columns:first-child {\n        margin-left: 0\n    }\n\n    .btlw .one.column,.btlw .one.columns {\n        width: 4.66666666667%\n    }\n\n    .btlw .two.columns {\n        width: 13.3333333333%\n    }\n\n    .btlw .three.columns {\n        width: 22%\n    }\n\n    .btlw .four.columns {\n        width: 30.6666666667%\n    }\n\n    .btlw .five.columns {\n        width: 39.3333333333%\n    }\n\n    .btlw .six.columns {\n        width: 48%\n    }\n\n    .btlw .seven.columns {\n        width: 56.6666666667%\n    }\n\n    .btlw .eight.columns {\n        width: 65.3333333333%\n    }\n\n    .btlw .nine.columns {\n        width: 74%\n    }\n\n    .btlw .ten.columns {\n        width: 82.6666666667%\n    }\n\n    .btlw .eleven.columns {\n        width: 91.3333333333%\n    }\n\n    .btlw .twelve.columns {\n        width: 100%;\n        margin-left: 0\n    }\n\n    .btlw .one-third.column {\n        width: 30.6666666667%\n    }\n\n    .btlw .two-thirds.column {\n        width: 65.3333333333%\n    }\n\n    .btlw .one-half.column {\n        width: 48%\n    }\n\n    .btlw .offset-by-one.column,.btlw .offset-by-one.columns {\n        margin-left: 8.66666666667%\n    }\n\n    .btlw .offset-by-two.column,.btlw .offset-by-two.columns {\n        margin-left: 17.3333333333%\n    }\n\n    .btlw .offset-by-three.column,.btlw .offset-by-three.columns {\n        margin-left: 26%\n    }\n\n    .btlw .offset-by-four.column,.btlw .offset-by-four.columns {\n        margin-left: 34.6666666667%\n    }\n\n    .btlw .offset-by-five.column,.btlw .offset-by-five.columns {\n        margin-left: 43.3333333333%\n    }\n\n    .btlw .offset-by-six.column,.btlw .offset-by-six.columns {\n        margin-left: 52%\n    }\n\n    .btlw .offset-by-seven.column,.btlw .offset-by-seven.columns {\n        margin-left: 60.6666666667%\n    }\n\n    .btlw .offset-by-eight.column,.btlw .offset-by-eight.columns {\n        margin-left: 69.3333333333%\n    }\n\n    .btlw .offset-by-nine.column,.btlw .offset-by-nine.columns {\n        margin-left: 78%\n    }\n\n    .btlw .offset-by-ten.column,.btlw .offset-by-ten.columns {\n        margin-left: 86.6666666667%\n    }\n\n    .btlw .offset-by-eleven.column,.btlw .offset-by-eleven.columns {\n        margin-left: 95.3333333333%\n    }\n\n    .btlw .offset-by-one-third.column,.btlw .offset-by-one-third.columns {\n        margin-left: 34.6666666667%\n    }\n\n    .btlw .offset-by-two-thirds.column,.btlw .offset-by-two-thirds.columns {\n        margin-left: 69.3333333333%\n    }\n\n    .btlw .offset-by-one-half.column,.btlw .offset-by-one-half.columns {\n        margin-left: 52%\n    }\n}\n\n.btlw  {\n    font-size: 62.5%\n}\n\n.btlw {\n    font-size: 1.5em;\n    line-height: 1.6;\n    font-weight: 400;\n    font-family: Raleway,HelveticaNeue,\"Helvetica Neue\",Helvetica,Arial,sans-serif;\n    color: #222\n}\n\n.btlw h1,.btlw h2,.btlw h3,.btlw h4,.btlw h5,.btlw h6 {\n    margin-top: 0;\n    margin-bottom: 2rem;\n    font-weight: 300\n}\n\n.btlw h1 {\n    font-size: 4rem;\n    line-height: 1.2;\n    letter-spacing: -.1rem\n}\n\n.btlw h2 {\n    font-size: 3.6rem;\n    line-height: 1.25;\n    letter-spacing: -.1rem\n}\n\n.btlw h3 {\n    font-size: 3rem;\n    line-height: 1.3;\n    letter-spacing: -.1rem\n}\n\n.btlw h4 {\n    font-size: 2.4rem;\n    line-height: 1.35;\n    letter-spacing: -.08rem\n}\n\n.btlw h5 {\n    font-size: 1.8rem;\n    line-height: 1.5;\n    letter-spacing: -.05rem\n}\n\n.btlw h6 {\n    font-size: 1.5rem;\n    line-height: 1.6;\n    letter-spacing: 0\n}\n\n@media (min-width: 550px) {\n    .btlw h1 {\n        font-size:5rem\n    }\n\n    .btlw h2 {\n        font-size: 4.2rem\n    }\n\n    .btlw h3 {\n        font-size: 3.6rem\n    }\n\n    .btlw h4 {\n        font-size: 3rem\n    }\n\n    .btlw h5 {\n        font-size: 2.4rem\n    }\n\n    .btlw h6 {\n        font-size: 1.5rem\n    }\n}\n\n.btlw p {\n    margin-top: 0\n}\n\n.btlw a {\n    color: #1EAEDB\n}\n\n.btlw a:hover {\n    color: #0FA0CE\n}\n\n.btlw .button,.btlw button,.btlw input[type=button],.btlw input[type=reset],.btlw input[type=submit] {\n    display: inline-block;\n    height: 38px;\n    padding: 0 30px;\n    color: #555;\n    text-align: center;\n    font-size: 11px;\n    font-weight: 600;\n    line-height: 38px;\n    letter-spacing: .1rem;\n    text-transform: uppercase;\n    text-decoration: none;\n    white-space: nowrap;\n    background-color: transparent;\n    border-radius: 4px;\n    border: 1px solid #bbb;\n    cursor: pointer;\n    box-sizing: border-box\n}\n\n.btlw .button:focus,.btlw .button:hover,.btlw button:focus,.btlw button:hover,.btlw input[type=button]:focus,.btlw input[type=button]:hover,.btlw input[type=reset]:focus,.btlw input[type=reset]:hover,.btlw input[type=submit]:focus,.btlw input[type=submit]:hover {\n    color: #333;\n    border-color: #888;\n    outline: 0\n}\n\n.btlw .button.button-primary,.btlw button.button-primary,.btlw input[type=button].button-primary,.btlw input[type=reset].button-primary,.btlw input[type=submit].button-primary {\n    color: #FFF;\n    background-color: #33C3F0;\n    border-color: #33C3F0\n}\n\n.btlw .button.button-primary:focus,.btlw .button.button-primary:hover,.btlw button.button-primary:focus,.btlw button.button-primary:hover,.btlw input[type=button].button-primary:focus,.btlw input[type=button].button-primary:hover,.btlw input[type=reset].button-primary:focus,.btlw input[type=reset].button-primary:hover,.btlw input[type=submit].button-primary:focus,.btlw input[type=submit].button-primary:hover {\n    color: #FFF;\n    background-color: #1EAEDB;\n    border-color: #1EAEDB\n}\n\n.btlw input[type=email],.btlw input[type=number],.btlw input[type=password],.btlw input[type=search],.btlw input[type=tel],.btlw input[type=text],.btlw input[type=url],.btlw select,.btlw textarea {\n    height: 38px;\n    padding: 6px 10px;\n    background-color: #fff;\n    border: 1px solid #D1D1D1;\n    border-radius: 4px;\n    box-shadow: none;\n    box-sizing: border-box\n}\n\n.btlw input[type=email],.btlw input[type=number],.btlw input[type=password],.btlw input[type=search],.btlw input[type=tel],.btlw input[type=text],.btlw input[type=url],.btlw textarea {\n    -webkit-appearance: none;\n    -moz-appearance: none;\n    appearance: none\n}\n\n.btlw textarea {\n    min-height: 65px;\n    padding-top: 6px;\n    padding-bottom: 6px\n}\n\n.btlw input[type=email]:focus,.btlw input[type=number]:focus,.btlw input[type=password]:focus,.btlw input[type=search]:focus,.btlw input[type=tel]:focus,.btlw input[type=text]:focus,.btlw input[type=url]:focus,.btlw select:focus,.btlw textarea:focus {\n    border: 1px solid #33C3F0;\n    outline: 0\n}\n\n.btlw label,.btlw legend {\n    display: block;\n    margin-bottom: .5rem;\n    font-weight: 600\n}\n\n.btlw fieldset {\n    padding: 0;\n    border-width: 0\n}\n\n.btlw input[type=checkbox],.btlw input[type=radio] {\n    display: inline\n}\n\n.btlw label>.label-body {\n    display: inline-block;\n    margin-left: .5rem;\n    font-weight: 400\n}\n\n.btlw ul {\n    list-style: circle inside\n}\n\n.btlw ol {\n    list-style: decimal inside\n}\n\n.btlw ol,.btlw ul {\n    padding-left: 0;\n    margin-top: 0\n}\n\n.btlw ol ol,.btlw ol ul,.btlw ul ol,.btlw ul ul {\n    margin: 1.5rem 0 1.5rem 3rem;\n    font-size: 90%\n}\n\n.btlw li {\n    margin-bottom: 1rem\n}\n\n.btlw code {\n    padding: .2rem .5rem;\n    margin: 0 .2rem;\n    font-size: 90%;\n    white-space: nowrap;\n    background: #F1F1F1;\n    border: 1px solid #E1E1E1;\n    border-radius: 4px\n}\n\n.btlw pre>code {\n    display: block;\n    padding: 1rem 1.5rem;\n    white-space: pre\n}\n\n.btlw td,.btlw th {\n    padding: 12px 15px;\n    text-align: left;\n    border-bottom: 1px solid #E1E1E1\n}\n\n.btlw td:first-child,.btlw th:first-child {\n    padding-left: 0\n}\n\n.btlw td:last-child,.btlw th:last-child {\n    padding-right: 0\n}\n\n.btlw .button,.btlw button {\n    margin-bottom: 1rem\n}\n\n.btlw fieldset,.btlw input,.btlw select,.btlw textarea {\n    margin-bottom: 1.5rem\n}\n\n.btlw blockquote,.btlw dl,.btlw figure,.btlw form,.btlw ol,.btlw p,.btlw pre,.btlw table,.btlw ul {\n    margin-bottom: 2.5rem\n}\n\n.btlw .u-full-width {\n    width: 100%;\n    box-sizing: border-box\n}\n\n.btlw .u-max-full-width {\n    max-width: 100%;\n    box-sizing: border-box\n}\n\n.btlw .u-pull-right {\n    float: right\n}\n\n.btlw .u-pull-left {\n    float: left\n}\n\n.btlw hr {\n    margin-top: 3rem;\n    margin-bottom: 3.5rem;\n    border-width: 0;\n    border-top: 1px solid #E1E1E1\n}\n\n.btlw .container:after,.btlw .row:after,.btlw .u-cf {\n    content: \"\";\n    display: table;\n    clear: both\n}\n"
RAW_FILES['98'] = RawFiles({"credit": "<a href='http://www.98.css' target='_blank'>98.css</a> by <code>jdan</code>"}, {}, {}, {})
RAW_FILES['98'].styles['98.css'] = "/*! 98.css v0.1.20 - https://github.com/jdan/98.css */@font-face{font-family:\"Pixelated MS Sans Serif\";font-style:normal;font-weight:400;src:url(ms_sans_serif.woff) format(\"woff\");src:url(ms_sans_serif.woff2) format(\"woff2\")}@font-face{font-family:\"Pixelated MS Sans Serif\";font-style:normal;font-weight:700;src:url(ms_sans_serif_bold.woff) format(\"woff\");src:url(ms_sans_serif_bold.woff2) format(\"woff2\")}body{color:#222;font-family:Arial;font-size:12px}.title-bar,.window,button,input,label,option,select,table,textarea,ul.tree-view{-webkit-font-smoothing:none;font-family:\"Pixelated MS Sans Serif\",Arial;font-size:11px}h1{font-size:5rem}h2{font-size:2.5rem}h3{font-size:2rem}h4{font-size:1.5rem}u{border-bottom:.5px solid #222;text-decoration:none}button,input[type=reset],input[type=submit]{border:none;border-radius:0;box-sizing:border-box;color:transparent;min-height:23px;min-width:75px;padding:0 12px;text-shadow:0 0 #222}.vertical-bar,button,input[type=reset],input[type=submit]{background:silver;box-shadow:inset -1px -1px #0a0a0a,inset 1px 1px #fff,inset -2px -2px grey,inset 2px 2px #dfdfdf}.vertical-bar{height:20px;width:4px}button:not(:disabled):active,input[type=reset]:not(:disabled):active,input[type=submit]:not(:disabled):active{box-shadow:inset -1px -1px #fff,inset 1px 1px #0a0a0a,inset -2px -2px #dfdfdf,inset 2px 2px grey;text-shadow:1px 1px #222}@media (not(hover)){button:not(:disabled):hover,input[type=reset]:not(:disabled):hover,input[type=submit]:not(:disabled):hover{box-shadow:inset -1px -1px #fff,inset 1px 1px #0a0a0a,inset -2px -2px #dfdfdf,inset 2px 2px grey}}button:focus,input[type=reset]:focus,input[type=submit]:focus{outline:1px dotted #000;outline-offset:-4px}button::-moz-focus-inner,input[type=reset]::-moz-focus-inner,input[type=submit]::-moz-focus-inner{border:0}:disabled,:disabled+label,input[readonly],input[readonly]+label{color:grey}:disabled+label,button:disabled,input[type=reset]:disabled,input[type=submit]:disabled{text-shadow:1px 1px 0 #fff}.window{background:silver;box-shadow:inset -1px -1px #0a0a0a,inset 1px 1px #dfdfdf,inset -2px -2px grey,inset 2px 2px #fff;padding:3px}.title-bar{align-items:center;background:linear-gradient(90deg,navy,#1084d0);display:flex;justify-content:space-between;padding:3px 2px 3px 3px}.title-bar.inactive{background:linear-gradient(90deg,grey,#b5b5b5)}.title-bar-text{color:#fff;font-weight:700;letter-spacing:0;margin-right:24px}.title-bar-controls{display:flex}.title-bar-controls button{display:block;min-height:14px;min-width:16px;padding:0}.title-bar-controls button:active{padding:0}.title-bar-controls button:focus{outline:none}.title-bar-controls button[aria-label=Minimize]{background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='6' height='2' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill='%23000' d='M0 0h6v2H0z'/%3E%3C/svg%3E\");background-position:bottom 3px left 4px;background-repeat:no-repeat}.title-bar-controls button[aria-label=Maximize]{background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='9' height='9' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M9 0H0v9h9V0zM8 2H1v6h7V2z' fill='%23000'/%3E%3C/svg%3E\");background-position:top 2px left 3px;background-repeat:no-repeat}.title-bar-controls button[aria-label=Restore]{background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='8' height='9' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill='%23000' d='M2 0h6v2H2zM7 2h1v4H7zM2 2h1v1H2zM6 5h1v1H6zM0 3h6v2H0zM5 5h1v4H5zM0 5h1v4H0zM1 8h4v1H1z'/%3E%3C/svg%3E\");background-position:top 2px left 3px;background-repeat:no-repeat}.title-bar-controls button[aria-label=Help]{background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='6' height='9' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill='%23000' d='M0 1h2v2H0zM1 0h4v1H1zM4 1h2v2H4zM3 3h2v1H3zM2 4h2v2H2zM2 7h2v2H2z'/%3E%3C/svg%3E\");background-position:top 2px left 5px;background-repeat:no-repeat}.title-bar-controls button[aria-label=Close]{background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='8' height='7' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M0 0h2v1h1v1h2V1h1V0h2v1H7v1H6v1H5v1h1v1h1v1h1v1H6V6H5V5H3v1H2v1H0V6h1V5h1V4h1V3H2V2H1V1H0V0z' fill='%23000'/%3E%3C/svg%3E\");background-position:top 3px left 4px;background-repeat:no-repeat;margin-left:2px}.status-bar{gap:1px;display:flex;margin:0 1px}.status-bar-field{box-shadow:inset -1px -1px #dfdfdf,inset 1px 1px grey;flex-grow:1;margin:0;padding:2px 3px}.window-body{margin:8px}fieldset{border-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='5' height='5' fill='gray' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M0 0h5v5H0V2h2v1h1V2H0' fill='%23fff'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M0 0h4v4H0V1h1v2h2V1H0'/%3E%3C/svg%3E\") 2;margin:0;padding:10px;padding-block-start:8px}legend{background:silver}.field-row{align-items:center;display:flex}[class^=field-row]+[class^=field-row]{margin-top:6px}.field-row>*+*{margin-left:6px}.field-row-stacked{display:flex;flex-direction:column}.field-row-stacked *+*{margin-top:6px}label{align-items:center;display:inline-flex}input[type=checkbox],input[type=radio]{appearance:none;-webkit-appearance:none;-moz-appearance:none;background:0;border:none;margin:0;opacity:0;position:fixed}input[type=checkbox]+label,input[type=radio]+label{line-height:13px}input[type=radio]+label{margin-left:18px;position:relative}input[type=radio]+label:before{background:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='12' height='12' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M8 0H4v1H2v1H1v2H0v4h1v2h1V8H1V4h1V2h2V1h4v1h2V1H8V0z' fill='gray'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M8 1H4v1H2v2H1v4h1v1h1V8H2V4h1V3h1V2h4v1h2V2H8V1z' fill='%23000'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M9 3h1v1H9V3zm1 5V4h1v4h-1zm-2 2V9h1V8h1v2H8zm-4 0v1h4v-1H4zm0 0V9H2v1h2z' fill='%23DFDFDF'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M11 2h-1v2h1v4h-1v2H8v1H4v-1H2v1h2v1h4v-1h2v-1h1V8h1V4h-1V2z' fill='%23fff'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M4 2h4v1h1v1h1v4H9v1H8v1H4V9H3V8H2V4h1V3h1V2z' fill='%23fff'/%3E%3C/svg%3E\");content:\"\";display:inline-block;height:12px;left:-18px;margin-right:6px;position:absolute;top:0;width:12px}input[type=radio]:active+label:before{background:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='12' height='12' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M8 0H4v1H2v1H1v2H0v4h1v2h1V8H1V4h1V2h2V1h4v1h2V1H8V0z' fill='gray'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M8 1H4v1H2v2H1v4h1v1h1V8H2V4h1V3h1V2h4v1h2V2H8V1z' fill='%23000'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M9 3h1v1H9V3zm1 5V4h1v4h-1zm-2 2V9h1V8h1v2H8zm-4 0v1h4v-1H4zm0 0V9H2v1h2z' fill='%23DFDFDF'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M11 2h-1v2h1v4h-1v2H8v1H4v-1H2v1h2v1h4v-1h2v-1h1V8h1V4h-1V2z' fill='%23fff'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M4 2h4v1h1v1h1v4H9v1H8v1H4V9H3V8H2V4h1V3h1V2z' fill='silver'/%3E%3C/svg%3E\")}input[type=radio]:checked+label:after{background:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='4' height='4' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M3 0H1v1H0v2h1v1h2V3h1V1H3V0z' fill='%23000'/%3E%3C/svg%3E\");content:\"\";display:block;height:4px;left:-14px;position:absolute;top:4px;width:4px}input[type=checkbox]:focus+label,input[type=radio]:focus+label{outline:1px dotted #000}input[type=radio][disabled]+label:before{background:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='12' height='12' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M8 0H4v1H2v1H1v2H0v4h1v2h1V8H1V4h1V2h2V1h4v1h2V1H8V0z' fill='gray'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M8 1H4v1H2v2H1v4h1v1h1V8H2V4h1V3h1V2h4v1h2V2H8V1z' fill='%23000'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M9 3h1v1H9V3zm1 5V4h1v4h-1zm-2 2V9h1V8h1v2H8zm-4 0v1h4v-1H4zm0 0V9H2v1h2z' fill='%23DFDFDF'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M11 2h-1v2h1v4h-1v2H8v1H4v-1H2v1h2v1h4v-1h2v-1h1V8h1V4h-1V2z' fill='%23fff'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M4 2h4v1h1v1h1v4H9v1H8v1H4V9H3V8H2V4h1V3h1V2z' fill='silver'/%3E%3C/svg%3E\")}input[type=radio][disabled]:checked+label:after{background:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='4' height='4' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M3 0H1v1H0v2h1v1h2V3h1V1H3V0z' fill='gray'/%3E%3C/svg%3E\")}input[type=checkbox]+label{margin-left:19px;position:relative}input[type=checkbox]+label:before{background:#fff;box-shadow:inset -1px -1px #fff,inset 1px 1px grey,inset -2px -2px #dfdfdf,inset 2px 2px #0a0a0a;content:\"\";display:inline-block;height:13px;left:-19px;margin-right:6px;position:absolute;width:13px}input[type=checkbox]:active+label:before{background:silver}input[type=checkbox]:checked+label:after{background:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='7' height='7' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M7 0H6v1H5v1H4v1H3v1H2V3H1V2H0v3h1v1h1v1h1V6h1V5h1V4h1V3h1V0z' fill='%23000'/%3E%3C/svg%3E\");content:\"\";display:block;height:7px;left:-16px;position:absolute;top:3px;width:7px}input[type=checkbox][disabled]+label:before{background:silver}input[type=checkbox][disabled]:checked+label:after{background:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='7' height='7' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M7 0H6v1H5v1H4v1H3v1H2V3H1V2H0v3h1v1h1v1h1V6h1V5h1V4h1V3h1V0z' fill='gray'/%3E%3C/svg%3E\")}input[type=email],input[type=number],input[type=password],input[type=text]{-webkit-appearance:none;-moz-appearance:none;appearance:none;border:none;border-radius:0}input[type=email],input[type=number],input[type=password],input[type=text],select{background-color:#fff;box-shadow:inset -1px -1px #fff,inset 1px 1px grey,inset -2px -2px #dfdfdf,inset 2px 2px #0a0a0a;box-sizing:border-box;padding:3px 4px}select,textarea{border:none}textarea{-webkit-appearance:none;-moz-appearance:none;appearance:none;background-color:#fff;border-radius:0;box-shadow:inset -1px -1px #fff,inset 1px 1px grey,inset -2px -2px #dfdfdf,inset 2px 2px #0a0a0a;box-sizing:border-box;padding:3px 4px}input[type=email],input[type=password],input[type=text],select{height:21px}input[type=number]{height:22px}input[type=email],input[type=number],input[type=password],input[type=text]{line-height:2}input[type=email]:disabled,input[type=email]:read-only,input[type=number]:disabled,input[type=number]:read-only,input[type=password]:disabled,input[type=password]:read-only,input[type=text]:disabled,input[type=text]:read-only,textarea:disabled{background-color:silver}select{appearance:none;-webkit-appearance:none;-moz-appearance:none;background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='16' height='17' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M15 0H0v16h1V1h14V0z' fill='%23DFDFDF'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M2 1H1v14h1V2h12V1H2z' fill='%23fff'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M16 17H0v-1h15V0h1v17z' fill='%23000'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M15 1h-1v14H1v1h14V1z' fill='gray'/%3E%3Cpath fill='silver' d='M2 2h12v13H2z'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M11 6H4v1h1v1h1v1h1v1h1V9h1V8h1V7h1V6z' fill='%23000'/%3E%3C/svg%3E\");background-position:top 2px right 2px;background-repeat:no-repeat;border-radius:0;padding-right:32px;position:relative}input[type=email]:focus,input[type=number]:focus,input[type=password]:focus,input[type=text]:focus,select:focus,textarea:focus{outline:none}input[type=range]{-webkit-appearance:none;background:transparent;width:100%}input[type=range]:focus{outline:none}input[type=range]::-webkit-slider-thumb{-webkit-appearance:none;background:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='11' height='21' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M0 0v16h2v2h2v2h1v-1H3v-2H1V1h9V0z' fill='%23fff'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M1 1v15h1v1h1v1h1v1h2v-1h1v-1h1v-1h1V1z' fill='%23C0C7C8'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M9 1h1v15H8v2H6v2H5v-1h2v-2h2z' fill='%2387888F'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M10 0h1v16H9v2H7v2H5v1h1v-2h2v-2h2z' fill='%23000'/%3E%3C/svg%3E\");border:none;box-shadow:none;height:21px;transform:translateY(-8px);width:11px}input[type=range].has-box-indicator::-webkit-slider-thumb{background:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='11' height='21' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M0 0v20h1V1h9V0z' fill='%23fff'/%3E%3Cpath fill='%23C0C7C8' d='M1 1h8v18H1z'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M9 1h1v19H1v-1h8z' fill='%2387888F'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M10 0h1v21H0v-1h10z' fill='%23000'/%3E%3C/svg%3E\");transform:translateY(-10px)}input[type=range]::-moz-range-thumb{background:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='11' height='21' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M0 0v16h2v2h2v2h1v-1H3v-2H1V1h9V0z' fill='%23fff'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M1 1v15h1v1h1v1h1v1h2v-1h1v-1h1v-1h1V1z' fill='%23C0C7C8'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M9 1h1v15H8v2H6v2H5v-1h2v-2h2z' fill='%2387888F'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M10 0h1v16H9v2H7v2H5v1h1v-2h2v-2h2z' fill='%23000'/%3E%3C/svg%3E\");border:0;border-radius:0;height:21px;transform:translateY(2px);width:11px}input[type=range].has-box-indicator::-moz-range-thumb{background:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='11' height='21' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M0 0v20h1V1h9V0z' fill='%23fff'/%3E%3Cpath fill='%23C0C7C8' d='M1 1h8v18H1z'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M9 1h1v19H1v-1h8z' fill='%2387888F'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M10 0h1v21H0v-1h10z' fill='%23000'/%3E%3C/svg%3E\");transform:translateY(0)}input[type=range]::-webkit-slider-runnable-track{background:#000;border-bottom:1px solid grey;border-right:1px solid grey;box-shadow:1px 0 0 #fff,1px 1px 0 #fff,0 1px 0 #fff,-1px 0 0 #a9a9a9,-1px -1px 0 #a9a9a9,0 -1px 0 #a9a9a9,-1px 1px 0 #fff,1px -1px #a9a9a9;box-sizing:border-box;height:2px;width:100%}input[type=range]::-moz-range-track{background:#000;border-bottom:1px solid grey;border-right:1px solid grey;box-shadow:1px 0 0 #fff,1px 1px 0 #fff,0 1px 0 #fff,-1px 0 0 #a9a9a9,-1px -1px 0 #a9a9a9,0 -1px 0 #a9a9a9,-1px 1px 0 #fff,1px -1px #a9a9a9;box-sizing:border-box;height:2px;width:100%}.is-vertical{display:inline-block;height:150px;transform:translateY(50%);width:4px}.is-vertical>input[type=range]{height:4px;margin:0 16px 0 10px;transform:rotate(270deg) translateX(calc(-50% + 8px));transform-origin:left;width:150px}.is-vertical>input[type=range]::-webkit-slider-runnable-track{border-bottom:1px solid grey;border-left:1px solid grey;border-right:0;box-shadow:-1px 0 0 #fff,-1px 1px 0 #fff,0 1px 0 #fff,1px 0 0 #a9a9a9,1px -1px 0 #a9a9a9,0 -1px 0 #a9a9a9,1px 1px 0 #fff,-1px -1px #a9a9a9}.is-vertical>input[type=range]::-moz-range-track{border-bottom:1px solid grey;border-left:1px solid grey;border-right:0;box-shadow:-1px 0 0 #fff,-1px 1px 0 #fff,0 1px 0 #fff,1px 0 0 #a9a9a9,1px -1px 0 #a9a9a9,0 -1px 0 #a9a9a9,1px 1px 0 #fff,-1px -1px #a9a9a9}.is-vertical>input[type=range]::-webkit-slider-thumb{transform:translateY(-8px) scaleX(-1)}.is-vertical>input[type=range].has-box-indicator::-webkit-slider-thumb{transform:translateY(-10px) scaleX(-1)}.is-vertical>input[type=range]::-moz-range-thumb{transform:translateY(2px) scaleX(-1)}.is-vertical>input[type=range].has-box-indicator::-moz-range-thumb{transform:translateY(0) scaleX(-1)}select:focus{background-color:navy;color:#fff}select:focus option{background-color:#fff;color:#000}select:active{background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='16' height='17' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M0 0h16v17H0V0zm1 16h14V1H1v15z' fill='gray'/%3E%3Cpath fill='silver' d='M1 1h14v15H1z'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M12 7H5v1h1v1h1v1h1v1h1v-1h1V9h1V8h1V7z' fill='%23000'/%3E%3C/svg%3E\")}a{color:#00f}a:focus{outline:1px dotted #00f}ul.tree-view{background:#fff;box-shadow:inset -1px -1px #fff,inset 1px 1px grey,inset -2px -2px #dfdfdf,inset 2px 2px #0a0a0a;display:block;margin:0;padding:6px}ul.tree-view li{list-style-type:none}ul.tree-view a{color:#000;text-decoration:none}ul.tree-view a:focus{background-color:navy;color:#fff}ul.tree-view li,ul.tree-view ul{margin-top:3px}ul.tree-view ul{border-left:1px dotted grey;margin-left:16px;padding-left:16px}ul.tree-view ul>li{position:relative}ul.tree-view ul>li:before{border-bottom:1px dotted grey;content:\"\";display:block;left:-16px;position:absolute;top:6px;width:12px}ul.tree-view ul>li:last-child:after{background:#fff;bottom:0;content:\"\";display:block;left:-20px;position:absolute;top:7px;width:8px}ul.tree-view details{margin-top:0}ul.tree-view details[open] summary{margin-bottom:0}ul.tree-view ul details>summary:before{margin-left:-22px;position:relative;z-index:1}ul.tree-view details>summary:before{background-color:#fff;border:1px solid grey;content:\"+\";display:block;float:left;height:9px;line-height:8px;margin-right:5px;padding-left:1px;text-align:center;width:8px}ul.tree-view details[open]>summary:before{content:\"-\"}ul.tree-view details>summary::-webkit-details-marker,ul.tree-view details>summary::marker{content:\"\"}pre{background:#fff;box-shadow:inset -1px -1px #fff,inset 1px 1px grey,inset -2px -2px #dfdfdf,inset 2px 2px #0a0a0a;display:block;margin:0;padding:12px 8px}code,code *{font-family:monospace}summary:focus{outline:1px dotted #000}::-webkit-scrollbar{width:16px}::-webkit-scrollbar:horizontal{height:17px}::-webkit-scrollbar-corner{background:#dfdfdf}::-webkit-scrollbar-track{background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='2' height='2' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M1 0H0v1h1v1h1V1H1V0z' fill='silver'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M2 0H1v1H0v1h1V1h1V0z' fill='%23fff'/%3E%3C/svg%3E\")}::-webkit-scrollbar-thumb{background-color:#dfdfdf;box-shadow:inset -1px -1px #0a0a0a,inset 1px 1px #fff,inset -2px -2px grey,inset 2px 2px #dfdfdf}::-webkit-scrollbar-button:horizontal:end:increment,::-webkit-scrollbar-button:horizontal:start:decrement,::-webkit-scrollbar-button:vertical:end:increment,::-webkit-scrollbar-button:vertical:start:decrement{display:block}::-webkit-scrollbar-button:vertical:start{background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='16' height='17' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M15 0H0v16h1V1h14V0z' fill='%23DFDFDF'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M2 1H1v14h1V2h12V1H2z' fill='%23fff'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M16 17H0v-1h15V0h1v17z' fill='%23000'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M15 1h-1v14H1v1h14V1z' fill='gray'/%3E%3Cpath fill='silver' d='M2 2h12v13H2z'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M8 6H7v1H6v1H5v1H4v1h7V9h-1V8H9V7H8V6z' fill='%23000'/%3E%3C/svg%3E\");height:17px}::-webkit-scrollbar-button:vertical:end{background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='16' height='17' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M15 0H0v16h1V1h14V0z' fill='%23DFDFDF'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M2 1H1v14h1V2h12V1H2z' fill='%23fff'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M16 17H0v-1h15V0h1v17z' fill='%23000'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M15 1h-1v14H1v1h14V1z' fill='gray'/%3E%3Cpath fill='silver' d='M2 2h12v13H2z'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M11 6H4v1h1v1h1v1h1v1h1V9h1V8h1V7h1V6z' fill='%23000'/%3E%3C/svg%3E\");height:17px}::-webkit-scrollbar-button:horizontal:start{background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='16' height='17' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M15 0H0v16h1V1h14V0z' fill='%23DFDFDF'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M2 1H1v14h1V2h12V1H2z' fill='%23fff'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M16 17H0v-1h15V0h1v17z' fill='%23000'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M15 1h-1v14H1v1h14V1z' fill='gray'/%3E%3Cpath fill='silver' d='M2 2h12v13H2z'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M9 4H8v1H7v1H6v1H5v1h1v1h1v1h1v1h1V4z' fill='%23000'/%3E%3C/svg%3E\");width:16px}::-webkit-scrollbar-button:horizontal:end{background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='16' height='17' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M15 0H0v16h1V1h14V0z' fill='%23DFDFDF'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M2 1H1v14h1V2h12V1H2z' fill='%23fff'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M16 17H0v-1h15V0h1v17z' fill='%23000'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M15 1h-1v14H1v1h14V1z' fill='gray'/%3E%3Cpath fill='silver' d='M2 2h12v13H2z'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M7 4H6v7h1v-1h1V9h1V8h1V7H9V6H8V5H7V4z' fill='%23000'/%3E%3C/svg%3E\");width:16px}.window[role=tabpanel]{position:relative;z-index:2}menu[role=tablist]{display:flex;list-style-type:none;margin:0 0 -2px;padding-left:3px;position:relative;text-indent:0}menu[role=tablist]>li{border-top-left-radius:3px;border-top-right-radius:3px;box-shadow:inset -1px 0 #0a0a0a,inset 1px 1px #dfdfdf,inset -2px 0 grey,inset 2px 2px #fff;z-index:1}menu[role=tablist]>li[aria-selected=true]{background-color:silver;margin-left:-3px;margin-top:-2px;padding-bottom:2px;position:relative;z-index:8}menu[role=tablist]>li>a{color:#222;display:block;margin:6px;text-decoration:none}menu[role=tablist]>li[aria-selected=true]>a:focus{outline:none}menu[role=tablist]>li>a:focus{outline:1px dotted #222}menu[role=tablist].multirows>li{flex-grow:1;text-align:center}.sunken-panel{border:2px groove transparent;border-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='5' height='5' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill='gray' d='M0 0h4v1H0z'/%3E%3Cpath fill='gray' d='M0 0h1v4H0z'/%3E%3Cpath fill='%230a0a0a' d='M1 1h2v1H1z'/%3E%3Cpath fill='%230a0a0a' d='M1 1h1v2H1z'/%3E%3Cpath fill='%23fff' d='M0 4h5v1H0z'/%3E%3Cpath fill='%23fff' d='M4 0h1v5H4z'/%3E%3Cpath fill='%23dfdfdf' d='M3 1h1v3H3z'/%3E%3Cpath fill='%23dfdfdf' d='M1 3h3v1H1z'/%3E%3C/svg%3E\") 2;box-sizing:border-box;overflow:auto}.sunken-panel,table{background-color:#fff}table{border-collapse:collapse;position:relative;text-align:left;white-space:nowrap}table>thead>tr>*{background:silver;box-shadow:inset -1px -1px #0a0a0a,inset 1px 1px #fff,inset -2px -2px grey,inset 2px 2px #dfdfdf;box-sizing:border-box;font-weight:400;height:17px;padding:0 6px;position:sticky;top:0}table.interactive>tbody>tr{cursor:pointer}table>tbody>tr.highlighted{background-color:navy;color:#fff}table>tbody>tr>*{height:14px;padding:0 6px}\n/*# sourceMappingURL=98.css.map */"
RAW_FILES['XP'] = RawFiles({"credit": "<a href='http://botoxparty.github.io/XP.css/' target='_blank'>XP.css</a> by <code>botoxparty</code>"}, {}, {}, {})
RAW_FILES['XP'].styles['XP.css'] = "/*! XP.css v0.2.6 - https://botoxparty.github.io/XP.css/ */@font-face{font-family:\"Pixelated MS Sans Serif\";src:url(ms_sans_serif.woff) format(\"woff\");src:url(ms_sans_serif.woff2) format(\"woff2\");font-weight:400;font-style:normal}@font-face{font-family:\"Pixelated MS Sans Serif\";src:url(ms_sans_serif_bold.woff) format(\"woff\");src:url(ms_sans_serif_bold.woff2) format(\"woff2\");font-weight:700;font-style:normal}@font-face{font-family:Perfect DOS VGA\\ 437 Win;src:url(PerfectDOSVGA437Win.woff2) format(\"woff2\"),url(PerfectDOSVGA437Win.woff) format(\"woff\");font-weight:400;font-style:normal}body{font-family:Arial;font-size:12px;color:#222}.surface{background:#ece9d8}h1{font-size:5rem}h2{font-size:2.5rem}h3{font-size:2rem}h4{font-size:1.5rem}u{text-decoration:none;border-bottom:.5px solid #222}a{color:#00f}a:focus{outline:1px dotted #00f}code,code *{font-family:monospace}pre{display:block;padding:12px 8px;background-color:#000;color:silver;font-size:1rem;margin:0}pre,pre *{font-family:Perfect DOS VGA\\ 437 Win}summary:focus{outline:1px dotted #000}::-webkit-scrollbar{width:16px}::-webkit-scrollbar:horizontal{height:17px}::-webkit-scrollbar-track{background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='2' height='2' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M1 0H0v1h1v1h1V1H1V0z' fill='silver'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M2 0H1v1H0v1h1V1h1V0z' fill='%23fff'/%3E%3C/svg%3E\")}::-webkit-scrollbar-thumb{background-color:#dfdfdf;box-shadow:inset -1px -1px #0a0a0a,inset 1px 1px #fff,inset -2px -2px grey,inset 2px 2px #dfdfdf}::-webkit-scrollbar-button:horizontal:end:increment,::-webkit-scrollbar-button:horizontal:start:decrement,::-webkit-scrollbar-button:vertical:end:increment,::-webkit-scrollbar-button:vertical:start:decrement{display:block}::-webkit-scrollbar-button:vertical:start{background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='16' height='17' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M15 0H0v16h1V1h14V0z' fill='%23DFDFDF'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M2 1H1v14h1V2h12V1H2z' fill='%23fff'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M16 17H0v-1h15V0h1v17z' fill='%23000'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M15 1h-1v14H1v1h14V1z' fill='gray'/%3E%3Cpath fill='silver' d='M2 2h12v13H2z'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M8 6H7v1H6v1H5v1H4v1h7V9h-1V8H9V7H8V6z' fill='%23000'/%3E%3C/svg%3E\")}::-webkit-scrollbar-button:vertical:end{background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='16' height='17' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M15 0H0v16h1V1h14V0z' fill='%23DFDFDF'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M2 1H1v14h1V2h12V1H2z' fill='%23fff'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M16 17H0v-1h15V0h1v17z' fill='%23000'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M15 1h-1v14H1v1h14V1z' fill='gray'/%3E%3Cpath fill='silver' d='M2 2h12v13H2z'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M11 6H4v1h1v1h1v1h1v1h1V9h1V8h1V7h1V6z' fill='%23000'/%3E%3C/svg%3E\")}::-webkit-scrollbar-button:horizontal:start{width:16px;background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='16' height='17' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M15 0H0v16h1V1h14V0z' fill='%23DFDFDF'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M2 1H1v14h1V2h12V1H2z' fill='%23fff'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M16 17H0v-1h15V0h1v17z' fill='%23000'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M15 1h-1v14H1v1h14V1z' fill='gray'/%3E%3Cpath fill='silver' d='M2 2h12v13H2z'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M9 4H8v1H7v1H6v1H5v1h1v1h1v1h1v1h1V4z' fill='%23000'/%3E%3C/svg%3E\")}::-webkit-scrollbar-button:horizontal:end{width:16px;background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='16' height='17' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M15 0H0v16h1V1h14V0z' fill='%23DFDFDF'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M2 1H1v14h1V2h12V1H2z' fill='%23fff'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M16 17H0v-1h15V0h1v17z' fill='%23000'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M15 1h-1v14H1v1h14V1z' fill='gray'/%3E%3Cpath fill='silver' d='M2 2h12v13H2z'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M7 4H6v7h1v-1h1V9h1V8h1V7H9V6H8V5H7V4z' fill='%23000'/%3E%3C/svg%3E\")}button{border:none;background:#ece9d8;box-shadow:inset -1px -1px #0a0a0a,inset 1px 1px #fff,inset -2px -2px grey,inset 2px 2px #dfdfdf;border-radius:0;min-width:75px;min-height:23px;padding:0 12px}button:not(:disabled).active,button:not(:disabled):active{box-shadow:inset -1px -1px #fff,inset 1px 1px #0a0a0a,inset -2px -2px #dfdfdf,inset 2px 2px grey}button.focused,button:focus{outline:1px dotted #000;outline-offset:-4px}label{display:inline-flex;align-items:center}input,label,textarea{font-family:\"Pixelated MS Sans Serif\",Arial;font-size:11px}textarea{padding:3px 4px;border:none;background-color:#fff;box-sizing:border-box;-webkit-appearance:none;-moz-appearance:none;appearance:none;border-radius:0}textarea:focus{outline:none}select{font-family:\"Pixelated MS Sans Serif\",Arial;font-size:11px;border:none;background-color:#fff;box-sizing:border-box;height:21px;appearance:none;-webkit-appearance:none;-moz-appearance:none;position:relative;padding:3px 32px 3px 4px;background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='16' height='17' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M15 0H0v16h1V1h14V0z' fill='%23DFDFDF'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M2 1H1v14h1V2h12V1H2z' fill='%23fff'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M16 17H0v-1h15V0h1v17z' fill='%23000'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M15 1h-1v14H1v1h14V1z' fill='gray'/%3E%3Cpath fill='silver' d='M2 2h12v13H2z'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M11 6H4v1h1v1h1v1h1v1h1V9h1V8h1V7h1V6z' fill='%23000'/%3E%3C/svg%3E\");background-position:top 2px right 2px;background-repeat:no-repeat;border-radius:0}select:focus{outline:none;color:#fff;background-color:#2267cb}select:focus option{color:#000;background-color:#fff}select:active{background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='16' height='17' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M0 0h16v17H0V0zm1 16h14V1H1v15z' fill='gray'/%3E%3Cpath fill='silver' d='M1 1h14v15H1z'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M12 7H5v1h1v1h1v1h1v1h1v-1h1V9h1V8h1V7z' fill='%23000'/%3E%3C/svg%3E\")}option{font-family:\"Pixelated MS Sans Serif\",Arial;font-size:11px}.vertical-bar{width:4px;height:20px;background:silver;box-shadow:inset -1px -1px #0a0a0a,inset 1px 1px #fff,inset -2px -2px grey,inset 2px 2px #dfdfdf}&:disabled,&:disabled+label{color:grey;text-shadow:1px 1px 0 #fff}input[type=radio]+label{line-height:13px;position:relative;margin-left:19px}input[type=radio]+label:before{content:\"\";position:absolute;top:0;left:-19px;display:inline-block;width:13px;height:13px;margin-right:6px;background:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='12' height='12' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M8 0H4v1H2v1H1v2H0v4h1v2h1V8H1V4h1V2h2V1h4v1h2V1H8V0z' fill='gray'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M8 1H4v1H2v2H1v4h1v1h1V8H2V4h1V3h1V2h4v1h2V2H8V1z' fill='%23000'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M9 3h1v1H9V3zm1 5V4h1v4h-1zm-2 2V9h1V8h1v2H8zm-4 0v1h4v-1H4zm0 0V9H2v1h2z' fill='%23DFDFDF'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M11 2h-1v2h1v4h-1v2H8v1H4v-1H2v1h2v1h4v-1h2v-1h1V8h1V4h-1V2z' fill='%23fff'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M4 2h4v1h1v1h1v4H9v1H8v1H4V9H3V8H2V4h1V3h1V2z' fill='%23fff'/%3E%3C/svg%3E\")}input[type=radio]:active+label:before{background:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='12' height='12' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M8 0H4v1H2v1H1v2H0v4h1v2h1V8H1V4h1V2h2V1h4v1h2V1H8V0z' fill='gray'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M8 1H4v1H2v2H1v4h1v1h1V8H2V4h1V3h1V2h4v1h2V2H8V1z' fill='%23000'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M9 3h1v1H9V3zm1 5V4h1v4h-1zm-2 2V9h1V8h1v2H8zm-4 0v1h4v-1H4zm0 0V9H2v1h2z' fill='%23DFDFDF'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M11 2h-1v2h1v4h-1v2H8v1H4v-1H2v1h2v1h4v-1h2v-1h1V8h1V4h-1V2z' fill='%23fff'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M4 2h4v1h1v1h1v4H9v1H8v1H4V9H3V8H2V4h1V3h1V2z' fill='silver'/%3E%3C/svg%3E\")}input[type=radio]:checked+label:after{content:\"\";display:block;width:5px;height:5px;top:5px;left:-14px;position:absolute;background:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='4' height='4' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M3 0H1v1H0v2h1v1h2V3h1V1H3V0z' fill='%23000'/%3E%3C/svg%3E\")}input[type=checkbox]{appearance:none;-webkit-appearance:none;-moz-appearance:none;margin:0;background:0;position:fixed;opacity:0;border:none}input[type=checkbox]+label{line-height:13px;position:relative;margin-left:19px}input[type=checkbox]+label:before{content:\"\";position:absolute;left:-19px;display:inline-block;width:13px;height:13px;background:#fff;box-shadow:inset -1px -1px #fff,inset 1px 1px grey,inset -2px -2px #dfdfdf,inset 2px 2px #0a0a0a;margin-right:6px}input[type=checkbox]:focus+label{outline:1px dotted #000}input[type=checkbox]:active+label:before{background:#ece9d8}input[type=checkbox]:checked+label:after{content:\"\";display:block;width:11px;height:11px;position:absolute;background:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='7' height='7' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M7 0H6v1H5v1H4v1H3v1H2V3H1V2H0v3h1v1h1v1h1V6h1V5h1V4h1V3h1V0z' fill='%23000'/%3E%3C/svg%3E\");top:1px}input[type=radio][disabled]+label:before{background:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='12' height='12' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M8 0H4v1H2v1H1v2H0v4h1v2h1V8H1V4h1V2h2V1h4v1h2V1H8V0z' fill='gray'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M8 1H4v1H2v2H1v4h1v1h1V8H2V4h1V3h1V2h4v1h2V2H8V1z' fill='%23000'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M9 3h1v1H9V3zm1 5V4h1v4h-1zm-2 2V9h1V8h1v2H8zm-4 0v1h4v-1H4zm0 0V9H2v1h2z' fill='%23DFDFDF'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M11 2h-1v2h1v4h-1v2H8v1H4v-1H2v1h2v1h4v-1h2v-1h1V8h1V4h-1V2z' fill='%23fff'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M4 2h4v1h1v1h1v4H9v1H8v1H4V9H3V8H2V4h1V3h1V2z' fill='silver'/%3E%3C/svg%3E\")}input[type=radio][disabled]:checked+label:after{background:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='4' height='4' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M3 0H1v1H0v2h1v1h2V3h1V1H3V0z' fill='gray'/%3E%3C/svg%3E\")}input[type=checkbox][disabled]+label:before{background:#ece9d8}input[type=checkbox][disabled]:checked+label:after{background:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='7' height='7' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M7 0H6v1H5v1H4v1H3v1H2V3H1V2H0v3h1v1h1v1h1V6h1V5h1V4h1V3h1V0z' fill='gray'/%3E%3C/svg%3E\")}input[type=email],input[type=password],input[type=text]{padding:3px 4px;border:1px solid #7f9db9;background-color:#fff;box-sizing:border-box;-webkit-appearance:none;-moz-appearance:none;appearance:none;border-radius:0;height:21px;line-height:2}input[type=email]:focus,input[type=password]:focus,input[type=text]:focus{outline:none}input[type=range]{-webkit-appearance:none;width:100%;background:transparent}input[type=range]:focus{outline:none}input[type=range]::-webkit-slider-thumb{-webkit-appearance:none;background:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='11' height='21' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M0 0v16h2v2h2v2h1v-1H3v-2H1V1h9V0z' fill='%23fff'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M1 1v15h1v1h1v1h1v1h2v-1h1v-1h1v-1h1V1z' fill='%23C0C7C8'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M9 1h1v15H8v2H6v2H5v-1h2v-2h2z' fill='%2387888F'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M10 0h1v16H9v2H7v2H5v1h1v-2h2v-2h2z' fill='%23000'/%3E%3C/svg%3E\")}input[type=range]::-moz-range-thumb{background:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='11' height='21' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M0 0v16h2v2h2v2h1v-1H3v-2H1V1h9V0z' fill='%23fff'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M1 1v15h1v1h1v1h1v1h2v-1h1v-1h1v-1h1V1z' fill='%23C0C7C8'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M9 1h1v15H8v2H6v2H5v-1h2v-2h2z' fill='%2387888F'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M10 0h1v16H9v2H7v2H5v1h1v-2h2v-2h2z' fill='%23000'/%3E%3C/svg%3E\")}input[type=range]::-webkit-slider-runnable-track{background:#000;border-right:1px solid grey;border-bottom:1px solid grey;box-shadow:1px 0 0 #fff,1px 1px 0 #fff,0 1px 0 #fff,-1px 0 0 #a9a9a9,-1px -1px 0 #a9a9a9,0 -1px 0 #a9a9a9,-1px 1px 0 #fff,1px -1px #a9a9a9}input[type=range]::-moz-range-track{background:#000;border-right:1px solid grey;border-bottom:1px solid grey;box-shadow:1px 0 0 #fff,1px 1px 0 #fff,0 1px 0 #fff,-1px 0 0 #a9a9a9,-1px -1px 0 #a9a9a9,0 -1px 0 #a9a9a9,-1px 1px 0 #fff,1px -1px #a9a9a9}input[type=range].has-box-indicator::-webkit-slider-thumb{background:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='11' height='21' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M0 0v20h1V1h9V0z' fill='%23fff'/%3E%3Cpath fill='%23C0C7C8' d='M1 1h8v18H1z'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M9 1h1v19H1v-1h8z' fill='%2387888F'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M10 0h1v21H0v-1h10z' fill='%23000'/%3E%3C/svg%3E\")}input[type=range].has-box-indicator::-moz-range-thumb{background:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='11' height='21' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M0 0v20h1V1h9V0z' fill='%23fff'/%3E%3Cpath fill='%23C0C7C8' d='M1 1h8v18H1z'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M9 1h1v19H1v-1h8z' fill='%2387888F'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M10 0h1v21H0v-1h10z' fill='%23000'/%3E%3C/svg%3E\")}.is-vertical{display:inline-block;width:4px;height:150px;transform:translateY(50%)}.is-vertical>input[type=range]{width:150px;height:4px;margin:0 16px 0 10px;transform-origin:left;transform:rotate(270deg) translateX(calc(-50% + 8px))}.is-vertical>input[type=range]::-webkit-slider-runnable-track{border-left:1px solid grey;border-bottom:1px solid grey;box-shadow:-1px 0 0 #fff,-1px 1px 0 #fff,0 1px 0 #fff,1px 0 0 #a9a9a9,1px -1px 0 #a9a9a9,0 -1px 0 #a9a9a9,1px 1px 0 #fff,-1px -1px #a9a9a9}.is-vertical>input[type=range]::-moz-range-track{border-left:1px solid grey;border-bottom:1px solid grey;box-shadow:-1px 0 0 #fff,-1px 1px 0 #fff,0 1px 0 #fff,1px 0 0 #a9a9a9,1px -1px 0 #a9a9a9,0 -1px 0 #a9a9a9,1px 1px 0 #fff,-1px -1px #a9a9a9}.is-vertical>input[type=range]::-webkit-slider-thumb{transform:translateY(-8px) scaleX(-1)}.is-vertical>input[type=range]::-moz-range-thumb{transform:translateY(2px) scaleX(-1)}.is-vertical>input[type=range].has-box-indicator::-webkit-slider-thumb{transform:translateY(-10px) scaleX(-1)}.is-vertical>input[type=range].has-box-indicator::-moz-range-thumb{transform:translateY(0) scaleX(-1)}.window{font-size:11px;box-shadow:inset -1px -1px #0a0a0a,inset 1px 1px #dfdfdf,inset -2px -2px grey,inset 2px 2px #fff;background:#ece9d8;padding:3px}.window fieldset{margin-bottom:9px}.title-bar{font-size:11px;background:#000;padding:3px 2px 3px 3px;display:flex;justify-content:space-between;align-items:center}.title-bar-text{font-weight:700;color:#fff;letter-spacing:0;margin-right:24px}.title-bar-controls button{padding:0;display:block;min-width:16px;min-height:14px}.title-bar-controls button:focus{outline:none}.window-body{margin:8px}.window-body pre{margin:-8px}.status-bar{margin:0 1px;display:flex;gap:1px}.status-bar-field{box-shadow:inset -1px -1px #dfdfdf,inset 1px 1px grey;flex-grow:1;padding:2px 3px;margin:0}ul.tree-view{font-family:\"Pixelated MS Sans Serif\",Arial;font-size:11px;display:block;background:#fff;padding:6px;margin:0}ul.tree-view li{list-style-type:none;margin-top:3px}ul.tree-view a{text-decoration:none;color:#000}ul.tree-view a:focus{background-color:#2267cb;color:#fff}ul.tree-view ul{margin-top:3px;margin-left:16px;padding-left:16px;border-left:1px dotted grey}ul.tree-view ul>li{position:relative}ul.tree-view ul>li:before{content:\"\";display:block;position:absolute;left:-16px;top:6px;width:12px;border-bottom:1px dotted grey}ul.tree-view ul>li:last-child:after{content:\"\";display:block;position:absolute;left:-20px;top:7px;bottom:0;width:8px;background:#fff}ul.tree-view ul details>summary:before{margin-left:-22px;position:relative;z-index:1}ul.tree-view details{margin-top:0}ul.tree-view details>summary:before{text-align:center;display:block;float:left;content:\"+\";border:1px solid grey;width:8px;height:9px;line-height:9px;margin-right:5px;padding-left:1px;background-color:#fff}ul.tree-view details[open] summary{margin-bottom:0}ul.tree-view details[open]>summary:before{content:\"-\"}fieldset{border-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='5' height='5' fill='gray' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M0 0h5v5H0V2h2v1h1V2H0' fill='%23fff'/%3E%3Cpath fill-rule='evenodd' clip-rule='evenodd' d='M0 0h4v4H0V1h1v2h2V1H0'/%3E%3C/svg%3E\") 2;padding:10px;padding-block-start:8px;margin:0}legend{background:#ece9d8}menu[role=tablist]{position:relative;margin:0 0 -2px;text-indent:0;list-style-type:none;display:flex;padding-left:3px}menu[role=tablist] button{z-index:1;display:block;color:#222;text-decoration:none;min-width:unset}menu[role=tablist] button[aria-selected=true]{padding-bottom:2px;margin-top:-2px;background-color:#ece9d8;position:relative;z-index:8;margin-left:-3px;margin-bottom:1px}menu[role=tablist] button:focus{outline:1px dotted #222;outline-offset:-4px}menu[role=tablist].justified button{flex-grow:1;text-align:center}[role=tabpanel]{padding:14px;clear:both;background:linear-gradient(180deg,#fcfcfe,#f4f3ee);border:1px solid #919b9c;position:relative;z-index:2;margin-bottom:9px}::-webkit-scrollbar{width:17px}::-webkit-scrollbar-corner{background:#dfdfdf}::-webkit-scrollbar-track:vertical{background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 -0.5 17 1' shape-rendering='crispEdges'%3E%3Cpath stroke='%23eeede5' d='M0 0h1m15 0h1'/%3E%3Cpath stroke='%23f3f1ec' d='M1 0h1'/%3E%3Cpath stroke='%23f4f1ec' d='M2 0h1'/%3E%3Cpath stroke='%23f4f3ee' d='M3 0h1'/%3E%3Cpath stroke='%23f5f4ef' d='M4 0h1'/%3E%3Cpath stroke='%23f6f5f0' d='M5 0h1'/%3E%3Cpath stroke='%23f7f7f3' d='M6 0h1'/%3E%3Cpath stroke='%23f9f8f4' d='M7 0h1'/%3E%3Cpath stroke='%23f9f9f7' d='M8 0h1'/%3E%3Cpath stroke='%23fbfbf8' d='M9 0h1'/%3E%3Cpath stroke='%23fbfbf9' d='M10 0h2'/%3E%3Cpath stroke='%23fdfdfa' d='M12 0h1'/%3E%3Cpath stroke='%23fefefb' d='M13 0h3'/%3E%3C/svg%3E\")}::-webkit-scrollbar-track:horizontal{background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 -0.5 1 17' shape-rendering='crispEdges'%3E%3Cpath stroke='%23eeede5' d='M0 0h1M0 16h1'/%3E%3Cpath stroke='%23f3f1ec' d='M0 1h1'/%3E%3Cpath stroke='%23f4f1ec' d='M0 2h1'/%3E%3Cpath stroke='%23f4f3ee' d='M0 3h1'/%3E%3Cpath stroke='%23f5f4ef' d='M0 4h1'/%3E%3Cpath stroke='%23f6f5f0' d='M0 5h1'/%3E%3Cpath stroke='%23f7f7f3' d='M0 6h1'/%3E%3Cpath stroke='%23f9f8f4' d='M0 7h1'/%3E%3Cpath stroke='%23f9f9f7' d='M0 8h1'/%3E%3Cpath stroke='%23fbfbf8' d='M0 9h1'/%3E%3Cpath stroke='%23fbfbf9' d='M0 10h1m-1 1h1'/%3E%3Cpath stroke='%23fdfdfa' d='M0 12h1'/%3E%3Cpath stroke='%23fefefb' d='M0 13h1m-1 1h1m-1 1h1'/%3E%3C/svg%3E\")}::-webkit-scrollbar-thumb{background-position:50%;background-repeat:no-repeat;background-color:#c8d6fb;background-size:7px;border:1px solid #fff;border-radius:2px;box-shadow:inset -3px 0 #bad1fc,inset 1px 1px #b7caf5}::-webkit-scrollbar-thumb:vertical{background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 -0.5 7 8' shape-rendering='crispEdges'%3E%3Cpath stroke='%23eef4fe' d='M0 0h6M0 2h6M0 4h6M0 6h6'/%3E%3Cpath stroke='%23bad1fc' d='M6 0h1M6 2h1M6 4h1'/%3E%3Cpath stroke='%23c8d6fb' d='M0 1h1M0 3h1M0 5h1M0 7h1'/%3E%3Cpath stroke='%238cb0f8' d='M1 1h6M1 3h6M1 5h6M1 7h6'/%3E%3Cpath stroke='%23bad3fc' d='M6 6h1'/%3E%3C/svg%3E\")}::-webkit-scrollbar-thumb:horizontal{background-size:8px;background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 -0.5 8 7' shape-rendering='crispEdges'%3E%3Cpath stroke='%23eef4fe' d='M0 0h1m1 0h1m1 0h1m1 0h1M0 1h1m1 0h1m1 0h1m1 0h1M0 2h1m1 0h1m1 0h1m1 0h1M0 3h1m1 0h1m1 0h1m1 0h1M0 4h1m1 0h1m1 0h1m1 0h1M0 5h1m1 0h1m1 0h1m1 0h1'/%3E%3Cpath stroke='%23c8d6fb' d='M1 0h1m1 0h1m1 0h1m1 0h1'/%3E%3Cpath stroke='%238cb0f8' d='M1 1h1m1 0h1m1 0h1m1 0h1M1 2h1m1 0h1m1 0h1m1 0h1M1 3h1m1 0h1m1 0h1m1 0h1M1 4h1m1 0h1m1 0h1m1 0h1M1 5h1m1 0h1m1 0h1m1 0h1M1 6h1m1 0h1m1 0h1m1 0h1'/%3E%3Cpath stroke='%23bad1fc' d='M0 6h1m1 0h1'/%3E%3Cpath stroke='%23bad3fc' d='M4 6h1m1 0h1'/%3E%3C/svg%3E\")}::-webkit-scrollbar-button:vertical:start{height:17px;background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 -0.5 17 17' shape-rendering='crispEdges'%3E%3Cpath stroke='%23eeede5' d='M0 0h1m15 0h1M0 1h1M0 2h1M0 3h1M0 4h1M0 5h1M0 6h1M0 7h1M0 8h1M0 9h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m15 0h1M0 16h1m15 0h1'/%3E%3Cpath stroke='%23fdfdfa' d='M1 0h1'/%3E%3Cpath stroke='%23fff' d='M2 0h14M1 1h1m13 0h1M1 2h1m13 0h1M1 3h1m13 0h1M1 4h1m13 0h1M1 5h1m13 0h1M1 6h1m13 0h1M1 7h1m13 0h1M1 8h1m13 0h1M1 9h1m13 0h1M1 10h1m13 0h1M1 11h1m13 0h1M1 12h1m13 0h1M1 13h1m13 0h1M1 14h1m13 0h1M2 15h13'/%3E%3Cpath stroke='%23e6eefc' d='M2 1h1'/%3E%3Cpath stroke='%23d0dffc' d='M3 1h1M2 2h1'/%3E%3Cpath stroke='%23cad8f9' d='M4 1h1M2 3h1'/%3E%3Cpath stroke='%23c4d2f7' d='M5 1h1'/%3E%3Cpath stroke='%23c0d0f7' d='M6 1h1'/%3E%3Cpath stroke='%23bdcef7' d='M7 1h1M2 6h1'/%3E%3Cpath stroke='%23bbcdf5' d='M8 1h1'/%3E%3Cpath stroke='%23b8cbf6' d='M9 1h1M2 7h1'/%3E%3Cpath stroke='%23b7caf5' d='M10 1h1M2 8h1'/%3E%3Cpath stroke='%23b5c8f7' d='M11 1h1'/%3E%3Cpath stroke='%23b3c7f5' d='M12 1h1'/%3E%3Cpath stroke='%23afc5f4' d='M13 1h1'/%3E%3Cpath stroke='%23dce6f9' d='M14 1h1'/%3E%3Cpath stroke='%23dfe2e1' d='M16 1h1'/%3E%3Cpath stroke='%23e1eafe' d='M3 2h1'/%3E%3Cpath stroke='%23dae6fe' d='M4 2h1M3 3h1'/%3E%3Cpath stroke='%23d4e1fc' d='M5 2h1M3 4h1'/%3E%3Cpath stroke='%23d1e0fd' d='M6 2h1M4 4h1'/%3E%3Cpath stroke='%23d0ddfc' d='M7 2h1M3 5h1'/%3E%3Cpath stroke='%23cedbfd' d='M8 2h1M6 3h1'/%3E%3Cpath stroke='%23cad9fd' d='M9 2h1M7 3h1M5 5h1'/%3E%3Cpath stroke='%23c8d8fb' d='M10 2h1'/%3E%3Cpath stroke='%23c5d6fc' d='M11 2h1m-8 8h1m1 0h1'/%3E%3Cpath stroke='%23c2d3fc' d='M12 2h1m-2 1h1m-9 7h1m0 1h1'/%3E%3Cpath stroke='%23bccefa' d='M13 2h1m-1 2h1m-9 9h2'/%3E%3Cpath stroke='%23b9c9f3' d='M14 2h1M5 14h3'/%3E%3Cpath stroke='%23cfd7dd' d='M16 2h1'/%3E%3Cpath stroke='%23d8e3fc' d='M4 3h1'/%3E%3Cpath stroke='%23d1defd' d='M5 3h1'/%3E%3Cpath stroke='%23c9d8fc' d='M8 3h1M6 4h2M5 6h2M3 7h1'/%3E%3Cpath stroke='%23c5d5fc' d='M9 3h1M3 9h1m3 0h1'/%3E%3Cpath stroke='%23c5d3fc' d='M10 3h1'/%3E%3Cpath stroke='%23bed0fc' d='M12 3h1M9 4h1m-7 7h1m0 1h1'/%3E%3Cpath stroke='%23bccdfa' d='M13 3h1'/%3E%3Cpath stroke='%23baccf4' d='M14 3h1'/%3E%3Cpath stroke='%23bdcbda' d='M16 3h1'/%3E%3Cpath stroke='%23c4d4f7' d='M2 4h1'/%3E%3Cpath stroke='%23cddbfc' d='M5 4h1M3 6h1'/%3E%3Cpath stroke='%23c8d5fb' d='M8 4h1'/%3E%3Cpath stroke='%23bbcefd' d='M10 4h3M9 5h1'/%3E%3Cpath stroke='%23bcccf3' d='M14 4h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23b1c2d5' d='M16 4h1'/%3E%3Cpath stroke='%23bed0f8' d='M2 5h1'/%3E%3Cpath stroke='%23ceddfd' d='M4 5h1'/%3E%3Cpath stroke='%23c8d6fb' d='M6 5h2M3 8h2'/%3E%3Cpath stroke='%234d6185' d='M8 5h1M7 6h3M6 7h5M5 8h3m1 0h3M4 9h3m3 0h3m-8 1h1m5 0h1'/%3E%3Cpath stroke='%23bacdfc' d='M10 5h1m1 0h2M3 12h1'/%3E%3Cpath stroke='%23b9cdfb' d='M11 5h1m-2 1h1m1 0h2m-1 1h1'/%3E%3Cpath stroke='%23a8bbd4' d='M16 5h1'/%3E%3Cpath stroke='%23cddafc' d='M4 6h1'/%3E%3Cpath stroke='%23b7cdfc' d='M11 6h1m0 1h1'/%3E%3Cpath stroke='%23a4b8d3' d='M16 6h1'/%3E%3Cpath stroke='%23cad8fd' d='M4 7h2'/%3E%3Cpath stroke='%23b6cefb' d='M11 7h1m0 1h1'/%3E%3Cpath stroke='%23bacbf4' d='M14 7h1'/%3E%3Cpath stroke='%23a0b5d3' d='M16 7h1m-1 1h1m-1 5h1'/%3E%3Cpath stroke='%23c1d3fb' d='M8 8h1'/%3E%3Cpath stroke='%23b6cdfb' d='M13 8h1m-5 5h1'/%3E%3Cpath stroke='%23b9cbf3' d='M14 8h1'/%3E%3Cpath stroke='%23b4c8f6' d='M2 9h1'/%3E%3Cpath stroke='%23c2d5fc' d='M8 9h1m-1 1h1m-3 1h2'/%3E%3Cpath stroke='%23bdd3fb' d='M9 9h1m-2 3h1'/%3E%3Cpath stroke='%23b5cdfa' d='M13 9h1'/%3E%3Cpath stroke='%23b5c9f3' d='M14 9h1'/%3E%3Cpath stroke='%239fb5d2' d='M16 9h1m-1 1h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23b1c7f6' d='M2 10h1'/%3E%3Cpath stroke='%23c3d5fd' d='M7 10h1'/%3E%3Cpath stroke='%23bad4fc' d='M9 10h1m-1 1h1'/%3E%3Cpath stroke='%23b2cffb' d='M10 10h1m1 0h1m-2 2h1'/%3E%3Cpath stroke='%23b1cbfa' d='M13 10h1'/%3E%3Cpath stroke='%23b3c8f5' d='M14 10h1m-6 4h2'/%3E%3Cpath stroke='%23adc3f6' d='M2 11h1'/%3E%3Cpath stroke='%23c3d3fd' d='M5 11h1'/%3E%3Cpath stroke='%23c1d5fb' d='M8 11h1'/%3E%3Cpath stroke='%23b7d3fc' d='M10 11h1m-2 1h1'/%3E%3Cpath stroke='%23b3d1fc' d='M11 11h1'/%3E%3Cpath stroke='%23afcefb' d='M12 11h1'/%3E%3Cpath stroke='%23aecafa' d='M13 11h1'/%3E%3Cpath stroke='%23b1c8f3' d='M14 11h1'/%3E%3Cpath stroke='%23acc2f5' d='M2 12h1'/%3E%3Cpath stroke='%23c1d2fb' d='M5 12h1'/%3E%3Cpath stroke='%23bed1fc' d='M6 12h2'/%3E%3Cpath stroke='%23b6d1fb' d='M10 12h1'/%3E%3Cpath stroke='%23afccfb' d='M12 12h1'/%3E%3Cpath stroke='%23adc9f9' d='M13 12h1m-2 1h1'/%3E%3Cpath stroke='%23b1c5f3' d='M14 12h1'/%3E%3Cpath stroke='%23aac0f3' d='M2 13h1'/%3E%3Cpath stroke='%23b7cbf9' d='M3 13h1'/%3E%3Cpath stroke='%23b9cefb' d='M4 13h1'/%3E%3Cpath stroke='%23bbcef9' d='M7 13h1'/%3E%3Cpath stroke='%23b9cffb' d='M8 13h1'/%3E%3Cpath stroke='%23b2cdfb' d='M10 13h1'/%3E%3Cpath stroke='%23b0cbf9' d='M11 13h1'/%3E%3Cpath stroke='%23aec8f7' d='M13 13h1'/%3E%3Cpath stroke='%23b0c5f2' d='M14 13h1'/%3E%3Cpath stroke='%23dbe3f8' d='M2 14h1'/%3E%3Cpath stroke='%23b7c6f1' d='M3 14h1'/%3E%3Cpath stroke='%23b8c9f2' d='M4 14h1m3 0h1'/%3E%3Cpath stroke='%23b2c8f4' d='M11 14h1'/%3E%3Cpath stroke='%23b1c6f3' d='M12 14h1'/%3E%3Cpath stroke='%23b0c4f2' d='M13 14h1'/%3E%3Cpath stroke='%23d9e3f6' d='M14 14h1'/%3E%3Cpath stroke='%23aec0d6' d='M16 14h1'/%3E%3Cpath stroke='%23c3d4e7' d='M1 15h1'/%3E%3Cpath stroke='%23aec4e5' d='M15 15h1'/%3E%3Cpath stroke='%23edf1f3' d='M1 16h1'/%3E%3Cpath stroke='%23aac0e1' d='M2 16h1'/%3E%3Cpath stroke='%2394b1d9' d='M3 16h1'/%3E%3Cpath stroke='%2388a7d8' d='M4 16h1'/%3E%3Cpath stroke='%2383a4d3' d='M5 16h1'/%3E%3Cpath stroke='%237da0d4' d='M6 16h1m3 0h3'/%3E%3Cpath stroke='%237e9fd2' d='M7 16h1'/%3E%3Cpath stroke='%237c9fd3' d='M8 16h2'/%3E%3Cpath stroke='%2382a4d6' d='M13 16h1'/%3E%3Cpath stroke='%2394b0dd' d='M14 16h1'/%3E%3Cpath stroke='%23ecf2f7' d='M15 16h1'/%3E%3C/svg%3E\")}::-webkit-scrollbar-button:vertical:end{height:17px;background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 -0.5 17 17' shape-rendering='crispEdges'%3E%3Cpath stroke='%23eeede5' d='M0 0h1m15 0h1M0 1h1M0 2h1M0 3h1M0 4h1M0 5h1M0 6h1M0 7h1M0 8h1M0 9h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m15 0h1M0 16h1m15 0h1'/%3E%3Cpath stroke='%23fdfdfa' d='M1 0h1'/%3E%3Cpath stroke='%23fff' d='M2 0h14M1 1h1m13 0h1M1 2h1m13 0h1M1 3h1m13 0h1M1 4h1m13 0h1M1 5h1m13 0h1M1 6h1m13 0h1M1 7h1m13 0h1M1 8h1m13 0h1M1 9h1m13 0h1M1 10h1m13 0h1M1 11h1m13 0h1M1 12h1m13 0h1M1 13h1m13 0h1M1 14h1m13 0h1M2 15h13'/%3E%3Cpath stroke='%23e6eefc' d='M2 1h1'/%3E%3Cpath stroke='%23d0dffc' d='M3 1h1M2 2h1'/%3E%3Cpath stroke='%23cad8f9' d='M4 1h1M2 3h1'/%3E%3Cpath stroke='%23c4d2f7' d='M5 1h1'/%3E%3Cpath stroke='%23c0d0f7' d='M6 1h1'/%3E%3Cpath stroke='%23bdcef7' d='M7 1h1M2 6h1'/%3E%3Cpath stroke='%23bbcdf5' d='M8 1h1'/%3E%3Cpath stroke='%23b8cbf6' d='M9 1h1M2 7h1'/%3E%3Cpath stroke='%23b7caf5' d='M10 1h1M2 8h1'/%3E%3Cpath stroke='%23b5c8f7' d='M11 1h1'/%3E%3Cpath stroke='%23b3c7f5' d='M12 1h1'/%3E%3Cpath stroke='%23afc5f4' d='M13 1h1'/%3E%3Cpath stroke='%23dce6f9' d='M14 1h1'/%3E%3Cpath stroke='%23dfe2e1' d='M16 1h1'/%3E%3Cpath stroke='%23e1eafe' d='M3 2h1'/%3E%3Cpath stroke='%23dae6fe' d='M4 2h1M3 3h1'/%3E%3Cpath stroke='%23d4e1fc' d='M5 2h1M3 4h1'/%3E%3Cpath stroke='%23d1e0fd' d='M6 2h1M4 4h1'/%3E%3Cpath stroke='%23d0ddfc' d='M7 2h1M3 5h1'/%3E%3Cpath stroke='%23cedbfd' d='M8 2h1M6 3h1'/%3E%3Cpath stroke='%23cad9fd' d='M9 2h1M7 3h1M5 5h1'/%3E%3Cpath stroke='%23c8d8fb' d='M10 2h1'/%3E%3Cpath stroke='%23c5d6fc' d='M11 2h1m-8 8h3'/%3E%3Cpath stroke='%23c2d3fc' d='M12 2h1m-2 1h1m-9 7h1m0 1h1'/%3E%3Cpath stroke='%23bccefa' d='M13 2h1m-1 2h1m-9 9h2'/%3E%3Cpath stroke='%23b9c9f3' d='M14 2h1M5 14h3'/%3E%3Cpath stroke='%23cfd7dd' d='M16 2h1'/%3E%3Cpath stroke='%23d8e3fc' d='M4 3h1'/%3E%3Cpath stroke='%23d1defd' d='M5 3h1'/%3E%3Cpath stroke='%23c9d8fc' d='M8 3h1M6 4h2M6 6h2M3 7h1'/%3E%3Cpath stroke='%23c5d5fc' d='M9 3h1M3 9h3'/%3E%3Cpath stroke='%23c5d3fc' d='M10 3h1'/%3E%3Cpath stroke='%23bed0fc' d='M12 3h1M9 4h1m-7 7h1m0 1h1'/%3E%3Cpath stroke='%23bccdfa' d='M13 3h1'/%3E%3Cpath stroke='%23baccf4' d='M14 3h1'/%3E%3Cpath stroke='%23bdcbda' d='M16 3h1'/%3E%3Cpath stroke='%23c4d4f7' d='M2 4h1'/%3E%3Cpath stroke='%23cddbfc' d='M5 4h1M3 6h1'/%3E%3Cpath stroke='%23c8d5fb' d='M8 4h1'/%3E%3Cpath stroke='%23bbcefd' d='M10 4h3M9 5h1M8 6h1'/%3E%3Cpath stroke='%23bcccf3' d='M14 4h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23b1c2d5' d='M16 4h1'/%3E%3Cpath stroke='%23bed0f8' d='M2 5h1'/%3E%3Cpath stroke='%23ceddfd' d='M4 5h1'/%3E%3Cpath stroke='%23c8d6fb' d='M6 5h3M3 8h2'/%3E%3Cpath stroke='%23bacdfc' d='M10 5h1m1 0h2M3 12h1'/%3E%3Cpath stroke='%23b9cdfb' d='M11 5h1M9 6h2m1 0h2m-1 1h1'/%3E%3Cpath stroke='%23a8bbd4' d='M16 5h1'/%3E%3Cpath stroke='%23cddafc' d='M4 6h1'/%3E%3Cpath stroke='%234d6185' d='M5 6h1m5 0h1M4 7h3m3 0h3M5 8h3m1 0h3M6 9h5m-4 1h3m-2 1h1'/%3E%3Cpath stroke='%23a4b8d3' d='M16 6h1'/%3E%3Cpath stroke='%23c1d3fb' d='M7 7h2M8 8h1'/%3E%3Cpath stroke='%23b6cefb' d='M9 7h1m2 1h1m-2 1h2'/%3E%3Cpath stroke='%23bacbf4' d='M14 7h1'/%3E%3Cpath stroke='%23a0b5d3' d='M16 7h1m-1 1h1m-1 5h1'/%3E%3Cpath stroke='%23b6cdfb' d='M13 8h1m-5 5h1'/%3E%3Cpath stroke='%23b9cbf3' d='M14 8h1'/%3E%3Cpath stroke='%23b4c8f6' d='M2 9h1'/%3E%3Cpath stroke='%23b5cdfa' d='M13 9h1'/%3E%3Cpath stroke='%23b5c9f3' d='M14 9h1'/%3E%3Cpath stroke='%239fb5d2' d='M16 9h1m-1 1h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23b1c7f6' d='M2 10h1'/%3E%3Cpath stroke='%23b2cffb' d='M10 10h3m-2 2h1'/%3E%3Cpath stroke='%23b1cbfa' d='M13 10h1'/%3E%3Cpath stroke='%23b3c8f5' d='M14 10h1m-6 4h2'/%3E%3Cpath stroke='%23adc3f6' d='M2 11h1'/%3E%3Cpath stroke='%23c3d3fd' d='M5 11h1'/%3E%3Cpath stroke='%23c2d5fc' d='M6 11h2'/%3E%3Cpath stroke='%23bad4fc' d='M9 11h1'/%3E%3Cpath stroke='%23b7d3fc' d='M10 11h1m-2 1h1'/%3E%3Cpath stroke='%23b3d1fc' d='M11 11h1'/%3E%3Cpath stroke='%23afcefb' d='M12 11h1'/%3E%3Cpath stroke='%23aecafa' d='M13 11h1'/%3E%3Cpath stroke='%23b1c8f3' d='M14 11h1'/%3E%3Cpath stroke='%23acc2f5' d='M2 12h1'/%3E%3Cpath stroke='%23c1d2fb' d='M5 12h1'/%3E%3Cpath stroke='%23bed1fc' d='M6 12h2'/%3E%3Cpath stroke='%23bdd3fb' d='M8 12h1'/%3E%3Cpath stroke='%23b6d1fb' d='M10 12h1'/%3E%3Cpath stroke='%23afccfb' d='M12 12h1'/%3E%3Cpath stroke='%23adc9f9' d='M13 12h1m-2 1h1'/%3E%3Cpath stroke='%23b1c5f3' d='M14 12h1'/%3E%3Cpath stroke='%23aac0f3' d='M2 13h1'/%3E%3Cpath stroke='%23b7cbf9' d='M3 13h1'/%3E%3Cpath stroke='%23b9cefb' d='M4 13h1'/%3E%3Cpath stroke='%23bbcef9' d='M7 13h1'/%3E%3Cpath stroke='%23b9cffb' d='M8 13h1'/%3E%3Cpath stroke='%23b2cdfb' d='M10 13h1'/%3E%3Cpath stroke='%23b0cbf9' d='M11 13h1'/%3E%3Cpath stroke='%23aec8f7' d='M13 13h1'/%3E%3Cpath stroke='%23b0c5f2' d='M14 13h1'/%3E%3Cpath stroke='%23dbe3f8' d='M2 14h1'/%3E%3Cpath stroke='%23b7c6f1' d='M3 14h1'/%3E%3Cpath stroke='%23b8c9f2' d='M4 14h1m3 0h1'/%3E%3Cpath stroke='%23b2c8f4' d='M11 14h1'/%3E%3Cpath stroke='%23b1c6f3' d='M12 14h1'/%3E%3Cpath stroke='%23b0c4f2' d='M13 14h1'/%3E%3Cpath stroke='%23d9e3f6' d='M14 14h1'/%3E%3Cpath stroke='%23aec0d6' d='M16 14h1'/%3E%3Cpath stroke='%23c3d4e7' d='M1 15h1'/%3E%3Cpath stroke='%23aec4e5' d='M15 15h1'/%3E%3Cpath stroke='%23edf1f3' d='M1 16h1'/%3E%3Cpath stroke='%23aac0e1' d='M2 16h1'/%3E%3Cpath stroke='%2394b1d9' d='M3 16h1'/%3E%3Cpath stroke='%2388a7d8' d='M4 16h1'/%3E%3Cpath stroke='%2383a4d3' d='M5 16h1'/%3E%3Cpath stroke='%237da0d4' d='M6 16h1m3 0h3'/%3E%3Cpath stroke='%237e9fd2' d='M7 16h1'/%3E%3Cpath stroke='%237c9fd3' d='M8 16h2'/%3E%3Cpath stroke='%2382a4d6' d='M13 16h1'/%3E%3Cpath stroke='%2394b0dd' d='M14 16h1'/%3E%3Cpath stroke='%23ecf2f7' d='M15 16h1'/%3E%3C/svg%3E\")}::-webkit-scrollbar-button:horizontal:start{width:17px;background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 -0.5 17 17' shape-rendering='crispEdges'%3E%3Cpath stroke='%23eeede5' d='M0 0h17m-1 1h1m-1 14h1m-1 1h1'/%3E%3Cpath stroke='%23fdfdfa' d='M0 1h1'/%3E%3Cpath stroke='%23fff' d='M1 1h15M0 2h1m14 0h1M0 3h1m14 0h1M0 4h1m14 0h1M0 5h1m14 0h1M0 6h1m14 0h1M0 7h1m14 0h1M0 8h1m14 0h1M0 9h1m14 0h1M0 10h1m14 0h1M0 11h1m14 0h1M0 12h1m14 0h1M0 13h1m14 0h1M0 14h1m14 0h1M1 15h14'/%3E%3Cpath stroke='%23e6eefc' d='M1 2h1'/%3E%3Cpath stroke='%23d0dffc' d='M2 2h1M1 3h1'/%3E%3Cpath stroke='%23cad8f9' d='M3 2h1M1 4h1'/%3E%3Cpath stroke='%23c4d2f7' d='M4 2h1'/%3E%3Cpath stroke='%23c0d0f7' d='M5 2h1'/%3E%3Cpath stroke='%23bdcef7' d='M6 2h1M1 7h1'/%3E%3Cpath stroke='%23bbcdf5' d='M7 2h2'/%3E%3Cpath stroke='%23b8cbf6' d='M9 2h1M1 8h1'/%3E%3Cpath stroke='%23b7caf5' d='M10 2h1M1 9h1'/%3E%3Cpath stroke='%23b5c8f7' d='M11 2h1'/%3E%3Cpath stroke='%23b3c7f5' d='M12 2h1'/%3E%3Cpath stroke='%23afc5f4' d='M13 2h1'/%3E%3Cpath stroke='%23dce6f9' d='M14 2h1'/%3E%3Cpath stroke='%23dfe2e1' d='M16 2h1'/%3E%3Cpath stroke='%23e1eafe' d='M2 3h1'/%3E%3Cpath stroke='%23dae6fe' d='M3 3h1M2 4h1'/%3E%3Cpath stroke='%23d4e1fc' d='M4 3h1M2 5h1'/%3E%3Cpath stroke='%23d1e0fd' d='M5 3h1M3 5h1'/%3E%3Cpath stroke='%23d0ddfc' d='M6 3h1M2 6h1'/%3E%3Cpath stroke='%23cedbfd' d='M7 3h1M5 4h1'/%3E%3Cpath stroke='%23cddbfc' d='M8 3h1M4 5h1M2 7h1'/%3E%3Cpath stroke='%23cad9fd' d='M9 3h1M6 4h1M4 6h1'/%3E%3Cpath stroke='%23c8d8fb' d='M10 3h1'/%3E%3Cpath stroke='%23c5d6fc' d='M11 3h1m-9 7h3'/%3E%3Cpath stroke='%23c2d3fc' d='M12 3h1m-2 1h1M2 10h1m0 1h1'/%3E%3Cpath stroke='%23bccefa' d='M13 3h1m-1 2h1M4 13h2'/%3E%3Cpath stroke='%23b9c9f3' d='M14 3h1M4 14h3'/%3E%3Cpath stroke='%23cfd7dd' d='M16 3h1'/%3E%3Cpath stroke='%23d8e3fc' d='M3 4h1'/%3E%3Cpath stroke='%23d1defd' d='M4 4h1'/%3E%3Cpath stroke='%23c9d8fc' d='M7 4h1M5 5h2M4 7h1M2 8h1'/%3E%3Cpath stroke='%234d6185' d='M8 4h1M7 5h3M6 6h3M5 7h3M4 8h3M5 9h3m-2 1h3m-2 1h3m-2 1h1'/%3E%3Cpath stroke='%23c5d5fc' d='M9 4h1'/%3E%3Cpath stroke='%23c5d3fc' d='M10 4h1'/%3E%3Cpath stroke='%23bed0fc' d='M12 4h1M2 11h1m0 1h1'/%3E%3Cpath stroke='%23bccdfa' d='M13 4h1'/%3E%3Cpath stroke='%23baccf4' d='M14 4h1'/%3E%3Cpath stroke='%23bdcbda' d='M16 4h1'/%3E%3Cpath stroke='%23c4d4f7' d='M1 5h1'/%3E%3Cpath stroke='%23bbcefd' d='M10 5h3M9 6h1'/%3E%3Cpath stroke='%23bcccf3' d='M14 5h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23b1c2d5' d='M16 5h1'/%3E%3Cpath stroke='%23bed0f8' d='M1 6h1'/%3E%3Cpath stroke='%23ceddfd' d='M3 6h1'/%3E%3Cpath stroke='%23c8d6fb' d='M5 6h1M2 9h3'/%3E%3Cpath stroke='%23bacdfc' d='M10 6h1m1 0h2M2 12h1'/%3E%3Cpath stroke='%23b9cdfb' d='M11 6h1M8 7h3m1 0h2m-1 1h1'/%3E%3Cpath stroke='%23a8bbd4' d='M16 6h1'/%3E%3Cpath stroke='%23cddafc' d='M3 7h1'/%3E%3Cpath stroke='%23b7cdfc' d='M11 7h1m0 1h1'/%3E%3Cpath stroke='%23a4b8d3' d='M16 7h1'/%3E%3Cpath stroke='%23cad8fd' d='M3 8h1'/%3E%3Cpath stroke='%23c1d3fb' d='M7 8h2'/%3E%3Cpath stroke='%23b6cefb' d='M9 8h3M9 9h4'/%3E%3Cpath stroke='%23bacbf4' d='M14 8h1'/%3E%3Cpath stroke='%23a0b5d3' d='M16 8h1m-1 1h1m-1 4h1'/%3E%3Cpath stroke='%23bdd3fb' d='M8 9h1m-2 3h1'/%3E%3Cpath stroke='%23b6cdfb' d='M13 9h1m-5 4h1'/%3E%3Cpath stroke='%23b9cbf3' d='M14 9h1'/%3E%3Cpath stroke='%23b1c7f6' d='M1 10h1'/%3E%3Cpath stroke='%23bad4fc' d='M9 10h1'/%3E%3Cpath stroke='%23b2cffb' d='M10 10h3m-2 2h1'/%3E%3Cpath stroke='%23b1cbfa' d='M13 10h1'/%3E%3Cpath stroke='%23b3c8f5' d='M14 10h1m-6 4h2'/%3E%3Cpath stroke='%239fb5d2' d='M16 10h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23adc3f6' d='M1 11h1'/%3E%3Cpath stroke='%23c3d3fd' d='M4 11h1'/%3E%3Cpath stroke='%23c2d5fc' d='M5 11h2'/%3E%3Cpath stroke='%23b7d3fc' d='M10 11h1m-2 1h1'/%3E%3Cpath stroke='%23b3d1fc' d='M11 11h1'/%3E%3Cpath stroke='%23afcefb' d='M12 11h1'/%3E%3Cpath stroke='%23aecafa' d='M13 11h1'/%3E%3Cpath stroke='%23b1c8f3' d='M14 11h1'/%3E%3Cpath stroke='%23acc2f5' d='M1 12h1'/%3E%3Cpath stroke='%23c1d2fb' d='M4 12h1'/%3E%3Cpath stroke='%23bed1fc' d='M5 12h2'/%3E%3Cpath stroke='%23b6d1fb' d='M10 12h1'/%3E%3Cpath stroke='%23afccfb' d='M12 12h1'/%3E%3Cpath stroke='%23adc9f9' d='M13 12h1m-2 1h1'/%3E%3Cpath stroke='%23b1c5f3' d='M14 12h1'/%3E%3Cpath stroke='%23aac0f3' d='M1 13h1'/%3E%3Cpath stroke='%23b7cbf9' d='M2 13h1'/%3E%3Cpath stroke='%23b9cefb' d='M3 13h1'/%3E%3Cpath stroke='%23bbcef9' d='M6 13h1'/%3E%3Cpath stroke='%23b9cffb' d='M7 13h1'/%3E%3Cpath stroke='%23b8cffa' d='M8 13h1'/%3E%3Cpath stroke='%23b2cdfb' d='M10 13h1'/%3E%3Cpath stroke='%23b0cbf9' d='M11 13h1'/%3E%3Cpath stroke='%23aec8f7' d='M13 13h1'/%3E%3Cpath stroke='%23b0c5f2' d='M14 13h1'/%3E%3Cpath stroke='%23dbe3f8' d='M1 14h1'/%3E%3Cpath stroke='%23b7c6f1' d='M2 14h1'/%3E%3Cpath stroke='%23b8c9f2' d='M3 14h1m3 0h2'/%3E%3Cpath stroke='%23b2c8f4' d='M11 14h1'/%3E%3Cpath stroke='%23b1c6f3' d='M12 14h1'/%3E%3Cpath stroke='%23b0c4f2' d='M13 14h1'/%3E%3Cpath stroke='%23d9e3f6' d='M14 14h1'/%3E%3Cpath stroke='%23aec0d6' d='M16 14h1'/%3E%3Cpath stroke='%23c3d4e7' d='M0 15h1'/%3E%3Cpath stroke='%23aec4e5' d='M15 15h1'/%3E%3Cpath stroke='%23edf1f3' d='M0 16h1'/%3E%3Cpath stroke='%23aac0e1' d='M1 16h1'/%3E%3Cpath stroke='%2394b1d9' d='M2 16h1'/%3E%3Cpath stroke='%2388a7d8' d='M3 16h1'/%3E%3Cpath stroke='%2383a4d3' d='M4 16h1'/%3E%3Cpath stroke='%237da0d4' d='M5 16h1m4 0h3'/%3E%3Cpath stroke='%237e9fd2' d='M6 16h1'/%3E%3Cpath stroke='%237c9fd3' d='M7 16h3'/%3E%3Cpath stroke='%2382a4d6' d='M13 16h1'/%3E%3Cpath stroke='%2394b0dd' d='M14 16h1'/%3E%3Cpath stroke='%23ecf2f7' d='M15 16h1'/%3E%3C/svg%3E\")}::-webkit-scrollbar-button:horizontal:end{width:17px;background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 -0.5 17 17' shape-rendering='crispEdges'%3E%3Cpath stroke='%23eeede5' d='M0 0h17m-1 1h1m-1 14h1m-1 1h1'/%3E%3Cpath stroke='%23fdfdfa' d='M0 1h1'/%3E%3Cpath stroke='%23fff' d='M1 1h15M0 2h1m14 0h1M0 3h1m14 0h1M0 4h1m14 0h1M0 5h1m14 0h1M0 6h1m14 0h1M0 7h1m14 0h1M0 8h1m14 0h1M0 9h1m14 0h1M0 10h1m14 0h1M0 11h1m14 0h1M0 12h1m14 0h1M0 13h1m14 0h1M0 14h1m14 0h1M1 15h14'/%3E%3Cpath stroke='%23e6eefc' d='M1 2h1'/%3E%3Cpath stroke='%23d0dffc' d='M2 2h1M1 3h1'/%3E%3Cpath stroke='%23cad8f9' d='M3 2h1M1 4h1'/%3E%3Cpath stroke='%23c4d2f7' d='M4 2h1'/%3E%3Cpath stroke='%23c0d0f7' d='M5 2h1'/%3E%3Cpath stroke='%23bdcef7' d='M6 2h1M1 7h1'/%3E%3Cpath stroke='%23bbcdf5' d='M7 2h2'/%3E%3Cpath stroke='%23b8cbf6' d='M9 2h1M1 8h1'/%3E%3Cpath stroke='%23b7caf5' d='M10 2h1'/%3E%3Cpath stroke='%23b5c8f7' d='M11 2h1'/%3E%3Cpath stroke='%23b3c7f5' d='M12 2h1'/%3E%3Cpath stroke='%23afc5f4' d='M13 2h1'/%3E%3Cpath stroke='%23dce6f9' d='M14 2h1'/%3E%3Cpath stroke='%23dfe2e1' d='M16 2h1'/%3E%3Cpath stroke='%23e1eafe' d='M2 3h1'/%3E%3Cpath stroke='%23dae6fe' d='M3 3h1M2 4h1'/%3E%3Cpath stroke='%23d4e1fc' d='M4 3h1M2 5h1'/%3E%3Cpath stroke='%23d1e0fd' d='M5 3h1M3 5h1'/%3E%3Cpath stroke='%23d0ddfc' d='M6 3h1M2 6h1'/%3E%3Cpath stroke='%23cedbfd' d='M7 3h1M5 4h1'/%3E%3Cpath stroke='%23cddbfc' d='M8 3h1M4 5h1M2 7h1'/%3E%3Cpath stroke='%23cad9fd' d='M9 3h1M6 4h1M4 6h1'/%3E%3Cpath stroke='%23c8d8fb' d='M10 3h1'/%3E%3Cpath stroke='%23c5d6fc' d='M11 3h1m-9 7h3'/%3E%3Cpath stroke='%23c2d3fc' d='M12 3h1m-2 1h1M2 10h1m0 1h1'/%3E%3Cpath stroke='%23bccefa' d='M13 3h1m-1 2h1M4 13h2'/%3E%3Cpath stroke='%23b9c9f3' d='M14 3h1M4 14h3'/%3E%3Cpath stroke='%23cfd7dd' d='M16 3h1'/%3E%3Cpath stroke='%23d8e3fc' d='M3 4h1'/%3E%3Cpath stroke='%23d1defd' d='M4 4h1'/%3E%3Cpath stroke='%234d6185' d='M7 4h1M6 5h3M7 6h3M8 7h3M9 8h3M8 9h3m-4 1h3m-4 1h3m-2 1h1'/%3E%3Cpath stroke='%23c8d6fb' d='M8 4h1M5 6h2'/%3E%3Cpath stroke='%23c5d5fc' d='M9 4h1M2 9h5'/%3E%3Cpath stroke='%23c5d3fc' d='M10 4h1'/%3E%3Cpath stroke='%23bed0fc' d='M12 4h1M9 5h1m-8 6h1m0 1h1'/%3E%3Cpath stroke='%23bccdfa' d='M13 4h1'/%3E%3Cpath stroke='%23baccf4' d='M14 4h1'/%3E%3Cpath stroke='%23bdcbda' d='M16 4h1'/%3E%3Cpath stroke='%23c4d4f7' d='M1 5h1'/%3E%3Cpath stroke='%23c9d8fc' d='M5 5h1M4 7h3M2 8h1'/%3E%3Cpath stroke='%23bbcefd' d='M10 5h3M7 7h1'/%3E%3Cpath stroke='%23bcccf3' d='M14 5h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23b1c2d5' d='M16 5h1'/%3E%3Cpath stroke='%23bed0f8' d='M1 6h1'/%3E%3Cpath stroke='%23ceddfd' d='M3 6h1'/%3E%3Cpath stroke='%23bacdfc' d='M10 6h1m1 0h2M2 12h1'/%3E%3Cpath stroke='%23b9cdfb' d='M11 6h1m0 1h2m-1 1h1'/%3E%3Cpath stroke='%23a8bbd4' d='M16 6h1'/%3E%3Cpath stroke='%23cddafc' d='M3 7h1'/%3E%3Cpath stroke='%23b7cdfc' d='M11 7h1m0 1h1'/%3E%3Cpath stroke='%23a4b8d3' d='M16 7h1'/%3E%3Cpath stroke='%23cad8fd' d='M3 8h3'/%3E%3Cpath stroke='%23c1d3fb' d='M6 8h3'/%3E%3Cpath stroke='%23bacbf4' d='M14 8h1'/%3E%3Cpath stroke='%23a0b5d3' d='M16 8h1m-1 5h1'/%3E%3Cpath stroke='%23b4c8f6' d='M1 9h1'/%3E%3Cpath stroke='%23c2d5fc' d='M7 9h1m-3 2h1'/%3E%3Cpath stroke='%23b6cefb' d='M11 9h2'/%3E%3Cpath stroke='%23b5cdfa' d='M13 9h1'/%3E%3Cpath stroke='%23b5c9f3' d='M14 9h1'/%3E%3Cpath stroke='%239fb5d2' d='M16 9h1m-1 1h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23b1c7f6' d='M1 10h1'/%3E%3Cpath stroke='%23c3d5fd' d='M6 10h1'/%3E%3Cpath stroke='%23b2cffb' d='M10 10h3m-2 2h1'/%3E%3Cpath stroke='%23b1cbfa' d='M13 10h1'/%3E%3Cpath stroke='%23b3c8f5' d='M14 10h1m-6 4h2'/%3E%3Cpath stroke='%23adc3f6' d='M1 11h1'/%3E%3Cpath stroke='%23c3d3fd' d='M4 11h1'/%3E%3Cpath stroke='%23bad4fc' d='M9 11h1'/%3E%3Cpath stroke='%23b7d3fc' d='M10 11h1m-2 1h1'/%3E%3Cpath stroke='%23b3d1fc' d='M11 11h1'/%3E%3Cpath stroke='%23afcefb' d='M12 11h1'/%3E%3Cpath stroke='%23aecafa' d='M13 11h1'/%3E%3Cpath stroke='%23b1c8f3' d='M14 11h1'/%3E%3Cpath stroke='%23acc2f5' d='M1 12h1'/%3E%3Cpath stroke='%23c1d2fb' d='M4 12h1'/%3E%3Cpath stroke='%23bed1fc' d='M5 12h2'/%3E%3Cpath stroke='%23bbd3fd' d='M8 12h1'/%3E%3Cpath stroke='%23b6d1fb' d='M10 12h1'/%3E%3Cpath stroke='%23afccfb' d='M12 12h1'/%3E%3Cpath stroke='%23adc9f9' d='M13 12h1m-2 1h1'/%3E%3Cpath stroke='%23b1c5f3' d='M14 12h1'/%3E%3Cpath stroke='%23aac0f3' d='M1 13h1'/%3E%3Cpath stroke='%23b7cbf9' d='M2 13h1'/%3E%3Cpath stroke='%23b9cefb' d='M3 13h1'/%3E%3Cpath stroke='%23bbcef9' d='M6 13h1'/%3E%3Cpath stroke='%23b9cffb' d='M7 13h1'/%3E%3Cpath stroke='%23b8cffa' d='M8 13h1'/%3E%3Cpath stroke='%23b6cdfb' d='M9 13h1'/%3E%3Cpath stroke='%23b2cdfb' d='M10 13h1'/%3E%3Cpath stroke='%23b0cbf9' d='M11 13h1'/%3E%3Cpath stroke='%23aec8f7' d='M13 13h1'/%3E%3Cpath stroke='%23b0c5f2' d='M14 13h1'/%3E%3Cpath stroke='%23dbe3f8' d='M1 14h1'/%3E%3Cpath stroke='%23b7c6f1' d='M2 14h1'/%3E%3Cpath stroke='%23b8c9f2' d='M3 14h1m3 0h2'/%3E%3Cpath stroke='%23b2c8f4' d='M11 14h1'/%3E%3Cpath stroke='%23b1c6f3' d='M12 14h1'/%3E%3Cpath stroke='%23b0c4f2' d='M13 14h1'/%3E%3Cpath stroke='%23d9e3f6' d='M14 14h1'/%3E%3Cpath stroke='%23aec0d6' d='M16 14h1'/%3E%3Cpath stroke='%23c3d4e7' d='M0 15h1'/%3E%3Cpath stroke='%23aec4e5' d='M15 15h1'/%3E%3Cpath stroke='%23edf1f3' d='M0 16h1'/%3E%3Cpath stroke='%23aac0e1' d='M1 16h1'/%3E%3Cpath stroke='%2394b1d9' d='M2 16h1'/%3E%3Cpath stroke='%2388a7d8' d='M3 16h1'/%3E%3Cpath stroke='%2383a4d3' d='M4 16h1'/%3E%3Cpath stroke='%237da0d4' d='M5 16h1m4 0h3'/%3E%3Cpath stroke='%237e9fd2' d='M6 16h1'/%3E%3Cpath stroke='%237c9fd3' d='M7 16h3'/%3E%3Cpath stroke='%2382a4d6' d='M13 16h1'/%3E%3Cpath stroke='%2394b0dd' d='M14 16h1'/%3E%3Cpath stroke='%23ecf2f7' d='M15 16h1'/%3E%3C/svg%3E\")}.window{box-shadow:inset -1px -1px #00138c,inset 1px 1px #0831d9,inset -2px -2px #001ea0,inset 2px 2px #166aee,inset -3px -3px #003bda,inset 3px 3px #0855dd;border-top-left-radius:8px;border-top-right-radius:8px;padding:0 0 3px;-webkit-font-smoothing:antialiased}.title-bar{font-family:Trebuchet MS;background:linear-gradient(180deg,#0997ff,#0053ee 8%,#0050ee 40%,#06f 88%,#06f 93%,#005bff 95%,#003dd7 96%,#003dd7);padding:3px 5px 3px 3px;border-top:1px solid #0831d9;border-left:1px solid #0831d9;border-right:1px solid #001ea0;border-top-left-radius:8px;border-top-right-radius:7px;font-size:13px;text-shadow:1px 1px #0f1089;height:21px}.title-bar-text{padding-left:3px}.title-bar-controls{display:flex}.title-bar-controls button{min-width:21px;min-height:21px;margin-left:2px;background-repeat:no-repeat;background-position:50%;box-shadow:none;background-color:#0050ee;transition:background .1s;border:none}.title-bar-controls button:active,.title-bar-controls button:focus,.title-bar-controls button:hover{box-shadow:none!important}.title-bar-controls button[aria-label=Minimize]{background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 -0.5 21 21' shape-rendering='crispEdges'%3E%3Cpath stroke='%236696eb' d='M1 0h1m17 0h1'/%3E%3Cpath stroke='%23e5edfb' d='M2 0h1'/%3E%3Cpath stroke='%23fff' d='M3 0h16M0 2h1M0 3h1m19 0h1M0 4h1m19 0h1M0 5h1m19 0h1M0 6h1m19 0h1M0 7h1m19 0h1M0 8h1m19 0h1M0 9h1m19 0h1M0 10h1m19 0h1M0 11h1m19 0h1M0 12h1m19 0h1M0 13h1m4 0h7m8 0h1M0 14h1m4 0h7m8 0h1M0 15h1m4 0h7m8 0h1M0 16h1m19 0h1M0 17h1m19 0h1m-1 1h1M2 20h16'/%3E%3Cpath stroke='%236693e9' d='M0 1h1m19 0h1'/%3E%3Cpath stroke='%23dce5fd' d='M1 1h1'/%3E%3Cpath stroke='%23739af8' d='M2 1h1'/%3E%3Cpath stroke='%23608cf7' d='M3 1h1M2 8h1'/%3E%3Cpath stroke='%235584f6' d='M4 1h1'/%3E%3Cpath stroke='%234d7ef6' d='M5 1h1M1 6h1m5 4h1'/%3E%3Cpath stroke='%23487af5' d='M6 1h1'/%3E%3Cpath stroke='%234276f5' d='M7 1h1M3 14h1'/%3E%3Cpath stroke='%234478f5' d='M8 1h1m5 3h1M2 12h1'/%3E%3Cpath stroke='%233e73f5' d='M9 1h2'/%3E%3Cpath stroke='%233b71f5' d='M11 1h2'/%3E%3Cpath stroke='%23336cf4' d='M13 1h2'/%3E%3Cpath stroke='%23306af4' d='M15 1h1'/%3E%3Cpath stroke='%232864f4' d='M16 1h1'/%3E%3Cpath stroke='%231f5def' d='M17 1h1'/%3E%3Cpath stroke='%233467e0' d='M18 1h1'/%3E%3Cpath stroke='%23d2dbf2' d='M19 1h1'/%3E%3Cpath stroke='%23769cf8' d='M1 2h1'/%3E%3Cpath stroke='%2390aff9' d='M2 2h1'/%3E%3Cpath stroke='%2394b2f9' d='M3 2h1'/%3E%3Cpath stroke='%2385a7f8' d='M4 2h1'/%3E%3Cpath stroke='%23759cf8' d='M5 2h1'/%3E%3Cpath stroke='%236e97f8' d='M6 2h1M2 6h1'/%3E%3Cpath stroke='%236892f7' d='M7 2h1'/%3E%3Cpath stroke='%236690f7' d='M8 2h1'/%3E%3Cpath stroke='%23628ef7' d='M9 2h1m0 1h1'/%3E%3Cpath stroke='%235f8cf7' d='M10 2h1'/%3E%3Cpath stroke='%235e8bf7' d='M11 2h1'/%3E%3Cpath stroke='%235988f6' d='M12 2h1'/%3E%3Cpath stroke='%235685f6' d='M13 2h1'/%3E%3Cpath stroke='%235082f6' d='M14 2h1'/%3E%3Cpath stroke='%23497cf5' d='M15 2h1'/%3E%3Cpath stroke='%233f75f5' d='M16 2h1m-2 2h1'/%3E%3Cpath stroke='%23326bf2' d='M17 2h1'/%3E%3Cpath stroke='%23235ce3' d='M18 2h1'/%3E%3Cpath stroke='%23305cc5' d='M19 2h1'/%3E%3Cpath stroke='%23e5ecfb' d='M20 2h1'/%3E%3Cpath stroke='%236590f7' d='M1 3h1'/%3E%3Cpath stroke='%2397b4f9' d='M2 3h1'/%3E%3Cpath stroke='%239ab7fa' d='M3 3h1'/%3E%3Cpath stroke='%2389aaf9' d='M4 3h1M2 4h1'/%3E%3Cpath stroke='%237aa0f8' d='M5 3h1'/%3E%3Cpath stroke='%23729af8' d='M6 3h1'/%3E%3Cpath stroke='%236d95f8' d='M7 3h1'/%3E%3Cpath stroke='%236892f8' d='M8 3h1M2 7h1'/%3E%3Cpath stroke='%23658ff7' d='M9 3h1'/%3E%3Cpath stroke='%23618df7' d='M11 3h1'/%3E%3Cpath stroke='%235d8af7' d='M12 3h1M3 9h1'/%3E%3Cpath stroke='%235987f6' d='M13 3h1M2 9h1'/%3E%3Cpath stroke='%235283f6' d='M14 3h1'/%3E%3Cpath stroke='%234c7ef6' d='M15 3h1'/%3E%3Cpath stroke='%234377f5' d='M16 3h1'/%3E%3Cpath stroke='%23376ef2' d='M17 3h1'/%3E%3Cpath stroke='%23285fe3' d='M18 3h1'/%3E%3Cpath stroke='%231546b9' d='M19 3h1'/%3E%3Cpath stroke='%235886f6' d='M1 4h1'/%3E%3Cpath stroke='%238dadf9' d='M3 4h1'/%3E%3Cpath stroke='%237fa3f8' d='M4 4h1'/%3E%3Cpath stroke='%237199f8' d='M5 4h1M4 5h1'/%3E%3Cpath stroke='%236a93f8' d='M6 4h1M4 6h1M3 7h1'/%3E%3Cpath stroke='%23648ef7' d='M7 4h1'/%3E%3Cpath stroke='%235e8af7' d='M8 4h1'/%3E%3Cpath stroke='%235986f7' d='M9 4h1M5 9h1m-2 1h1'/%3E%3Cpath stroke='%235482f6' d='M10 4h1'/%3E%3Cpath stroke='%235180f6' d='M11 4h1'/%3E%3Cpath stroke='%234b7cf5' d='M12 4h1'/%3E%3Cpath stroke='%234a7cf5' d='M13 4h1'/%3E%3Cpath stroke='%233a72f4' d='M16 4h1'/%3E%3Cpath stroke='%23346cf2' d='M17 4h1'/%3E%3Cpath stroke='%232a61e3' d='M18 4h1'/%3E%3Cpath stroke='%231848bb' d='M19 4h1'/%3E%3Cpath stroke='%235282f6' d='M1 5h1m4 6h1m-3 1h1'/%3E%3Cpath stroke='%23799ff8' d='M2 5h1'/%3E%3Cpath stroke='%237ca1f8' d='M3 5h1'/%3E%3Cpath stroke='%236791f8' d='M5 5h1'/%3E%3Cpath stroke='%23608bf7' d='M6 5h1M4 8h1'/%3E%3Cpath stroke='%235985f7' d='M7 5h1'/%3E%3Cpath stroke='%235381f6' d='M8 5h1M6 9h1'/%3E%3Cpath stroke='%234d7bf6' d='M9 5h1M8 6h1'/%3E%3Cpath stroke='%234677f5' d='M10 5h1'/%3E%3Cpath stroke='%234173f5' d='M11 5h1'/%3E%3Cpath stroke='%233a6ff4' d='M12 5h1'/%3E%3Cpath stroke='%23386ef4' d='M13 5h1'/%3E%3Cpath stroke='%23346cf4' d='M14 5h1'/%3E%3Cpath stroke='%23326cf4' d='M15 5h1'/%3E%3Cpath stroke='%23316bf4' d='M16 5h1M3 16h1'/%3E%3Cpath stroke='%233069f1' d='M17 5h1'/%3E%3Cpath stroke='%232c62e4' d='M18 5h1'/%3E%3Cpath stroke='%231d4cbc' d='M19 5h1m-1 1h1'/%3E%3Cpath stroke='%237099f8' d='M3 6h1'/%3E%3Cpath stroke='%23628cf8' d='M5 6h1'/%3E%3Cpath stroke='%235b86f7' d='M6 6h1'/%3E%3Cpath stroke='%235480f7' d='M7 6h1'/%3E%3Cpath stroke='%234777f6' d='M9 6h1'/%3E%3Cpath stroke='%234072f5' d='M10 6h1'/%3E%3Cpath stroke='%233a6ff5' d='M11 6h1'/%3E%3Cpath stroke='%23346df4' d='M12 6h1'/%3E%3Cpath stroke='%23306bf4' d='M13 6h1'/%3E%3Cpath stroke='%232d69f4' d='M14 6h1'/%3E%3Cpath stroke='%232c69f5' d='M15 6h1'/%3E%3Cpath stroke='%232d69f5' d='M16 6h1'/%3E%3Cpath stroke='%232e69f2' d='M17 6h1'/%3E%3Cpath stroke='%232c63e5' d='M18 6h1'/%3E%3Cpath stroke='%234679f5' d='M1 7h1M1 8h1'/%3E%3Cpath stroke='%23658ff8' d='M4 7h1'/%3E%3Cpath stroke='%235e89f7' d='M5 7h1'/%3E%3Cpath stroke='%235783f7' d='M6 7h1'/%3E%3Cpath stroke='%23507ef6' d='M7 7h1'/%3E%3Cpath stroke='%234a79f6' d='M8 7h1'/%3E%3Cpath stroke='%234375f5' d='M9 7h1'/%3E%3Cpath stroke='%233d71f5' d='M10 7h1'/%3E%3Cpath stroke='%23366ef4' d='M11 7h1M2 14h1'/%3E%3Cpath stroke='%232f6bf5' d='M12 7h1'/%3E%3Cpath stroke='%232b69f5' d='M13 7h1'/%3E%3Cpath stroke='%232867f5' d='M14 7h1'/%3E%3Cpath stroke='%232766f5' d='M15 7h1'/%3E%3Cpath stroke='%232a68f5' d='M16 7h1'/%3E%3Cpath stroke='%232c69f2' d='M17 7h1'/%3E%3Cpath stroke='%232a62e4' d='M18 7h1'/%3E%3Cpath stroke='%231c4cbd' d='M19 7h1'/%3E%3Cpath stroke='%23628df8' d='M3 8h1'/%3E%3Cpath stroke='%235b87f7' d='M5 8h1'/%3E%3Cpath stroke='%235482f7' d='M6 8h1'/%3E%3Cpath stroke='%234e7cf6' d='M7 8h1'/%3E%3Cpath stroke='%234778f6' d='M8 8h1'/%3E%3Cpath stroke='%234174f5' d='M9 8h1'/%3E%3Cpath stroke='%233a71f5' d='M10 8h1'/%3E%3Cpath stroke='%23346ef4' d='M11 8h1'/%3E%3Cpath stroke='%232d6bf5' d='M12 8h1'/%3E%3Cpath stroke='%232869f5' d='M13 8h1'/%3E%3Cpath stroke='%232467f5' d='M14 8h1'/%3E%3Cpath stroke='%232266f5' d='M15 8h1'/%3E%3Cpath stroke='%232567f5' d='M16 8h1'/%3E%3Cpath stroke='%232968f2' d='M17 8h1'/%3E%3Cpath stroke='%232963e4' d='M18 8h1'/%3E%3Cpath stroke='%231b4bbd' d='M19 8h1'/%3E%3Cpath stroke='%233c72f4' d='M1 9h1'/%3E%3Cpath stroke='%235d89f7' d='M4 9h1'/%3E%3Cpath stroke='%234e7ef6' d='M7 9h1'/%3E%3Cpath stroke='%23477af5' d='M8 9h1'/%3E%3Cpath stroke='%234178f5' d='M9 9h1'/%3E%3Cpath stroke='%233a74f5' d='M10 9h1'/%3E%3Cpath stroke='%233472f5' d='M11 9h1'/%3E%3Cpath stroke='%232c6ff5' d='M12 9h1'/%3E%3Cpath stroke='%23276cf5' d='M13 9h1'/%3E%3Cpath stroke='%23236af6' d='M14 9h1'/%3E%3Cpath stroke='%232069f6' d='M15 9h1'/%3E%3Cpath stroke='%232268f5' d='M16 9h1'/%3E%3Cpath stroke='%232569f2' d='M17 9h1'/%3E%3Cpath stroke='%232562e6' d='M18 9h1'/%3E%3Cpath stroke='%23194bbe' d='M19 9h1'/%3E%3Cpath stroke='%23376ef4' d='M1 10h1'/%3E%3Cpath stroke='%235181f6' d='M2 10h1'/%3E%3Cpath stroke='%235785f7' d='M3 10h1m1 0h1'/%3E%3Cpath stroke='%235281f6' d='M6 10h1'/%3E%3Cpath stroke='%23477bf6' d='M8 10h1'/%3E%3Cpath stroke='%234179f6' d='M9 10h1'/%3E%3Cpath stroke='%233b77f5' d='M10 10h1'/%3E%3Cpath stroke='%233474f5' d='M11 10h1'/%3E%3Cpath stroke='%232c72f6' d='M12 10h1'/%3E%3Cpath stroke='%23266ff6' d='M13 10h1'/%3E%3Cpath stroke='%23226df6' d='M14 10h1'/%3E%3Cpath stroke='%231e6bf6' d='M15 10h1'/%3E%3Cpath stroke='%231f6af6' d='M16 10h1'/%3E%3Cpath stroke='%23216af3' d='M17 10h1'/%3E%3Cpath stroke='%232162e6' d='M18 10h1'/%3E%3Cpath stroke='%231649be' d='M19 10h1'/%3E%3Cpath stroke='%23326bf4' d='M1 11h1'/%3E%3Cpath stroke='%234b7df5' d='M2 11h1'/%3E%3Cpath stroke='%235483f6' d='M3 11h1'/%3E%3Cpath stroke='%235684f7' d='M4 11h1'/%3E%3Cpath stroke='%235583f7' d='M5 11h1'/%3E%3Cpath stroke='%234d80f6' d='M7 11h1'/%3E%3Cpath stroke='%23487df6' d='M8 11h1'/%3E%3Cpath stroke='%23427cf6' d='M9 11h1'/%3E%3Cpath stroke='%233c7af6' d='M10 11h1'/%3E%3Cpath stroke='%233478f6' d='M11 11h1'/%3E%3Cpath stroke='%232d76f6' d='M12 11h1'/%3E%3Cpath stroke='%232673f7' d='M13 11h1'/%3E%3Cpath stroke='%232171f7' d='M14 11h1'/%3E%3Cpath stroke='%231c6ff6' d='M15 11h1'/%3E%3Cpath stroke='%231c6df6' d='M16 11h1'/%3E%3Cpath stroke='%231c6af4' d='M17 11h1'/%3E%3Cpath stroke='%231c61e6' d='M18 11h1'/%3E%3Cpath stroke='%231248bf' d='M19 11h1'/%3E%3Cpath stroke='%232b66f4' d='M1 12h1'/%3E%3Cpath stroke='%234e7ff6' d='M3 12h1'/%3E%3Cpath stroke='%235383f6' d='M5 12h1'/%3E%3Cpath stroke='%235182f6' d='M6 12h1'/%3E%3Cpath stroke='%234d81f7' d='M7 12h1'/%3E%3Cpath stroke='%23487ff6' d='M8 12h1'/%3E%3Cpath stroke='%23437ff6' d='M9 12h1'/%3E%3Cpath stroke='%233d7ef6' d='M10 12h1'/%3E%3Cpath stroke='%23357cf6' d='M11 12h1'/%3E%3Cpath stroke='%232d7af7' d='M12 12h1'/%3E%3Cpath stroke='%232677f7' d='M13 12h1'/%3E%3Cpath stroke='%232174f7' d='M14 12h1'/%3E%3Cpath stroke='%231b71f7' d='M15 12h1'/%3E%3Cpath stroke='%23186ef7' d='M16 12h1'/%3E%3Cpath stroke='%23186af4' d='M17 12h1'/%3E%3Cpath stroke='%23165fe7' d='M18 12h1'/%3E%3Cpath stroke='%230f47c0' d='M19 12h1'/%3E%3Cpath stroke='%232562f3' d='M1 13h1'/%3E%3Cpath stroke='%233d73f4' d='M2 13h1'/%3E%3Cpath stroke='%23487bf5' d='M3 13h1'/%3E%3Cpath stroke='%234e80f6' d='M4 13h1'/%3E%3Cpath stroke='%232d7cf7' d='M12 13h1'/%3E%3Cpath stroke='%232679f8' d='M13 13h1'/%3E%3Cpath stroke='%232077f7' d='M14 13h1'/%3E%3Cpath stroke='%231973f7' d='M15 13h1'/%3E%3Cpath stroke='%23166ff7' d='M16 13h1'/%3E%3Cpath stroke='%231369f4' d='M17 13h1'/%3E%3Cpath stroke='%23105de8' d='M18 13h1'/%3E%3Cpath stroke='%230a44bf' d='M19 13h1'/%3E%3Cpath stroke='%231e5df3' d='M1 14h1'/%3E%3Cpath stroke='%23497bf5' d='M4 14h1'/%3E%3Cpath stroke='%232d7df7' d='M12 14h1'/%3E%3Cpath stroke='%23257af8' d='M13 14h1'/%3E%3Cpath stroke='%231e77f8' d='M14 14h1'/%3E%3Cpath stroke='%231773f8' d='M15 14h1'/%3E%3Cpath stroke='%23116df7' d='M16 14h1'/%3E%3Cpath stroke='%230d66f4' d='M17 14h1m-3 3h1'/%3E%3Cpath stroke='%230b59e7' d='M18 14h1'/%3E%3Cpath stroke='%230641c0' d='M19 14h1m-6 5h1'/%3E%3Cpath stroke='%231859f3' d='M1 15h1'/%3E%3Cpath stroke='%232e68f4' d='M2 15h1'/%3E%3Cpath stroke='%233a71f4' d='M3 15h1'/%3E%3Cpath stroke='%234277f5' d='M4 15h1'/%3E%3Cpath stroke='%232a7cf8' d='M12 15h1'/%3E%3Cpath stroke='%23247af8' d='M13 15h1'/%3E%3Cpath stroke='%231d77f8' d='M14 15h1'/%3E%3Cpath stroke='%231573f8' d='M15 15h1'/%3E%3Cpath stroke='%230e6cf8' d='M16 15h1'/%3E%3Cpath stroke='%230963f4' d='M17 15h1'/%3E%3Cpath stroke='%230556e7' d='M18 15h1'/%3E%3Cpath stroke='%23023fbf' d='M19 15h1'/%3E%3Cpath stroke='%231456f3' d='M1 16h1'/%3E%3Cpath stroke='%232562f4' d='M2 16h1'/%3E%3Cpath stroke='%233971f4' d='M4 16h1'/%3E%3Cpath stroke='%233d74f5' d='M5 16h1'/%3E%3Cpath stroke='%233d74f6' d='M6 16h1'/%3E%3Cpath stroke='%233b75f5' d='M7 16h1'/%3E%3Cpath stroke='%233976f5' d='M8 16h1'/%3E%3Cpath stroke='%233777f5' d='M9 16h1'/%3E%3Cpath stroke='%233278f6' d='M10 16h1'/%3E%3Cpath stroke='%232c78f7' d='M11 16h1'/%3E%3Cpath stroke='%232577f7' d='M12 16h1'/%3E%3Cpath stroke='%231f76f7' d='M13 16h1'/%3E%3Cpath stroke='%231972f7' d='M14 16h1'/%3E%3Cpath stroke='%23116ef8' d='M15 16h1'/%3E%3Cpath stroke='%230b68f7' d='M16 16h1'/%3E%3Cpath stroke='%230560f4' d='M17 16h1'/%3E%3Cpath stroke='%230253e6' d='M18 16h1'/%3E%3Cpath stroke='%23013dbe' d='M19 16h1'/%3E%3Cpath stroke='%230e50ed' d='M1 17h1'/%3E%3Cpath stroke='%231c5bef' d='M2 17h1'/%3E%3Cpath stroke='%232863f0' d='M3 17h1'/%3E%3Cpath stroke='%232f68f0' d='M4 17h1'/%3E%3Cpath stroke='%23336bf1' d='M5 17h1'/%3E%3Cpath stroke='%23346cf1' d='M6 17h1'/%3E%3Cpath stroke='%23316cf2' d='M7 17h1'/%3E%3Cpath stroke='%23316df2' d='M8 17h1'/%3E%3Cpath stroke='%232e6ff2' d='M9 17h1'/%3E%3Cpath stroke='%232a70f2' d='M10 17h1'/%3E%3Cpath stroke='%232570f3' d='M11 17h1'/%3E%3Cpath stroke='%231f6ff3' d='M12 17h1'/%3E%3Cpath stroke='%23196df4' d='M13 17h1'/%3E%3Cpath stroke='%23136af4' d='M14 17h1'/%3E%3Cpath stroke='%230760f3' d='M16 17h1'/%3E%3Cpath stroke='%23025af0' d='M17 17h1'/%3E%3Cpath stroke='%23004de2' d='M18 17h1'/%3E%3Cpath stroke='%23003ab9' d='M19 17h1'/%3E%3Cpath stroke='%23e5eefd' d='M0 18h1'/%3E%3Cpath stroke='%23285edf' d='M1 18h1'/%3E%3Cpath stroke='%23134fdf' d='M2 18h1'/%3E%3Cpath stroke='%231b55df' d='M3 18h1'/%3E%3Cpath stroke='%23215ae2' d='M4 18h1'/%3E%3Cpath stroke='%23255ce1' d='M5 18h1'/%3E%3Cpath stroke='%23265de0' d='M6 18h1'/%3E%3Cpath stroke='%23245ce1' d='M7 18h1'/%3E%3Cpath stroke='%23235ee2' d='M8 18h1'/%3E%3Cpath stroke='%23215ee2' d='M9 18h1'/%3E%3Cpath stroke='%231e5ee2' d='M10 18h1'/%3E%3Cpath stroke='%231b5fe5' d='M11 18h1'/%3E%3Cpath stroke='%23165ee5' d='M12 18h1'/%3E%3Cpath stroke='%23135de6' d='M13 18h1'/%3E%3Cpath stroke='%230e5be5' d='M14 18h1'/%3E%3Cpath stroke='%230958e6' d='M15 18h1'/%3E%3Cpath stroke='%230454e6' d='M16 18h1'/%3E%3Cpath stroke='%23014ee2' d='M17 18h1'/%3E%3Cpath stroke='%230045d3' d='M18 18h1'/%3E%3Cpath stroke='%231f4eb8' d='M19 18h1'/%3E%3Cpath stroke='%23679ef6' d='M0 19h1m19 0h1'/%3E%3Cpath stroke='%23d0daf1' d='M1 19h1'/%3E%3Cpath stroke='%232856c3' d='M2 19h1'/%3E%3Cpath stroke='%230d3fb6' d='M3 19h1'/%3E%3Cpath stroke='%231144bd' d='M4 19h1'/%3E%3Cpath stroke='%231245bb' d='M5 19h1'/%3E%3Cpath stroke='%231445b9' d='M6 19h1'/%3E%3Cpath stroke='%231244b9' d='M7 19h1'/%3E%3Cpath stroke='%231345bc' d='M8 19h1'/%3E%3Cpath stroke='%231346bd' d='M9 19h1'/%3E%3Cpath stroke='%231045be' d='M10 19h1'/%3E%3Cpath stroke='%230d45c0' d='M11 19h1'/%3E%3Cpath stroke='%230a45c1' d='M12 19h1'/%3E%3Cpath stroke='%230844c3' d='M13 19h1'/%3E%3Cpath stroke='%23033fc0' d='M15 19h1'/%3E%3Cpath stroke='%23013fc3' d='M16 19h1'/%3E%3Cpath stroke='%23003bbe' d='M17 19h1'/%3E%3Cpath stroke='%231f4eb9' d='M18 19h1'/%3E%3Cpath stroke='%23cfd8ed' d='M19 19h1'/%3E%3Cpath stroke='%23669bf5' d='M1 20h1m17 0h1'/%3E%3Cpath stroke='%23e5edfd' d='M18 20h1'/%3E%3C/svg%3E\")}.title-bar-controls button[aria-label=Minimize]:hover{background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 -0.5 21 21' shape-rendering='crispEdges'%3E%3Cpath stroke='%2393b1ed' d='M1 0h1m17 0h1'/%3E%3Cpath stroke='%23f3f6fd' d='M2 0h1m17 2h1M0 18h1'/%3E%3Cpath stroke='%23fff' d='M3 0h15M0 3h1m19 0h1M0 4h1m19 0h1M0 5h1m19 0h1M0 6h1m19 0h1M0 7h1m19 0h1M0 8h1m19 0h1M0 9h1m19 0h1M0 10h1m19 0h1M0 11h1m19 0h1M0 12h1m19 0h1M0 13h1m4 0h7m8 0h1M0 14h1m4 0h7m8 0h1M0 15h1m4 0h7m8 0h1M0 16h1m19 0h1M0 17h1m19 0h1M3 20h11'/%3E%3Cpath stroke='%23f5f7fd' d='M18 0h1M0 2h1m19 16h1M2 20h1'/%3E%3Cpath stroke='%2393b0ec' d='M0 1h1m19 0h1'/%3E%3Cpath stroke='%23dce7ff' d='M1 1h1'/%3E%3Cpath stroke='%2372a1ff' d='M2 1h1m4 3h1M5 6h1'/%3E%3Cpath stroke='%236a9cff' d='M3 1h1'/%3E%3Cpath stroke='%235f94ff' d='M4 1h1M4 11h2'/%3E%3Cpath stroke='%23558eff' d='M5 1h1M3 12h1'/%3E%3Cpath stroke='%23518bff' d='M6 1h1m3 4h1'/%3E%3Cpath stroke='%234a86ff' d='M7 1h1'/%3E%3Cpath stroke='%234b87ff' d='M8 1h1m2 4h1M2 12h1'/%3E%3Cpath stroke='%234684ff' d='M9 1h2'/%3E%3Cpath stroke='%234482ff' d='M11 1h1m4 1h1m-5 3h1M1 9h1m0 4h1'/%3E%3Cpath stroke='%234080ff' d='M12 1h1M3 15h1'/%3E%3Cpath stroke='%233b7cff' d='M13 1h1'/%3E%3Cpath stroke='%233a7bff' d='M14 1h1'/%3E%3Cpath stroke='%233678ff' d='M15 1h1'/%3E%3Cpath stroke='%232e73ff' d='M16 1h1'/%3E%3Cpath stroke='%23276cf9' d='M17 1h1'/%3E%3Cpath stroke='%233a73e7' d='M18 1h1'/%3E%3Cpath stroke='%23d3ddf3' d='M19 1h1'/%3E%3Cpath stroke='%2373a1ff' d='M1 2h1'/%3E%3Cpath stroke='%2397b9ff' d='M2 2h1'/%3E%3Cpath stroke='%239cbdff' d='M3 2h1'/%3E%3Cpath stroke='%2390b5ff' d='M4 2h1'/%3E%3Cpath stroke='%2382acff' d='M5 2h1M5 4h1'/%3E%3Cpath stroke='%237ba7ff' d='M6 2h1M2 6h1'/%3E%3Cpath stroke='%2375a3ff' d='M7 2h1'/%3E%3Cpath stroke='%236f9fff' d='M8 2h1M3 8h1'/%3E%3Cpath stroke='%236c9dff' d='M9 2h1M1 3h1'/%3E%3Cpath stroke='%23689bff' d='M10 2h1M5 8h1M3 9h1'/%3E%3Cpath stroke='%236599ff' d='M11 2h1m0 1h1M5 9h1'/%3E%3Cpath stroke='%236095ff' d='M12 2h1m0 1h1M8 5h1'/%3E%3Cpath stroke='%235d93ff' d='M13 2h1'/%3E%3Cpath stroke='%23568eff' d='M14 2h1'/%3E%3Cpath stroke='%234f8aff' d='M15 2h1M3 13h1m0 1h1'/%3E%3Cpath stroke='%233878fb' d='M17 2h1'/%3E%3Cpath stroke='%232969eb' d='M18 2h1'/%3E%3Cpath stroke='%233566cb' d='M19 2h1'/%3E%3Cpath stroke='%239ebeff' d='M2 3h1'/%3E%3Cpath stroke='%23a4c2ff' d='M3 3h1'/%3E%3Cpath stroke='%2399baff' d='M4 3h1M3 4h1'/%3E%3Cpath stroke='%238ab0ff' d='M5 3h1'/%3E%3Cpath stroke='%2382abff' d='M6 3h1'/%3E%3Cpath stroke='%2379a6ff' d='M7 3h1'/%3E%3Cpath stroke='%2374a3ff' d='M8 3h1'/%3E%3Cpath stroke='%2371a0ff' d='M9 3h1'/%3E%3Cpath stroke='%236d9eff' d='M10 3h1M5 7h1M4 8h1'/%3E%3Cpath stroke='%23699bff' d='M11 3h1'/%3E%3Cpath stroke='%235a91ff' d='M14 3h1M2 10h1m1 2h1'/%3E%3Cpath stroke='%23538cff' d='M15 3h1M2 11h1'/%3E%3Cpath stroke='%234986ff' d='M16 3h1'/%3E%3Cpath stroke='%233d7cfc' d='M17 3h1'/%3E%3Cpath stroke='%232e6cea' d='M18 3h1'/%3E%3Cpath stroke='%231b52c2' d='M19 3h1'/%3E%3Cpath stroke='%236296ff' d='M1 4h1'/%3E%3Cpath stroke='%2391b5ff' d='M2 4h1'/%3E%3Cpath stroke='%238fb4ff' d='M4 4h1'/%3E%3Cpath stroke='%237aa6ff' d='M6 4h1'/%3E%3Cpath stroke='%236b9dff' d='M8 4h1'/%3E%3Cpath stroke='%236598ff' d='M9 4h1'/%3E%3Cpath stroke='%235f95ff' d='M10 4h1M7 7h1m-2 3h1'/%3E%3Cpath stroke='%235b92ff' d='M11 4h1'/%3E%3Cpath stroke='%23548dff' d='M12 4h1M1 6h1m2 7h1'/%3E%3Cpath stroke='%23528cff' d='M13 4h1'/%3E%3Cpath stroke='%234c88ff' d='M14 4h1m-5 2h1'/%3E%3Cpath stroke='%234785ff' d='M15 4h1'/%3E%3Cpath stroke='%234280ff' d='M16 4h1'/%3E%3Cpath stroke='%233b7afb' d='M17 4h1'/%3E%3Cpath stroke='%23316fec' d='M18 4h1'/%3E%3Cpath stroke='%231f55c3' d='M19 4h1'/%3E%3Cpath stroke='%235990ff' d='M1 5h1m7 0h1'/%3E%3Cpath stroke='%2385adff' d='M2 5h1'/%3E%3Cpath stroke='%238bb1ff' d='M3 5h1'/%3E%3Cpath stroke='%2384acff' d='M4 5h1'/%3E%3Cpath stroke='%2378a5ff' d='M5 5h1'/%3E%3Cpath stroke='%2370a0ff' d='M6 5h1'/%3E%3Cpath stroke='%23679aff' d='M7 5h1'/%3E%3Cpath stroke='%234180ff' d='M13 5h1'/%3E%3Cpath stroke='%233d7eff' d='M14 5h1'/%3E%3Cpath stroke='%233b7bff' d='M15 5h1'/%3E%3Cpath stroke='%23397aff' d='M16 5h1M1 11h1'/%3E%3Cpath stroke='%233979fc' d='M17 5h1'/%3E%3Cpath stroke='%233370ec' d='M18 5h1m-1 1h1'/%3E%3Cpath stroke='%232357c3' d='M19 5h1'/%3E%3Cpath stroke='%2381aaff' d='M3 6h1'/%3E%3Cpath stroke='%237aa7ff' d='M4 6h1'/%3E%3Cpath stroke='%236b9cff' d='M6 6h1'/%3E%3Cpath stroke='%236297ff' d='M7 6h1m-3 4h1'/%3E%3Cpath stroke='%235c93ff' d='M8 6h1M7 8h1m-2 3h1'/%3E%3Cpath stroke='%23548eff' d='M9 6h1'/%3E%3Cpath stroke='%234483ff' d='M11 6h1M5 16h1'/%3E%3Cpath stroke='%233d7fff' d='M12 6h1'/%3E%3Cpath stroke='%23387bff' d='M13 6h1'/%3E%3Cpath stroke='%233679ff' d='M14 6h1m1 0h1'/%3E%3Cpath stroke='%233579ff' d='M15 6h1'/%3E%3Cpath stroke='%233879fc' d='M17 6h1'/%3E%3Cpath stroke='%232358c5' d='M19 6h1'/%3E%3Cpath stroke='%234e89ff' d='M1 7h1'/%3E%3Cpath stroke='%2371a1ff' d='M2 7h1'/%3E%3Cpath stroke='%2377a5ff' d='M3 7h1'/%3E%3Cpath stroke='%2374a2ff' d='M4 7h1'/%3E%3Cpath stroke='%23669aff' d='M6 7h1'/%3E%3Cpath stroke='%235890ff' d='M8 7h1'/%3E%3Cpath stroke='%23508dff' d='M9 7h1'/%3E%3Cpath stroke='%234989ff' d='M10 7h1'/%3E%3Cpath stroke='%234183ff' d='M11 7h1'/%3E%3Cpath stroke='%233a7fff' d='M12 7h1'/%3E%3Cpath stroke='%23357bff' d='M13 7h1'/%3E%3Cpath stroke='%23317aff' d='M14 7h2'/%3E%3Cpath stroke='%23337aff' d='M16 7h1'/%3E%3Cpath stroke='%23367bfc' d='M17 7h1'/%3E%3Cpath stroke='%233372ed' d='M18 7h1'/%3E%3Cpath stroke='%232359c5' d='M19 7h1'/%3E%3Cpath stroke='%234d88ff' d='M1 8h1'/%3E%3Cpath stroke='%23699cff' d='M2 8h1'/%3E%3Cpath stroke='%236398ff' d='M6 8h1'/%3E%3Cpath stroke='%23548fff' d='M8 8h1'/%3E%3Cpath stroke='%234d8cff' d='M9 8h1'/%3E%3Cpath stroke='%23468aff' d='M10 8h1'/%3E%3Cpath stroke='%233f86ff' d='M11 8h1'/%3E%3Cpath stroke='%233983ff' d='M12 8h1'/%3E%3Cpath stroke='%233380ff' d='M13 8h1'/%3E%3Cpath stroke='%232f7fff' d='M14 8h2'/%3E%3Cpath stroke='%233280ff' d='M16 8h1'/%3E%3Cpath stroke='%233580fc' d='M17 8h1'/%3E%3Cpath stroke='%233276ed' d='M18 8h1'/%3E%3Cpath stroke='%23235ac6' d='M19 8h1'/%3E%3Cpath stroke='%236196ff' d='M2 9h1m3 0h1m-4 1h1'/%3E%3Cpath stroke='%23689aff' d='M4 9h1'/%3E%3Cpath stroke='%235b93ff' d='M7 9h1'/%3E%3Cpath stroke='%235491ff' d='M8 9h1'/%3E%3Cpath stroke='%234f90ff' d='M9 9h1'/%3E%3Cpath stroke='%234890ff' d='M10 9h1'/%3E%3Cpath stroke='%23428eff' d='M11 9h1'/%3E%3Cpath stroke='%233b8dff' d='M12 9h1'/%3E%3Cpath stroke='%23348aff' d='M13 9h1'/%3E%3Cpath stroke='%233189ff' d='M14 9h1'/%3E%3Cpath stroke='%232f88ff' d='M15 9h1'/%3E%3Cpath stroke='%233188ff' d='M16 9h1'/%3E%3Cpath stroke='%233385fc' d='M17 9h1'/%3E%3Cpath stroke='%233079ed' d='M18 9h1'/%3E%3Cpath stroke='%23215cc8' d='M19 9h1'/%3E%3Cpath stroke='%233f7fff' d='M1 10h1'/%3E%3Cpath stroke='%236397ff' d='M4 10h1'/%3E%3Cpath stroke='%235993ff' d='M7 10h1'/%3E%3Cpath stroke='%235492ff' d='M8 10h1'/%3E%3Cpath stroke='%235093ff' d='M9 10h1'/%3E%3Cpath stroke='%234a95ff' d='M10 10h1'/%3E%3Cpath stroke='%234496ff' d='M11 10h1'/%3E%3Cpath stroke='%233d96ff' d='M12 10h1'/%3E%3Cpath stroke='%233694ff' d='M13 10h1'/%3E%3Cpath stroke='%233193ff' d='M14 10h1'/%3E%3Cpath stroke='%232f92ff' d='M15 10h1'/%3E%3Cpath stroke='%233090ff' d='M16 10h1'/%3E%3Cpath stroke='%23328cfc' d='M17 10h1'/%3E%3Cpath stroke='%232e7def' d='M18 10h1'/%3E%3Cpath stroke='%231e5dc9' d='M19 10h1'/%3E%3Cpath stroke='%235c92ff' d='M3 11h1m1 1h1'/%3E%3Cpath stroke='%235792ff' d='M7 11h1m-1 1h1'/%3E%3Cpath stroke='%235594ff' d='M8 11h1'/%3E%3Cpath stroke='%235298ff' d='M9 11h1'/%3E%3Cpath stroke='%234d9cff' d='M10 11h1'/%3E%3Cpath stroke='%23479eff' d='M11 11h1'/%3E%3Cpath stroke='%23409fff' d='M12 11h1'/%3E%3Cpath stroke='%23379fff' d='M13 11h1'/%3E%3Cpath stroke='%23339dff' d='M14 11h1'/%3E%3Cpath stroke='%232f9bff' d='M15 11h1'/%3E%3Cpath stroke='%232e97ff' d='M16 11h1'/%3E%3Cpath stroke='%232e91fc' d='M17 11h1'/%3E%3Cpath stroke='%232a80f0' d='M18 11h1'/%3E%3Cpath stroke='%231b5dcb' d='M19 11h1'/%3E%3Cpath stroke='%233275ff' d='M1 12h1'/%3E%3Cpath stroke='%235991ff' d='M6 12h1'/%3E%3Cpath stroke='%235596ff' d='M8 12h1'/%3E%3Cpath stroke='%23529cff' d='M9 12h1'/%3E%3Cpath stroke='%234fa1ff' d='M10 12h1'/%3E%3Cpath stroke='%234aa6ff' d='M11 12h1'/%3E%3Cpath stroke='%2342a9ff' d='M12 12h1'/%3E%3Cpath stroke='%233aa9ff' d='M13 12h1'/%3E%3Cpath stroke='%2334a7ff' d='M14 12h1'/%3E%3Cpath stroke='%2330a5ff' d='M15 12h1'/%3E%3Cpath stroke='%232ca0ff' d='M16 12h1'/%3E%3Cpath stroke='%232a96fd' d='M17 12h1'/%3E%3Cpath stroke='%232581f1' d='M18 12h1'/%3E%3Cpath stroke='%23185dcc' d='M19 12h1'/%3E%3Cpath stroke='%232d72ff' d='M1 13h1m0 3h1'/%3E%3Cpath stroke='%2344afff' d='M12 13h1'/%3E%3Cpath stroke='%233eb1ff' d='M13 13h1'/%3E%3Cpath stroke='%2337afff' d='M14 13h1'/%3E%3Cpath stroke='%232fabff' d='M15 13h1'/%3E%3Cpath stroke='%2329a4ff' d='M16 13h1'/%3E%3Cpath stroke='%232599fd' d='M17 13h1'/%3E%3Cpath stroke='%231e80f2' d='M18 13h1'/%3E%3Cpath stroke='%23145bcd' d='M19 13h1'/%3E%3Cpath stroke='%23276eff' d='M1 14h1'/%3E%3Cpath stroke='%233d7dff' d='M2 14h1'/%3E%3Cpath stroke='%234985ff' d='M3 14h1'/%3E%3Cpath stroke='%2343b1ff' d='M12 14h1'/%3E%3Cpath stroke='%233eb4ff' d='M13 14h1'/%3E%3Cpath stroke='%2335b2ff' d='M14 14h1'/%3E%3Cpath stroke='%232caeff' d='M15 14h1'/%3E%3Cpath stroke='%2324a5ff' d='M16 14h1'/%3E%3Cpath stroke='%231f97fd' d='M17 14h1'/%3E%3Cpath stroke='%231980f3' d='M18 14h1'/%3E%3Cpath stroke='%23105ace' d='M19 14h1'/%3E%3Cpath stroke='%23216aff' d='M1 15h1'/%3E%3Cpath stroke='%233578ff' d='M2 15h1'/%3E%3Cpath stroke='%234885ff' d='M4 15h1'/%3E%3Cpath stroke='%2341afff' d='M12 15h1'/%3E%3Cpath stroke='%233bb2ff' d='M13 15h1'/%3E%3Cpath stroke='%2333b1ff' d='M14 15h1'/%3E%3Cpath stroke='%232aadff' d='M15 15h1'/%3E%3Cpath stroke='%2321a3ff' d='M16 15h1'/%3E%3Cpath stroke='%231a95fd' d='M17 15h1'/%3E%3Cpath stroke='%23137cf2' d='M18 15h1'/%3E%3Cpath stroke='%230c59cf' d='M19 15h1'/%3E%3Cpath stroke='%231c66ff' d='M1 16h1'/%3E%3Cpath stroke='%233879ff' d='M3 16h1'/%3E%3Cpath stroke='%233f7eff' d='M4 16h1'/%3E%3Cpath stroke='%234584ff' d='M6 16h1'/%3E%3Cpath stroke='%234587ff' d='M7 16h1'/%3E%3Cpath stroke='%23468eff' d='M8 16h1'/%3E%3Cpath stroke='%234696ff' d='M9 16h1'/%3E%3Cpath stroke='%23439cff' d='M10 16h1'/%3E%3Cpath stroke='%233fa3ff' d='M11 16h1'/%3E%3Cpath stroke='%233ba8ff' d='M12 16h1'/%3E%3Cpath stroke='%233af' d='M13 16h1'/%3E%3Cpath stroke='%232da9ff' d='M14 16h1'/%3E%3Cpath stroke='%2324a6ff' d='M15 16h1'/%3E%3Cpath stroke='%231d9eff' d='M16 16h1'/%3E%3Cpath stroke='%231690fd' d='M17 16h1'/%3E%3Cpath stroke='%231078f1' d='M18 16h1'/%3E%3Cpath stroke='%230b57ce' d='M19 16h1'/%3E%3Cpath stroke='%231761f9' d='M1 17h1'/%3E%3Cpath stroke='%23246bfa' d='M2 17h1'/%3E%3Cpath stroke='%232f72fb' d='M3 17h1'/%3E%3Cpath stroke='%233676fb' d='M4 17h1'/%3E%3Cpath stroke='%233a7afb' d='M5 17h1'/%3E%3Cpath stroke='%233b7bfc' d='M6 17h1'/%3E%3Cpath stroke='%233b7efc' d='M7 17h1'/%3E%3Cpath stroke='%233c84fc' d='M8 17h1'/%3E%3Cpath stroke='%233b8afc' d='M9 17h1'/%3E%3Cpath stroke='%233990fc' d='M10 17h1'/%3E%3Cpath stroke='%233695fc' d='M11 17h1'/%3E%3Cpath stroke='%233299fc' d='M12 17h1'/%3E%3Cpath stroke='%232c9cfd' d='M13 17h1'/%3E%3Cpath stroke='%23259bfd' d='M14 17h1'/%3E%3Cpath stroke='%231e97fd' d='M15 17h1'/%3E%3Cpath stroke='%231790fc' d='M16 17h1'/%3E%3Cpath stroke='%231184fa' d='M17 17h1'/%3E%3Cpath stroke='%230c6ded' d='M18 17h1'/%3E%3Cpath stroke='%230850c8' d='M19 17h1'/%3E%3Cpath stroke='%232f6ae4' d='M1 18h1'/%3E%3Cpath stroke='%231b5fe9' d='M2 18h1'/%3E%3Cpath stroke='%232163e8' d='M3 18h1'/%3E%3Cpath stroke='%232868eb' d='M4 18h1'/%3E%3Cpath stroke='%232c6aea' d='M5 18h1'/%3E%3Cpath stroke='%232e6dea' d='M6 18h1'/%3E%3Cpath stroke='%232d6deb' d='M7 18h1'/%3E%3Cpath stroke='%232c71ec' d='M8 18h1'/%3E%3Cpath stroke='%232c76ec' d='M9 18h1'/%3E%3Cpath stroke='%232a79ed' d='M10 18h1'/%3E%3Cpath stroke='%23287eef' d='M11 18h1'/%3E%3Cpath stroke='%232481f1' d='M12 18h1'/%3E%3Cpath stroke='%232182f1' d='M13 18h1'/%3E%3Cpath stroke='%231c80f1' d='M14 18h1'/%3E%3Cpath stroke='%231880f3' d='M15 18h1'/%3E%3Cpath stroke='%23117af2' d='M16 18h1'/%3E%3Cpath stroke='%230c6eed' d='M17 18h1'/%3E%3Cpath stroke='%230a5ddd' d='M18 18h1'/%3E%3Cpath stroke='%23265dc1' d='M19 18h1'/%3E%3Cpath stroke='%2393b4f2' d='M0 19h1m19 0h1'/%3E%3Cpath stroke='%23d1ddf4' d='M1 19h1'/%3E%3Cpath stroke='%232e61ca' d='M2 19h1'/%3E%3Cpath stroke='%23134bbf' d='M3 19h1'/%3E%3Cpath stroke='%23164fc2' d='M4 19h1'/%3E%3Cpath stroke='%231950c1' d='M5 19h1'/%3E%3Cpath stroke='%231b52c1' d='M6 19h1'/%3E%3Cpath stroke='%231a52c3' d='M7 19h1'/%3E%3Cpath stroke='%231954c6' d='M8 19h1'/%3E%3Cpath stroke='%231b58c9' d='M9 19h1'/%3E%3Cpath stroke='%231858c8' d='M10 19h1'/%3E%3Cpath stroke='%23165bcd' d='M11 19h1'/%3E%3Cpath stroke='%23145cd0' d='M12 19h1'/%3E%3Cpath stroke='%23135cd0' d='M13 19h1'/%3E%3Cpath stroke='%230f58cc' d='M14 19h1'/%3E%3Cpath stroke='%230d5ad2' d='M15 19h1'/%3E%3Cpath stroke='%230b58d1' d='M16 19h1'/%3E%3Cpath stroke='%230951cb' d='M17 19h1'/%3E%3Cpath stroke='%23265ec3' d='M18 19h1'/%3E%3Cpath stroke='%23d0daee' d='M19 19h1'/%3E%3Cpath stroke='%2393b3f2' d='M1 20h1m17 0h1'/%3E%3Cpath stroke='%23fefefe' d='M14 20h1'/%3E%3Cpath stroke='%23fdfdfd' d='M15 20h1m1 0h1'/%3E%3Cpath stroke='%23fcfcfc' d='M16 20h1'/%3E%3Cpath stroke='%23f2f5fc' d='M18 20h1'/%3E%3C/svg%3E\")}.title-bar-controls button[aria-label=Minimize]:not(:disabled):active{background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 -0.5 21 21' shape-rendering='crispEdges'%3E%3Cpath stroke='%2393b1ed' d='M1 0h1m17 0h1'/%3E%3Cpath stroke='%23f4f6fd' d='M2 0h1m15 0h1M0 2h1m19 0h1M0 18h1m19 0h1M2 20h1m15 0h1'/%3E%3Cpath stroke='%23fff' d='M3 0h15M0 3h1m19 0h1M0 4h1m19 0h1M0 5h1m19 0h1M0 6h1m19 0h1M0 7h1m19 0h1M0 8h1m19 0h1M0 9h1m19 0h1M0 10h1m19 0h1M0 11h1m19 0h1M0 12h1m19 0h1M0 13h1m19 0h1M0 14h1m19 0h1M0 15h1m19 0h1M0 16h1m19 0h1M0 17h1m19 0h1M3 20h15'/%3E%3Cpath stroke='%23a7bcee' d='M0 1h1m19 0h1'/%3E%3Cpath stroke='%23cfd3da' d='M1 1h1'/%3E%3Cpath stroke='%231f3b5f' d='M2 1h1M1 2h1'/%3E%3Cpath stroke='%23002453' d='M3 1h1M1 4h1'/%3E%3Cpath stroke='%23002557' d='M4 1h1'/%3E%3Cpath stroke='%23002658' d='M5 1h1'/%3E%3Cpath stroke='%2300285c' d='M6 1h1'/%3E%3Cpath stroke='%23002a61' d='M7 1h1'/%3E%3Cpath stroke='%23002d67' d='M8 1h1'/%3E%3Cpath stroke='%23002f6b' d='M9 1h1'/%3E%3Cpath stroke='%23002f6c' d='M10 1h1M1 10h1'/%3E%3Cpath stroke='%23003273' d='M11 1h1'/%3E%3Cpath stroke='%23003478' d='M12 1h1M5 2h1'/%3E%3Cpath stroke='%2300357b' d='M13 1h1M2 5h1m-2 8h1'/%3E%3Cpath stroke='%2300377f' d='M14 1h1M6 2h1'/%3E%3Cpath stroke='%23003780' d='M15 1h1'/%3E%3Cpath stroke='%23003984' d='M16 1h1'/%3E%3Cpath stroke='%23003882' d='M17 1h1M3 3h1'/%3E%3Cpath stroke='%231f5295' d='M18 1h1'/%3E%3Cpath stroke='%23cfdae9' d='M19 1h1'/%3E%3Cpath stroke='%23002a62' d='M2 2h1'/%3E%3Cpath stroke='%23003070' d='M3 2h1'/%3E%3Cpath stroke='%23003275' d='M4 2h1'/%3E%3Cpath stroke='%23003883' d='M7 2h1M1 17h1'/%3E%3Cpath stroke='%23003a88' d='M8 2h1'/%3E%3Cpath stroke='%23003d8f' d='M9 2h1M2 9h1'/%3E%3Cpath stroke='%23003e90' d='M10 2h1'/%3E%3Cpath stroke='%23004094' d='M11 2h1'/%3E%3Cpath stroke='%23004299' d='M12 2h1M2 12h1'/%3E%3Cpath stroke='%2300439b' d='M13 2h1'/%3E%3Cpath stroke='%2300449e' d='M14 2h1M2 14h1'/%3E%3Cpath stroke='%2300459f' d='M15 2h1'/%3E%3Cpath stroke='%230045a1' d='M16 2h1m1 0h1M2 17h1'/%3E%3Cpath stroke='%230045a0' d='M17 2h1M2 15h1'/%3E%3Cpath stroke='%231f5aa8' d='M19 2h1'/%3E%3Cpath stroke='%23002452' d='M1 3h1'/%3E%3Cpath stroke='%23003170' d='M2 3h1'/%3E%3Cpath stroke='%23003b8b' d='M4 3h1M3 4h1'/%3E%3Cpath stroke='%23003c8f' d='M5 3h1'/%3E%3Cpath stroke='%23003e94' d='M6 3h1'/%3E%3Cpath stroke='%23004099' d='M7 3h1'/%3E%3Cpath stroke='%2300429d' d='M8 3h1'/%3E%3Cpath stroke='%230044a2' d='M9 3h1'/%3E%3Cpath stroke='%230046a5' d='M10 3h1'/%3E%3Cpath stroke='%230048a8' d='M11 3h1'/%3E%3Cpath stroke='%230049ab' d='M12 3h1m-3 2h1'/%3E%3Cpath stroke='%23004aac' d='M13 3h1'/%3E%3Cpath stroke='%23004aad' d='M14 3h1'/%3E%3Cpath stroke='%23004bae' d='M15 3h2m1 0h1M3 14h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23004baf' d='M17 3h1m-5 2h1m-7 5h1m-5 7h1m-1 1h1'/%3E%3Cpath stroke='%23004bad' d='M19 3h1M3 13h1m-1 6h1'/%3E%3Cpath stroke='%23037' d='M2 4h1m-2 8h1'/%3E%3Cpath stroke='%23003d92' d='M4 4h1'/%3E%3Cpath stroke='%23003f97' d='M5 4h1M4 5h1'/%3E%3Cpath stroke='%2300419d' d='M6 4h1M4 6h1'/%3E%3Cpath stroke='%230043a1' d='M7 4h1'/%3E%3Cpath stroke='%230045a4' d='M8 4h1'/%3E%3Cpath stroke='%230047a8' d='M9 4h1M4 9h1'/%3E%3Cpath stroke='%230048ab' d='M10 4h1m-7 6h1'/%3E%3Cpath stroke='%230049ad' d='M11 4h1m-2 2h1m-6 5h1'/%3E%3Cpath stroke='%23004aae' d='M12 4h1m-1 1h1m-2 1h1m-6 5h1m-3 1h2'/%3E%3Cpath stroke='%23004cb0' d='M13 4h1m0 1h1m-8 6h1m-4 2h1'/%3E%3Cpath stroke='%23004db1' d='M14 4h3m-2 1h2m-4 1h4M7 12h1m-4 2h1m-1 1h1m-1 1h2'/%3E%3Cpath stroke='%23004db2' d='M17 4h3m-3 1h3m-2 1h2m-8 1h1m6 0h1m-9 1h1m-4 3h1m-5 6h2m-2 1h4m-4 1h4'/%3E%3Cpath stroke='%23002555' d='M1 5h1'/%3E%3Cpath stroke='%23003d90' d='M3 5h1'/%3E%3Cpath stroke='%2300409c' d='M5 5h1'/%3E%3Cpath stroke='%230042a1' d='M6 5h1M5 6h1'/%3E%3Cpath stroke='%230044a5' d='M7 5h1M6 6h1'/%3E%3Cpath stroke='%230046a8' d='M8 5h1M5 8h1'/%3E%3Cpath stroke='%230047aa' d='M9 5h1'/%3E%3Cpath stroke='%230049ac' d='M11 5h1m-7 5h1m-2 1h1m-2 1h1'/%3E%3Cpath stroke='%2300275a' d='M1 6h1'/%3E%3Cpath stroke='%23003781' d='M2 6h1m-2 9h1'/%3E%3Cpath stroke='%23003f95' d='M3 6h1'/%3E%3Cpath stroke='%230045a9' d='M7 6h1'/%3E%3Cpath stroke='%230046aa' d='M8 6h1M6 7h1'/%3E%3Cpath stroke='%230047ac' d='M9 6h1M7 7h1'/%3E%3Cpath stroke='%23004bb0' d='M12 6h1M8 9h1m-3 3h1'/%3E%3Cpath stroke='%23004eb3' d='M17 6h1m-5 1h1m4 0h1m0 1h1M10 9h1m-2 1h1m-3 6h1m-2 1h2m0 2h1'/%3E%3Cpath stroke='%2300295f' d='M1 7h1'/%3E%3Cpath stroke='%23003985' d='M2 7h1'/%3E%3Cpath stroke='%2300419b' d='M3 7h1'/%3E%3Cpath stroke='%230043a2' d='M4 7h1'/%3E%3Cpath stroke='%230044a6' d='M5 7h1'/%3E%3Cpath stroke='%230048ad' d='M8 7h1M6 9h1'/%3E%3Cpath stroke='%230049ae' d='M9 7h1M7 8h2m-3 2h1'/%3E%3Cpath stroke='%23004aaf' d='M10 7h1M9 8h1M7 9h1'/%3E%3Cpath stroke='%23004cb1' d='M11 7h1m-2 1h1M9 9h1m-2 1h1'/%3E%3Cpath stroke='%23004fb3' d='M14 7h1'/%3E%3Cpath stroke='%23004fb4' d='M15 7h3m-6 1h1m5 0h1m0 1h1M8 12h1m-1 6h1m0 1h1'/%3E%3Cpath stroke='%23002b63' d='M1 8h1'/%3E%3Cpath stroke='%23003b8a' d='M2 8h1'/%3E%3Cpath stroke='%2300439f' d='M3 8h1'/%3E%3Cpath stroke='%230045a5' d='M4 8h1'/%3E%3Cpath stroke='%230047ab' d='M6 8h1M5 9h1'/%3E%3Cpath stroke='%230050b5' d='M13 8h2m1 0h2m-7 1h1m-2 1h1m8 0h1M9 11h1m-2 5h1m-1 1h1m1 2h1'/%3E%3Cpath stroke='%230051b6' d='M15 8h1m2 1h1m0 2h1m-1 1h1m-1 5h1M9 18h1m1 1h1'/%3E%3Cpath stroke='%23002d68' d='M1 9h1'/%3E%3Cpath stroke='%230045a3' d='M3 9h1'/%3E%3Cpath stroke='%230052b7' d='M12 9h1m-2 1h1m-2 1h1m-2 1h1m9 1h1m-8 6h2m3 0h1'/%3E%3Cpath stroke='%230053b8' d='M13 9h1m2 0h2m0 1h1m0 4h1M9 16h1m9 0h1M9 17h1m0 1h1m3 1h1m1 0h1'/%3E%3Cpath stroke='%230054b9' d='M14 9h2m2 9h1m-4 1h1'/%3E%3Cpath stroke='%23003f93' d='M2 10h1'/%3E%3Cpath stroke='%230047a7' d='M3 10h1'/%3E%3Cpath stroke='%230055ba' d='M12 10h1m4 0h1m-7 1h1m6 0h1m-9 6h1m0 1h1'/%3E%3Cpath stroke='%230056bb' d='M13 10h1m2 0h1m1 2h1m-9 4h1'/%3E%3Cpath stroke='%230057bc' d='M14 10h2m-5 2h1m6 5h1m-7 1h1m4 0h1'/%3E%3Cpath stroke='%23003172' d='M1 11h1'/%3E%3Cpath stroke='%23004095' d='M2 11h1'/%3E%3Cpath stroke='%230048aa' d='M3 11h1'/%3E%3Cpath stroke='%230058bd' d='M12 11h1m4 0h1m0 2h1m-6 5h1'/%3E%3Cpath stroke='%230059be' d='M13 11h1m2 0h1m-6 5h1m6 0h1m-5 2h1m1 0h1'/%3E%3Cpath stroke='%23005abf' d='M14 11h2m-4 1h1m4 0h1m-6 5h1m2 1h1'/%3E%3Cpath stroke='%230055b9' d='M10 12h1'/%3E%3Cpath stroke='%23005cc1' d='M13 12h1m2 0h1m-5 1h1m4 0h1m-5 4h1'/%3E%3Cpath stroke='%23005dc2' d='M14 12h1m-3 2h1m4 0h1m-6 1h1m4 1h1m-4 1h1m1 0h1'/%3E%3Cpath stroke='%23005ec3' d='M15 12h1m-3 1h1m2 0h1m0 2h1m-5 1h1m1 1h1'/%3E%3Cpath stroke='%2300449d' d='M2 13h1'/%3E%3Cpath stroke='%2378a2d8' d='M5 13h7m-7 1h7m-7 1h7'/%3E%3Cpath stroke='%23005fc4' d='M14 13h1m-2 1h1m2 0h1m-4 1h1'/%3E%3Cpath stroke='%230060c5' d='M15 13h1m-2 1h1m1 1h1m-2 1h1'/%3E%3Cpath stroke='%2300367e' d='M1 14h1'/%3E%3Cpath stroke='%230061c6' d='M15 14h1m-2 1h1'/%3E%3Cpath stroke='%230059bd' d='M18 14h1'/%3E%3Cpath stroke='%230062c6' d='M15 15h1'/%3E%3Cpath stroke='%23005abe' d='M18 15h1'/%3E%3Cpath stroke='%230054b8' d='M19 15h1'/%3E%3Cpath stroke='%23003881' d='M1 16h1'/%3E%3Cpath stroke='%230046a1' d='M2 16h1'/%3E%3Cpath stroke='%23004eb2' d='M6 16h1'/%3E%3Cpath stroke='%23005cc0' d='M12 16h1'/%3E%3Cpath stroke='%23005fc3' d='M14 16h1'/%3E%3Cpath stroke='%230060c4' d='M16 16h1'/%3E%3Cpath stroke='%230058bc' d='M11 17h1'/%3E%3Cpath stroke='%23005bc0' d='M17 17h1'/%3E%3Cpath stroke='%231f5294' d='M1 18h1'/%3E%3Cpath stroke='%230046a2' d='M2 18h1'/%3E%3Cpath stroke='%231f66be' d='M19 18h1'/%3E%3Cpath stroke='%23a7bef0' d='M0 19h1m0 1h1m17 0h1'/%3E%3Cpath stroke='%23cfdae8' d='M1 19h1'/%3E%3Cpath stroke='%231f5ba9' d='M2 19h1'/%3E%3Cpath stroke='%231f66bf' d='M18 19h1'/%3E%3Cpath stroke='%23cfdef1' d='M19 19h1'/%3E%3Cpath stroke='%2393b4f2' d='M20 19h1'/%3E%3C/svg%3E\")}.title-bar-controls button[aria-label=Maximize]{background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 -0.5 21 21' shape-rendering='crispEdges'%3E%3Cpath stroke='%236696eb' d='M1 0h1'/%3E%3Cpath stroke='%23e5edfb' d='M2 0h1'/%3E%3Cpath stroke='%23fff' d='M3 0h16M0 2h1M0 3h1m19 0h1M0 4h1m19 0h1M0 5h1m4 0h11m4 0h1M0 6h1m4 0h11m4 0h1M0 7h1m4 0h11m4 0h1M0 8h1m4 0h1m9 0h1m4 0h1M0 9h1m4 0h1m9 0h1m4 0h1M0 10h1m4 0h1m9 0h1m4 0h1M0 11h1m4 0h1m9 0h1m4 0h1M0 12h1m4 0h1m9 0h1m4 0h1M0 13h1m4 0h1m9 0h1m4 0h1M0 14h1m4 0h1m9 0h1m4 0h1M0 15h1m4 0h11m4 0h1M0 16h1m19 0h1M0 17h1m19 0h1m-1 1h1M2 20h16'/%3E%3Cpath stroke='%236694eb' d='M19 0h1'/%3E%3Cpath stroke='%236693e9' d='M0 1h1m19 0h1'/%3E%3Cpath stroke='%23dce5fd' d='M1 1h1'/%3E%3Cpath stroke='%23739af8' d='M2 1h1'/%3E%3Cpath stroke='%23608cf7' d='M3 1h1M2 8h1'/%3E%3Cpath stroke='%235584f6' d='M4 1h1'/%3E%3Cpath stroke='%234d7ef6' d='M5 1h1M1 6h1m5 4h1'/%3E%3Cpath stroke='%23487af5' d='M6 1h1'/%3E%3Cpath stroke='%234276f5' d='M7 1h1M3 14h1'/%3E%3Cpath stroke='%234478f5' d='M8 1h1m5 3h1M2 12h1'/%3E%3Cpath stroke='%233e73f5' d='M9 1h2'/%3E%3Cpath stroke='%233b71f5' d='M11 1h2'/%3E%3Cpath stroke='%23336cf4' d='M13 1h2'/%3E%3Cpath stroke='%23306af4' d='M15 1h1'/%3E%3Cpath stroke='%232864f4' d='M16 1h1'/%3E%3Cpath stroke='%231f5def' d='M17 1h1'/%3E%3Cpath stroke='%233467e0' d='M18 1h1'/%3E%3Cpath stroke='%23d2dbf2' d='M19 1h1'/%3E%3Cpath stroke='%23769cf8' d='M1 2h1'/%3E%3Cpath stroke='%2390aff9' d='M2 2h1'/%3E%3Cpath stroke='%2394b2f9' d='M3 2h1'/%3E%3Cpath stroke='%2385a7f8' d='M4 2h1'/%3E%3Cpath stroke='%23759cf8' d='M5 2h1'/%3E%3Cpath stroke='%236e97f8' d='M6 2h1M2 6h1'/%3E%3Cpath stroke='%236892f7' d='M7 2h1'/%3E%3Cpath stroke='%236690f7' d='M8 2h1'/%3E%3Cpath stroke='%23628ef7' d='M9 2h1m0 1h1'/%3E%3Cpath stroke='%235f8cf7' d='M10 2h1'/%3E%3Cpath stroke='%235e8bf7' d='M11 2h1'/%3E%3Cpath stroke='%235988f6' d='M12 2h1'/%3E%3Cpath stroke='%235685f6' d='M13 2h1'/%3E%3Cpath stroke='%235082f6' d='M14 2h1'/%3E%3Cpath stroke='%23497cf5' d='M15 2h1'/%3E%3Cpath stroke='%233f75f5' d='M16 2h1m-2 2h1'/%3E%3Cpath stroke='%23326bf2' d='M17 2h1'/%3E%3Cpath stroke='%23235ce3' d='M18 2h1'/%3E%3Cpath stroke='%23305cc5' d='M19 2h1'/%3E%3Cpath stroke='%23e5ecfb' d='M20 2h1'/%3E%3Cpath stroke='%236590f7' d='M1 3h1'/%3E%3Cpath stroke='%2397b4f9' d='M2 3h1'/%3E%3Cpath stroke='%239ab7fa' d='M3 3h1'/%3E%3Cpath stroke='%2389aaf9' d='M4 3h1M2 4h1'/%3E%3Cpath stroke='%237aa0f8' d='M5 3h1'/%3E%3Cpath stroke='%23729af8' d='M6 3h1'/%3E%3Cpath stroke='%236d95f8' d='M7 3h1'/%3E%3Cpath stroke='%236892f8' d='M8 3h1M2 7h1'/%3E%3Cpath stroke='%23658ff7' d='M9 3h1'/%3E%3Cpath stroke='%23618df7' d='M11 3h1'/%3E%3Cpath stroke='%235d8af7' d='M12 3h1M3 9h1'/%3E%3Cpath stroke='%235987f6' d='M13 3h1M2 9h1'/%3E%3Cpath stroke='%235283f6' d='M14 3h1'/%3E%3Cpath stroke='%234c7ef6' d='M15 3h1'/%3E%3Cpath stroke='%234377f5' d='M16 3h1'/%3E%3Cpath stroke='%23376ef2' d='M17 3h1'/%3E%3Cpath stroke='%23285fe3' d='M18 3h1'/%3E%3Cpath stroke='%231546b9' d='M19 3h1'/%3E%3Cpath stroke='%235886f6' d='M1 4h1'/%3E%3Cpath stroke='%238dadf9' d='M3 4h1'/%3E%3Cpath stroke='%237fa3f8' d='M4 4h1'/%3E%3Cpath stroke='%237199f8' d='M5 4h1M4 5h1'/%3E%3Cpath stroke='%236a93f8' d='M6 4h1M4 6h1M3 7h1'/%3E%3Cpath stroke='%23648ef7' d='M7 4h1'/%3E%3Cpath stroke='%235e8af7' d='M8 4h1'/%3E%3Cpath stroke='%235986f7' d='M9 4h1m-6 6h1'/%3E%3Cpath stroke='%235482f6' d='M10 4h1'/%3E%3Cpath stroke='%235180f6' d='M11 4h1'/%3E%3Cpath stroke='%234b7cf5' d='M12 4h1'/%3E%3Cpath stroke='%234a7cf5' d='M13 4h1'/%3E%3Cpath stroke='%233a72f4' d='M16 4h1'/%3E%3Cpath stroke='%23346cf2' d='M17 4h1'/%3E%3Cpath stroke='%232a61e3' d='M18 4h1'/%3E%3Cpath stroke='%231848bb' d='M19 4h1'/%3E%3Cpath stroke='%235282f6' d='M1 5h1m4 6h1m-3 1h1'/%3E%3Cpath stroke='%23799ff8' d='M2 5h1'/%3E%3Cpath stroke='%237ca1f8' d='M3 5h1'/%3E%3Cpath stroke='%23316bf4' d='M16 5h1M3 16h1'/%3E%3Cpath stroke='%233069f1' d='M17 5h1'/%3E%3Cpath stroke='%232c62e4' d='M18 5h1'/%3E%3Cpath stroke='%231d4cbc' d='M19 5h1m-1 1h1'/%3E%3Cpath stroke='%237099f8' d='M3 6h1'/%3E%3Cpath stroke='%232d69f5' d='M16 6h1'/%3E%3Cpath stroke='%232e69f2' d='M17 6h1'/%3E%3Cpath stroke='%232c63e5' d='M18 6h1'/%3E%3Cpath stroke='%234679f5' d='M1 7h1M1 8h1'/%3E%3Cpath stroke='%23658ff8' d='M4 7h1'/%3E%3Cpath stroke='%232a68f5' d='M16 7h1'/%3E%3Cpath stroke='%232c69f2' d='M17 7h1'/%3E%3Cpath stroke='%232a62e4' d='M18 7h1'/%3E%3Cpath stroke='%231c4cbd' d='M19 7h1'/%3E%3Cpath stroke='%23628df8' d='M3 8h1'/%3E%3Cpath stroke='%23608bf7' d='M4 8h1'/%3E%3Cpath stroke='%235482f7' d='M6 8h1'/%3E%3Cpath stroke='%234e7cf6' d='M7 8h1'/%3E%3Cpath stroke='%234778f6' d='M8 8h1'/%3E%3Cpath stroke='%234174f5' d='M9 8h1'/%3E%3Cpath stroke='%233a71f5' d='M10 8h1'/%3E%3Cpath stroke='%23346ef4' d='M11 8h1'/%3E%3Cpath stroke='%232d6bf5' d='M12 8h1'/%3E%3Cpath stroke='%232869f5' d='M13 8h1'/%3E%3Cpath stroke='%232467f5' d='M14 8h1'/%3E%3Cpath stroke='%232567f5' d='M16 8h1'/%3E%3Cpath stroke='%232968f2' d='M17 8h1'/%3E%3Cpath stroke='%232963e4' d='M18 8h1'/%3E%3Cpath stroke='%231b4bbd' d='M19 8h1'/%3E%3Cpath stroke='%233c72f4' d='M1 9h1'/%3E%3Cpath stroke='%235d89f7' d='M4 9h1'/%3E%3Cpath stroke='%235381f6' d='M6 9h1'/%3E%3Cpath stroke='%234e7ef6' d='M7 9h1'/%3E%3Cpath stroke='%23477af5' d='M8 9h1'/%3E%3Cpath stroke='%234178f5' d='M9 9h1'/%3E%3Cpath stroke='%233a74f5' d='M10 9h1'/%3E%3Cpath stroke='%233472f5' d='M11 9h1'/%3E%3Cpath stroke='%232c6ff5' d='M12 9h1'/%3E%3Cpath stroke='%23276cf5' d='M13 9h1'/%3E%3Cpath stroke='%23236af6' d='M14 9h1'/%3E%3Cpath stroke='%232268f5' d='M16 9h1'/%3E%3Cpath stroke='%232569f2' d='M17 9h1'/%3E%3Cpath stroke='%232562e6' d='M18 9h1'/%3E%3Cpath stroke='%23194bbe' d='M19 9h1'/%3E%3Cpath stroke='%23376ef4' d='M1 10h1'/%3E%3Cpath stroke='%235181f6' d='M2 10h1'/%3E%3Cpath stroke='%235785f7' d='M3 10h1'/%3E%3Cpath stroke='%235281f6' d='M6 10h1'/%3E%3Cpath stroke='%23477bf6' d='M8 10h1'/%3E%3Cpath stroke='%234179f6' d='M9 10h1'/%3E%3Cpath stroke='%233b77f5' d='M10 10h1'/%3E%3Cpath stroke='%233474f5' d='M11 10h1'/%3E%3Cpath stroke='%232c72f6' d='M12 10h1'/%3E%3Cpath stroke='%23266ff6' d='M13 10h1'/%3E%3Cpath stroke='%23226df6' d='M14 10h1'/%3E%3Cpath stroke='%231f6af6' d='M16 10h1'/%3E%3Cpath stroke='%23216af3' d='M17 10h1'/%3E%3Cpath stroke='%232162e6' d='M18 10h1'/%3E%3Cpath stroke='%231649be' d='M19 10h1'/%3E%3Cpath stroke='%23326bf4' d='M1 11h1'/%3E%3Cpath stroke='%234b7df5' d='M2 11h1'/%3E%3Cpath stroke='%235483f6' d='M3 11h1'/%3E%3Cpath stroke='%235684f7' d='M4 11h1'/%3E%3Cpath stroke='%234d80f6' d='M7 11h1'/%3E%3Cpath stroke='%23487df6' d='M8 11h1'/%3E%3Cpath stroke='%23427cf6' d='M9 11h1'/%3E%3Cpath stroke='%233c7af6' d='M10 11h1'/%3E%3Cpath stroke='%233478f6' d='M11 11h1'/%3E%3Cpath stroke='%232d76f6' d='M12 11h1'/%3E%3Cpath stroke='%232673f7' d='M13 11h1'/%3E%3Cpath stroke='%232171f7' d='M14 11h1'/%3E%3Cpath stroke='%231c6df6' d='M16 11h1'/%3E%3Cpath stroke='%231c6af4' d='M17 11h1'/%3E%3Cpath stroke='%231c61e6' d='M18 11h1'/%3E%3Cpath stroke='%231248bf' d='M19 11h1'/%3E%3Cpath stroke='%232b66f4' d='M1 12h1'/%3E%3Cpath stroke='%234e7ff6' d='M3 12h1'/%3E%3Cpath stroke='%235182f6' d='M6 12h1'/%3E%3Cpath stroke='%234d81f7' d='M7 12h1'/%3E%3Cpath stroke='%23487ff6' d='M8 12h1'/%3E%3Cpath stroke='%23437ff6' d='M9 12h1'/%3E%3Cpath stroke='%233d7ef6' d='M10 12h1'/%3E%3Cpath stroke='%23357cf6' d='M11 12h1'/%3E%3Cpath stroke='%232d7af7' d='M12 12h1'/%3E%3Cpath stroke='%232677f7' d='M13 12h1'/%3E%3Cpath stroke='%232174f7' d='M14 12h1'/%3E%3Cpath stroke='%23186ef7' d='M16 12h1'/%3E%3Cpath stroke='%23186af4' d='M17 12h1'/%3E%3Cpath stroke='%23165fe7' d='M18 12h1'/%3E%3Cpath stroke='%230f47c0' d='M19 12h1'/%3E%3Cpath stroke='%232562f3' d='M1 13h1'/%3E%3Cpath stroke='%233d73f4' d='M2 13h1'/%3E%3Cpath stroke='%23487bf5' d='M3 13h1'/%3E%3Cpath stroke='%234e80f6' d='M4 13h1'/%3E%3Cpath stroke='%234e81f6' d='M6 13h1'/%3E%3Cpath stroke='%234b80f6' d='M7 13h1'/%3E%3Cpath stroke='%23477ff6' d='M8 13h1'/%3E%3Cpath stroke='%23427ff6' d='M9 13h1'/%3E%3Cpath stroke='%233c7ff6' d='M10 13h1'/%3E%3Cpath stroke='%23367ff7' d='M11 13h1'/%3E%3Cpath stroke='%232d7cf7' d='M12 13h1'/%3E%3Cpath stroke='%232679f8' d='M13 13h1'/%3E%3Cpath stroke='%232077f7' d='M14 13h1'/%3E%3Cpath stroke='%23166ff7' d='M16 13h1'/%3E%3Cpath stroke='%231369f4' d='M17 13h1'/%3E%3Cpath stroke='%23105de8' d='M18 13h1'/%3E%3Cpath stroke='%230a44bf' d='M19 13h1'/%3E%3Cpath stroke='%231e5df3' d='M1 14h1'/%3E%3Cpath stroke='%23366ef4' d='M2 14h1'/%3E%3Cpath stroke='%23497bf5' d='M4 14h1'/%3E%3Cpath stroke='%234a7ef7' d='M6 14h1'/%3E%3Cpath stroke='%23487ef6' d='M7 14h1'/%3E%3Cpath stroke='%23457ff6' d='M8 14h1'/%3E%3Cpath stroke='%234180f6' d='M9 14h1'/%3E%3Cpath stroke='%233b7ff6' d='M10 14h1'/%3E%3Cpath stroke='%23357ff7' d='M11 14h1'/%3E%3Cpath stroke='%232d7df7' d='M12 14h1'/%3E%3Cpath stroke='%23257af8' d='M13 14h1'/%3E%3Cpath stroke='%231e77f8' d='M14 14h1'/%3E%3Cpath stroke='%23116df7' d='M16 14h1'/%3E%3Cpath stroke='%230d66f4' d='M17 14h1m-3 3h1'/%3E%3Cpath stroke='%230b59e7' d='M18 14h1'/%3E%3Cpath stroke='%230641c0' d='M19 14h1m-6 5h1'/%3E%3Cpath stroke='%231859f3' d='M1 15h1'/%3E%3Cpath stroke='%232e68f4' d='M2 15h1'/%3E%3Cpath stroke='%233a71f4' d='M3 15h1'/%3E%3Cpath stroke='%234277f5' d='M4 15h1'/%3E%3Cpath stroke='%230e6cf8' d='M16 15h1'/%3E%3Cpath stroke='%230963f4' d='M17 15h1'/%3E%3Cpath stroke='%230556e7' d='M18 15h1'/%3E%3Cpath stroke='%23023fbf' d='M19 15h1'/%3E%3Cpath stroke='%231456f3' d='M1 16h1'/%3E%3Cpath stroke='%232562f4' d='M2 16h1'/%3E%3Cpath stroke='%233971f4' d='M4 16h1'/%3E%3Cpath stroke='%233d74f5' d='M5 16h1'/%3E%3Cpath stroke='%233d74f6' d='M6 16h1'/%3E%3Cpath stroke='%233b75f5' d='M7 16h1'/%3E%3Cpath stroke='%233976f5' d='M8 16h1'/%3E%3Cpath stroke='%233777f5' d='M9 16h1'/%3E%3Cpath stroke='%233278f6' d='M10 16h1'/%3E%3Cpath stroke='%232c78f7' d='M11 16h1'/%3E%3Cpath stroke='%232577f7' d='M12 16h1'/%3E%3Cpath stroke='%231f76f7' d='M13 16h1'/%3E%3Cpath stroke='%231972f7' d='M14 16h1'/%3E%3Cpath stroke='%23116ef8' d='M15 16h1'/%3E%3Cpath stroke='%230b68f7' d='M16 16h1'/%3E%3Cpath stroke='%230560f4' d='M17 16h1'/%3E%3Cpath stroke='%230253e6' d='M18 16h1'/%3E%3Cpath stroke='%23013dbe' d='M19 16h1'/%3E%3Cpath stroke='%230e50ed' d='M1 17h1'/%3E%3Cpath stroke='%231c5bef' d='M2 17h1'/%3E%3Cpath stroke='%232863f0' d='M3 17h1'/%3E%3Cpath stroke='%232f68f0' d='M4 17h1'/%3E%3Cpath stroke='%23336bf1' d='M5 17h1'/%3E%3Cpath stroke='%23346cf1' d='M6 17h1'/%3E%3Cpath stroke='%23316cf2' d='M7 17h1'/%3E%3Cpath stroke='%23316df2' d='M8 17h1'/%3E%3Cpath stroke='%232e6ff2' d='M9 17h1'/%3E%3Cpath stroke='%232a70f2' d='M10 17h1'/%3E%3Cpath stroke='%232570f3' d='M11 17h1'/%3E%3Cpath stroke='%231f6ff3' d='M12 17h1'/%3E%3Cpath stroke='%23196df4' d='M13 17h1'/%3E%3Cpath stroke='%23136af4' d='M14 17h1'/%3E%3Cpath stroke='%230760f3' d='M16 17h1'/%3E%3Cpath stroke='%23025af0' d='M17 17h1'/%3E%3Cpath stroke='%23004de2' d='M18 17h1'/%3E%3Cpath stroke='%23003ab9' d='M19 17h1'/%3E%3Cpath stroke='%23e5eefd' d='M0 18h1'/%3E%3Cpath stroke='%23285edf' d='M1 18h1'/%3E%3Cpath stroke='%23134fdf' d='M2 18h1'/%3E%3Cpath stroke='%231b55df' d='M3 18h1'/%3E%3Cpath stroke='%23215ae2' d='M4 18h1'/%3E%3Cpath stroke='%23255ce1' d='M5 18h1'/%3E%3Cpath stroke='%23265de0' d='M6 18h1'/%3E%3Cpath stroke='%23245ce1' d='M7 18h1'/%3E%3Cpath stroke='%23235ee2' d='M8 18h1'/%3E%3Cpath stroke='%23215ee2' d='M9 18h1'/%3E%3Cpath stroke='%231e5ee2' d='M10 18h1'/%3E%3Cpath stroke='%231b5fe5' d='M11 18h1'/%3E%3Cpath stroke='%23165ee5' d='M12 18h1'/%3E%3Cpath stroke='%23135de6' d='M13 18h1'/%3E%3Cpath stroke='%230e5be5' d='M14 18h1'/%3E%3Cpath stroke='%230958e6' d='M15 18h1'/%3E%3Cpath stroke='%230454e6' d='M16 18h1'/%3E%3Cpath stroke='%23014ee2' d='M17 18h1'/%3E%3Cpath stroke='%230045d3' d='M18 18h1'/%3E%3Cpath stroke='%231f4eb8' d='M19 18h1'/%3E%3Cpath stroke='%23679ef6' d='M0 19h1'/%3E%3Cpath stroke='%23d0daf1' d='M1 19h1'/%3E%3Cpath stroke='%232856c3' d='M2 19h1'/%3E%3Cpath stroke='%230d3fb6' d='M3 19h1'/%3E%3Cpath stroke='%231144bd' d='M4 19h1'/%3E%3Cpath stroke='%231245bb' d='M5 19h1'/%3E%3Cpath stroke='%231445b9' d='M6 19h1'/%3E%3Cpath stroke='%231244b9' d='M7 19h1'/%3E%3Cpath stroke='%231345bc' d='M8 19h1'/%3E%3Cpath stroke='%231346bd' d='M9 19h1'/%3E%3Cpath stroke='%231045be' d='M10 19h1'/%3E%3Cpath stroke='%230d45c0' d='M11 19h1'/%3E%3Cpath stroke='%230a45c1' d='M12 19h1'/%3E%3Cpath stroke='%230844c3' d='M13 19h1'/%3E%3Cpath stroke='%23033fc0' d='M15 19h1'/%3E%3Cpath stroke='%23013fc3' d='M16 19h1'/%3E%3Cpath stroke='%23003bbe' d='M17 19h1'/%3E%3Cpath stroke='%231f4eb9' d='M18 19h1'/%3E%3Cpath stroke='%23cfd8ed' d='M19 19h1'/%3E%3Cpath stroke='%23669bf5' d='M20 19h1M1 20h1'/%3E%3Cpath stroke='%23e5edfd' d='M18 20h1'/%3E%3Cpath stroke='%236699f3' d='M19 20h1'/%3E%3C/svg%3E\")}.title-bar-controls button[aria-label=Maximize]:hover{background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 -0.5 21 21' shape-rendering='crispEdges'%3E%3Cpath stroke='%23afc2ef' d='M1 0h1m17 0h1M0 1h1m19 0h1M0 19h1m19 0h1M1 20h1m17 0h1'/%3E%3Cpath stroke='%23f4f6fd' d='M2 0h1m17 2h1M0 18h1m17 2h1'/%3E%3Cpath stroke='%23fff' d='M3 0h15M0 3h1m19 0h1M0 4h1m19 0h1M0 5h1m4 0h11m4 0h1M0 6h1m4 0h11m4 0h1M0 7h1m4 0h11m4 0h1M0 8h1m4 0h1m9 0h1m4 0h1M0 9h1m4 0h1m9 0h1m4 0h1M0 10h1m4 0h1m9 0h1m4 0h1M0 11h1m4 0h1m9 0h1m4 0h1M0 12h1m4 0h1m9 0h1m4 0h1M0 13h1m4 0h1m9 0h1m4 0h1M0 14h1m4 0h1m9 0h1m4 0h1M0 15h1m4 0h11m4 0h1M0 16h1m19 0h1M0 17h1m19 0h1M3 20h15'/%3E%3Cpath stroke='%23f5f7fd' d='M18 0h1M0 2h1m19 16h1M2 20h1'/%3E%3Cpath stroke='%23dce7ff' d='M1 1h1'/%3E%3Cpath stroke='%2372a1ff' d='M2 1h1m4 3h1'/%3E%3Cpath stroke='%236a9cff' d='M3 1h1'/%3E%3Cpath stroke='%235f94ff' d='M4 1h1M4 11h1'/%3E%3Cpath stroke='%23558eff' d='M5 1h1M3 12h1'/%3E%3Cpath stroke='%23518bff' d='M6 1h1'/%3E%3Cpath stroke='%234a86ff' d='M7 1h1'/%3E%3Cpath stroke='%234b87ff' d='M8 1h1M2 12h1'/%3E%3Cpath stroke='%234684ff' d='M9 1h2'/%3E%3Cpath stroke='%234482ff' d='M11 1h1m4 1h1M1 9h1m0 4h1'/%3E%3Cpath stroke='%234080ff' d='M12 1h1M3 15h1'/%3E%3Cpath stroke='%233b7cff' d='M13 1h1'/%3E%3Cpath stroke='%233a7bff' d='M14 1h1'/%3E%3Cpath stroke='%233678ff' d='M15 1h1'/%3E%3Cpath stroke='%232e73ff' d='M16 1h1'/%3E%3Cpath stroke='%23276cf9' d='M17 1h1'/%3E%3Cpath stroke='%233a73e7' d='M18 1h1'/%3E%3Cpath stroke='%23d3ddf3' d='M19 1h1'/%3E%3Cpath stroke='%2373a1ff' d='M1 2h1'/%3E%3Cpath stroke='%2397b9ff' d='M2 2h1'/%3E%3Cpath stroke='%239cbdff' d='M3 2h1'/%3E%3Cpath stroke='%2390b5ff' d='M4 2h1'/%3E%3Cpath stroke='%2382acff' d='M5 2h1M5 4h1'/%3E%3Cpath stroke='%237ba7ff' d='M6 2h1M2 6h1'/%3E%3Cpath stroke='%2375a3ff' d='M7 2h1'/%3E%3Cpath stroke='%236f9fff' d='M8 2h1M3 8h1'/%3E%3Cpath stroke='%236c9dff' d='M9 2h1M1 3h1'/%3E%3Cpath stroke='%23689bff' d='M10 2h1M3 9h1'/%3E%3Cpath stroke='%236599ff' d='M11 2h1m0 1h1'/%3E%3Cpath stroke='%236095ff' d='M12 2h1m0 1h1'/%3E%3Cpath stroke='%235d93ff' d='M13 2h1'/%3E%3Cpath stroke='%23568eff' d='M14 2h1'/%3E%3Cpath stroke='%234f8aff' d='M15 2h1M3 13h1m0 1h1'/%3E%3Cpath stroke='%233878fb' d='M17 2h1'/%3E%3Cpath stroke='%232969eb' d='M18 2h1'/%3E%3Cpath stroke='%233566cb' d='M19 2h1'/%3E%3Cpath stroke='%239ebeff' d='M2 3h1'/%3E%3Cpath stroke='%23a4c2ff' d='M3 3h1'/%3E%3Cpath stroke='%2399baff' d='M4 3h1M3 4h1'/%3E%3Cpath stroke='%238ab0ff' d='M5 3h1'/%3E%3Cpath stroke='%2382abff' d='M6 3h1'/%3E%3Cpath stroke='%2379a6ff' d='M7 3h1'/%3E%3Cpath stroke='%2374a3ff' d='M8 3h1'/%3E%3Cpath stroke='%2371a0ff' d='M9 3h1'/%3E%3Cpath stroke='%236d9eff' d='M10 3h1M4 8h1'/%3E%3Cpath stroke='%23699bff' d='M11 3h1'/%3E%3Cpath stroke='%235a91ff' d='M14 3h1M2 10h1m1 2h1'/%3E%3Cpath stroke='%23538cff' d='M15 3h1M2 11h1'/%3E%3Cpath stroke='%234986ff' d='M16 3h1'/%3E%3Cpath stroke='%233d7cfc' d='M17 3h1'/%3E%3Cpath stroke='%232e6cea' d='M18 3h1'/%3E%3Cpath stroke='%231b52c2' d='M19 3h1'/%3E%3Cpath stroke='%236296ff' d='M1 4h1'/%3E%3Cpath stroke='%2391b5ff' d='M2 4h1'/%3E%3Cpath stroke='%238fb4ff' d='M4 4h1'/%3E%3Cpath stroke='%237aa6ff' d='M6 4h1'/%3E%3Cpath stroke='%236b9dff' d='M8 4h1'/%3E%3Cpath stroke='%236598ff' d='M9 4h1'/%3E%3Cpath stroke='%235f95ff' d='M10 4h1m-5 6h1'/%3E%3Cpath stroke='%235b92ff' d='M11 4h1'/%3E%3Cpath stroke='%23548dff' d='M12 4h1M1 6h1m2 7h1'/%3E%3Cpath stroke='%23528cff' d='M13 4h1'/%3E%3Cpath stroke='%234c88ff' d='M14 4h1'/%3E%3Cpath stroke='%234785ff' d='M15 4h1'/%3E%3Cpath stroke='%234280ff' d='M16 4h1'/%3E%3Cpath stroke='%233b7afb' d='M17 4h1'/%3E%3Cpath stroke='%23316fec' d='M18 4h1'/%3E%3Cpath stroke='%231f55c3' d='M19 4h1'/%3E%3Cpath stroke='%235990ff' d='M1 5h1'/%3E%3Cpath stroke='%2385adff' d='M2 5h1'/%3E%3Cpath stroke='%238bb1ff' d='M3 5h1'/%3E%3Cpath stroke='%2384acff' d='M4 5h1'/%3E%3Cpath stroke='%23397aff' d='M16 5h1M1 11h1'/%3E%3Cpath stroke='%233979fc' d='M17 5h1'/%3E%3Cpath stroke='%233370ec' d='M18 5h1m-1 1h1'/%3E%3Cpath stroke='%232357c3' d='M19 5h1'/%3E%3Cpath stroke='%2381aaff' d='M3 6h1'/%3E%3Cpath stroke='%237aa7ff' d='M4 6h1'/%3E%3Cpath stroke='%233679ff' d='M16 6h1'/%3E%3Cpath stroke='%233879fc' d='M17 6h1'/%3E%3Cpath stroke='%232358c5' d='M19 6h1'/%3E%3Cpath stroke='%234e89ff' d='M1 7h1'/%3E%3Cpath stroke='%2371a1ff' d='M2 7h1'/%3E%3Cpath stroke='%2377a5ff' d='M3 7h1'/%3E%3Cpath stroke='%2374a2ff' d='M4 7h1'/%3E%3Cpath stroke='%23337aff' d='M16 7h1'/%3E%3Cpath stroke='%23367bfc' d='M17 7h1'/%3E%3Cpath stroke='%233372ed' d='M18 7h1'/%3E%3Cpath stroke='%232359c5' d='M19 7h1'/%3E%3Cpath stroke='%234d88ff' d='M1 8h1'/%3E%3Cpath stroke='%23699cff' d='M2 8h1'/%3E%3Cpath stroke='%236398ff' d='M6 8h1'/%3E%3Cpath stroke='%235c93ff' d='M7 8h1m-2 3h1'/%3E%3Cpath stroke='%23548fff' d='M8 8h1'/%3E%3Cpath stroke='%234d8cff' d='M9 8h1'/%3E%3Cpath stroke='%23468aff' d='M10 8h1'/%3E%3Cpath stroke='%233f86ff' d='M11 8h1'/%3E%3Cpath stroke='%233983ff' d='M12 8h1'/%3E%3Cpath stroke='%233380ff' d='M13 8h1'/%3E%3Cpath stroke='%232f7fff' d='M14 8h1'/%3E%3Cpath stroke='%233280ff' d='M16 8h1'/%3E%3Cpath stroke='%233580fc' d='M17 8h1'/%3E%3Cpath stroke='%233276ed' d='M18 8h1'/%3E%3Cpath stroke='%23235ac6' d='M19 8h1'/%3E%3Cpath stroke='%236196ff' d='M2 9h1m3 0h1m-4 1h1'/%3E%3Cpath stroke='%23689aff' d='M4 9h1'/%3E%3Cpath stroke='%235b93ff' d='M7 9h1'/%3E%3Cpath stroke='%235491ff' d='M8 9h1'/%3E%3Cpath stroke='%234f90ff' d='M9 9h1'/%3E%3Cpath stroke='%234890ff' d='M10 9h1'/%3E%3Cpath stroke='%23428eff' d='M11 9h1'/%3E%3Cpath stroke='%233b8dff' d='M12 9h1'/%3E%3Cpath stroke='%23348aff' d='M13 9h1'/%3E%3Cpath stroke='%233189ff' d='M14 9h1'/%3E%3Cpath stroke='%233188ff' d='M16 9h1'/%3E%3Cpath stroke='%233385fc' d='M17 9h1'/%3E%3Cpath stroke='%233079ed' d='M18 9h1'/%3E%3Cpath stroke='%23215cc8' d='M19 9h1'/%3E%3Cpath stroke='%233f7fff' d='M1 10h1'/%3E%3Cpath stroke='%236397ff' d='M4 10h1'/%3E%3Cpath stroke='%235993ff' d='M7 10h1'/%3E%3Cpath stroke='%235492ff' d='M8 10h1'/%3E%3Cpath stroke='%235093ff' d='M9 10h1'/%3E%3Cpath stroke='%234a95ff' d='M10 10h1'/%3E%3Cpath stroke='%234496ff' d='M11 10h1'/%3E%3Cpath stroke='%233d96ff' d='M12 10h1'/%3E%3Cpath stroke='%233694ff' d='M13 10h1'/%3E%3Cpath stroke='%233193ff' d='M14 10h1'/%3E%3Cpath stroke='%233090ff' d='M16 10h1'/%3E%3Cpath stroke='%23328cfc' d='M17 10h1'/%3E%3Cpath stroke='%232e7def' d='M18 10h1'/%3E%3Cpath stroke='%231e5dc9' d='M19 10h1'/%3E%3Cpath stroke='%235c92ff' d='M3 11h1'/%3E%3Cpath stroke='%235792ff' d='M7 11h1m-1 1h1'/%3E%3Cpath stroke='%235594ff' d='M8 11h1'/%3E%3Cpath stroke='%235298ff' d='M9 11h1'/%3E%3Cpath stroke='%234d9cff' d='M10 11h1'/%3E%3Cpath stroke='%23479eff' d='M11 11h1'/%3E%3Cpath stroke='%23409fff' d='M12 11h1'/%3E%3Cpath stroke='%23379fff' d='M13 11h1'/%3E%3Cpath stroke='%23339dff' d='M14 11h1'/%3E%3Cpath stroke='%232e97ff' d='M16 11h1'/%3E%3Cpath stroke='%232e91fc' d='M17 11h1'/%3E%3Cpath stroke='%232a80f0' d='M18 11h1'/%3E%3Cpath stroke='%231b5dcb' d='M19 11h1'/%3E%3Cpath stroke='%233275ff' d='M1 12h1'/%3E%3Cpath stroke='%235991ff' d='M6 12h1'/%3E%3Cpath stroke='%235596ff' d='M8 12h1'/%3E%3Cpath stroke='%23529cff' d='M9 12h1'/%3E%3Cpath stroke='%234fa1ff' d='M10 12h1'/%3E%3Cpath stroke='%234aa6ff' d='M11 12h1'/%3E%3Cpath stroke='%2342a9ff' d='M12 12h1'/%3E%3Cpath stroke='%233aa9ff' d='M13 12h1'/%3E%3Cpath stroke='%2334a7ff' d='M14 12h1'/%3E%3Cpath stroke='%232ca0ff' d='M16 12h1'/%3E%3Cpath stroke='%232a96fd' d='M17 12h1'/%3E%3Cpath stroke='%232581f1' d='M18 12h1'/%3E%3Cpath stroke='%23185dcc' d='M19 12h1'/%3E%3Cpath stroke='%232d72ff' d='M1 13h1m0 3h1'/%3E%3Cpath stroke='%235790ff' d='M6 13h1'/%3E%3Cpath stroke='%235490ff' d='M7 13h1'/%3E%3Cpath stroke='%235597ff' d='M8 13h1'/%3E%3Cpath stroke='%23539fff' d='M9 13h1'/%3E%3Cpath stroke='%234fa4ff' d='M10 13h1'/%3E%3Cpath stroke='%234aaaff' d='M11 13h1'/%3E%3Cpath stroke='%2344afff' d='M12 13h1'/%3E%3Cpath stroke='%233eb1ff' d='M13 13h1'/%3E%3Cpath stroke='%2337afff' d='M14 13h1'/%3E%3Cpath stroke='%2329a4ff' d='M16 13h1'/%3E%3Cpath stroke='%232599fd' d='M17 13h1'/%3E%3Cpath stroke='%231e80f2' d='M18 13h1'/%3E%3Cpath stroke='%23145bcd' d='M19 13h1'/%3E%3Cpath stroke='%23276eff' d='M1 14h1'/%3E%3Cpath stroke='%233d7dff' d='M2 14h1'/%3E%3Cpath stroke='%234985ff' d='M3 14h1'/%3E%3Cpath stroke='%23528dff' d='M6 14h1'/%3E%3Cpath stroke='%23518fff' d='M7 14h1'/%3E%3Cpath stroke='%235196ff' d='M8 14h1'/%3E%3Cpath stroke='%23509fff' d='M9 14h1'/%3E%3Cpath stroke='%234ea6ff' d='M10 14h1'/%3E%3Cpath stroke='%2349acff' d='M11 14h1'/%3E%3Cpath stroke='%2343b1ff' d='M12 14h1'/%3E%3Cpath stroke='%233eb4ff' d='M13 14h1'/%3E%3Cpath stroke='%2335b2ff' d='M14 14h1'/%3E%3Cpath stroke='%2324a5ff' d='M16 14h1'/%3E%3Cpath stroke='%231f97fd' d='M17 14h1'/%3E%3Cpath stroke='%231980f3' d='M18 14h1'/%3E%3Cpath stroke='%23105ace' d='M19 14h1'/%3E%3Cpath stroke='%23216aff' d='M1 15h1'/%3E%3Cpath stroke='%233578ff' d='M2 15h1'/%3E%3Cpath stroke='%234885ff' d='M4 15h1'/%3E%3Cpath stroke='%2321a3ff' d='M16 15h1'/%3E%3Cpath stroke='%231a95fd' d='M17 15h1'/%3E%3Cpath stroke='%23137cf2' d='M18 15h1'/%3E%3Cpath stroke='%230c59cf' d='M19 15h1'/%3E%3Cpath stroke='%231c66ff' d='M1 16h1'/%3E%3Cpath stroke='%233879ff' d='M3 16h1'/%3E%3Cpath stroke='%233f7eff' d='M4 16h1'/%3E%3Cpath stroke='%234483ff' d='M5 16h1'/%3E%3Cpath stroke='%234584ff' d='M6 16h1'/%3E%3Cpath stroke='%234587ff' d='M7 16h1'/%3E%3Cpath stroke='%23468eff' d='M8 16h1'/%3E%3Cpath stroke='%234696ff' d='M9 16h1'/%3E%3Cpath stroke='%23439cff' d='M10 16h1'/%3E%3Cpath stroke='%233fa3ff' d='M11 16h1'/%3E%3Cpath stroke='%233ba8ff' d='M12 16h1'/%3E%3Cpath stroke='%233af' d='M13 16h1'/%3E%3Cpath stroke='%232da9ff' d='M14 16h1'/%3E%3Cpath stroke='%2324a6ff' d='M15 16h1'/%3E%3Cpath stroke='%231d9eff' d='M16 16h1'/%3E%3Cpath stroke='%231690fd' d='M17 16h1'/%3E%3Cpath stroke='%231078f1' d='M18 16h1'/%3E%3Cpath stroke='%230b57ce' d='M19 16h1'/%3E%3Cpath stroke='%231761f9' d='M1 17h1'/%3E%3Cpath stroke='%23246bfa' d='M2 17h1'/%3E%3Cpath stroke='%232f72fb' d='M3 17h1'/%3E%3Cpath stroke='%233676fb' d='M4 17h1'/%3E%3Cpath stroke='%233a7afb' d='M5 17h1'/%3E%3Cpath stroke='%233b7bfc' d='M6 17h1'/%3E%3Cpath stroke='%233b7efc' d='M7 17h1'/%3E%3Cpath stroke='%233c84fc' d='M8 17h1'/%3E%3Cpath stroke='%233b8afc' d='M9 17h1'/%3E%3Cpath stroke='%233990fc' d='M10 17h1'/%3E%3Cpath stroke='%233695fc' d='M11 17h1'/%3E%3Cpath stroke='%233299fc' d='M12 17h1'/%3E%3Cpath stroke='%232c9cfd' d='M13 17h1'/%3E%3Cpath stroke='%23259bfd' d='M14 17h1'/%3E%3Cpath stroke='%231e97fd' d='M15 17h1'/%3E%3Cpath stroke='%231790fc' d='M16 17h1'/%3E%3Cpath stroke='%231184fa' d='M17 17h1'/%3E%3Cpath stroke='%230c6ded' d='M18 17h1'/%3E%3Cpath stroke='%230850c8' d='M19 17h1'/%3E%3Cpath stroke='%232f6ae4' d='M1 18h1'/%3E%3Cpath stroke='%231b5fe9' d='M2 18h1'/%3E%3Cpath stroke='%232163e8' d='M3 18h1'/%3E%3Cpath stroke='%232868eb' d='M4 18h1'/%3E%3Cpath stroke='%232c6aea' d='M5 18h1'/%3E%3Cpath stroke='%232e6dea' d='M6 18h1'/%3E%3Cpath stroke='%232d6deb' d='M7 18h1'/%3E%3Cpath stroke='%232c71ec' d='M8 18h1'/%3E%3Cpath stroke='%232c76ec' d='M9 18h1'/%3E%3Cpath stroke='%232a79ed' d='M10 18h1'/%3E%3Cpath stroke='%23287eef' d='M11 18h1'/%3E%3Cpath stroke='%232481f1' d='M12 18h1'/%3E%3Cpath stroke='%232182f1' d='M13 18h1'/%3E%3Cpath stroke='%231c80f1' d='M14 18h1'/%3E%3Cpath stroke='%231880f3' d='M15 18h1'/%3E%3Cpath stroke='%23117af2' d='M16 18h1'/%3E%3Cpath stroke='%230c6eed' d='M17 18h1'/%3E%3Cpath stroke='%230a5ddd' d='M18 18h1'/%3E%3Cpath stroke='%23265dc1' d='M19 18h1'/%3E%3Cpath stroke='%23d1ddf4' d='M1 19h1'/%3E%3Cpath stroke='%232e61ca' d='M2 19h1'/%3E%3Cpath stroke='%23134bbf' d='M3 19h1'/%3E%3Cpath stroke='%23164fc2' d='M4 19h1'/%3E%3Cpath stroke='%231950c1' d='M5 19h1'/%3E%3Cpath stroke='%231b52c1' d='M6 19h1'/%3E%3Cpath stroke='%231a52c3' d='M7 19h1'/%3E%3Cpath stroke='%231954c6' d='M8 19h1'/%3E%3Cpath stroke='%231b58c9' d='M9 19h1'/%3E%3Cpath stroke='%231858c8' d='M10 19h1'/%3E%3Cpath stroke='%23165bcd' d='M11 19h1'/%3E%3Cpath stroke='%23145cd0' d='M12 19h1'/%3E%3Cpath stroke='%23135cd0' d='M13 19h1'/%3E%3Cpath stroke='%230f58cc' d='M14 19h1'/%3E%3Cpath stroke='%230d5ad2' d='M15 19h1'/%3E%3Cpath stroke='%230b58d1' d='M16 19h1'/%3E%3Cpath stroke='%230951cb' d='M17 19h1'/%3E%3Cpath stroke='%23265ec3' d='M18 19h1'/%3E%3Cpath stroke='%23d0daee' d='M19 19h1'/%3E%3C/svg%3E\")}.title-bar-controls button[aria-label=Maximize]:not(:disabled):active{background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 -0.5 21 21' shape-rendering='crispEdges'%3E%3Cpath stroke='%23b3c4ef' d='M1 0h1m17 0h1M0 1h1m19 0h1M0 19h1m19 0h1M1 20h1'/%3E%3Cpath stroke='%23f4f6fd' d='M2 0h1m17 2h1M0 18h1m17 2h1'/%3E%3Cpath stroke='%23fff' d='M3 0h15M0 3h1m19 0h1M0 4h1m19 0h1M0 5h1m19 0h1M0 6h1m19 0h1M0 7h1m19 0h1M0 8h1m19 0h1M0 9h1m19 0h1M0 10h1m19 0h1M0 11h1m19 0h1M0 12h1m19 0h1M0 13h1m19 0h1M0 14h1m19 0h1M0 15h1m19 0h1M0 16h1m19 0h1M0 17h1m19 0h1M3 20h15'/%3E%3Cpath stroke='%23f5f7fd' d='M18 0h1M0 2h1m19 16h1M2 20h1'/%3E%3Cpath stroke='%23cfd3da' d='M1 1h1'/%3E%3Cpath stroke='%231f3b5f' d='M2 1h1M1 2h1'/%3E%3Cpath stroke='%23002453' d='M3 1h1M1 4h1'/%3E%3Cpath stroke='%23002557' d='M4 1h1'/%3E%3Cpath stroke='%23002658' d='M5 1h1'/%3E%3Cpath stroke='%2300285c' d='M6 1h1'/%3E%3Cpath stroke='%23002a61' d='M7 1h1'/%3E%3Cpath stroke='%23002d67' d='M8 1h1'/%3E%3Cpath stroke='%23002f6b' d='M9 1h1'/%3E%3Cpath stroke='%23002f6c' d='M10 1h1M1 10h1'/%3E%3Cpath stroke='%23003273' d='M11 1h1'/%3E%3Cpath stroke='%23003478' d='M12 1h1M5 2h1'/%3E%3Cpath stroke='%2300357b' d='M13 1h1M2 5h1m-2 8h1'/%3E%3Cpath stroke='%2300377f' d='M14 1h1M6 2h1'/%3E%3Cpath stroke='%23003780' d='M15 1h1'/%3E%3Cpath stroke='%23003984' d='M16 1h1'/%3E%3Cpath stroke='%23003882' d='M17 1h1M3 3h1'/%3E%3Cpath stroke='%231f5295' d='M18 1h1'/%3E%3Cpath stroke='%23cfdae9' d='M19 1h1'/%3E%3Cpath stroke='%23002a62' d='M2 2h1'/%3E%3Cpath stroke='%23003070' d='M3 2h1'/%3E%3Cpath stroke='%23003275' d='M4 2h1'/%3E%3Cpath stroke='%23003883' d='M7 2h1M1 17h1'/%3E%3Cpath stroke='%23003a88' d='M8 2h1'/%3E%3Cpath stroke='%23003d8f' d='M9 2h1M2 9h1'/%3E%3Cpath stroke='%23003e90' d='M10 2h1'/%3E%3Cpath stroke='%23004094' d='M11 2h1'/%3E%3Cpath stroke='%23004299' d='M12 2h1M2 12h1'/%3E%3Cpath stroke='%2300439b' d='M13 2h1'/%3E%3Cpath stroke='%2300449e' d='M14 2h1M2 14h1'/%3E%3Cpath stroke='%2300459f' d='M15 2h1'/%3E%3Cpath stroke='%230045a1' d='M16 2h1m1 0h1M2 17h1'/%3E%3Cpath stroke='%230045a0' d='M17 2h1M2 15h1'/%3E%3Cpath stroke='%231f5aa8' d='M19 2h1'/%3E%3Cpath stroke='%23002452' d='M1 3h1'/%3E%3Cpath stroke='%23003170' d='M2 3h1'/%3E%3Cpath stroke='%23003b8b' d='M4 3h1M3 4h1'/%3E%3Cpath stroke='%23003c8f' d='M5 3h1'/%3E%3Cpath stroke='%23003e94' d='M6 3h1'/%3E%3Cpath stroke='%23004099' d='M7 3h1'/%3E%3Cpath stroke='%2300429d' d='M8 3h1'/%3E%3Cpath stroke='%230044a2' d='M9 3h1'/%3E%3Cpath stroke='%230046a5' d='M10 3h1'/%3E%3Cpath stroke='%230048a8' d='M11 3h1'/%3E%3Cpath stroke='%230049ab' d='M12 3h1'/%3E%3Cpath stroke='%23004aac' d='M13 3h1'/%3E%3Cpath stroke='%23004aad' d='M14 3h1'/%3E%3Cpath stroke='%23004bae' d='M15 3h2m1 0h1M3 14h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23004baf' d='M17 3h1M7 10h1m-5 7h1m-1 1h1'/%3E%3Cpath stroke='%23004bad' d='M19 3h1M3 13h1m-1 6h1'/%3E%3Cpath stroke='%23037' d='M2 4h1m-2 8h1'/%3E%3Cpath stroke='%23003d92' d='M4 4h1'/%3E%3Cpath stroke='%23003f97' d='M5 4h1M4 5h1'/%3E%3Cpath stroke='%2300419d' d='M6 4h1M4 6h1'/%3E%3Cpath stroke='%230043a1' d='M7 4h1'/%3E%3Cpath stroke='%230045a4' d='M8 4h1'/%3E%3Cpath stroke='%230047a8' d='M9 4h1M4 9h1'/%3E%3Cpath stroke='%230048ab' d='M10 4h1m-7 6h1'/%3E%3Cpath stroke='%230049ad' d='M11 4h1'/%3E%3Cpath stroke='%23004aae' d='M12 4h1m-7 7h1m-3 1h1'/%3E%3Cpath stroke='%23004cb0' d='M13 4h1m-7 7h1m-4 2h1'/%3E%3Cpath stroke='%23004db1' d='M14 4h3m-1 1h1m-1 1h1M7 12h1m-2 1h1m-3 1h1m1 0h1m-3 1h1m-1 1h2'/%3E%3Cpath stroke='%23004db2' d='M17 4h3m-3 1h3m-2 1h2m-1 1h1m-9 1h1m-4 3h1m-5 6h2m-2 1h4m-4 1h4'/%3E%3Cpath stroke='%23002555' d='M1 5h1'/%3E%3Cpath stroke='%23003d90' d='M3 5h1'/%3E%3Cpath stroke='%2378a2d8' d='M5 5h11M5 6h11M5 7h11M5 8h1m9 0h1M5 9h1m9 0h1M5 10h1m9 0h1M5 11h1m9 0h1M5 12h1m9 0h1M5 13h1m9 0h1M5 14h1m9 0h1M5 15h11'/%3E%3Cpath stroke='%2300275a' d='M1 6h1'/%3E%3Cpath stroke='%23003781' d='M2 6h1m-2 9h1'/%3E%3Cpath stroke='%23003f95' d='M3 6h1'/%3E%3Cpath stroke='%23004eb3' d='M17 6h1m0 1h1m0 1h1M10 9h1m-2 1h1m-3 6h1m-2 1h2m0 2h1'/%3E%3Cpath stroke='%2300295f' d='M1 7h1'/%3E%3Cpath stroke='%23003985' d='M2 7h1'/%3E%3Cpath stroke='%2300419b' d='M3 7h1'/%3E%3Cpath stroke='%230043a2' d='M4 7h1'/%3E%3Cpath stroke='%23004fb4' d='M16 7h2m-6 1h1m5 0h1m0 1h1M8 12h1m-1 6h1m0 1h1'/%3E%3Cpath stroke='%23002b63' d='M1 8h1'/%3E%3Cpath stroke='%23003b8a' d='M2 8h1'/%3E%3Cpath stroke='%2300439f' d='M3 8h1'/%3E%3Cpath stroke='%230045a5' d='M4 8h1'/%3E%3Cpath stroke='%230047ab' d='M6 8h1'/%3E%3Cpath stroke='%230049ae' d='M7 8h2m-3 2h1'/%3E%3Cpath stroke='%23004aaf' d='M9 8h1M7 9h1'/%3E%3Cpath stroke='%23004cb1' d='M10 8h1M9 9h1m-2 1h1'/%3E%3Cpath stroke='%230050b5' d='M13 8h2m1 0h2m-7 1h1m-2 1h1m8 0h1M9 11h1m-2 2h1m-1 3h1m-1 1h1m1 2h1'/%3E%3Cpath stroke='%23002d68' d='M1 9h1'/%3E%3Cpath stroke='%230045a3' d='M3 9h1'/%3E%3Cpath stroke='%230048ad' d='M6 9h1'/%3E%3Cpath stroke='%23004bb0' d='M8 9h1m-3 3h1'/%3E%3Cpath stroke='%230052b7' d='M12 9h1m-2 1h1m-2 1h1m-2 1h1m9 1h1m-8 6h2m3 0h1'/%3E%3Cpath stroke='%230053b8' d='M13 9h1m2 0h2m0 1h1M9 13h1m9 1h1M9 16h1m9 0h1M9 17h1m0 1h1m3 1h1m1 0h1'/%3E%3Cpath stroke='%230054b9' d='M14 9h1m-6 5h1m8 4h1m-4 1h1'/%3E%3Cpath stroke='%230051b6' d='M18 9h1m0 2h1m-1 1h1M8 14h1m10 3h1M9 18h1m1 1h1'/%3E%3Cpath stroke='%23003f93' d='M2 10h1'/%3E%3Cpath stroke='%230047a7' d='M3 10h1'/%3E%3Cpath stroke='%230055ba' d='M12 10h1m4 0h1m-7 1h1m6 0h1m-9 6h1m0 1h1'/%3E%3Cpath stroke='%230056bb' d='M13 10h1m2 0h1m1 2h1m-9 1h1m-1 3h1'/%3E%3Cpath stroke='%230057bc' d='M14 10h1m-4 2h1m-2 2h1m7 3h1m-7 1h1m4 0h1'/%3E%3Cpath stroke='%23003172' d='M1 11h1'/%3E%3Cpath stroke='%23004095' d='M2 11h1'/%3E%3Cpath stroke='%230048aa' d='M3 11h1'/%3E%3Cpath stroke='%230049ac' d='M4 11h1m-2 1h1'/%3E%3Cpath stroke='%230058bd' d='M12 11h1m4 0h1m0 2h1m-6 5h1'/%3E%3Cpath stroke='%230059be' d='M13 11h1m2 0h1m-6 2h1m-1 3h1m6 0h1m-5 2h1m1 0h1'/%3E%3Cpath stroke='%23005abf' d='M14 11h1m-3 1h1m4 0h1m-7 2h1m0 3h1m2 1h1'/%3E%3Cpath stroke='%230055b9' d='M10 12h1'/%3E%3Cpath stroke='%23005cc1' d='M13 12h1m2 0h1m-5 1h1m4 0h1m-5 4h1'/%3E%3Cpath stroke='%23005dc2' d='M14 12h1m-3 2h1m4 0h1m-1 2h1m-4 1h1m1 0h1'/%3E%3Cpath stroke='%2300449d' d='M2 13h1'/%3E%3Cpath stroke='%23004eb2' d='M7 13h1m-2 3h1'/%3E%3Cpath stroke='%23005ec3' d='M13 13h1m2 0h1m0 2h1m-5 1h1m1 1h1'/%3E%3Cpath stroke='%23005fc4' d='M14 13h1m-2 1h1m2 0h1'/%3E%3Cpath stroke='%2300367e' d='M1 14h1'/%3E%3Cpath stroke='%23004fb3' d='M7 14h1'/%3E%3Cpath stroke='%230060c5' d='M14 14h1m1 1h1m-2 1h1'/%3E%3Cpath stroke='%230059bd' d='M18 14h1'/%3E%3Cpath stroke='%23005abe' d='M18 15h1'/%3E%3Cpath stroke='%230054b8' d='M19 15h1'/%3E%3Cpath stroke='%23003881' d='M1 16h1'/%3E%3Cpath stroke='%230046a1' d='M2 16h1'/%3E%3Cpath stroke='%23005cc0' d='M12 16h1'/%3E%3Cpath stroke='%23005fc3' d='M14 16h1'/%3E%3Cpath stroke='%230060c4' d='M16 16h1'/%3E%3Cpath stroke='%230058bc' d='M11 17h1'/%3E%3Cpath stroke='%23005bc0' d='M17 17h1'/%3E%3Cpath stroke='%231f5294' d='M1 18h1'/%3E%3Cpath stroke='%230046a2' d='M2 18h1'/%3E%3Cpath stroke='%231f66be' d='M19 18h1'/%3E%3Cpath stroke='%23cfdae8' d='M1 19h1'/%3E%3Cpath stroke='%231f5ba9' d='M2 19h1'/%3E%3Cpath stroke='%231f66bf' d='M18 19h1'/%3E%3Cpath stroke='%23cfdef1' d='M19 19h1'/%3E%3Cpath stroke='%23b2c3ee' d='M19 20h1'/%3E%3C/svg%3E\")}.title-bar-controls button[aria-label=Restore]{background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 -0.5 21 21' shape-rendering='crispEdges'%3E%3Cpath stroke='%236696eb' d='M1 0h1m17 0h1'/%3E%3Cpath stroke='%23e5edfb' d='M2 0h1'/%3E%3Cpath stroke='%23fff' d='M3 0h16M0 2h1M0 3h1m19 0h1M0 4h1m19 0h1M0 5h1m19 0h1M0 6h1m19 0h1M0 7h1m19 0h1M0 8h1m19 0h1M0 9h1m19 0h1M0 10h1m19 0h1M0 11h1m19 0h1M0 12h1m19 0h1M0 13h1m4 0h7m8 0h1M0 14h1m4 0h7m8 0h1M0 15h1m4 0h7m8 0h1M0 16h1m19 0h1M0 17h1m19 0h1m-1 1h1M2 20h16'/%3E%3Cpath stroke='%236693e9' d='M0 1h1m19 0h1'/%3E%3Cpath stroke='%23dce5fd' d='M1 1h1'/%3E%3Cpath stroke='%23739af8' d='M2 1h1'/%3E%3Cpath stroke='%23608cf7' d='M3 1h1M2 8h1'/%3E%3Cpath stroke='%235584f6' d='M4 1h1'/%3E%3Cpath stroke='%234d7ef6' d='M5 1h1M1 6h1m5 4h1'/%3E%3Cpath stroke='%23487af5' d='M6 1h1'/%3E%3Cpath stroke='%234276f5' d='M7 1h1M3 14h1'/%3E%3Cpath stroke='%234478f5' d='M8 1h1m5 3h1M2 12h1'/%3E%3Cpath stroke='%233e73f5' d='M9 1h2'/%3E%3Cpath stroke='%233b71f5' d='M11 1h2'/%3E%3Cpath stroke='%23336cf4' d='M13 1h2'/%3E%3Cpath stroke='%23306af4' d='M15 1h1'/%3E%3Cpath stroke='%232864f4' d='M16 1h1'/%3E%3Cpath stroke='%231f5def' d='M17 1h1'/%3E%3Cpath stroke='%233467e0' d='M18 1h1'/%3E%3Cpath stroke='%23d2dbf2' d='M19 1h1'/%3E%3Cpath stroke='%23769cf8' d='M1 2h1'/%3E%3Cpath stroke='%2390aff9' d='M2 2h1'/%3E%3Cpath stroke='%2394b2f9' d='M3 2h1'/%3E%3Cpath stroke='%2385a7f8' d='M4 2h1'/%3E%3Cpath stroke='%23759cf8' d='M5 2h1'/%3E%3Cpath stroke='%236e97f8' d='M6 2h1M2 6h1'/%3E%3Cpath stroke='%236892f7' d='M7 2h1'/%3E%3Cpath stroke='%236690f7' d='M8 2h1'/%3E%3Cpath stroke='%23628ef7' d='M9 2h1m0 1h1'/%3E%3Cpath stroke='%235f8cf7' d='M10 2h1'/%3E%3Cpath stroke='%235e8bf7' d='M11 2h1'/%3E%3Cpath stroke='%235988f6' d='M12 2h1'/%3E%3Cpath stroke='%235685f6' d='M13 2h1'/%3E%3Cpath stroke='%235082f6' d='M14 2h1'/%3E%3Cpath stroke='%23497cf5' d='M15 2h1'/%3E%3Cpath stroke='%233f75f5' d='M16 2h1m-2 2h1'/%3E%3Cpath stroke='%23326bf2' d='M17 2h1'/%3E%3Cpath stroke='%23235ce3' d='M18 2h1'/%3E%3Cpath stroke='%23305cc5' d='M19 2h1'/%3E%3Cpath stroke='%23e5ecfb' d='M20 2h1'/%3E%3Cpath stroke='%236590f7' d='M1 3h1'/%3E%3Cpath stroke='%2397b4f9' d='M2 3h1'/%3E%3Cpath stroke='%239ab7fa' d='M3 3h1'/%3E%3Cpath stroke='%2389aaf9' d='M4 3h1M2 4h1'/%3E%3Cpath stroke='%237aa0f8' d='M5 3h1'/%3E%3Cpath stroke='%23729af8' d='M6 3h1'/%3E%3Cpath stroke='%236d95f8' d='M7 3h1'/%3E%3Cpath stroke='%236892f8' d='M8 3h1M2 7h1'/%3E%3Cpath stroke='%23658ff7' d='M9 3h1'/%3E%3Cpath stroke='%23618df7' d='M11 3h1'/%3E%3Cpath stroke='%235d8af7' d='M12 3h1M3 9h1'/%3E%3Cpath stroke='%235987f6' d='M13 3h1M2 9h1'/%3E%3Cpath stroke='%235283f6' d='M14 3h1'/%3E%3Cpath stroke='%234c7ef6' d='M15 3h1'/%3E%3Cpath stroke='%234377f5' d='M16 3h1'/%3E%3Cpath stroke='%23376ef2' d='M17 3h1'/%3E%3Cpath stroke='%23285fe3' d='M18 3h1'/%3E%3Cpath stroke='%231546b9' d='M19 3h1'/%3E%3Cpath stroke='%235886f6' d='M1 4h1'/%3E%3Cpath stroke='%238dadf9' d='M3 4h1'/%3E%3Cpath stroke='%237fa3f8' d='M4 4h1'/%3E%3Cpath stroke='%237199f8' d='M5 4h1M4 5h1'/%3E%3Cpath stroke='%236a93f8' d='M6 4h1M4 6h1M3 7h1'/%3E%3Cpath stroke='%23648ef7' d='M7 4h1'/%3E%3Cpath stroke='%235e8af7' d='M8 4h1'/%3E%3Cpath stroke='%235986f7' d='M9 4h1M5 9h1m-2 1h1'/%3E%3Cpath stroke='%235482f6' d='M10 4h1'/%3E%3Cpath stroke='%235180f6' d='M11 4h1'/%3E%3Cpath stroke='%234b7cf5' d='M12 4h1'/%3E%3Cpath stroke='%234a7cf5' d='M13 4h1'/%3E%3Cpath stroke='%233a72f4' d='M16 4h1'/%3E%3Cpath stroke='%23346cf2' d='M17 4h1'/%3E%3Cpath stroke='%232a61e3' d='M18 4h1'/%3E%3Cpath stroke='%231848bb' d='M19 4h1'/%3E%3Cpath stroke='%235282f6' d='M1 5h1m4 6h1m-3 1h1'/%3E%3Cpath stroke='%23799ff8' d='M2 5h1'/%3E%3Cpath stroke='%237ca1f8' d='M3 5h1'/%3E%3Cpath stroke='%236791f8' d='M5 5h1'/%3E%3Cpath stroke='%23608bf7' d='M6 5h1M4 8h1'/%3E%3Cpath stroke='%23FFF' d='M7 5h1M8 5h1M6 9h1M9 5h1M8 6h1M10 5h1M11 5h1M12 5h1M13 5h1M14 5h1M15 5h1'/%3E%3Cpath stroke='%23316bf4' d='M16 5h1M3 16h1'/%3E%3Cpath stroke='%233069f1' d='M17 5h1'/%3E%3Cpath stroke='%232c62e4' d='M18 5h1'/%3E%3Cpath stroke='%231d4cbc' d='M19 5h1m-1 1h1'/%3E%3Cpath stroke='%237099f8' d='M3 6h1'/%3E%3Cpath stroke='%23628cf8' d='M5 6h1'/%3E%3Cpath stroke='%235b86f7' d='M6 6h1'/%3E%3Cpath stroke='%23FFF' d='M7 6h1M8 6h1M9 6h1M10 6h1M11 6h1M12 6h1M13 6h1M14 6h1M15 6h1'/%3E%3Cpath stroke='%232d69f5' d='M16 6h1'/%3E%3Cpath stroke='%232e69f2' d='M17 6h1'/%3E%3Cpath stroke='%232c63e5' d='M18 6h1'/%3E%3Cpath stroke='%234679f5' d='M1 7h1M1 8h1'/%3E%3Cpath stroke='%23658ff8' d='M4 7h1'/%3E%3Cpath stroke='%235e89f7' d='M5 7h1'/%3E%3Cpath stroke='%235783f7' d='M6 7h1'/%3E%3Cpath stroke='%23FFF' d='M7 7h1'/%3E%3Cpath stroke='%234375f5' d='M8 7h1M9 7h1'/%3E%3Cpath stroke='%233d71f5' d='M10 7h1'/%3E%3Cpath stroke='%23366ef4' d='M11 7h1M2 14h1'/%3E%3Cpath stroke='%232f6bf5' d='M12 7h1'/%3E%3Cpath stroke='%232b69f5' d='M13 7h1'/%3E%3Cpath stroke='%232867f5' d='M14 7h1'/%3E%3Cpath stroke='%23FFF' d='M15 7h1'/%3E%3Cpath stroke='%232a68f5' d='M16 7h1'/%3E%3Cpath stroke='%232c69f2' d='M17 7h1'/%3E%3Cpath stroke='%232a62e4' d='M18 7h1'/%3E%3Cpath stroke='%231c4cbd' d='M19 7h1'/%3E%3Cpath stroke='%23628df8' d='M3 8h1'/%3E%3Cpath stroke='%235b87f7' d='M5 8h1'/%3E%3Cpath stroke='%235482f7' d='M6 8h1'/%3E%3Cpath stroke='%23FFF' d='M7 8h1'/%3E%3Cpath stroke='%234174f5' d='M8 8h1M9 8h1'/%3E%3Cpath stroke='%233a71f5' d='M10 8h1'/%3E%3Cpath stroke='%23346ef4' d='M11 8h1'/%3E%3Cpath stroke='%232d6bf5' d='M12 8h1'/%3E%3Cpath stroke='%232869f5' d='M13 8h1'/%3E%3Cpath stroke='%232467f5' d='M14 8h1'/%3E%3Cpath stroke='%23FFF' d='M15 8h1'/%3E%3Cpath stroke='%232567f5' d='M16 8h1'/%3E%3Cpath stroke='%232968f2' d='M17 8h1'/%3E%3Cpath stroke='%232963e4' d='M18 8h1'/%3E%3Cpath stroke='%231b4bbd' d='M19 8h1'/%3E%3Cpath stroke='%233c72f4' d='M1 9h1'/%3E%3Cpath stroke='%235d89f7' d='M4 9h1'/%3E%3Cpath stroke='%23FFF' d='M5 9h1M6 9h1M7 9h1M8 9h1M9 9h1M10 9h1M11 9h1M12 9h1M13 9h1'/%3E%3Cpath stroke='%23236af6' d='M14 9h1'/%3E%3Cpath stroke='%23FFF' d='M15 9h1'/%3E%3Cpath stroke='%232268f5' d='M16 9h1'/%3E%3Cpath stroke='%232569f2' d='M17 9h1'/%3E%3Cpath stroke='%232562e6' d='M18 9h1'/%3E%3Cpath stroke='%23194bbe' d='M19 9h1'/%3E%3Cpath stroke='%23376ef4' d='M1 10h1'/%3E%3Cpath stroke='%235181f6' d='M2 10h1'/%3E%3Cpath stroke='%235785f7' d='M3 10h1M4 10h1'/%3E%3Cpath stroke='%23FFF' d='M5 10h1M6 10h1M7 10h1M8 10h1M9 10h1M10 10h1M11 10h1M12 10h1M13 10h1'/%3E%3Cpath stroke='%23226df6' d='M14 10h1'/%3E%3Cpath stroke='%23FFF' d='M15 10h1'/%3E%3Cpath stroke='%231f6af6' d='M16 10h1'/%3E%3Cpath stroke='%23216af3' d='M17 10h1'/%3E%3Cpath stroke='%232162e6' d='M18 10h1'/%3E%3Cpath stroke='%231649be' d='M19 10h1'/%3E%3Cpath stroke='%23326bf4' d='M1 11h1'/%3E%3Cpath stroke='%234b7df5' d='M2 11h1'/%3E%3Cpath stroke='%235483f6' d='M3 11h1'/%3E%3Cpath stroke='%235684f7' d='M4 11h1'/%3E%3Cpath stroke='%23FFF' d='M5 11h1'/%3E%3Cpath stroke='%234d80f6' d='M7 11h1'/%3E%3Cpath stroke='%23487df6' d='M8 11h1'/%3E%3Cpath stroke='%23427cf6' d='M9 11h1'/%3E%3Cpath stroke='%233c7af6' d='M10 11h1'/%3E%3Cpath stroke='%233478f6' d='M11 11h1'/%3E%3Cpath stroke='%232673f7' d='M12 11h1'/%3E%3Cpath stroke='%23FFF' d='M13 11h1M14 11h1M15 11h1'/%3E%3Cpath stroke='%231c6df6' d='M16 11h1'/%3E%3Cpath stroke='%231c6af4' d='M17 11h1'/%3E%3Cpath stroke='%231c61e6' d='M18 11h1'/%3E%3Cpath stroke='%231248bf' d='M19 11h1'/%3E%3Cpath stroke='%232b66f4' d='M1 12h1'/%3E%3Cpath stroke='%234e7ff6' d='M3 12h1'/%3E%3Cpath stroke='%23FFF' d='M5 12h1'/%3E%3Cpath stroke='%235182f6' d='M6 12h1'/%3E%3Cpath stroke='%234d81f7' d='M7 12h1'/%3E%3Cpath stroke='%23487ff6' d='M8 12h1'/%3E%3Cpath stroke='%23437ff6' d='M9 12h1'/%3E%3Cpath stroke='%233d7ef6' d='M10 12h1'/%3E%3Cpath stroke='%23357cf6' d='M11 12h1'/%3E%3Cpath stroke='%232677f7' d='M12 12h1'/%3E%3Cpath stroke='%23FFF' d='M13 12h1'/%3E%3Cpath stroke='%232174f7' d='M14 12h1'/%3E%3Cpath stroke='%231b71f7' d='M15 12h1'/%3E%3Cpath stroke='%23186ef7' d='M16 12h1'/%3E%3Cpath stroke='%23186af4' d='M17 12h1'/%3E%3Cpath stroke='%23165fe7' d='M18 12h1'/%3E%3Cpath stroke='%230f47c0' d='M19 12h1'/%3E%3Cpath stroke='%232562f3' d='M1 13h1'/%3E%3Cpath stroke='%233d73f4' d='M2 13h1'/%3E%3Cpath stroke='%23487bf5' d='M3 13h1'/%3E%3Cpath stroke='%234e80f6' d='M4 13h1M6 13h1M7 13h1'/%3E%3Cpath stroke='%23437ff6' d='M8 13h1'/%3E%3Cpath stroke='%232d7df7' d='M9 13h1'/%3E%3Cpath stroke='%232d7cf7' d='M10 13h1M11 13h1'/%3E%3Cpath stroke='%232679f8' d='M12 13h1'/%3E%3Cpath stroke='%23FFF' d='M13 13h1'/%3E%3Cpath stroke='%232077f7' d='M14 13h1'/%3E%3Cpath stroke='%231973f7' d='M15 13h1'/%3E%3Cpath stroke='%23166ff7' d='M16 13h1'/%3E%3Cpath stroke='%231369f4' d='M17 13h1'/%3E%3Cpath stroke='%23105de8' d='M18 13h1'/%3E%3Cpath stroke='%230a44bf' d='M19 13h1'/%3E%3Cpath stroke='%231e5df3' d='M1 14h1'/%3E%3Cpath stroke='%23497bf5' d='M4 14h1M6 14h1'/%3E%3Cpath stroke='%232d7df7' d='M7 14h1M8 14h1M9 14h1M10 14h1M11 14h1'/%3E%3Cpath stroke='%23257af8' d='M12 14h1'/%3E%3Cpath stroke='%23FFF' d='M13 14h1'/%3E%3Cpath stroke='%231e77f8' d='M14 14h1'/%3E%3Cpath stroke='%231773f8' d='M15 14h1'/%3E%3Cpath stroke='%23116df7' d='M16 14h1'/%3E%3Cpath stroke='%230d66f4' d='M17 14h1m-3 3h1'/%3E%3Cpath stroke='%230b59e7' d='M18 14h1'/%3E%3Cpath stroke='%230641c0' d='M19 14h1m-6 5h1'/%3E%3Cpath stroke='%231859f3' d='M1 15h1'/%3E%3Cpath stroke='%232e68f4' d='M2 15h1'/%3E%3Cpath stroke='%233a71f4' d='M3 15h1'/%3E%3Cpath stroke='%234277f5' d='M4 15h1'/%3E%3Cpath stroke='%23FFF' d='M11 15h1M12 15h1M13 15h1'/%3E%3Cpath stroke='%231d77f8' d='M14 15h1'/%3E%3Cpath stroke='%231573f8' d='M15 15h1'/%3E%3Cpath stroke='%230e6cf8' d='M16 15h1'/%3E%3Cpath stroke='%230963f4' d='M17 15h1'/%3E%3Cpath stroke='%230556e7' d='M18 15h1'/%3E%3Cpath stroke='%23023fbf' d='M19 15h1'/%3E%3Cpath stroke='%231456f3' d='M1 16h1'/%3E%3Cpath stroke='%232562f4' d='M2 16h1'/%3E%3Cpath stroke='%233971f4' d='M4 16h1'/%3E%3Cpath stroke='%233d74f5' d='M5 16h1'/%3E%3Cpath stroke='%233d74f6' d='M6 16h1'/%3E%3Cpath stroke='%233b75f5' d='M7 16h1'/%3E%3Cpath stroke='%233976f5' d='M8 16h1'/%3E%3Cpath stroke='%233777f5' d='M9 16h1'/%3E%3Cpath stroke='%233278f6' d='M10 16h1'/%3E%3Cpath stroke='%232c78f7' d='M11 16h1'/%3E%3Cpath stroke='%232577f7' d='M12 16h1'/%3E%3Cpath stroke='%231f76f7' d='M13 16h1'/%3E%3Cpath stroke='%231972f7' d='M14 16h1'/%3E%3Cpath stroke='%23116ef8' d='M15 16h1'/%3E%3Cpath stroke='%230b68f7' d='M16 16h1'/%3E%3Cpath stroke='%230560f4' d='M17 16h1'/%3E%3Cpath stroke='%230253e6' d='M18 16h1'/%3E%3Cpath stroke='%23013dbe' d='M19 16h1'/%3E%3Cpath stroke='%230e50ed' d='M1 17h1'/%3E%3Cpath stroke='%231c5bef' d='M2 17h1'/%3E%3Cpath stroke='%232863f0' d='M3 17h1'/%3E%3Cpath stroke='%232f68f0' d='M4 17h1'/%3E%3Cpath stroke='%23336bf1' d='M5 17h1'/%3E%3Cpath stroke='%23346cf1' d='M6 17h1'/%3E%3Cpath stroke='%23316cf2' d='M7 17h1'/%3E%3Cpath stroke='%23316df2' d='M8 17h1'/%3E%3Cpath stroke='%232e6ff2' d='M9 17h1'/%3E%3Cpath stroke='%232a70f2' d='M10 17h1'/%3E%3Cpath stroke='%232570f3' d='M11 17h1'/%3E%3Cpath stroke='%231f6ff3' d='M12 17h1'/%3E%3Cpath stroke='%23196df4' d='M13 17h1'/%3E%3Cpath stroke='%23136af4' d='M14 17h1'/%3E%3Cpath stroke='%230760f3' d='M16 17h1'/%3E%3Cpath stroke='%23025af0' d='M17 17h1'/%3E%3Cpath stroke='%23004de2' d='M18 17h1'/%3E%3Cpath stroke='%23003ab9' d='M19 17h1'/%3E%3Cpath stroke='%23e5eefd' d='M0 18h1'/%3E%3Cpath stroke='%23285edf' d='M1 18h1'/%3E%3Cpath stroke='%23134fdf' d='M2 18h1'/%3E%3Cpath stroke='%231b55df' d='M3 18h1'/%3E%3Cpath stroke='%23215ae2' d='M4 18h1'/%3E%3Cpath stroke='%23255ce1' d='M5 18h1'/%3E%3Cpath stroke='%23265de0' d='M6 18h1'/%3E%3Cpath stroke='%23245ce1' d='M7 18h1'/%3E%3Cpath stroke='%23235ee2' d='M8 18h1'/%3E%3Cpath stroke='%23215ee2' d='M9 18h1'/%3E%3Cpath stroke='%231e5ee2' d='M10 18h1'/%3E%3Cpath stroke='%231b5fe5' d='M11 18h1'/%3E%3Cpath stroke='%23165ee5' d='M12 18h1'/%3E%3Cpath stroke='%23135de6' d='M13 18h1'/%3E%3Cpath stroke='%230e5be5' d='M14 18h1'/%3E%3Cpath stroke='%230958e6' d='M15 18h1'/%3E%3Cpath stroke='%230454e6' d='M16 18h1'/%3E%3Cpath stroke='%23014ee2' d='M17 18h1'/%3E%3Cpath stroke='%230045d3' d='M18 18h1'/%3E%3Cpath stroke='%231f4eb8' d='M19 18h1'/%3E%3Cpath stroke='%23679ef6' d='M0 19h1m19 0h1'/%3E%3Cpath stroke='%23d0daf1' d='M1 19h1'/%3E%3Cpath stroke='%232856c3' d='M2 19h1'/%3E%3Cpath stroke='%230d3fb6' d='M3 19h1'/%3E%3Cpath stroke='%231144bd' d='M4 19h1'/%3E%3Cpath stroke='%231245bb' d='M5 19h1'/%3E%3Cpath stroke='%231445b9' d='M6 19h1'/%3E%3Cpath stroke='%231244b9' d='M7 19h1'/%3E%3Cpath stroke='%231345bc' d='M8 19h1'/%3E%3Cpath stroke='%231346bd' d='M9 19h1'/%3E%3Cpath stroke='%231045be' d='M10 19h1'/%3E%3Cpath stroke='%230d45c0' d='M11 19h1'/%3E%3Cpath stroke='%230a45c1' d='M12 19h1'/%3E%3Cpath stroke='%230844c3' d='M13 19h1'/%3E%3Cpath stroke='%23033fc0' d='M15 19h1'/%3E%3Cpath stroke='%23013fc3' d='M16 19h1'/%3E%3Cpath stroke='%23003bbe' d='M17 19h1'/%3E%3Cpath stroke='%231f4eb9' d='M18 19h1'/%3E%3Cpath stroke='%23cfd8ed' d='M19 19h1'/%3E%3Cpath stroke='%23669bf5' d='M1 20h1m17 0h1'/%3E%3Cpath stroke='%23e5edfd' d='M18 20h1'/%3E%3Cpath stroke='%23FFF' d='M5 15h9M5 9h9M5 10h9M5.5 8.5v7M13.5 8.5v7M7 5h9M7 6h9M14 11h2M7.5 5v4M15.5 5v6'/%3E%3C/svg%3E\")}.title-bar-controls button[aria-label=Restore]:hover{background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 -0.5 21 21' shape-rendering='crispEdges'%3E%3Cpath stroke='%2393b1ed' d='M1 0h1m17 0h1'/%3E%3Cpath stroke='%23f3f6fd' d='M2 0h1m17 2h1M0 18h1'/%3E%3Cpath stroke='%23fff' d='M3 0h15M0 3h1m19 0h1M0 4h1m19 0h1M0 5h1m19 0h1M0 6h1m19 0h1M0 7h1m19 0h1M0 8h1m19 0h1M0 9h1m19 0h1M0 10h1m19 0h1M0 11h1m19 0h1M0 12h1m19 0h1M0 13h1m4 0h7m8 0h1M0 14h1m4 0h7m8 0h1M0 15h1m4 0h7m8 0h1M0 16h1m19 0h1M0 17h1m19 0h1M3 20h11'/%3E%3Cpath stroke='%23f5f7fd' d='M18 0h1M0 2h1m19 16h1M2 20h1'/%3E%3Cpath stroke='%2393b0ec' d='M0 1h1m19 0h1'/%3E%3Cpath stroke='%23dce7ff' d='M1 1h1'/%3E%3Cpath stroke='%2372a1ff' d='M2 1h1m4 3h1M5 6h1'/%3E%3Cpath stroke='%236a9cff' d='M3 1h1'/%3E%3Cpath stroke='%235f94ff' d='M4 1h1M4 11h2'/%3E%3Cpath stroke='%23558eff' d='M5 1h1M3 12h1'/%3E%3Cpath stroke='%23518bff' d='M6 1h1m3 4h1'/%3E%3Cpath stroke='%234a86ff' d='M7 1h1'/%3E%3Cpath stroke='%234b87ff' d='M8 1h1m2 4h1M2 12h1'/%3E%3Cpath stroke='%234684ff' d='M9 1h2'/%3E%3Cpath stroke='%234482ff' d='M11 1h1m4 1h1m-5 3h1M1 9h1m0 4h1'/%3E%3Cpath stroke='%234080ff' d='M12 1h1M3 15h1'/%3E%3Cpath stroke='%233b7cff' d='M13 1h1'/%3E%3Cpath stroke='%233a7bff' d='M14 1h1'/%3E%3Cpath stroke='%233678ff' d='M15 1h1'/%3E%3Cpath stroke='%232e73ff' d='M16 1h1'/%3E%3Cpath stroke='%23276cf9' d='M17 1h1'/%3E%3Cpath stroke='%233a73e7' d='M18 1h1'/%3E%3Cpath stroke='%23d3ddf3' d='M19 1h1'/%3E%3Cpath stroke='%2373a1ff' d='M1 2h1'/%3E%3Cpath stroke='%2397b9ff' d='M2 2h1'/%3E%3Cpath stroke='%239cbdff' d='M3 2h1'/%3E%3Cpath stroke='%2390b5ff' d='M4 2h1'/%3E%3Cpath stroke='%2382acff' d='M5 2h1M5 4h1'/%3E%3Cpath stroke='%237ba7ff' d='M6 2h1M2 6h1'/%3E%3Cpath stroke='%2375a3ff' d='M7 2h1'/%3E%3Cpath stroke='%236f9fff' d='M8 2h1M3 8h1'/%3E%3Cpath stroke='%236c9dff' d='M9 2h1M1 3h1'/%3E%3Cpath stroke='%23689bff' d='M10 2h1M5 8h1M3 9h1'/%3E%3Cpath stroke='%236599ff' d='M11 2h1m0 1h1M5 9h1'/%3E%3Cpath stroke='%236095ff' d='M12 2h1m0 1h1M8 5h1'/%3E%3Cpath stroke='%235d93ff' d='M13 2h1'/%3E%3Cpath stroke='%23568eff' d='M14 2h1'/%3E%3Cpath stroke='%234f8aff' d='M15 2h1M3 13h1m0 1h1'/%3E%3Cpath stroke='%233878fb' d='M17 2h1'/%3E%3Cpath stroke='%232969eb' d='M18 2h1'/%3E%3Cpath stroke='%233566cb' d='M19 2h1'/%3E%3Cpath stroke='%239ebeff' d='M2 3h1'/%3E%3Cpath stroke='%23a4c2ff' d='M3 3h1'/%3E%3Cpath stroke='%2399baff' d='M4 3h1M3 4h1'/%3E%3Cpath stroke='%238ab0ff' d='M5 3h1'/%3E%3Cpath stroke='%2382abff' d='M6 3h1'/%3E%3Cpath stroke='%2379a6ff' d='M7 3h1'/%3E%3Cpath stroke='%2374a3ff' d='M8 3h1'/%3E%3Cpath stroke='%2371a0ff' d='M9 3h1'/%3E%3Cpath stroke='%236d9eff' d='M10 3h1M5 7h1M4 8h1'/%3E%3Cpath stroke='%23699bff' d='M11 3h1'/%3E%3Cpath stroke='%235a91ff' d='M14 3h1M2 10h1m1 2h1'/%3E%3Cpath stroke='%23538cff' d='M15 3h1M2 11h1'/%3E%3Cpath stroke='%234986ff' d='M16 3h1'/%3E%3Cpath stroke='%233d7cfc' d='M17 3h1'/%3E%3Cpath stroke='%232e6cea' d='M18 3h1'/%3E%3Cpath stroke='%231b52c2' d='M19 3h1'/%3E%3Cpath stroke='%236296ff' d='M1 4h1'/%3E%3Cpath stroke='%2391b5ff' d='M2 4h1'/%3E%3Cpath stroke='%238fb4ff' d='M4 4h1'/%3E%3Cpath stroke='%237aa6ff' d='M6 4h1'/%3E%3Cpath stroke='%236b9dff' d='M8 4h1'/%3E%3Cpath stroke='%236598ff' d='M9 4h1'/%3E%3Cpath stroke='%235f95ff' d='M10 4h1M7 7h1m-2 3h1'/%3E%3Cpath stroke='%235b92ff' d='M11 4h1'/%3E%3Cpath stroke='%23548dff' d='M12 4h1M1 6h1m2 7h1'/%3E%3Cpath stroke='%23528cff' d='M13 4h1'/%3E%3Cpath stroke='%234c88ff' d='M14 4h1m-5 2h1'/%3E%3Cpath stroke='%234785ff' d='M15 4h1'/%3E%3Cpath stroke='%234280ff' d='M16 4h1'/%3E%3Cpath stroke='%233b7afb' d='M17 4h1'/%3E%3Cpath stroke='%23316fec' d='M18 4h1'/%3E%3Cpath stroke='%231f55c3' d='M19 4h1'/%3E%3Cpath stroke='%235990ff' d='M1 5h1m7 0h1'/%3E%3Cpath stroke='%2385adff' d='M2 5h1'/%3E%3Cpath stroke='%238bb1ff' d='M3 5h1'/%3E%3Cpath stroke='%2384acff' d='M4 5h1'/%3E%3Cpath stroke='%2378a5ff' d='M5 5h1'/%3E%3Cpath stroke='%2370a0ff' d='M6 5h1'/%3E%3Cpath stroke='%23679aff' d='M7 5h1'/%3E%3Cpath stroke='%234180ff' d='M13 5h1'/%3E%3Cpath stroke='%233d7eff' d='M14 5h1'/%3E%3Cpath stroke='%233b7bff' d='M15 5h1'/%3E%3Cpath stroke='%23397aff' d='M16 5h1M1 11h1'/%3E%3Cpath stroke='%233979fc' d='M17 5h1'/%3E%3Cpath stroke='%233370ec' d='M18 5h1m-1 1h1'/%3E%3Cpath stroke='%232357c3' d='M19 5h1'/%3E%3Cpath stroke='%2381aaff' d='M3 6h1'/%3E%3Cpath stroke='%237aa7ff' d='M4 6h1'/%3E%3Cpath stroke='%236b9cff' d='M6 6h1'/%3E%3Cpath stroke='%236297ff' d='M7 6h1m-3 4h1'/%3E%3Cpath stroke='%235c93ff' d='M8 6h1M7 8h1m-2 3h1'/%3E%3Cpath stroke='%23548eff' d='M9 6h1'/%3E%3Cpath stroke='%234483ff' d='M11 6h1M5 16h1'/%3E%3Cpath stroke='%233d7fff' d='M12 6h1'/%3E%3Cpath stroke='%23387bff' d='M13 6h1'/%3E%3Cpath stroke='%233679ff' d='M14 6h1m1 0h1'/%3E%3Cpath stroke='%233579ff' d='M15 6h1'/%3E%3Cpath stroke='%233879fc' d='M17 6h1'/%3E%3Cpath stroke='%232358c5' d='M19 6h1'/%3E%3Cpath stroke='%234e89ff' d='M1 7h1'/%3E%3Cpath stroke='%2371a1ff' d='M2 7h1'/%3E%3Cpath stroke='%2377a5ff' d='M3 7h1'/%3E%3Cpath stroke='%2374a2ff' d='M4 7h1'/%3E%3Cpath stroke='%23669aff' d='M6 7h1'/%3E%3Cpath stroke='%235890ff' d='M8 7h1'/%3E%3Cpath stroke='%23508dff' d='M9 7h1'/%3E%3Cpath stroke='%234989ff' d='M10 7h1'/%3E%3Cpath stroke='%234183ff' d='M11 7h1'/%3E%3Cpath stroke='%233a7fff' d='M12 7h1'/%3E%3Cpath stroke='%23357bff' d='M13 7h1'/%3E%3Cpath stroke='%23317aff' d='M14 7h2'/%3E%3Cpath stroke='%23337aff' d='M16 7h1'/%3E%3Cpath stroke='%23367bfc' d='M17 7h1'/%3E%3Cpath stroke='%233372ed' d='M18 7h1'/%3E%3Cpath stroke='%232359c5' d='M19 7h1'/%3E%3Cpath stroke='%234d88ff' d='M1 8h1'/%3E%3Cpath stroke='%23699cff' d='M2 8h1'/%3E%3Cpath stroke='%236398ff' d='M6 8h1'/%3E%3Cpath stroke='%23548fff' d='M8 8h1'/%3E%3Cpath stroke='%234d8cff' d='M9 8h1'/%3E%3Cpath stroke='%23468aff' d='M10 8h1'/%3E%3Cpath stroke='%233f86ff' d='M11 8h1'/%3E%3Cpath stroke='%233983ff' d='M12 8h1'/%3E%3Cpath stroke='%233380ff' d='M13 8h1'/%3E%3Cpath stroke='%232f7fff' d='M14 8h2'/%3E%3Cpath stroke='%233280ff' d='M16 8h1'/%3E%3Cpath stroke='%233580fc' d='M17 8h1'/%3E%3Cpath stroke='%233276ed' d='M18 8h1'/%3E%3Cpath stroke='%23235ac6' d='M19 8h1'/%3E%3Cpath stroke='%236196ff' d='M2 9h1m3 0h1m-4 1h1'/%3E%3Cpath stroke='%23689aff' d='M4 9h1'/%3E%3Cpath stroke='%235b93ff' d='M7 9h1'/%3E%3Cpath stroke='%235491ff' d='M8 9h1'/%3E%3Cpath stroke='%234f90ff' d='M9 9h1'/%3E%3Cpath stroke='%234890ff' d='M10 9h1'/%3E%3Cpath stroke='%23428eff' d='M11 9h1'/%3E%3Cpath stroke='%233b8dff' d='M12 9h1'/%3E%3Cpath stroke='%23348aff' d='M13 9h1'/%3E%3Cpath stroke='%233189ff' d='M14 9h1'/%3E%3Cpath stroke='%232f88ff' d='M15 9h1'/%3E%3Cpath stroke='%233188ff' d='M16 9h1'/%3E%3Cpath stroke='%233385fc' d='M17 9h1'/%3E%3Cpath stroke='%233079ed' d='M18 9h1'/%3E%3Cpath stroke='%23215cc8' d='M19 9h1'/%3E%3Cpath stroke='%233f7fff' d='M1 10h1'/%3E%3Cpath stroke='%236397ff' d='M4 10h1'/%3E%3Cpath stroke='%235993ff' d='M7 10h1'/%3E%3Cpath stroke='%235492ff' d='M8 10h1'/%3E%3Cpath stroke='%235093ff' d='M9 10h1'/%3E%3Cpath stroke='%234a95ff' d='M10 10h1'/%3E%3Cpath stroke='%234496ff' d='M11 10h1'/%3E%3Cpath stroke='%233d96ff' d='M12 10h1'/%3E%3Cpath stroke='%233694ff' d='M13 10h1'/%3E%3Cpath stroke='%233193ff' d='M14 10h1'/%3E%3Cpath stroke='%232f92ff' d='M15 10h1'/%3E%3Cpath stroke='%233090ff' d='M16 10h1'/%3E%3Cpath stroke='%23328cfc' d='M17 10h1'/%3E%3Cpath stroke='%232e7def' d='M18 10h1'/%3E%3Cpath stroke='%231e5dc9' d='M19 10h1'/%3E%3Cpath stroke='%235c92ff' d='M3 11h1m1 1h1'/%3E%3Cpath stroke='%235792ff' d='M7 11h1m-1 1h1'/%3E%3Cpath stroke='%235594ff' d='M8 11h1'/%3E%3Cpath stroke='%235298ff' d='M9 11h1'/%3E%3Cpath stroke='%234d9cff' d='M10 11h1'/%3E%3Cpath stroke='%23479eff' d='M11 11h1'/%3E%3Cpath stroke='%23409fff' d='M12 11h1'/%3E%3Cpath stroke='%23379fff' d='M13 11h1'/%3E%3Cpath stroke='%23339dff' d='M14 11h1'/%3E%3Cpath stroke='%232f9bff' d='M15 11h1'/%3E%3Cpath stroke='%232e97ff' d='M16 11h1'/%3E%3Cpath stroke='%232e91fc' d='M17 11h1'/%3E%3Cpath stroke='%232a80f0' d='M18 11h1'/%3E%3Cpath stroke='%231b5dcb' d='M19 11h1'/%3E%3Cpath stroke='%233275ff' d='M1 12h1'/%3E%3Cpath stroke='%235991ff' d='M6 12h1'/%3E%3Cpath stroke='%235596ff' d='M8 12h1'/%3E%3Cpath stroke='%23529cff' d='M9 12h1'/%3E%3Cpath stroke='%234fa1ff' d='M10 12h1'/%3E%3Cpath stroke='%234aa6ff' d='M11 12h1'/%3E%3Cpath stroke='%2342a9ff' d='M12 12h1'/%3E%3Cpath stroke='%233aa9ff' d='M13 12h1'/%3E%3Cpath stroke='%2334a7ff' d='M14 12h1'/%3E%3Cpath stroke='%2330a5ff' d='M15 12h1'/%3E%3Cpath stroke='%232ca0ff' d='M16 12h1'/%3E%3Cpath stroke='%232a96fd' d='M17 12h1'/%3E%3Cpath stroke='%232581f1' d='M18 12h1'/%3E%3Cpath stroke='%23185dcc' d='M19 12h1'/%3E%3Cpath stroke='%232d72ff' d='M1 13h1m0 3h1'/%3E%3Cpath stroke='%23548DFF' d='M5 13h1'/%3E%3Cpath stroke='%235991FF' d='M6 13h1'/%3E%3Cpath stroke='%235792FF' d='M7 13h1'/%3E%3Cpath stroke='%235496FF' d='M8 13h1'/%3E%3Cpath stroke='%23539CFF' d='M9 13h1'/%3E%3Cpath stroke='%234FA1FF' d='M10 13h1'/%3E%3Cpath stroke='%2344AFFE' d='M11 13h1'/%3E%3Cpath stroke='%2344afff' d='M12 13h1'/%3E%3Cpath stroke='%233eb1ff' d='M13 13h1'/%3E%3Cpath stroke='%2337afff' d='M14 13h1'/%3E%3Cpath stroke='%232fabff' d='M15 13h1'/%3E%3Cpath stroke='%2329a4ff' d='M16 13h1'/%3E%3Cpath stroke='%232599fd' d='M17 13h1'/%3E%3Cpath stroke='%231e80f2' d='M18 13h1'/%3E%3Cpath stroke='%23145bcd' d='M19 13h1'/%3E%3Cpath stroke='%23276eff' d='M1 14h1'/%3E%3Cpath stroke='%233d7dff' d='M2 14h1'/%3E%3Cpath stroke='%234985ff' d='M3 14h1'/%3E%3Cpath stroke='%23548DFF' d='M5 14h1'/%3E%3Cpath stroke='%235991FF' d='M6 14h1'/%3E%3Cpath stroke='%235792FF' d='M7 14h1'/%3E%3Cpath stroke='%235496FF' d='M8 14h1'/%3E%3Cpath stroke='%23539CFF' d='M9 14h1'/%3E%3Cpath stroke='%234FA1FF' d='M10 14h1'/%3E%3Cpath stroke='%2344AFFE' d='M11 14h1'/%3E%3Cpath stroke='%2343b1ff' d='M12 14h1'/%3E%3Cpath stroke='%233eb4ff' d='M13 14h1'/%3E%3Cpath stroke='%2335b2ff' d='M14 14h1'/%3E%3Cpath stroke='%232caeff' d='M15 14h1'/%3E%3Cpath stroke='%2324a5ff' d='M16 14h1'/%3E%3Cpath stroke='%231f97fd' d='M17 14h1'/%3E%3Cpath stroke='%231980f3' d='M18 14h1'/%3E%3Cpath stroke='%23105ace' d='M19 14h1'/%3E%3Cpath stroke='%23216aff' d='M1 15h1'/%3E%3Cpath stroke='%233578ff' d='M2 15h1'/%3E%3Cpath stroke='%234885ff' d='M4 15h1'/%3E%3Cpath stroke='%2341afff' d='M12 15h1'/%3E%3Cpath stroke='%233bb2ff' d='M13 15h1'/%3E%3Cpath stroke='%2333b1ff' d='M14 15h1'/%3E%3Cpath stroke='%232aadff' d='M15 15h1'/%3E%3Cpath stroke='%2321a3ff' d='M16 15h1'/%3E%3Cpath stroke='%231a95fd' d='M17 15h1'/%3E%3Cpath stroke='%23137cf2' d='M18 15h1'/%3E%3Cpath stroke='%230c59cf' d='M19 15h1'/%3E%3Cpath stroke='%231c66ff' d='M1 16h1'/%3E%3Cpath stroke='%233879ff' d='M3 16h1'/%3E%3Cpath stroke='%233f7eff' d='M4 16h1'/%3E%3Cpath stroke='%234584ff' d='M6 16h1'/%3E%3Cpath stroke='%234587ff' d='M7 16h1'/%3E%3Cpath stroke='%23468eff' d='M8 16h1'/%3E%3Cpath stroke='%234696ff' d='M9 16h1'/%3E%3Cpath stroke='%23439cff' d='M10 16h1'/%3E%3Cpath stroke='%233fa3ff' d='M11 16h1'/%3E%3Cpath stroke='%233ba8ff' d='M12 16h1'/%3E%3Cpath stroke='%233af' d='M13 16h1'/%3E%3Cpath stroke='%232da9ff' d='M14 16h1'/%3E%3Cpath stroke='%2324a6ff' d='M15 16h1'/%3E%3Cpath stroke='%231d9eff' d='M16 16h1'/%3E%3Cpath stroke='%231690fd' d='M17 16h1'/%3E%3Cpath stroke='%231078f1' d='M18 16h1'/%3E%3Cpath stroke='%230b57ce' d='M19 16h1'/%3E%3Cpath stroke='%231761f9' d='M1 17h1'/%3E%3Cpath stroke='%23246bfa' d='M2 17h1'/%3E%3Cpath stroke='%232f72fb' d='M3 17h1'/%3E%3Cpath stroke='%233676fb' d='M4 17h1'/%3E%3Cpath stroke='%233a7afb' d='M5 17h1'/%3E%3Cpath stroke='%233b7bfc' d='M6 17h1'/%3E%3Cpath stroke='%233b7efc' d='M7 17h1'/%3E%3Cpath stroke='%233c84fc' d='M8 17h1'/%3E%3Cpath stroke='%233b8afc' d='M9 17h1'/%3E%3Cpath stroke='%233990fc' d='M10 17h1'/%3E%3Cpath stroke='%233695fc' d='M11 17h1'/%3E%3Cpath stroke='%233299fc' d='M12 17h1'/%3E%3Cpath stroke='%232c9cfd' d='M13 17h1'/%3E%3Cpath stroke='%23259bfd' d='M14 17h1'/%3E%3Cpath stroke='%231e97fd' d='M15 17h1'/%3E%3Cpath stroke='%231790fc' d='M16 17h1'/%3E%3Cpath stroke='%231184fa' d='M17 17h1'/%3E%3Cpath stroke='%230c6ded' d='M18 17h1'/%3E%3Cpath stroke='%230850c8' d='M19 17h1'/%3E%3Cpath stroke='%232f6ae4' d='M1 18h1'/%3E%3Cpath stroke='%231b5fe9' d='M2 18h1'/%3E%3Cpath stroke='%232163e8' d='M3 18h1'/%3E%3Cpath stroke='%232868eb' d='M4 18h1'/%3E%3Cpath stroke='%232c6aea' d='M5 18h1'/%3E%3Cpath stroke='%232e6dea' d='M6 18h1'/%3E%3Cpath stroke='%232d6deb' d='M7 18h1'/%3E%3Cpath stroke='%232c71ec' d='M8 18h1'/%3E%3Cpath stroke='%232c76ec' d='M9 18h1'/%3E%3Cpath stroke='%232a79ed' d='M10 18h1'/%3E%3Cpath stroke='%23287eef' d='M11 18h1'/%3E%3Cpath stroke='%232481f1' d='M12 18h1'/%3E%3Cpath stroke='%232182f1' d='M13 18h1'/%3E%3Cpath stroke='%231c80f1' d='M14 18h1'/%3E%3Cpath stroke='%231880f3' d='M15 18h1'/%3E%3Cpath stroke='%23117af2' d='M16 18h1'/%3E%3Cpath stroke='%230c6eed' d='M17 18h1'/%3E%3Cpath stroke='%230a5ddd' d='M18 18h1'/%3E%3Cpath stroke='%23265dc1' d='M19 18h1'/%3E%3Cpath stroke='%2393b4f2' d='M0 19h1m19 0h1'/%3E%3Cpath stroke='%23d1ddf4' d='M1 19h1'/%3E%3Cpath stroke='%232e61ca' d='M2 19h1'/%3E%3Cpath stroke='%23134bbf' d='M3 19h1'/%3E%3Cpath stroke='%23164fc2' d='M4 19h1'/%3E%3Cpath stroke='%231950c1' d='M5 19h1'/%3E%3Cpath stroke='%231b52c1' d='M6 19h1'/%3E%3Cpath stroke='%231a52c3' d='M7 19h1'/%3E%3Cpath stroke='%231954c6' d='M8 19h1'/%3E%3Cpath stroke='%231b58c9' d='M9 19h1'/%3E%3Cpath stroke='%231858c8' d='M10 19h1'/%3E%3Cpath stroke='%23165bcd' d='M11 19h1'/%3E%3Cpath stroke='%23145cd0' d='M12 19h1'/%3E%3Cpath stroke='%23135cd0' d='M13 19h1'/%3E%3Cpath stroke='%230f58cc' d='M14 19h1'/%3E%3Cpath stroke='%230d5ad2' d='M15 19h1'/%3E%3Cpath stroke='%230b58d1' d='M16 19h1'/%3E%3Cpath stroke='%230951cb' d='M17 19h1'/%3E%3Cpath stroke='%23265ec3' d='M18 19h1'/%3E%3Cpath stroke='%23d0daee' d='M19 19h1'/%3E%3Cpath stroke='%2393b3f2' d='M1 20h1m17 0h1'/%3E%3Cpath stroke='%23fefefe' d='M14 20h1'/%3E%3Cpath stroke='%23fdfdfd' d='M15 20h1m1 0h1'/%3E%3Cpath stroke='%23fcfcfc' d='M16 20h1'/%3E%3Cpath stroke='%23f2f5fc' d='M18 20h1M5 15h9M5 9h9M5 10h9M5.5 8.5v7M13.5 8.5v7M7 5h9M7 6h9M14 11h2M7.5 5v4M15.5 5v6'/%3E%3C/svg%3E\")}.title-bar-controls button[aria-label=Restore]:not(:disabled):active{background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 -0.5 21 21' shape-rendering='crispEdges'%3E%3Cpath stroke='%2393b1ed' d='M1 0h1m17 0h1'/%3E%3Cpath stroke='%23f4f6fd' d='M2 0h1m15 0h1M0 2h1m19 0h1M0 18h1m19 0h1M2 20h1m15 0h1'/%3E%3Cpath stroke='%23fff' d='M3 0h15M0 3h1m19 0h1M0 4h1m19 0h1M0 5h1m19 0h1M0 6h1m19 0h1M0 7h1m19 0h1M0 8h1m19 0h1M0 9h1m19 0h1M0 10h1m19 0h1M0 11h1m19 0h1M0 12h1m19 0h1M0 13h1m19 0h1M0 14h1m19 0h1M0 15h1m19 0h1M0 16h1m19 0h1M0 17h1m19 0h1M3 20h15'/%3E%3Cpath stroke='%23a7bcee' d='M0 1h1m19 0h1'/%3E%3Cpath stroke='%23cfd3da' d='M1 1h1'/%3E%3Cpath stroke='%231f3b5f' d='M2 1h1M1 2h1'/%3E%3Cpath stroke='%23002453' d='M3 1h1M1 4h1'/%3E%3Cpath stroke='%23002557' d='M4 1h1'/%3E%3Cpath stroke='%23002658' d='M5 1h1'/%3E%3Cpath stroke='%2300285c' d='M6 1h1'/%3E%3Cpath stroke='%23002a61' d='M7 1h1'/%3E%3Cpath stroke='%23002d67' d='M8 1h1'/%3E%3Cpath stroke='%23002f6b' d='M9 1h1'/%3E%3Cpath stroke='%23002f6c' d='M10 1h1M1 10h1'/%3E%3Cpath stroke='%23003273' d='M11 1h1'/%3E%3Cpath stroke='%23003478' d='M12 1h1M5 2h1'/%3E%3Cpath stroke='%2300357b' d='M13 1h1M2 5h1m-2 8h1'/%3E%3Cpath stroke='%2300377f' d='M14 1h1M6 2h1'/%3E%3Cpath stroke='%23003780' d='M15 1h1'/%3E%3Cpath stroke='%23003984' d='M16 1h1'/%3E%3Cpath stroke='%23003882' d='M17 1h1M3 3h1'/%3E%3Cpath stroke='%231f5295' d='M18 1h1'/%3E%3Cpath stroke='%23cfdae9' d='M19 1h1'/%3E%3Cpath stroke='%23002a62' d='M2 2h1'/%3E%3Cpath stroke='%23003070' d='M3 2h1'/%3E%3Cpath stroke='%23003275' d='M4 2h1'/%3E%3Cpath stroke='%23003883' d='M7 2h1M1 17h1'/%3E%3Cpath stroke='%23003a88' d='M8 2h1'/%3E%3Cpath stroke='%23003d8f' d='M9 2h1M2 9h1'/%3E%3Cpath stroke='%23003e90' d='M10 2h1'/%3E%3Cpath stroke='%23004094' d='M11 2h1'/%3E%3Cpath stroke='%23004299' d='M12 2h1M2 12h1'/%3E%3Cpath stroke='%2300439b' d='M13 2h1'/%3E%3Cpath stroke='%2300449e' d='M14 2h1M2 14h1'/%3E%3Cpath stroke='%2300459f' d='M15 2h1'/%3E%3Cpath stroke='%230045a1' d='M16 2h1m1 0h1M2 17h1'/%3E%3Cpath stroke='%230045a0' d='M17 2h1M2 15h1'/%3E%3Cpath stroke='%231f5aa8' d='M19 2h1'/%3E%3Cpath stroke='%23002452' d='M1 3h1'/%3E%3Cpath stroke='%23003170' d='M2 3h1'/%3E%3Cpath stroke='%23003b8b' d='M4 3h1M3 4h1'/%3E%3Cpath stroke='%23003c8f' d='M5 3h1'/%3E%3Cpath stroke='%23003e94' d='M6 3h1'/%3E%3Cpath stroke='%23004099' d='M7 3h1'/%3E%3Cpath stroke='%2300429d' d='M8 3h1'/%3E%3Cpath stroke='%230044a2' d='M9 3h1'/%3E%3Cpath stroke='%230046a5' d='M10 3h1'/%3E%3Cpath stroke='%230048a8' d='M11 3h1'/%3E%3Cpath stroke='%230049ab' d='M12 3h1m-3 2h1'/%3E%3Cpath stroke='%23004aac' d='M13 3h1'/%3E%3Cpath stroke='%23004aad' d='M14 3h1'/%3E%3Cpath stroke='%23004bae' d='M15 3h2m1 0h1M3 14h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23004baf' d='M17 3h1m-5 2h1m-7 5h1m-5 7h1m-1 1h1'/%3E%3Cpath stroke='%23004bad' d='M19 3h1M3 13h1m-1 6h1'/%3E%3Cpath stroke='%23037' d='M2 4h1m-2 8h1'/%3E%3Cpath stroke='%23003d92' d='M4 4h1'/%3E%3Cpath stroke='%23003f97' d='M5 4h1M4 5h1'/%3E%3Cpath stroke='%2300419d' d='M6 4h1M4 6h1'/%3E%3Cpath stroke='%230043a1' d='M7 4h1'/%3E%3Cpath stroke='%230045a4' d='M8 4h1'/%3E%3Cpath stroke='%230047a8' d='M9 4h1M4 9h1'/%3E%3Cpath stroke='%230048ab' d='M10 4h1m-7 6h1'/%3E%3Cpath stroke='%230049ad' d='M11 4h1m-2 2h1m-6 5h1'/%3E%3Cpath stroke='%23004aae' d='M12 4h1m-1 1h1m-2 1h1m-6 5h1m-3 1h2'/%3E%3Cpath stroke='%23004cb0' d='M13 4h1m0 1h1m-8 6h1m-4 2h1'/%3E%3Cpath stroke='%23004db1' d='M14 4h3m-2 1h2m-4 1h4M7 12h1m-4 2h1m-1 1h1m-1 1h2'/%3E%3Cpath stroke='%23004db2' d='M17 4h3m-3 1h3m-2 1h2m-8 1h1m6 0h1m-9 1h1m-4 3h1m-5 6h2m-2 1h4m-4 1h4'/%3E%3Cpath stroke='%23002555' d='M1 5h1'/%3E%3Cpath stroke='%23003d90' d='M3 5h1'/%3E%3Cpath stroke='%2300409c' d='M5 5h1'/%3E%3Cpath stroke='%230042a1' d='M6 5h1M5 6h1'/%3E%3Cpath stroke='%230044a5' d='M7 5h1M6 6h1'/%3E%3Cpath stroke='%230046a8' d='M8 5h1M5 8h1'/%3E%3Cpath stroke='%230047aa' d='M9 5h1'/%3E%3Cpath stroke='%230049ac' d='M11 5h1m-7 5h1m-2 1h1m-2 1h1'/%3E%3Cpath stroke='%2300275a' d='M1 6h1'/%3E%3Cpath stroke='%23003781' d='M2 6h1m-2 9h1'/%3E%3Cpath stroke='%23003f95' d='M3 6h1'/%3E%3Cpath stroke='%230045a9' d='M7 6h1'/%3E%3Cpath stroke='%230046aa' d='M8 6h1M6 7h1'/%3E%3Cpath stroke='%230047ac' d='M9 6h1M7 7h1'/%3E%3Cpath stroke='%23004bb0' d='M12 6h1M8 9h1m-3 3h1'/%3E%3Cpath stroke='%23004eb3' d='M17 6h1m-5 1h1m4 0h1m0 1h1M10 9h1m-2 1h1m-3 6h1m-2 1h2m0 2h1'/%3E%3Cpath stroke='%2300295f' d='M1 7h1'/%3E%3Cpath stroke='%23003985' d='M2 7h1'/%3E%3Cpath stroke='%2300419b' d='M3 7h1'/%3E%3Cpath stroke='%230043a2' d='M4 7h1'/%3E%3Cpath stroke='%230044a6' d='M5 7h1'/%3E%3Cpath stroke='%230048ad' d='M8 7h1M6 9h1'/%3E%3Cpath stroke='%230049ae' d='M9 7h1M7 8h2m-3 2h1'/%3E%3Cpath stroke='%23004aaf' d='M10 7h1M9 8h1M7 9h1'/%3E%3Cpath stroke='%23004cb1' d='M11 7h1m-2 1h1M9 9h1m-2 1h1'/%3E%3Cpath stroke='%23004fb3' d='M14 7h1'/%3E%3Cpath stroke='%23004fb4' d='M15 7h3m-6 1h1m5 0h1m0 1h1M8 12h1m-1 6h1m0 1h1'/%3E%3Cpath stroke='%23002b63' d='M1 8h1'/%3E%3Cpath stroke='%23003b8a' d='M2 8h1'/%3E%3Cpath stroke='%2300439f' d='M3 8h1'/%3E%3Cpath stroke='%230045a5' d='M4 8h1'/%3E%3Cpath stroke='%230047ab' d='M6 8h1M5 9h1'/%3E%3Cpath stroke='%230050b5' d='M13 8h2m1 0h2m-7 1h1m-2 1h1m8 0h1M9 11h1m-2 5h1m-1 1h1m1 2h1'/%3E%3Cpath stroke='%230051b6' d='M15 8h1m2 1h1m0 2h1m-1 1h1m-1 5h1M9 18h1m1 1h1'/%3E%3Cpath stroke='%23002d68' d='M1 9h1'/%3E%3Cpath stroke='%230045a3' d='M3 9h1'/%3E%3Cpath stroke='%230052b7' d='M12 9h1m-2 1h1m-2 1h1m-2 1h1m9 1h1m-8 6h2m3 0h1'/%3E%3Cpath stroke='%230053b8' d='M13 9h1m2 0h2m0 1h1m0 4h1M9 16h1m9 0h1M9 17h1m0 1h1m3 1h1m1 0h1'/%3E%3Cpath stroke='%230054b9' d='M14 9h2m2 9h1m-4 1h1'/%3E%3Cpath stroke='%23003f93' d='M2 10h1'/%3E%3Cpath stroke='%230047a7' d='M3 10h1'/%3E%3Cpath stroke='%230055ba' d='M12 10h1m4 0h1m-7 1h1m6 0h1m-9 6h1m0 1h1'/%3E%3Cpath stroke='%230056bb' d='M13 10h1m2 0h1m1 2h1m-9 4h1'/%3E%3Cpath stroke='%230057bc' d='M14 10h2m-5 2h1m6 5h1m-7 1h1m4 0h1'/%3E%3Cpath stroke='%23003172' d='M1 11h1'/%3E%3Cpath stroke='%23004095' d='M2 11h1'/%3E%3Cpath stroke='%230048aa' d='M3 11h1'/%3E%3Cpath stroke='%230058bd' d='M12 11h1m4 0h1m0 2h1m-6 5h1'/%3E%3Cpath stroke='%230059be' d='M13 11h1m2 0h1m-6 5h1m6 0h1m-5 2h1m1 0h1'/%3E%3Cpath stroke='%23005abf' d='M12 12h1m4 0h1m-6 5h1m2 1h1'/%3E%3Cpath stroke='%230055b9' d='M10 12h1'/%3E%3Cpath stroke='%23005cc1' d='M13 12h1m2 0h1m-5 1h1m4 0h1m-5 4h1'/%3E%3Cpath stroke='%23005dc2' d='M14 12h1m-3 2h1m4 0h1m-6 1h1m4 1h1m-4 1h1m1 0h1'/%3E%3Cpath stroke='%23005ec3' d='M15 12h1m-3 1h1m2 0h1m0 2h1m-5 1h1m1 1h1'/%3E%3Cpath stroke='%2300449d' d='M2 13h1'/%3E%3Cpath stroke='%2378a2d8' d='M5 13h7m-7 1h7m-7 1h7M5 13h1'/%3E%3Cpath stroke='%23004BB0' d='M6 13h1'/%3E%3Cpath stroke='%23004DB1' d='M7 13h1'/%3E%3Cpath stroke='%23004FB4' d='M8 13h1'/%3E%3Cpath stroke='%230052B7' d='M9 13h1'/%3E%3Cpath stroke='%230055B9' d='M10 13h1'/%3E%3Cpath stroke='%230157BC' d='M11 13h1'/%3E%3Cpath stroke='%2378a2d8' d='M13 13h1'/%3E%3Cpath stroke='%23005fc4' d='M14 13h1m1 1h1'/%3E%3Cpath stroke='%230060c5' d='M15 13h1m-2 1h1m1 1h1m-2 1h1'/%3E%3Cpath stroke='%2300367e' d='M1 14h1'/%3E%3Cpath stroke='%230061c6' d='M15 14h1m-2 1h1'/%3E%3Cpath stroke='%23004BB0' d='M6 14h1'/%3E%3Cpath stroke='%23004DB1' d='M7 14h1'/%3E%3Cpath stroke='%23004FB4' d='M8 14h1'/%3E%3Cpath stroke='%230052B7' d='M9 14h1'/%3E%3Cpath stroke='%230055B9' d='M10 14h1'/%3E%3Cpath stroke='%230157BC' d='M11 14h1'/%3E%3Cpath stroke='%2378a2d8' d='M13 14h1'/%3E%3Cpath stroke='%230059bd' d='M18 14h1'/%3E%3Cpath stroke='%2378a2d8' d='M12 15h1M13 15h1'/%3E%3Cpath stroke='%230062c6' d='M15 15h1'/%3E%3Cpath stroke='%23005abe' d='M18 15h1'/%3E%3Cpath stroke='%230054b8' d='M19 15h1'/%3E%3Cpath stroke='%23003881' d='M1 16h1'/%3E%3Cpath stroke='%230046a1' d='M2 16h1'/%3E%3Cpath stroke='%23004eb2' d='M6 16h1'/%3E%3Cpath stroke='%23005cc0' d='M12 16h1'/%3E%3Cpath stroke='%23005fc3' d='M14 16h1'/%3E%3Cpath stroke='%230060c4' d='M16 16h1'/%3E%3Cpath stroke='%230058bc' d='M11 17h1'/%3E%3Cpath stroke='%23005bc0' d='M17 17h1'/%3E%3Cpath stroke='%231f5294' d='M1 18h1'/%3E%3Cpath stroke='%230046a2' d='M2 18h1'/%3E%3Cpath stroke='%231f66be' d='M19 18h1'/%3E%3Cpath stroke='%23a7bef0' d='M0 19h1m0 1h1m17 0h1'/%3E%3Cpath stroke='%23cfdae8' d='M1 19h1'/%3E%3Cpath stroke='%231f5ba9' d='M2 19h1'/%3E%3Cpath stroke='%231f66bf' d='M18 19h1'/%3E%3Cpath stroke='%23cfdef1' d='M19 19h1'/%3E%3Cpath stroke='%2393b4f2' d='M20 19h1'/%3E%3Cpath stroke='%2378a2d8' d='M5 15h9M5 9h9M5 10h9M5.5 8.5v7M13.5 8.5v7M7 5h9M7 6h9M14 11h2M7.5 5v4M15.5 5v6'/%3E%3C/svg%3E\")}.title-bar-controls button[aria-label=Help]{background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 -0.5 21 21' shape-rendering='crispEdges'%3E%3Cpath stroke='%23b5c6ef' d='M1 0h1m17 0h1M0 1h1m19 0h1M0 19h1m19 0h1M1 20h1m17 0h1'/%3E%3Cpath stroke='%23f4f6fd' d='M2 0h1m17 2h1M0 18h1m17 2h1'/%3E%3Cpath stroke='%23fff' d='M3 0h16M0 2h1M0 3h1m19 0h1M0 4h1m8 0h3m8 0h1M0 5h1m7 0h1m3 0h1m7 0h1M0 6h1m6 0h1m5 0h1m6 0h1M0 7h1m12 0h1m6 0h1M0 8h1m12 0h1m6 0h1M0 9h1m12 0h1m6 0h1M0 10h1m10 0h2m7 0h1M0 11h1m9 0h1m9 0h1M0 12h1m9 0h1m9 0h1M0 13h1m19 0h1M0 14h1m19 0h1M0 15h1m9 0h1m9 0h1M0 16h1m9 0h1m9 0h1M0 17h1m19 0h1m-1 1h1M2 20h16'/%3E%3Cpath stroke='%23dce5fd' d='M1 1h1'/%3E%3Cpath stroke='%23739af8' d='M2 1h1'/%3E%3Cpath stroke='%23608cf7' d='M3 1h1M2 8h1'/%3E%3Cpath stroke='%235584f6' d='M4 1h1'/%3E%3Cpath stroke='%234d7ef6' d='M5 1h1M1 6h1m5 4h1'/%3E%3Cpath stroke='%23487af5' d='M6 1h1'/%3E%3Cpath stroke='%234276f5' d='M7 1h1M3 14h1'/%3E%3Cpath stroke='%234478f5' d='M8 1h1m5 3h1M2 12h1'/%3E%3Cpath stroke='%233e73f5' d='M9 1h2'/%3E%3Cpath stroke='%233b71f5' d='M11 1h2'/%3E%3Cpath stroke='%23336cf4' d='M13 1h2'/%3E%3Cpath stroke='%23306af4' d='M15 1h1'/%3E%3Cpath stroke='%232864f4' d='M16 1h1'/%3E%3Cpath stroke='%231f5def' d='M17 1h1'/%3E%3Cpath stroke='%233467e0' d='M18 1h1'/%3E%3Cpath stroke='%23d2dbf2' d='M19 1h1'/%3E%3Cpath stroke='%23769cf8' d='M1 2h1'/%3E%3Cpath stroke='%2390aff9' d='M2 2h1'/%3E%3Cpath stroke='%2394b2f9' d='M3 2h1'/%3E%3Cpath stroke='%2385a7f8' d='M4 2h1'/%3E%3Cpath stroke='%23759cf8' d='M5 2h1'/%3E%3Cpath stroke='%236e97f8' d='M6 2h1M2 6h1'/%3E%3Cpath stroke='%236892f7' d='M7 2h1'/%3E%3Cpath stroke='%236690f7' d='M8 2h1'/%3E%3Cpath stroke='%23628ef7' d='M9 2h1m0 1h1'/%3E%3Cpath stroke='%235f8cf7' d='M10 2h1'/%3E%3Cpath stroke='%235e8bf7' d='M11 2h1'/%3E%3Cpath stroke='%235988f6' d='M12 2h1'/%3E%3Cpath stroke='%235685f6' d='M13 2h1'/%3E%3Cpath stroke='%235082f6' d='M14 2h1'/%3E%3Cpath stroke='%23497cf5' d='M15 2h1'/%3E%3Cpath stroke='%233f75f5' d='M16 2h1m-2 2h1'/%3E%3Cpath stroke='%23326bf2' d='M17 2h1'/%3E%3Cpath stroke='%23235ce3' d='M18 2h1'/%3E%3Cpath stroke='%23305cc5' d='M19 2h1'/%3E%3Cpath stroke='%236590f7' d='M1 3h1'/%3E%3Cpath stroke='%2397b4f9' d='M2 3h1'/%3E%3Cpath stroke='%239ab7fa' d='M3 3h1'/%3E%3Cpath stroke='%2389aaf9' d='M4 3h1M2 4h1'/%3E%3Cpath stroke='%237aa0f8' d='M5 3h1'/%3E%3Cpath stroke='%23729af8' d='M6 3h1'/%3E%3Cpath stroke='%236d95f8' d='M7 3h1'/%3E%3Cpath stroke='%236892f8' d='M8 3h1M2 7h1'/%3E%3Cpath stroke='%23658ff7' d='M9 3h1'/%3E%3Cpath stroke='%23618df7' d='M11 3h1'/%3E%3Cpath stroke='%235d8af7' d='M12 3h1M3 9h1'/%3E%3Cpath stroke='%235987f6' d='M13 3h1M2 9h1'/%3E%3Cpath stroke='%235283f6' d='M14 3h1'/%3E%3Cpath stroke='%234c7ef6' d='M15 3h1M5 14h1'/%3E%3Cpath stroke='%234377f5' d='M16 3h1'/%3E%3Cpath stroke='%23376ef2' d='M17 3h1'/%3E%3Cpath stroke='%23285fe3' d='M18 3h1'/%3E%3Cpath stroke='%231546b9' d='M19 3h1'/%3E%3Cpath stroke='%235886f6' d='M1 4h1'/%3E%3Cpath stroke='%238dadf9' d='M3 4h1'/%3E%3Cpath stroke='%237fa3f8' d='M4 4h1'/%3E%3Cpath stroke='%237199f8' d='M5 4h1M4 5h1'/%3E%3Cpath stroke='%236a93f8' d='M6 4h1M4 6h1M3 7h1'/%3E%3Cpath stroke='%2392aff9' d='M7 4h1'/%3E%3Cpath stroke='%23e1e9fd' d='M8 4h1'/%3E%3Cpath stroke='%23e0e8fd' d='M12 4h1'/%3E%3Cpath stroke='%2381a4f8' d='M13 4h1'/%3E%3Cpath stroke='%233a72f4' d='M16 4h1'/%3E%3Cpath stroke='%23346cf2' d='M17 4h1'/%3E%3Cpath stroke='%232a61e3' d='M18 4h1'/%3E%3Cpath stroke='%231848bb' d='M19 4h1'/%3E%3Cpath stroke='%235282f6' d='M1 5h1m4 6h1m-3 1h1'/%3E%3Cpath stroke='%23799ff8' d='M2 5h1'/%3E%3Cpath stroke='%237ca1f8' d='M3 5h1'/%3E%3Cpath stroke='%236791f8' d='M5 5h1'/%3E%3Cpath stroke='%238eacf9' d='M6 5h1'/%3E%3Cpath stroke='%23f3f6fe' d='M7 5h1'/%3E%3Cpath stroke='%23d8e2fd' d='M9 5h1'/%3E%3Cpath stroke='%23cfdcfc' d='M10 5h1'/%3E%3Cpath stroke='%23ecf1fe' d='M11 5h1'/%3E%3Cpath stroke='%23eff4fe' d='M13 5h1'/%3E%3Cpath stroke='%23749af7' d='M14 5h1'/%3E%3Cpath stroke='%23326cf4' d='M15 5h1'/%3E%3Cpath stroke='%23316bf4' d='M16 5h1M3 16h1'/%3E%3Cpath stroke='%233069f1' d='M17 5h1'/%3E%3Cpath stroke='%232c62e4' d='M18 5h1'/%3E%3Cpath stroke='%231d4cbc' d='M19 5h1m-1 1h1'/%3E%3Cpath stroke='%237099f8' d='M3 6h1'/%3E%3Cpath stroke='%23628cf8' d='M5 6h1'/%3E%3Cpath stroke='%23d3dffd' d='M6 6h1'/%3E%3Cpath stroke='%23b2c6fb' d='M8 6h1'/%3E%3Cpath stroke='%234777f6' d='M9 6h1'/%3E%3Cpath stroke='%234072f5' d='M10 6h1'/%3E%3Cpath stroke='%234a7bf6' d='M11 6h1'/%3E%3Cpath stroke='%23c8d7fc' d='M12 6h1'/%3E%3Cpath stroke='%23c6d6fc' d='M14 6h1'/%3E%3Cpath stroke='%232c69f5' d='M15 6h1'/%3E%3Cpath stroke='%232d69f5' d='M16 6h1'/%3E%3Cpath stroke='%232e69f2' d='M17 6h1'/%3E%3Cpath stroke='%232c63e5' d='M18 6h1'/%3E%3Cpath stroke='%234679f5' d='M1 7h1M1 8h1'/%3E%3Cpath stroke='%23658ff8' d='M4 7h1'/%3E%3Cpath stroke='%235e89f7' d='M5 7h1'/%3E%3Cpath stroke='%23e6edfe' d='M6 7h1'/%3E%3Cpath stroke='%23e5ecfe' d='M7 7h1'/%3E%3Cpath stroke='%235a85f7' d='M8 7h1'/%3E%3Cpath stroke='%234375f5' d='M9 7h1'/%3E%3Cpath stroke='%233d71f5' d='M10 7h1'/%3E%3Cpath stroke='%23366ef4' d='M11 7h1M2 14h1'/%3E%3Cpath stroke='%236c97f8' d='M12 7h1'/%3E%3Cpath stroke='%23cfddfd' d='M14 7h1'/%3E%3Cpath stroke='%232766f5' d='M15 7h1'/%3E%3Cpath stroke='%232a68f5' d='M16 7h1'/%3E%3Cpath stroke='%232c69f2' d='M17 7h1'/%3E%3Cpath stroke='%232a62e4' d='M18 7h1'/%3E%3Cpath stroke='%231c4cbd' d='M19 7h1'/%3E%3Cpath stroke='%23628df8' d='M3 8h1'/%3E%3Cpath stroke='%23608bf7' d='M4 8h1'/%3E%3Cpath stroke='%235b87f7' d='M5 8h1'/%3E%3Cpath stroke='%235482f7' d='M6 8h1'/%3E%3Cpath stroke='%234e7cf6' d='M7 8h1'/%3E%3Cpath stroke='%234778f6' d='M8 8h1'/%3E%3Cpath stroke='%234174f5' d='M9 8h1'/%3E%3Cpath stroke='%233a71f5' d='M10 8h1'/%3E%3Cpath stroke='%23346ef4' d='M11 8h1'/%3E%3Cpath stroke='%2385a9f9' d='M12 8h1'/%3E%3Cpath stroke='%23cbdbfd' d='M14 8h1'/%3E%3Cpath stroke='%232266f5' d='M15 8h1'/%3E%3Cpath stroke='%232567f5' d='M16 8h1'/%3E%3Cpath stroke='%232968f2' d='M17 8h1'/%3E%3Cpath stroke='%232963e4' d='M18 8h1'/%3E%3Cpath stroke='%231b4bbd' d='M19 8h1'/%3E%3Cpath stroke='%233c72f4' d='M1 9h1'/%3E%3Cpath stroke='%235d89f7' d='M4 9h1'/%3E%3Cpath stroke='%235986f7' d='M5 9h1m-2 1h1'/%3E%3Cpath stroke='%235381f6' d='M6 9h1'/%3E%3Cpath stroke='%234e7ef6' d='M7 9h1'/%3E%3Cpath stroke='%23477af5' d='M8 9h1'/%3E%3Cpath stroke='%234178f5' d='M9 9h1'/%3E%3Cpath stroke='%233a74f5' d='M10 9h1'/%3E%3Cpath stroke='%2396b6fa' d='M11 9h1'/%3E%3Cpath stroke='%23f2f6fe' d='M12 9h1'/%3E%3Cpath stroke='%2393b6fb' d='M14 9h1'/%3E%3Cpath stroke='%232069f6' d='M15 9h1'/%3E%3Cpath stroke='%232268f5' d='M16 9h1'/%3E%3Cpath stroke='%232569f2' d='M17 9h1'/%3E%3Cpath stroke='%232562e6' d='M18 9h1'/%3E%3Cpath stroke='%23194bbe' d='M19 9h1'/%3E%3Cpath stroke='%23376ef4' d='M1 10h1'/%3E%3Cpath stroke='%235181f6' d='M2 10h1'/%3E%3Cpath stroke='%235785f7' d='M3 10h1m1 0h1'/%3E%3Cpath stroke='%235281f6' d='M6 10h1'/%3E%3Cpath stroke='%23477bf6' d='M8 10h1'/%3E%3Cpath stroke='%234e82f7' d='M9 10h1'/%3E%3Cpath stroke='%23cadafc' d='M10 10h1'/%3E%3Cpath stroke='%23a0c0fb' d='M13 10h1'/%3E%3Cpath stroke='%232a72f6' d='M14 10h1'/%3E%3Cpath stroke='%231e6bf6' d='M15 10h1'/%3E%3Cpath stroke='%231f6af6' d='M16 10h1'/%3E%3Cpath stroke='%23216af3' d='M17 10h1'/%3E%3Cpath stroke='%232162e6' d='M18 10h1'/%3E%3Cpath stroke='%231649be' d='M19 10h1'/%3E%3Cpath stroke='%23326bf4' d='M1 11h1'/%3E%3Cpath stroke='%234b7df5' d='M2 11h1'/%3E%3Cpath stroke='%235483f6' d='M3 11h1'/%3E%3Cpath stroke='%235684f7' d='M4 11h1'/%3E%3Cpath stroke='%235583f7' d='M5 11h1'/%3E%3Cpath stroke='%234d80f6' d='M7 11h1'/%3E%3Cpath stroke='%23487df6' d='M8 11h1'/%3E%3Cpath stroke='%23bcd1fc' d='M9 11h1'/%3E%3Cpath stroke='%23dde8fd' d='M11 11h1'/%3E%3Cpath stroke='%235f97f8' d='M12 11h1'/%3E%3Cpath stroke='%232673f7' d='M13 11h1'/%3E%3Cpath stroke='%232171f7' d='M14 11h1'/%3E%3Cpath stroke='%231c6ff6' d='M15 11h1'/%3E%3Cpath stroke='%231c6df6' d='M16 11h1'/%3E%3Cpath stroke='%231c6af4' d='M17 11h1'/%3E%3Cpath stroke='%231c61e6' d='M18 11h1'/%3E%3Cpath stroke='%231248bf' d='M19 11h1'/%3E%3Cpath stroke='%232b66f4' d='M1 12h1'/%3E%3Cpath stroke='%234e7ff6' d='M3 12h1'/%3E%3Cpath stroke='%235383f6' d='M5 12h1'/%3E%3Cpath stroke='%235182f6' d='M6 12h1'/%3E%3Cpath stroke='%234d81f7' d='M7 12h1'/%3E%3Cpath stroke='%23487ff6' d='M8 12h1'/%3E%3Cpath stroke='%23dfe9fd' d='M9 12h1'/%3E%3Cpath stroke='%234687f7' d='M11 12h1'/%3E%3Cpath stroke='%232d7af7' d='M12 12h1'/%3E%3Cpath stroke='%232677f7' d='M13 12h1'/%3E%3Cpath stroke='%232174f7' d='M14 12h1'/%3E%3Cpath stroke='%231b71f7' d='M15 12h1'/%3E%3Cpath stroke='%23186ef7' d='M16 12h1'/%3E%3Cpath stroke='%23186af4' d='M17 12h1'/%3E%3Cpath stroke='%23165fe7' d='M18 12h1'/%3E%3Cpath stroke='%230f47c0' d='M19 12h1'/%3E%3Cpath stroke='%232562f3' d='M1 13h1'/%3E%3Cpath stroke='%233d73f4' d='M2 13h1'/%3E%3Cpath stroke='%23487bf5' d='M3 13h1'/%3E%3Cpath stroke='%234e80f6' d='M4 13h1'/%3E%3Cpath stroke='%235081f6' d='M5 13h1'/%3E%3Cpath stroke='%234e81f6' d='M6 13h1'/%3E%3Cpath stroke='%234b80f6' d='M7 13h1'/%3E%3Cpath stroke='%23477ff6' d='M8 13h1'/%3E%3Cpath stroke='%23d2e0fd' d='M9 13h1'/%3E%3Cpath stroke='%23edf3fe' d='M10 13h1'/%3E%3Cpath stroke='%23367ff7' d='M11 13h1'/%3E%3Cpath stroke='%232d7cf7' d='M12 13h1'/%3E%3Cpath stroke='%232679f8' d='M13 13h1'/%3E%3Cpath stroke='%232077f7' d='M14 13h1'/%3E%3Cpath stroke='%231973f7' d='M15 13h1'/%3E%3Cpath stroke='%23166ff7' d='M16 13h1'/%3E%3Cpath stroke='%231369f4' d='M17 13h1'/%3E%3Cpath stroke='%23105de8' d='M18 13h1'/%3E%3Cpath stroke='%230a44bf' d='M19 13h1'/%3E%3Cpath stroke='%231e5df3' d='M1 14h1'/%3E%3Cpath stroke='%23497bf5' d='M4 14h1'/%3E%3Cpath stroke='%234a7ef7' d='M6 14h1'/%3E%3Cpath stroke='%23487ef6' d='M7 14h1'/%3E%3Cpath stroke='%23457ff6' d='M8 14h1'/%3E%3Cpath stroke='%234180f6' d='M9 14h1'/%3E%3Cpath stroke='%233b7ff6' d='M10 14h1'/%3E%3Cpath stroke='%23357ff7' d='M11 14h1'/%3E%3Cpath stroke='%232d7df7' d='M12 14h1'/%3E%3Cpath stroke='%23257af8' d='M13 14h1'/%3E%3Cpath stroke='%231e77f8' d='M14 14h1'/%3E%3Cpath stroke='%231773f8' d='M15 14h1'/%3E%3Cpath stroke='%23116df7' d='M16 14h1'/%3E%3Cpath stroke='%230d66f4' d='M17 14h1m-3 3h1'/%3E%3Cpath stroke='%230b59e7' d='M18 14h1'/%3E%3Cpath stroke='%230641c0' d='M19 14h1m-6 5h1'/%3E%3Cpath stroke='%231859f3' d='M1 15h1'/%3E%3Cpath stroke='%232e68f4' d='M2 15h1'/%3E%3Cpath stroke='%233a71f4' d='M3 15h1'/%3E%3Cpath stroke='%234277f5' d='M4 15h1'/%3E%3Cpath stroke='%23467af5' d='M5 15h1'/%3E%3Cpath stroke='%23457af6' d='M6 15h1'/%3E%3Cpath stroke='%23437bf6' d='M7 15h1'/%3E%3Cpath stroke='%23417cf6' d='M8 15h1'/%3E%3Cpath stroke='%23cbdcfd' d='M9 15h1'/%3E%3Cpath stroke='%23327df7' d='M11 15h1'/%3E%3Cpath stroke='%232a7cf8' d='M12 15h1'/%3E%3Cpath stroke='%23247af8' d='M13 15h1'/%3E%3Cpath stroke='%231d77f8' d='M14 15h1'/%3E%3Cpath stroke='%231573f8' d='M15 15h1'/%3E%3Cpath stroke='%230e6cf8' d='M16 15h1'/%3E%3Cpath stroke='%230963f4' d='M17 15h1'/%3E%3Cpath stroke='%230556e7' d='M18 15h1'/%3E%3Cpath stroke='%23023fbf' d='M19 15h1'/%3E%3Cpath stroke='%231456f3' d='M1 16h1'/%3E%3Cpath stroke='%232562f4' d='M2 16h1'/%3E%3Cpath stroke='%233971f4' d='M4 16h1'/%3E%3Cpath stroke='%233d74f5' d='M5 16h1'/%3E%3Cpath stroke='%233d74f6' d='M6 16h1'/%3E%3Cpath stroke='%233b75f5' d='M7 16h1'/%3E%3Cpath stroke='%233976f5' d='M8 16h1'/%3E%3Cpath stroke='%23f5f8fe' d='M9 16h1'/%3E%3Cpath stroke='%232c78f7' d='M11 16h1'/%3E%3Cpath stroke='%232577f7' d='M12 16h1'/%3E%3Cpath stroke='%231f76f7' d='M13 16h1'/%3E%3Cpath stroke='%231972f7' d='M14 16h1'/%3E%3Cpath stroke='%23116ef8' d='M15 16h1'/%3E%3Cpath stroke='%230b68f7' d='M16 16h1'/%3E%3Cpath stroke='%230560f4' d='M17 16h1'/%3E%3Cpath stroke='%230253e6' d='M18 16h1'/%3E%3Cpath stroke='%23013dbe' d='M19 16h1'/%3E%3Cpath stroke='%230e50ed' d='M1 17h1'/%3E%3Cpath stroke='%231c5bef' d='M2 17h1'/%3E%3Cpath stroke='%232863f0' d='M3 17h1'/%3E%3Cpath stroke='%232f68f0' d='M4 17h1'/%3E%3Cpath stroke='%23336bf1' d='M5 17h1'/%3E%3Cpath stroke='%23346cf1' d='M6 17h1'/%3E%3Cpath stroke='%23316cf2' d='M7 17h1'/%3E%3Cpath stroke='%23316df2' d='M8 17h1'/%3E%3Cpath stroke='%232e6ff2' d='M9 17h1'/%3E%3Cpath stroke='%232a70f2' d='M10 17h1'/%3E%3Cpath stroke='%232570f3' d='M11 17h1'/%3E%3Cpath stroke='%231f6ff3' d='M12 17h1'/%3E%3Cpath stroke='%23196df4' d='M13 17h1'/%3E%3Cpath stroke='%23136af4' d='M14 17h1'/%3E%3Cpath stroke='%230760f3' d='M16 17h1'/%3E%3Cpath stroke='%23025af0' d='M17 17h1'/%3E%3Cpath stroke='%23004de2' d='M18 17h1'/%3E%3Cpath stroke='%23003ab9' d='M19 17h1'/%3E%3Cpath stroke='%23285edf' d='M1 18h1'/%3E%3Cpath stroke='%23134fdf' d='M2 18h1'/%3E%3Cpath stroke='%231b55df' d='M3 18h1'/%3E%3Cpath stroke='%23215ae2' d='M4 18h1'/%3E%3Cpath stroke='%23255ce1' d='M5 18h1'/%3E%3Cpath stroke='%23265de0' d='M6 18h1'/%3E%3Cpath stroke='%23245ce1' d='M7 18h1'/%3E%3Cpath stroke='%23235ee2' d='M8 18h1'/%3E%3Cpath stroke='%23215ee2' d='M9 18h1'/%3E%3Cpath stroke='%231e5ee2' d='M10 18h1'/%3E%3Cpath stroke='%231b5fe5' d='M11 18h1'/%3E%3Cpath stroke='%23165ee5' d='M12 18h1'/%3E%3Cpath stroke='%23135de6' d='M13 18h1'/%3E%3Cpath stroke='%230e5be5' d='M14 18h1'/%3E%3Cpath stroke='%230958e6' d='M15 18h1'/%3E%3Cpath stroke='%230454e6' d='M16 18h1'/%3E%3Cpath stroke='%23014ee2' d='M17 18h1'/%3E%3Cpath stroke='%230045d3' d='M18 18h1'/%3E%3Cpath stroke='%231f4eb8' d='M19 18h1'/%3E%3Cpath stroke='%23d0daf1' d='M1 19h1'/%3E%3Cpath stroke='%232856c3' d='M2 19h1'/%3E%3Cpath stroke='%230d3fb6' d='M3 19h1'/%3E%3Cpath stroke='%231144bd' d='M4 19h1'/%3E%3Cpath stroke='%231245bb' d='M5 19h1'/%3E%3Cpath stroke='%231445b9' d='M6 19h1'/%3E%3Cpath stroke='%231244b9' d='M7 19h1'/%3E%3Cpath stroke='%231345bc' d='M8 19h1'/%3E%3Cpath stroke='%231346bd' d='M9 19h1'/%3E%3Cpath stroke='%231045be' d='M10 19h1'/%3E%3Cpath stroke='%230d45c0' d='M11 19h1'/%3E%3Cpath stroke='%230a45c1' d='M12 19h1'/%3E%3Cpath stroke='%230844c3' d='M13 19h1'/%3E%3Cpath stroke='%23033fc0' d='M15 19h1'/%3E%3Cpath stroke='%23013fc3' d='M16 19h1'/%3E%3Cpath stroke='%23003bbe' d='M17 19h1'/%3E%3Cpath stroke='%231f4eb9' d='M18 19h1'/%3E%3Cpath stroke='%23cfd8ed' d='M19 19h1'/%3E%3C/svg%3E\")}.title-bar-controls button[aria-label=Help]:hover{background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 -0.5 21 21' shape-rendering='crispEdges'%3E%3Cpath stroke='%2393b1ee' d='M1 0h1'/%3E%3Cpath stroke='%23f3f6fd' d='M2 0h1m17 2h1M0 18h1m17 2h1'/%3E%3Cpath stroke='%23fff' d='M3 0h15M0 3h1m19 0h1M0 4h1m8 0h3m8 0h1M0 5h1m7 0h1m3 0h1m7 0h1M0 6h1m6 0h1m5 0h1m6 0h1M0 7h1m12 0h1m6 0h1M0 8h1m12 0h1m6 0h1M0 9h1m12 0h1m6 0h1M0 10h1m10 0h2m7 0h1M0 11h1m9 0h1m9 0h1M0 12h1m9 0h1m9 0h1M0 13h1m19 0h1M0 14h1m19 0h1M0 15h1m9 0h1m9 0h1M0 16h1m9 0h1m9 0h1M0 17h1m19 0h1M3 20h15'/%3E%3Cpath stroke='%23f5f7fd' d='M18 0h1M0 2h1m19 16h1M2 20h1'/%3E%3Cpath stroke='%2393b1ed' d='M19 0h1M0 1h1'/%3E%3Cpath stroke='%23dce7ff' d='M1 1h1'/%3E%3Cpath stroke='%2372a1ff' d='M2 1h1m2 5h1'/%3E%3Cpath stroke='%236a9cff' d='M3 1h1'/%3E%3Cpath stroke='%235f94ff' d='M4 1h1M4 11h2'/%3E%3Cpath stroke='%23558eff' d='M5 1h1M3 12h1'/%3E%3Cpath stroke='%23518bff' d='M6 1h1'/%3E%3Cpath stroke='%234a86ff' d='M7 1h1'/%3E%3Cpath stroke='%234b87ff' d='M8 1h1M2 12h1'/%3E%3Cpath stroke='%234684ff' d='M9 1h2'/%3E%3Cpath stroke='%234482ff' d='M11 1h1m4 1h1M1 9h1m0 4h1'/%3E%3Cpath stroke='%234080ff' d='M12 1h1M3 15h1'/%3E%3Cpath stroke='%233b7cff' d='M13 1h1'/%3E%3Cpath stroke='%233a7bff' d='M14 1h1'/%3E%3Cpath stroke='%233678ff' d='M15 1h1'/%3E%3Cpath stroke='%232e73ff' d='M16 1h1'/%3E%3Cpath stroke='%23276cf9' d='M17 1h1'/%3E%3Cpath stroke='%233a73e7' d='M18 1h1'/%3E%3Cpath stroke='%23d3ddf3' d='M19 1h1'/%3E%3Cpath stroke='%2393b0ed' d='M20 1h1'/%3E%3Cpath stroke='%2373a1ff' d='M1 2h1'/%3E%3Cpath stroke='%2397b9ff' d='M2 2h1'/%3E%3Cpath stroke='%239cbdff' d='M3 2h1'/%3E%3Cpath stroke='%2390b5ff' d='M4 2h1'/%3E%3Cpath stroke='%2382acff' d='M5 2h1M5 4h1'/%3E%3Cpath stroke='%237ba7ff' d='M6 2h1M2 6h1'/%3E%3Cpath stroke='%2375a3ff' d='M7 2h1'/%3E%3Cpath stroke='%236f9fff' d='M8 2h1M3 8h1'/%3E%3Cpath stroke='%236c9dff' d='M9 2h1M1 3h1'/%3E%3Cpath stroke='%23689bff' d='M10 2h1M5 8h1M3 9h1'/%3E%3Cpath stroke='%236599ff' d='M11 2h1m0 1h1M5 9h1'/%3E%3Cpath stroke='%236095ff' d='M12 2h1m0 1h1'/%3E%3Cpath stroke='%235d93ff' d='M13 2h1'/%3E%3Cpath stroke='%23568eff' d='M14 2h1'/%3E%3Cpath stroke='%234f8aff' d='M15 2h1M3 13h1m0 1h1'/%3E%3Cpath stroke='%233878fb' d='M17 2h1'/%3E%3Cpath stroke='%232969eb' d='M18 2h1'/%3E%3Cpath stroke='%233566cb' d='M19 2h1'/%3E%3Cpath stroke='%239ebeff' d='M2 3h1'/%3E%3Cpath stroke='%23a4c2ff' d='M3 3h1'/%3E%3Cpath stroke='%2399baff' d='M4 3h1M3 4h1'/%3E%3Cpath stroke='%238ab0ff' d='M5 3h1'/%3E%3Cpath stroke='%2382abff' d='M6 3h1'/%3E%3Cpath stroke='%2379a6ff' d='M7 3h1'/%3E%3Cpath stroke='%2374a3ff' d='M8 3h1'/%3E%3Cpath stroke='%2371a0ff' d='M9 3h1'/%3E%3Cpath stroke='%236d9eff' d='M10 3h1M5 7h1M4 8h1'/%3E%3Cpath stroke='%23699bff' d='M11 3h1'/%3E%3Cpath stroke='%235a91ff' d='M14 3h1M2 10h1m1 2h1'/%3E%3Cpath stroke='%23538cff' d='M15 3h1M2 11h1'/%3E%3Cpath stroke='%234986ff' d='M16 3h1'/%3E%3Cpath stroke='%233d7cfc' d='M17 3h1'/%3E%3Cpath stroke='%232e6cea' d='M18 3h1'/%3E%3Cpath stroke='%231b52c2' d='M19 3h1'/%3E%3Cpath stroke='%236296ff' d='M1 4h1'/%3E%3Cpath stroke='%2391b5ff' d='M2 4h1'/%3E%3Cpath stroke='%238fb4ff' d='M4 4h1'/%3E%3Cpath stroke='%237aa6ff' d='M6 4h1m7 1h1'/%3E%3Cpath stroke='%239bbdff' d='M7 4h1'/%3E%3Cpath stroke='%23e3edff' d='M8 4h1'/%3E%3Cpath stroke='%23e1ebff' d='M12 4h1'/%3E%3Cpath stroke='%2387afff' d='M13 4h1'/%3E%3Cpath stroke='%234c88ff' d='M14 4h1m-5 2h1m-6 9h1'/%3E%3Cpath stroke='%234785ff' d='M15 4h1'/%3E%3Cpath stroke='%234280ff' d='M16 4h1'/%3E%3Cpath stroke='%233b7afb' d='M17 4h1'/%3E%3Cpath stroke='%23316fec' d='M18 4h1'/%3E%3Cpath stroke='%231f55c3' d='M19 4h1'/%3E%3Cpath stroke='%235990ff' d='M1 5h1'/%3E%3Cpath stroke='%2385adff' d='M2 5h1'/%3E%3Cpath stroke='%238bb1ff' d='M3 5h1'/%3E%3Cpath stroke='%2384acff' d='M4 5h1'/%3E%3Cpath stroke='%2378a5ff' d='M5 5h1'/%3E%3Cpath stroke='%239bf' d='M6 5h1'/%3E%3Cpath stroke='%23f4f7ff' d='M7 5h1'/%3E%3Cpath stroke='%23dbe7ff' d='M9 5h1'/%3E%3Cpath stroke='%23d2e1ff' d='M10 5h1'/%3E%3Cpath stroke='%23edf3ff' d='M11 5h1'/%3E%3Cpath stroke='%23f0f5ff' d='M13 5h1'/%3E%3Cpath stroke='%233b7bff' d='M15 5h1'/%3E%3Cpath stroke='%23397aff' d='M16 5h1M1 11h1'/%3E%3Cpath stroke='%233979fc' d='M17 5h1'/%3E%3Cpath stroke='%233370ec' d='M18 5h1m-1 1h1'/%3E%3Cpath stroke='%232357c3' d='M19 5h1'/%3E%3Cpath stroke='%23548dff' d='M1 6h1m2 7h1'/%3E%3Cpath stroke='%2381aaff' d='M3 6h1'/%3E%3Cpath stroke='%237aa7ff' d='M4 6h1'/%3E%3Cpath stroke='%23d8e5ff' d='M6 6h1'/%3E%3Cpath stroke='%23b9d0ff' d='M8 6h1'/%3E%3Cpath stroke='%23548eff' d='M9 6h1'/%3E%3Cpath stroke='%23538dff' d='M11 6h1'/%3E%3Cpath stroke='%23cbdcff' d='M12 6h1'/%3E%3Cpath stroke='%23c9dbff' d='M14 6h1'/%3E%3Cpath stroke='%233579ff' d='M15 6h1'/%3E%3Cpath stroke='%233679ff' d='M16 6h1'/%3E%3Cpath stroke='%233879fc' d='M17 6h1'/%3E%3Cpath stroke='%232358c5' d='M19 6h1'/%3E%3Cpath stroke='%234e89ff' d='M1 7h1'/%3E%3Cpath stroke='%2371a1ff' d='M2 7h1'/%3E%3Cpath stroke='%2377a5ff' d='M3 7h1'/%3E%3Cpath stroke='%2374a2ff' d='M4 7h1'/%3E%3Cpath stroke='%23e8f0ff' d='M6 7h1'/%3E%3Cpath stroke='%23e7efff' d='M7 7h1'/%3E%3Cpath stroke='%23679aff' d='M8 7h1'/%3E%3Cpath stroke='%23508dff' d='M9 7h1'/%3E%3Cpath stroke='%234989ff' d='M10 7h1'/%3E%3Cpath stroke='%234183ff' d='M11 7h1'/%3E%3Cpath stroke='%2374a5ff' d='M12 7h1'/%3E%3Cpath stroke='%23d1e1ff' d='M14 7h1'/%3E%3Cpath stroke='%23317aff' d='M15 7h1'/%3E%3Cpath stroke='%23337aff' d='M16 7h1'/%3E%3Cpath stroke='%23367bfc' d='M17 7h1'/%3E%3Cpath stroke='%233372ed' d='M18 7h1'/%3E%3Cpath stroke='%232359c5' d='M19 7h1'/%3E%3Cpath stroke='%234d88ff' d='M1 8h1'/%3E%3Cpath stroke='%23699cff' d='M2 8h1'/%3E%3Cpath stroke='%236398ff' d='M6 8h1'/%3E%3Cpath stroke='%235c93ff' d='M7 8h1m-2 3h1'/%3E%3Cpath stroke='%23548fff' d='M8 8h1'/%3E%3Cpath stroke='%234d8cff' d='M9 8h1'/%3E%3Cpath stroke='%23468aff' d='M10 8h1'/%3E%3Cpath stroke='%233f86ff' d='M11 8h1'/%3E%3Cpath stroke='%238cb7ff' d='M12 8h1'/%3E%3Cpath stroke='%23cde0ff' d='M14 8h1'/%3E%3Cpath stroke='%232f7fff' d='M15 8h1'/%3E%3Cpath stroke='%233280ff' d='M16 8h1'/%3E%3Cpath stroke='%233580fc' d='M17 8h1'/%3E%3Cpath stroke='%233276ed' d='M18 8h1'/%3E%3Cpath stroke='%23235ac6' d='M19 8h1'/%3E%3Cpath stroke='%236196ff' d='M2 9h1m3 0h1m-4 1h1'/%3E%3Cpath stroke='%23689aff' d='M4 9h1'/%3E%3Cpath stroke='%235b93ff' d='M7 9h1'/%3E%3Cpath stroke='%235491ff' d='M8 9h1'/%3E%3Cpath stroke='%234f90ff' d='M9 9h1'/%3E%3Cpath stroke='%234890ff' d='M10 9h1'/%3E%3Cpath stroke='%239dc5ff' d='M11 9h1'/%3E%3Cpath stroke='%23f3f8ff' d='M12 9h1'/%3E%3Cpath stroke='%239ac5ff' d='M14 9h1'/%3E%3Cpath stroke='%232f88ff' d='M15 9h1'/%3E%3Cpath stroke='%233188ff' d='M16 9h1'/%3E%3Cpath stroke='%233385fc' d='M17 9h1'/%3E%3Cpath stroke='%233079ed' d='M18 9h1'/%3E%3Cpath stroke='%23215cc8' d='M19 9h1'/%3E%3Cpath stroke='%233f7fff' d='M1 10h1'/%3E%3Cpath stroke='%236397ff' d='M4 10h1'/%3E%3Cpath stroke='%236297ff' d='M5 10h1'/%3E%3Cpath stroke='%235f95ff' d='M6 10h1'/%3E%3Cpath stroke='%235993ff' d='M7 10h1'/%3E%3Cpath stroke='%235492ff' d='M8 10h1'/%3E%3Cpath stroke='%235c9aff' d='M9 10h1'/%3E%3Cpath stroke='%23cee2ff' d='M10 10h1'/%3E%3Cpath stroke='%23a7d0ff' d='M13 10h1'/%3E%3Cpath stroke='%233897ff' d='M14 10h1'/%3E%3Cpath stroke='%232f92ff' d='M15 10h1'/%3E%3Cpath stroke='%233090ff' d='M16 10h1'/%3E%3Cpath stroke='%23328cfc' d='M17 10h1'/%3E%3Cpath stroke='%232e7def' d='M18 10h1'/%3E%3Cpath stroke='%231e5dc9' d='M19 10h1'/%3E%3Cpath stroke='%235c92ff' d='M3 11h1m1 1h1'/%3E%3Cpath stroke='%235792ff' d='M7 11h1m-1 1h1'/%3E%3Cpath stroke='%235594ff' d='M8 11h1'/%3E%3Cpath stroke='%23c2dbff' d='M9 11h1'/%3E%3Cpath stroke='%23e0efff' d='M11 11h1'/%3E%3Cpath stroke='%236eb6ff' d='M12 11h1'/%3E%3Cpath stroke='%23379fff' d='M13 11h1'/%3E%3Cpath stroke='%23339dff' d='M14 11h1'/%3E%3Cpath stroke='%232f9bff' d='M15 11h1'/%3E%3Cpath stroke='%232e97ff' d='M16 11h1'/%3E%3Cpath stroke='%232e91fc' d='M17 11h1'/%3E%3Cpath stroke='%232a80f0' d='M18 11h1'/%3E%3Cpath stroke='%231b5dcb' d='M19 11h1'/%3E%3Cpath stroke='%233275ff' d='M1 12h1'/%3E%3Cpath stroke='%235991ff' d='M6 12h1'/%3E%3Cpath stroke='%235596ff' d='M8 12h1'/%3E%3Cpath stroke='%23e2eeff' d='M9 12h1'/%3E%3Cpath stroke='%2359adff' d='M11 12h1'/%3E%3Cpath stroke='%2342a9ff' d='M12 12h1'/%3E%3Cpath stroke='%233aa9ff' d='M13 12h1'/%3E%3Cpath stroke='%2334a7ff' d='M14 12h1'/%3E%3Cpath stroke='%2330a5ff' d='M15 12h1'/%3E%3Cpath stroke='%232ca0ff' d='M16 12h1'/%3E%3Cpath stroke='%232a96fd' d='M17 12h1'/%3E%3Cpath stroke='%232581f1' d='M18 12h1'/%3E%3Cpath stroke='%23185dcc' d='M19 12h1'/%3E%3Cpath stroke='%232d72ff' d='M1 13h1m0 3h1'/%3E%3Cpath stroke='%235790ff' d='M5 13h2'/%3E%3Cpath stroke='%235490ff' d='M7 13h1'/%3E%3Cpath stroke='%235597ff' d='M8 13h1'/%3E%3Cpath stroke='%23d6e8ff' d='M9 13h1'/%3E%3Cpath stroke='%23eef6ff' d='M10 13h1'/%3E%3Cpath stroke='%234aaaff' d='M11 13h1'/%3E%3Cpath stroke='%2344afff' d='M12 13h1'/%3E%3Cpath stroke='%233eb1ff' d='M13 13h1'/%3E%3Cpath stroke='%2337afff' d='M14 13h1'/%3E%3Cpath stroke='%232fabff' d='M15 13h1'/%3E%3Cpath stroke='%2329a4ff' d='M16 13h1'/%3E%3Cpath stroke='%232599fd' d='M17 13h1'/%3E%3Cpath stroke='%231e80f2' d='M18 13h1'/%3E%3Cpath stroke='%23145bcd' d='M19 13h1'/%3E%3Cpath stroke='%23276eff' d='M1 14h1'/%3E%3Cpath stroke='%233d7dff' d='M2 14h1'/%3E%3Cpath stroke='%234985ff' d='M3 14h1'/%3E%3Cpath stroke='%23528cff' d='M5 14h1'/%3E%3Cpath stroke='%23528dff' d='M6 14h1'/%3E%3Cpath stroke='%23518fff' d='M7 14h1'/%3E%3Cpath stroke='%235196ff' d='M8 14h1'/%3E%3Cpath stroke='%23509fff' d='M9 14h1'/%3E%3Cpath stroke='%234ea6ff' d='M10 14h1'/%3E%3Cpath stroke='%2349acff' d='M11 14h1'/%3E%3Cpath stroke='%2343b1ff' d='M12 14h1'/%3E%3Cpath stroke='%233eb4ff' d='M13 14h1'/%3E%3Cpath stroke='%2335b2ff' d='M14 14h1'/%3E%3Cpath stroke='%232caeff' d='M15 14h1'/%3E%3Cpath stroke='%2324a5ff' d='M16 14h1'/%3E%3Cpath stroke='%231f97fd' d='M17 14h1'/%3E%3Cpath stroke='%231980f3' d='M18 14h1'/%3E%3Cpath stroke='%23105ace' d='M19 14h1'/%3E%3Cpath stroke='%23216aff' d='M1 15h1'/%3E%3Cpath stroke='%233578ff' d='M2 15h1'/%3E%3Cpath stroke='%234885ff' d='M4 15h1'/%3E%3Cpath stroke='%234d89ff' d='M6 15h1'/%3E%3Cpath stroke='%234c8cff' d='M7 15h1'/%3E%3Cpath stroke='%234d94ff' d='M8 15h1'/%3E%3Cpath stroke='%23cfe4ff' d='M9 15h1'/%3E%3Cpath stroke='%2347aaff' d='M11 15h1'/%3E%3Cpath stroke='%2341afff' d='M12 15h1'/%3E%3Cpath stroke='%233bb2ff' d='M13 15h1'/%3E%3Cpath stroke='%2333b1ff' d='M14 15h1'/%3E%3Cpath stroke='%232aadff' d='M15 15h1'/%3E%3Cpath stroke='%2321a3ff' d='M16 15h1'/%3E%3Cpath stroke='%231a95fd' d='M17 15h1'/%3E%3Cpath stroke='%23137cf2' d='M18 15h1'/%3E%3Cpath stroke='%230c59cf' d='M19 15h1'/%3E%3Cpath stroke='%231c66ff' d='M1 16h1'/%3E%3Cpath stroke='%233879ff' d='M3 16h1'/%3E%3Cpath stroke='%233f7eff' d='M4 16h1'/%3E%3Cpath stroke='%234483ff' d='M5 16h1'/%3E%3Cpath stroke='%234584ff' d='M6 16h1'/%3E%3Cpath stroke='%234587ff' d='M7 16h1'/%3E%3Cpath stroke='%23468eff' d='M8 16h1'/%3E%3Cpath stroke='%23f6faff' d='M9 16h1'/%3E%3Cpath stroke='%233fa3ff' d='M11 16h1'/%3E%3Cpath stroke='%233ba8ff' d='M12 16h1'/%3E%3Cpath stroke='%233af' d='M13 16h1'/%3E%3Cpath stroke='%232da9ff' d='M14 16h1'/%3E%3Cpath stroke='%2324a6ff' d='M15 16h1'/%3E%3Cpath stroke='%231d9eff' d='M16 16h1'/%3E%3Cpath stroke='%231690fd' d='M17 16h1'/%3E%3Cpath stroke='%231078f1' d='M18 16h1'/%3E%3Cpath stroke='%230b57ce' d='M19 16h1'/%3E%3Cpath stroke='%231761f9' d='M1 17h1'/%3E%3Cpath stroke='%23246bfa' d='M2 17h1'/%3E%3Cpath stroke='%232f72fb' d='M3 17h1'/%3E%3Cpath stroke='%233676fb' d='M4 17h1'/%3E%3Cpath stroke='%233a7afb' d='M5 17h1'/%3E%3Cpath stroke='%233b7bfc' d='M6 17h1'/%3E%3Cpath stroke='%233b7efc' d='M7 17h1'/%3E%3Cpath stroke='%233c84fc' d='M8 17h1'/%3E%3Cpath stroke='%233b8afc' d='M9 17h1'/%3E%3Cpath stroke='%233990fc' d='M10 17h1'/%3E%3Cpath stroke='%233695fc' d='M11 17h1'/%3E%3Cpath stroke='%233299fc' d='M12 17h1'/%3E%3Cpath stroke='%232c9cfd' d='M13 17h1'/%3E%3Cpath stroke='%23259bfd' d='M14 17h1'/%3E%3Cpath stroke='%231e97fd' d='M15 17h1'/%3E%3Cpath stroke='%231790fc' d='M16 17h1'/%3E%3Cpath stroke='%231184fa' d='M17 17h1'/%3E%3Cpath stroke='%230c6ded' d='M18 17h1'/%3E%3Cpath stroke='%230850c8' d='M19 17h1'/%3E%3Cpath stroke='%232f6ae4' d='M1 18h1'/%3E%3Cpath stroke='%231b5fe9' d='M2 18h1'/%3E%3Cpath stroke='%232163e8' d='M3 18h1'/%3E%3Cpath stroke='%232868eb' d='M4 18h1'/%3E%3Cpath stroke='%232c6aea' d='M5 18h1'/%3E%3Cpath stroke='%232e6dea' d='M6 18h1'/%3E%3Cpath stroke='%232d6deb' d='M7 18h1'/%3E%3Cpath stroke='%232c71ec' d='M8 18h1'/%3E%3Cpath stroke='%232c76ec' d='M9 18h1'/%3E%3Cpath stroke='%232a79ed' d='M10 18h1'/%3E%3Cpath stroke='%23287eef' d='M11 18h1'/%3E%3Cpath stroke='%232481f1' d='M12 18h1'/%3E%3Cpath stroke='%232182f1' d='M13 18h1'/%3E%3Cpath stroke='%231c80f1' d='M14 18h1'/%3E%3Cpath stroke='%231880f3' d='M15 18h1'/%3E%3Cpath stroke='%23117af2' d='M16 18h1'/%3E%3Cpath stroke='%230c6eed' d='M17 18h1'/%3E%3Cpath stroke='%230a5ddd' d='M18 18h1'/%3E%3Cpath stroke='%23265dc1' d='M19 18h1'/%3E%3Cpath stroke='%2393b4f2' d='M0 19h1'/%3E%3Cpath stroke='%23d1ddf4' d='M1 19h1'/%3E%3Cpath stroke='%232e61ca' d='M2 19h1'/%3E%3Cpath stroke='%23134bbf' d='M3 19h1'/%3E%3Cpath stroke='%23164fc2' d='M4 19h1'/%3E%3Cpath stroke='%231950c1' d='M5 19h1'/%3E%3Cpath stroke='%231b52c1' d='M6 19h1'/%3E%3Cpath stroke='%231a52c3' d='M7 19h1'/%3E%3Cpath stroke='%231954c6' d='M8 19h1'/%3E%3Cpath stroke='%231b58c9' d='M9 19h1'/%3E%3Cpath stroke='%231858c8' d='M10 19h1'/%3E%3Cpath stroke='%23165bcd' d='M11 19h1'/%3E%3Cpath stroke='%23145cd0' d='M12 19h1'/%3E%3Cpath stroke='%23135cd0' d='M13 19h1'/%3E%3Cpath stroke='%230f58cc' d='M14 19h1'/%3E%3Cpath stroke='%230d5ad2' d='M15 19h1'/%3E%3Cpath stroke='%230b58d1' d='M16 19h1'/%3E%3Cpath stroke='%230951cb' d='M17 19h1'/%3E%3Cpath stroke='%23265ec3' d='M18 19h1'/%3E%3Cpath stroke='%23d0daee' d='M19 19h1'/%3E%3Cpath stroke='%2393b3f2' d='M20 19h1M1 20h1'/%3E%3Cpath stroke='%2393b2f1' d='M19 20h1'/%3E%3C/svg%3E\")}.title-bar-controls button[aria-label=Help]:not(:disabled):active{background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 -0.5 21 21' shape-rendering='crispEdges'%3E%3Cpath stroke='%23a7bdef' d='M1 0h1'/%3E%3Cpath stroke='%23f4f6fd' d='M2 0h1m15 0h1M0 2h1m19 0h1M0 18h1m19 0h1M2 20h1m15 0h1'/%3E%3Cpath stroke='%23fff' d='M3 0h15M0 3h1m19 0h1M0 4h1m19 0h1M0 5h1m19 0h1M0 6h1m19 0h1M0 7h1m19 0h1M0 8h1m19 0h1M0 9h1m19 0h1M0 10h1m19 0h1M0 11h1m19 0h1M0 12h1m19 0h1M0 13h1m19 0h1M0 14h1m19 0h1M0 15h1m19 0h1M0 16h1m19 0h1M0 17h1m19 0h1M3 20h1m5 0h9'/%3E%3Cpath stroke='%23a7bdee' d='M19 0h1M0 1h1'/%3E%3Cpath stroke='%23cfd3da' d='M1 1h1'/%3E%3Cpath stroke='%231f3b5f' d='M2 1h1M1 2h1'/%3E%3Cpath stroke='%23002453' d='M3 1h1M1 4h1'/%3E%3Cpath stroke='%23002557' d='M4 1h1'/%3E%3Cpath stroke='%23002658' d='M5 1h1'/%3E%3Cpath stroke='%2300285c' d='M6 1h1'/%3E%3Cpath stroke='%23002a61' d='M7 1h1'/%3E%3Cpath stroke='%23002d67' d='M8 1h1'/%3E%3Cpath stroke='%23002f6b' d='M9 1h1'/%3E%3Cpath stroke='%23002f6c' d='M10 1h1M1 10h1'/%3E%3Cpath stroke='%23003273' d='M11 1h1'/%3E%3Cpath stroke='%23003478' d='M12 1h1M5 2h1'/%3E%3Cpath stroke='%2300357b' d='M13 1h1M2 5h1m-2 8h1'/%3E%3Cpath stroke='%2300377f' d='M14 1h1M6 2h1'/%3E%3Cpath stroke='%23003780' d='M15 1h1'/%3E%3Cpath stroke='%23003984' d='M16 1h1'/%3E%3Cpath stroke='%23003882' d='M17 1h1M3 3h1'/%3E%3Cpath stroke='%231f5295' d='M18 1h1'/%3E%3Cpath stroke='%23cfdae9' d='M19 1h1'/%3E%3Cpath stroke='%23a7bcee' d='M20 1h1'/%3E%3Cpath stroke='%23002a62' d='M2 2h1'/%3E%3Cpath stroke='%23003070' d='M3 2h1'/%3E%3Cpath stroke='%23003275' d='M4 2h1'/%3E%3Cpath stroke='%23003883' d='M7 2h1M1 17h1'/%3E%3Cpath stroke='%23003a88' d='M8 2h1'/%3E%3Cpath stroke='%23003d8f' d='M9 2h1M2 9h1'/%3E%3Cpath stroke='%23003e90' d='M10 2h1'/%3E%3Cpath stroke='%23004094' d='M11 2h1'/%3E%3Cpath stroke='%23004299' d='M12 2h1M2 12h1'/%3E%3Cpath stroke='%2300439b' d='M13 2h1'/%3E%3Cpath stroke='%2300449e' d='M14 2h1M2 14h1'/%3E%3Cpath stroke='%2300459f' d='M15 2h1'/%3E%3Cpath stroke='%230045a1' d='M16 2h1m1 0h1M2 17h1'/%3E%3Cpath stroke='%230045a0' d='M17 2h1M2 15h1'/%3E%3Cpath stroke='%231f5aa8' d='M19 2h1'/%3E%3Cpath stroke='%23002452' d='M1 3h1'/%3E%3Cpath stroke='%23003170' d='M2 3h1'/%3E%3Cpath stroke='%23003b8b' d='M4 3h1M3 4h1'/%3E%3Cpath stroke='%23003c8f' d='M5 3h1'/%3E%3Cpath stroke='%23003e94' d='M6 3h1'/%3E%3Cpath stroke='%23004099' d='M7 3h1'/%3E%3Cpath stroke='%2300429d' d='M8 3h1'/%3E%3Cpath stroke='%230044a2' d='M9 3h1'/%3E%3Cpath stroke='%230046a5' d='M10 3h1'/%3E%3Cpath stroke='%230048a8' d='M11 3h1'/%3E%3Cpath stroke='%230049ab' d='M12 3h1'/%3E%3Cpath stroke='%23004aac' d='M13 3h1'/%3E%3Cpath stroke='%23004aad' d='M14 3h1'/%3E%3Cpath stroke='%23004bae' d='M15 3h2m1 0h1M3 14h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23004baf' d='M17 3h1M7 10h1m-5 7h1m-1 1h1'/%3E%3Cpath stroke='%23004bad' d='M19 3h1M3 13h1m-1 6h1'/%3E%3Cpath stroke='%23037' d='M2 4h1m-2 8h1'/%3E%3Cpath stroke='%23003d92' d='M4 4h1'/%3E%3Cpath stroke='%23003f97' d='M5 4h1M4 5h1'/%3E%3Cpath stroke='%2300419d' d='M6 4h1M4 6h1'/%3E%3Cpath stroke='%230a4aa5' d='M7 4h1'/%3E%3Cpath stroke='%234e7ec0' d='M8 4h1'/%3E%3Cpath stroke='%23789ed1' d='M9 4h1'/%3E%3Cpath stroke='%23789ed3' d='M10 4h1'/%3E%3Cpath stroke='%23789fd4' d='M11 4h1m0 1h1'/%3E%3Cpath stroke='%235184c7' d='M12 4h1'/%3E%3Cpath stroke='%230b54b3' d='M13 4h1m0 1h1'/%3E%3Cpath stroke='%23004db1' d='M14 4h3m-2 1h2m-2 1h2M7 12h1m-2 1h1m-3 1h3m-3 1h2m-2 1h2'/%3E%3Cpath stroke='%23004db2' d='M17 4h3m-3 1h3m-2 1h2m-1 1h1m-9 1h1m-4 3h1m-5 6h2m-2 1h4m-4 1h4'/%3E%3Cpath stroke='%23002555' d='M1 5h1'/%3E%3Cpath stroke='%23003d90' d='M3 5h1'/%3E%3Cpath stroke='%2300409c' d='M5 5h1'/%3E%3Cpath stroke='%230949a4' d='M6 5h1'/%3E%3Cpath stroke='%23668ec8' d='M7 5h1'/%3E%3Cpath stroke='%23789dd1' d='M8 5h1M7 6h1'/%3E%3Cpath stroke='%23497cc1' d='M9 5h1'/%3E%3Cpath stroke='%234178c0' d='M10 5h1'/%3E%3Cpath stroke='%23608dcb' d='M11 5h1'/%3E%3Cpath stroke='%236693cf' d='M13 5h1'/%3E%3Cpath stroke='%2300275a' d='M1 6h1'/%3E%3Cpath stroke='%23003781' d='M2 6h1m-2 9h1'/%3E%3Cpath stroke='%23003f95' d='M3 6h1'/%3E%3Cpath stroke='%230042a1' d='M5 6h1'/%3E%3Cpath stroke='%234073bb' d='M6 6h1'/%3E%3Cpath stroke='%232661b6' d='M8 6h1'/%3E%3Cpath stroke='%230047ac' d='M9 6h1'/%3E%3Cpath stroke='%230049ad' d='M10 6h1m-6 5h1'/%3E%3Cpath stroke='%23004aae' d='M11 6h1m-6 5h1m-3 1h2'/%3E%3Cpath stroke='%234077c4' d='M12 6h1'/%3E%3Cpath stroke='%2378a1d6' d='M13 6h1'/%3E%3Cpath stroke='%234079c4' d='M14 6h1'/%3E%3Cpath stroke='%23004eb3' d='M17 6h1m0 1h1m0 1h1M10 9h1m-2 1h1m-3 6h1m-2 1h2m0 2h1'/%3E%3Cpath stroke='%2300295f' d='M1 7h1'/%3E%3Cpath stroke='%23003985' d='M2 7h1'/%3E%3Cpath stroke='%2300419b' d='M3 7h1'/%3E%3Cpath stroke='%230043a2' d='M4 7h1'/%3E%3Cpath stroke='%230044a6' d='M5 7h1'/%3E%3Cpath stroke='%235684c6' d='M6 7h1'/%3E%3Cpath stroke='%235686c8' d='M7 7h1'/%3E%3Cpath stroke='%230049ac' d='M8 7h1m-4 3h1m-2 1h1m-2 1h1'/%3E%3Cpath stroke='%230049ae' d='M9 7h1M7 8h2m-3 2h1'/%3E%3Cpath stroke='%23004aaf' d='M10 7h1M9 8h1M7 9h1'/%3E%3Cpath stroke='%23004cb1' d='M11 7h1m-2 1h1M9 9h1m-2 1h1'/%3E%3Cpath stroke='%230a53b5' d='M12 7h1'/%3E%3Cpath stroke='%2378a1d7' d='M13 7h1'/%3E%3Cpath stroke='%234881c8' d='M14 7h1'/%3E%3Cpath stroke='%23004fb4' d='M15 7h3m0 1h1m0 1h1M8 12h1m-2 3h1m0 3h1m0 1h1'/%3E%3Cpath stroke='%23002b63' d='M1 8h1'/%3E%3Cpath stroke='%23003b8a' d='M2 8h1'/%3E%3Cpath stroke='%2300439f' d='M3 8h1'/%3E%3Cpath stroke='%230045a5' d='M4 8h1'/%3E%3Cpath stroke='%230046a8' d='M5 8h1'/%3E%3Cpath stroke='%230047ab' d='M6 8h1M5 9h1'/%3E%3Cpath stroke='%23145db9' d='M12 8h1'/%3E%3Cpath stroke='%2378a2d8' d='M13 8h1'/%3E%3Cpath stroke='%23457fc8' d='M14 8h1'/%3E%3Cpath stroke='%230051b6' d='M15 8h1m2 1h1m0 2h1m-1 1h1M8 14h1m-1 1h1m10 2h1M9 18h1m1 1h1'/%3E%3Cpath stroke='%230050b5' d='M16 8h2m1 2h1M8 13h1m-1 3h1m-1 1h1m1 2h1'/%3E%3Cpath stroke='%23002d68' d='M1 9h1'/%3E%3Cpath stroke='%230045a3' d='M3 9h1'/%3E%3Cpath stroke='%230047a8' d='M4 9h1'/%3E%3Cpath stroke='%230048ad' d='M6 9h1'/%3E%3Cpath stroke='%23004bb0' d='M8 9h1m-3 3h1m-2 1h1'/%3E%3Cpath stroke='%231b62bd' d='M11 9h1'/%3E%3Cpath stroke='%236899d4' d='M12 9h1'/%3E%3Cpath stroke='%2378a4d9' d='M13 9h1'/%3E%3Cpath stroke='%231f68c1' d='M14 9h1'/%3E%3Cpath stroke='%230054b9' d='M15 9h1m-7 5h1m8 4h1m-4 1h1'/%3E%3Cpath stroke='%230053b8' d='M16 9h2m0 1h1m0 4h1m-1 2h1M9 17h1m0 1h1m3 1h1m1 0h1'/%3E%3Cpath stroke='%23003f93' d='M2 10h1'/%3E%3Cpath stroke='%230047a7' d='M3 10h1'/%3E%3Cpath stroke='%230048ab' d='M4 10h1'/%3E%3Cpath stroke='%23407cc7' d='M10 10h1'/%3E%3Cpath stroke='%2378a3d9' d='M11 10h1m-2 1h1'/%3E%3Cpath stroke='%2378a5da' d='M12 10h1m-3 2h1'/%3E%3Cpath stroke='%23256ec4' d='M13 10h1'/%3E%3Cpath stroke='%230057bb' d='M14 10h1'/%3E%3Cpath stroke='%230057bc' d='M15 10h1m-5 2h1m-2 2h1m7 3h1m-7 1h1m4 0h1'/%3E%3Cpath stroke='%230056bb' d='M16 10h1m1 2h1'/%3E%3Cpath stroke='%230055ba' d='M17 10h1m0 1h1m-9 6h1m0 1h1'/%3E%3Cpath stroke='%23003172' d='M1 11h1'/%3E%3Cpath stroke='%23004095' d='M2 11h1'/%3E%3Cpath stroke='%230048aa' d='M3 11h1'/%3E%3Cpath stroke='%23004cb0' d='M7 11h1m-4 2h1'/%3E%3Cpath stroke='%233272c4' d='M9 11h1'/%3E%3Cpath stroke='%23538cd0' d='M11 11h1'/%3E%3Cpath stroke='%23065cbf' d='M12 11h1'/%3E%3Cpath stroke='%230059be' d='M13 11h1m2 0h1m-6 2h1m-1 3h1m6 0h1m-5 2h1m1 0h1'/%3E%3Cpath stroke='%23005abf' d='M14 11h2m-4 1h1m4 0h1m-7 2h1m-1 1h1m0 2h1m2 1h1'/%3E%3Cpath stroke='%230058bd' d='M17 11h1m0 2h1m-6 5h1'/%3E%3Cpath stroke='%23538ace' d='M9 12h1'/%3E%3Cpath stroke='%23005cc1' d='M13 12h1m2 0h1m-5 1h1m4 0h1m-5 4h1'/%3E%3Cpath stroke='%23005dc2' d='M14 12h1m-3 2h1m4 0h1m-6 1h1m4 1h1m-4 1h1m1 0h1'/%3E%3Cpath stroke='%23005ec3' d='M15 12h1m-3 1h1m2 0h1m0 2h1m-5 1h1m1 1h1'/%3E%3Cpath stroke='%2300449d' d='M2 13h1'/%3E%3Cpath stroke='%23004eb2' d='M7 13h1m-2 2h1m-1 1h1'/%3E%3Cpath stroke='%234581cb' d='M9 13h1'/%3E%3Cpath stroke='%236297d5' d='M10 13h1'/%3E%3Cpath stroke='%23005fc4' d='M14 13h1m-2 1h1m2 0h1m-4 1h1'/%3E%3Cpath stroke='%230060c5' d='M15 13h1m-2 1h1m1 1h1m-2 1h1'/%3E%3Cpath stroke='%230052b7' d='M19 13h1m-8 6h2m3 0h1'/%3E%3Cpath stroke='%2300367e' d='M1 14h1'/%3E%3Cpath stroke='%23004fb3' d='M7 14h1'/%3E%3Cpath stroke='%230061c6' d='M15 14h1m-2 1h1'/%3E%3Cpath stroke='%230059bd' d='M18 14h1'/%3E%3Cpath stroke='%23407fca' d='M9 15h1'/%3E%3Cpath stroke='%2378a6dc' d='M10 15h1'/%3E%3Cpath stroke='%230062c6' d='M15 15h1'/%3E%3Cpath stroke='%23005abe' d='M18 15h1'/%3E%3Cpath stroke='%230054b8' d='M19 15h1'/%3E%3Cpath stroke='%23003881' d='M1 16h1'/%3E%3Cpath stroke='%230046a1' d='M2 16h1'/%3E%3Cpath stroke='%236c9bd5' d='M9 16h1'/%3E%3Cpath stroke='%2378a6db' d='M10 16h1'/%3E%3Cpath stroke='%23005cc0' d='M12 16h1'/%3E%3Cpath stroke='%23005fc3' d='M14 16h1'/%3E%3Cpath stroke='%230060c4' d='M16 16h1'/%3E%3Cpath stroke='%230058bc' d='M11 17h1'/%3E%3Cpath stroke='%23005bc0' d='M17 17h1'/%3E%3Cpath stroke='%231f5294' d='M1 18h1'/%3E%3Cpath stroke='%230046a2' d='M2 18h1'/%3E%3Cpath stroke='%231f66be' d='M19 18h1'/%3E%3Cpath stroke='%23a7bef0' d='M0 19h1m19 0h1M1 20h1'/%3E%3Cpath stroke='%23cfdae8' d='M1 19h1'/%3E%3Cpath stroke='%231f5ba9' d='M2 19h1'/%3E%3Cpath stroke='%231f66bf' d='M18 19h1'/%3E%3Cpath stroke='%23cfdef1' d='M19 19h1'/%3E%3Cpath stroke='%23fefefe' d='M4 20h1m3 0h1'/%3E%3Cpath stroke='%23fdfdfd' d='M5 20h1m1 0h1'/%3E%3Cpath stroke='%23fcfcfc' d='M6 20h1'/%3E%3Cpath stroke='%23a7bdf0' d='M19 20h1'/%3E%3C/svg%3E\")}.title-bar-controls button[aria-label=Close]{background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 -0.5 21 21' shape-rendering='crispEdges'%3E%3Cpath stroke='%23b3c4ef' d='M1 0h1m17 0h1M0 1h1m19 0h1M0 19h1m19 0h1M1 20h1m17 0h1'/%3E%3Cpath stroke='%23f4f6fd' d='M2 0h1m17 2h1M0 18h1m17 2h1'/%3E%3Cpath stroke='%23fff' d='M3 0h16M0 2h1M0 3h1m19 0h1M0 4h1m19 0h1M0 5h1m5 0h1m7 0h1m5 0h1M0 6h1m4 0h3m5 0h3m4 0h1M0 7h1m5 0h3m3 0h3m5 0h1M0 8h1m6 0h3m1 0h3m6 0h1M0 9h1m7 0h5m7 0h1M0 10h1m8 0h3m8 0h1M0 11h1m7 0h5m7 0h1M0 12h1m6 0h3m1 0h2m7 0h1M0 13h1m5 0h3m3 0h3m5 0h1M0 14h1m4 0h3m5 0h3m4 0h1M0 15h1m5 0h1m7 0h1m5 0h1M0 16h1m19 0h1M0 17h1m19 0h1m-1 1h1M2 20h16'/%3E%3Cpath stroke='%23fae1dc' d='M1 1h1'/%3E%3Cpath stroke='%23eb8b73' d='M2 1h1'/%3E%3Cpath stroke='%23e97b60' d='M3 1h1'/%3E%3Cpath stroke='%23e77155' d='M4 1h1'/%3E%3Cpath stroke='%23e66a4d' d='M5 1h1M1 6h1m5 4h1'/%3E%3Cpath stroke='%23e56648' d='M6 1h1'/%3E%3Cpath stroke='%23e46142' d='M7 1h1'/%3E%3Cpath stroke='%23e46344' d='M8 1h1m5 3h1M2 12h1'/%3E%3Cpath stroke='%23e45f3e' d='M9 1h2'/%3E%3Cpath stroke='%23e35c3b' d='M11 1h2'/%3E%3Cpath stroke='%23e25633' d='M13 1h2'/%3E%3Cpath stroke='%23e25330' d='M15 1h1'/%3E%3Cpath stroke='%23e04d28' d='M16 1h1'/%3E%3Cpath stroke='%23dc451f' d='M17 1h1'/%3E%3Cpath stroke='%23d05334' d='M18 1h1'/%3E%3Cpath stroke='%23efd8d2' d='M19 1h1'/%3E%3Cpath stroke='%23ec8d76' d='M1 2h1'/%3E%3Cpath stroke='%23efa390' d='M2 2h1'/%3E%3Cpath stroke='%23f0a694' d='M3 2h1'/%3E%3Cpath stroke='%23ee9a85' d='M4 2h1'/%3E%3Cpath stroke='%23eb8d75' d='M5 2h1'/%3E%3Cpath stroke='%23ea876e' d='M6 2h1'/%3E%3Cpath stroke='%23ea8168' d='M7 2h1'/%3E%3Cpath stroke='%23e97f66' d='M8 2h1'/%3E%3Cpath stroke='%23e97c62' d='M9 2h1m0 1h1'/%3E%3Cpath stroke='%23e8795f' d='M10 2h1'/%3E%3Cpath stroke='%23e8795e' d='M11 2h1'/%3E%3Cpath stroke='%23e87559' d='M12 2h1'/%3E%3Cpath stroke='%23e77256' d='M13 2h1'/%3E%3Cpath stroke='%23e66e50' d='M14 2h1'/%3E%3Cpath stroke='%23e56849' d='M15 2h1'/%3E%3Cpath stroke='%23e4603f' d='M16 2h1m-2 2h1'/%3E%3Cpath stroke='%23e05532' d='M17 2h1'/%3E%3Cpath stroke='%23d04623' d='M18 2h1'/%3E%3Cpath stroke='%23b64b30' d='M19 2h1'/%3E%3Cpath stroke='%23e97f65' d='M1 3h1'/%3E%3Cpath stroke='%23f0a997' d='M2 3h1'/%3E%3Cpath stroke='%23f1ac9a' d='M3 3h1'/%3E%3Cpath stroke='%23ee9d89' d='M4 3h1M2 4h1'/%3E%3Cpath stroke='%23ec917a' d='M5 3h1'/%3E%3Cpath stroke='%23eb8b72' d='M6 3h1'/%3E%3Cpath stroke='%23ea856d' d='M7 3h1'/%3E%3Cpath stroke='%23e98168' d='M8 3h1M2 7h1'/%3E%3Cpath stroke='%23e87e65' d='M9 3h1'/%3E%3Cpath stroke='%23e97b61' d='M11 3h1'/%3E%3Cpath stroke='%23e8775d' d='M12 3h1M3 9h1'/%3E%3Cpath stroke='%23e87459' d='M13 3h1M2 9h1'/%3E%3Cpath stroke='%23e66f52' d='M14 3h1'/%3E%3Cpath stroke='%23e56a4c' d='M15 3h1'/%3E%3Cpath stroke='%23e46343' d='M16 3h1'/%3E%3Cpath stroke='%23e15937' d='M17 3h1'/%3E%3Cpath stroke='%23d24a28' d='M18 3h1'/%3E%3Cpath stroke='%23aa3315' d='M19 3h1'/%3E%3Cpath stroke='%23e87458' d='M1 4h1'/%3E%3Cpath stroke='%23efa18d' d='M3 4h1'/%3E%3Cpath stroke='%23ed957f' d='M4 4h1'/%3E%3Cpath stroke='%23eb8a71' d='M5 4h1M4 5h1'/%3E%3Cpath stroke='%23ea836a' d='M6 4h1M4 6h1M3 7h1'/%3E%3Cpath stroke='%23e97d64' d='M7 4h1'/%3E%3Cpath stroke='%23e8785e' d='M8 4h1'/%3E%3Cpath stroke='%23e77359' d='M9 4h1'/%3E%3Cpath stroke='%23e76f54' d='M10 4h1'/%3E%3Cpath stroke='%23e66d51' d='M11 4h1'/%3E%3Cpath stroke='%23e5684b' d='M12 4h1'/%3E%3Cpath stroke='%23e5684a' d='M13 4h1'/%3E%3Cpath stroke='%23e35c3a' d='M16 4h1m-7 4h1m-8 7h1'/%3E%3Cpath stroke='%23e05634' d='M17 4h1'/%3E%3Cpath stroke='%23d24c2a' d='M18 4h1'/%3E%3Cpath stroke='%23ac3618' d='M19 4h1'/%3E%3Cpath stroke='%23e76f52' d='M1 5h1m4 6h1m-3 1h1'/%3E%3Cpath stroke='%23ec9179' d='M2 5h1'/%3E%3Cpath stroke='%23ec937c' d='M3 5h1'/%3E%3Cpath stroke='%23f7ccc2' d='M5 5h1'/%3E%3Cpath stroke='%23e77259' d='M7 5h1M5 9h1'/%3E%3Cpath stroke='%23e76d53' d='M8 5h1'/%3E%3Cpath stroke='%23e5684d' d='M9 5h1M8 6h1'/%3E%3Cpath stroke='%23e46446' d='M10 5h1'/%3E%3Cpath stroke='%23e45f41' d='M11 5h1'/%3E%3Cpath stroke='%23e35b3a' d='M12 5h1m-2 1h1'/%3E%3Cpath stroke='%23e35938' d='M13 5h1'/%3E%3Cpath stroke='%23f3bbad' d='M15 5h1'/%3E%3Cpath stroke='%23e25531' d='M16 5h1'/%3E%3Cpath stroke='%23df5330' d='M17 5h1'/%3E%3Cpath stroke='%23d34e2c' d='M18 5h1'/%3E%3Cpath stroke='%23ad3a1d' d='M19 5h1m-1 1h1'/%3E%3Cpath stroke='%23eb876e' d='M2 6h1'/%3E%3Cpath stroke='%23eb8a70' d='M3 6h1'/%3E%3Cpath stroke='%23e46447' d='M9 6h1'/%3E%3Cpath stroke='%23e45f40' d='M10 6h1'/%3E%3Cpath stroke='%23e25634' d='M12 6h1'/%3E%3Cpath stroke='%23e2522d' d='M16 6h1'/%3E%3Cpath stroke='%23df522e' d='M17 6h1'/%3E%3Cpath stroke='%23d34d2c' d='M18 6h1'/%3E%3Cpath stroke='%23e56546' d='M1 7h1M1 8h1'/%3E%3Cpath stroke='%23e97e65' d='M4 7h1'/%3E%3Cpath stroke='%23e8775e' d='M5 7h1'/%3E%3Cpath stroke='%23e46143' d='M9 7h1'/%3E%3Cpath stroke='%23e45d3d' d='M10 7h1'/%3E%3Cpath stroke='%23e35836' d='M11 7h1'/%3E%3Cpath stroke='%23e24e27' d='M15 7h1'/%3E%3Cpath stroke='%23e2502a' d='M16 7h1'/%3E%3Cpath stroke='%23e0512c' d='M17 7h1'/%3E%3Cpath stroke='%23d34d2a' d='M18 7h1'/%3E%3Cpath stroke='%23ad391c' d='M19 7h1'/%3E%3Cpath stroke='%23e87a60' d='M2 8h1m1 0h1'/%3E%3Cpath stroke='%23e87c62' d='M3 8h1'/%3E%3Cpath stroke='%23e8745b' d='M5 8h1'/%3E%3Cpath stroke='%23e76e54' d='M6 8h1'/%3E%3Cpath stroke='%23e24d24' d='M14 8h1'/%3E%3Cpath stroke='%23e24b22' d='M15 8h1'/%3E%3Cpath stroke='%23e24d25' d='M16 8h1'/%3E%3Cpath stroke='%23e05029' d='M17 8h1'/%3E%3Cpath stroke='%23d44c29' d='M18 8h1'/%3E%3Cpath stroke='%23ae391b' d='M19 8h1'/%3E%3Cpath stroke='%23e35d3c' d='M1 9h1'/%3E%3Cpath stroke='%23e8765d' d='M4 9h1'/%3E%3Cpath stroke='%23e66f53' d='M6 9h1'/%3E%3Cpath stroke='%23e56b4e' d='M7 9h1'/%3E%3Cpath stroke='%23e45127' d='M13 9h1'/%3E%3Cpath stroke='%23e44f23' d='M14 9h1'/%3E%3Cpath stroke='%23e34c20' d='M15 9h1'/%3E%3Cpath stroke='%23e34d22' d='M16 9h1'/%3E%3Cpath stroke='%23e14f25' d='M17 9h1'/%3E%3Cpath stroke='%23d54a25' d='M18 9h1'/%3E%3Cpath stroke='%23af3719' d='M19 9h1'/%3E%3Cpath stroke='%23e35937' d='M1 10h1'/%3E%3Cpath stroke='%23e76d51' d='M2 10h1'/%3E%3Cpath stroke='%23e87257' d='M3 10h1'/%3E%3Cpath stroke='%23e87359' d='M4 10h1'/%3E%3Cpath stroke='%23e77157' d='M5 10h1'/%3E%3Cpath stroke='%23e66e52' d='M6 10h1'/%3E%3Cpath stroke='%23e56747' d='M8 10h1'/%3E%3Cpath stroke='%23e5572c' d='M12 10h1'/%3E%3Cpath stroke='%23e55326' d='M13 10h1'/%3E%3Cpath stroke='%23e55022' d='M14 10h1'/%3E%3Cpath stroke='%23e54d1e' d='M15 10h1'/%3E%3Cpath stroke='%23e54d1f' d='M16 10h1'/%3E%3Cpath stroke='%23e24e21' d='M17 10h1'/%3E%3Cpath stroke='%23d64921' d='M18 10h1'/%3E%3Cpath stroke='%23af3516' d='M19 10h1'/%3E%3Cpath stroke='%23e25432' d='M1 11h1'/%3E%3Cpath stroke='%23e5694b' d='M2 11h1'/%3E%3Cpath stroke='%23e77054' d='M3 11h1'/%3E%3Cpath stroke='%23e77156' d='M4 11h1'/%3E%3Cpath stroke='%23e87055' d='M5 11h1'/%3E%3Cpath stroke='%23e66c4d' d='M7 11h1'/%3E%3Cpath stroke='%23e75526' d='M13 11h1'/%3E%3Cpath stroke='%23e75221' d='M14 11h1'/%3E%3Cpath stroke='%23e64e1c' d='M15 11h1'/%3E%3Cpath stroke='%23e64d1c' d='M16 11h1'/%3E%3Cpath stroke='%23e34c1c' d='M17 11h1'/%3E%3Cpath stroke='%23d6461c' d='M18 11h1'/%3E%3Cpath stroke='%23b03312' d='M19 11h1'/%3E%3Cpath stroke='%23e14f2b' d='M1 12h1'/%3E%3Cpath stroke='%23e66b4e' d='M3 12h1'/%3E%3Cpath stroke='%23e76f53' d='M5 12h1'/%3E%3Cpath stroke='%23e66e51' d='M6 12h1'/%3E%3Cpath stroke='%23e7653d' d='M10 12h1'/%3E%3Cpath stroke='%23fef5f1' d='M13 12h1'/%3E%3Cpath stroke='%23e85421' d='M14 12h1'/%3E%3Cpath stroke='%23e8501b' d='M15 12h1'/%3E%3Cpath stroke='%23e74d18' d='M16 12h1'/%3E%3Cpath stroke='%23e44a18' d='M17 12h1'/%3E%3Cpath stroke='%23d74216' d='M18 12h1'/%3E%3Cpath stroke='%23b2310f' d='M19 12h1'/%3E%3Cpath stroke='%23e04b25' d='M1 13h1m0 3h1'/%3E%3Cpath stroke='%23e35e3d' d='M2 13h1'/%3E%3Cpath stroke='%23e56748' d='M3 13h1'/%3E%3Cpath stroke='%23e66c4e' d='M4 13h1'/%3E%3Cpath stroke='%23e66d50' d='M5 13h1'/%3E%3Cpath stroke='%23e76842' d='M9 13h1'/%3E%3Cpath stroke='%23e7653c' d='M10 13h1'/%3E%3Cpath stroke='%23e86236' d='M11 13h1'/%3E%3Cpath stroke='%23e95019' d='M15 13h1m-2 3h1'/%3E%3Cpath stroke='%23e84c16' d='M16 13h1'/%3E%3Cpath stroke='%23e44713' d='M17 13h1'/%3E%3Cpath stroke='%23d83f10' d='M18 13h1'/%3E%3Cpath stroke='%23b12d0a' d='M19 13h1'/%3E%3Cpath stroke='%23df451e' d='M1 14h1'/%3E%3Cpath stroke='%23e25836' d='M2 14h1'/%3E%3Cpath stroke='%23e46242' d='M3 14h1m0 1h1'/%3E%3Cpath stroke='%23e56749' d='M4 14h1'/%3E%3Cpath stroke='%23e66845' d='M8 14h1'/%3E%3Cpath stroke='%23e76741' d='M9 14h1'/%3E%3Cpath stroke='%23e7643b' d='M10 14h1'/%3E%3Cpath stroke='%23e86235' d='M11 14h1'/%3E%3Cpath stroke='%23ea5e2d' d='M12 14h1'/%3E%3Cpath stroke='%23e94a11' d='M16 14h1m-2 2h1'/%3E%3Cpath stroke='%23e6440d' d='M17 14h1'/%3E%3Cpath stroke='%23d73b0b' d='M18 14h1'/%3E%3Cpath stroke='%23b12b06' d='M19 14h1'/%3E%3Cpath stroke='%23de4018' d='M1 15h1'/%3E%3Cpath stroke='%23e1512e' d='M2 15h1'/%3E%3Cpath stroke='%23f5c1b5' d='M5 15h1'/%3E%3Cpath stroke='%23e66543' d='M7 15h1'/%3E%3Cpath stroke='%23e66541' d='M8 15h1'/%3E%3Cpath stroke='%23e6643d' d='M9 15h1'/%3E%3Cpath stroke='%23e76238' d='M10 15h1'/%3E%3Cpath stroke='%23e86032' d='M11 15h1'/%3E%3Cpath stroke='%23e95c2a' d='M12 15h1'/%3E%3Cpath stroke='%23ea5924' d='M13 15h1'/%3E%3Cpath stroke='%23f7b8a1' d='M15 15h1'/%3E%3Cpath stroke='%23e9480e' d='M16 15h1'/%3E%3Cpath stroke='%23e54009' d='M17 15h1'/%3E%3Cpath stroke='%23d73605' d='M18 15h1'/%3E%3Cpath stroke='%23b02702' d='M19 15h1'/%3E%3Cpath stroke='%23dd3c14' d='M1 16h1'/%3E%3Cpath stroke='%23e15431' d='M3 16h1'/%3E%3Cpath stroke='%23e35b39' d='M4 16h1'/%3E%3Cpath stroke='%23e45e3d' d='M5 16h1'/%3E%3Cpath stroke='%23e45f3d' d='M6 16h1'/%3E%3Cpath stroke='%23e45e3b' d='M7 16h1'/%3E%3Cpath stroke='%23e55e39' d='M8 16h1'/%3E%3Cpath stroke='%23e55e37' d='M9 16h1'/%3E%3Cpath stroke='%23e65d32' d='M10 16h1'/%3E%3Cpath stroke='%23e75b2c' d='M11 16h1'/%3E%3Cpath stroke='%23e85725' d='M12 16h1'/%3E%3Cpath stroke='%23e9541f' d='M13 16h1'/%3E%3Cpath stroke='%23e8440b' d='M16 16h1'/%3E%3Cpath stroke='%23e43d05' d='M17 16h1'/%3E%3Cpath stroke='%23d63302' d='M18 16h1'/%3E%3Cpath stroke='%23af2601' d='M19 16h1'/%3E%3Cpath stroke='%23d8370e' d='M1 17h1'/%3E%3Cpath stroke='%23db431c' d='M2 17h1'/%3E%3Cpath stroke='%23dd4c28' d='M3 17h1'/%3E%3Cpath stroke='%23de522f' d='M4 17h1'/%3E%3Cpath stroke='%23df5533' d='M5 17h1'/%3E%3Cpath stroke='%23e05734' d='M6 17h1'/%3E%3Cpath stroke='%23e05531' d='M7 17h1'/%3E%3Cpath stroke='%23e05631' d='M8 17h1'/%3E%3Cpath stroke='%23e1562e' d='M9 17h1'/%3E%3Cpath stroke='%23e2552a' d='M10 17h1'/%3E%3Cpath stroke='%23e45325' d='M11 17h1'/%3E%3Cpath stroke='%23e4501f' d='M12 17h1'/%3E%3Cpath stroke='%23e54c19' d='M13 17h1'/%3E%3Cpath stroke='%23e54813' d='M14 17h1'/%3E%3Cpath stroke='%23e5430d' d='M15 17h1'/%3E%3Cpath stroke='%23e43e07' d='M16 17h1'/%3E%3Cpath stroke='%23e03802' d='M17 17h1'/%3E%3Cpath stroke='%23d12f00' d='M18 17h1'/%3E%3Cpath stroke='%23aa2300' d='M19 17h1'/%3E%3Cpath stroke='%23cd4928' d='M1 18h1'/%3E%3Cpath stroke='%23cc3813' d='M2 18h1'/%3E%3Cpath stroke='%23cc3e1b' d='M3 18h1'/%3E%3Cpath stroke='%23cf4421' d='M4 18h1'/%3E%3Cpath stroke='%23cf4725' d='M5 18h1'/%3E%3Cpath stroke='%23cf4726' d='M6 18h1'/%3E%3Cpath stroke='%23cf4624' d='M7 18h1'/%3E%3Cpath stroke='%23d04723' d='M8 18h1'/%3E%3Cpath stroke='%23d14621' d='M9 18h1'/%3E%3Cpath stroke='%23d2451e' d='M10 18h1'/%3E%3Cpath stroke='%23d5451b' d='M11 18h1'/%3E%3Cpath stroke='%23d54216' d='M12 18h1'/%3E%3Cpath stroke='%23d64013' d='M13 18h1'/%3E%3Cpath stroke='%23d53d0e' d='M14 18h1'/%3E%3Cpath stroke='%23d63909' d='M15 18h1'/%3E%3Cpath stroke='%23d53504' d='M16 18h1'/%3E%3Cpath stroke='%23d13001' d='M17 18h1'/%3E%3Cpath stroke='%23c22a00' d='M18 18h1'/%3E%3Cpath stroke='%23ab3c1f' d='M19 18h1'/%3E%3Cpath stroke='%23eed6d0' d='M1 19h1'/%3E%3Cpath stroke='%23b54428' d='M2 19h1'/%3E%3Cpath stroke='%23a62b0d' d='M3 19h1'/%3E%3Cpath stroke='%23ac3011' d='M4 19h1'/%3E%3Cpath stroke='%23ab3112' d='M5 19h1'/%3E%3Cpath stroke='%23a93214' d='M6 19h1'/%3E%3Cpath stroke='%23a93012' d='M7 19h1'/%3E%3Cpath stroke='%23ac3213' d='M8 19h1'/%3E%3Cpath stroke='%23ad3213' d='M9 19h1'/%3E%3Cpath stroke='%23ae3110' d='M10 19h1'/%3E%3Cpath stroke='%23b1300d' d='M11 19h1'/%3E%3Cpath stroke='%23b22e0a' d='M12 19h1'/%3E%3Cpath stroke='%23b42d08' d='M13 19h1'/%3E%3Cpath stroke='%23b12a06' d='M14 19h1'/%3E%3Cpath stroke='%23b12803' d='M15 19h1'/%3E%3Cpath stroke='%23b42701' d='M16 19h1'/%3E%3Cpath stroke='%23ae2400' d='M17 19h1'/%3E%3Cpath stroke='%23ac3c1f' d='M18 19h1'/%3E%3Cpath stroke='%23ead4cf' d='M19 19h1'/%3E%3C/svg%3E\")}.title-bar-controls button[aria-label=Close]:hover{background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 -0.5 21 21' shape-rendering='crispEdges'%3E%3Cpath stroke='%23b5c6ef' d='M1 0h1m17 0h1M0 1h1m19 0h1M0 19h1m19 0h1M1 20h1m17 0h1'/%3E%3Cpath stroke='%23f4f6fd' d='M2 0h1m17 2h1M0 18h1m17 2h1'/%3E%3Cpath stroke='%23fff' d='M3 0h15M0 3h1m19 0h1M0 4h1m19 0h1M0 5h1m5 0h1m7 0h1m5 0h1M0 6h1m4 0h3m5 0h3m4 0h1M0 7h1m5 0h3m3 0h3m5 0h1M0 8h1m6 0h3m1 0h3m6 0h1M0 9h1m7 0h5m7 0h1M0 10h1m8 0h3m8 0h1M0 11h1m7 0h5m7 0h1M0 12h1m6 0h3m1 0h2m7 0h1M0 13h1m5 0h3m3 0h3m5 0h1M0 14h1m4 0h3m5 0h3m4 0h1M0 15h1m5 0h1m7 0h1m5 0h1M0 16h1m19 0h1M0 17h1m19 0h1M3 20h3m5 0h7'/%3E%3Cpath stroke='%23f5f7fd' d='M18 0h1M0 2h1m19 16h1M2 20h1'/%3E%3Cpath stroke='%23ffe4e1' d='M1 1h1'/%3E%3Cpath stroke='%23ff9285' d='M2 1h1m4 3h1M2 7h1'/%3E%3Cpath stroke='%23ff8c7f' d='M3 1h1'/%3E%3Cpath stroke='%23ff8375' d='M4 1h1m5 3h1'/%3E%3Cpath stroke='%23ff7b6c' d='M5 1h1M3 12h1'/%3E%3Cpath stroke='%23ff7868' d='M6 1h1m3 4h1'/%3E%3Cpath stroke='%23ff7362' d='M7 1h1'/%3E%3Cpath stroke='%23ff7363' d='M8 1h1m2 4h1M2 12h1'/%3E%3Cpath stroke='%23ff705f' d='M9 1h1M6 16h1'/%3E%3Cpath stroke='%23ff6f5f' d='M10 1h1'/%3E%3Cpath stroke='%23ff6e5d' d='M11 1h1m4 1h1m-5 3h1M2 13h1'/%3E%3Cpath stroke='%23ff6b5a' d='M12 1h1M3 15h1'/%3E%3Cpath stroke='%23f65' d='M13 1h2'/%3E%3Cpath stroke='%23ff6250' d='M15 1h1M2 15h1'/%3E%3Cpath stroke='%23ff5d4a' d='M16 1h1'/%3E%3Cpath stroke='%23fa5643' d='M17 1h1'/%3E%3Cpath stroke='%23eb6151' d='M18 1h1'/%3E%3Cpath stroke='%23f5dad7' d='M19 1h1'/%3E%3Cpath stroke='%23ff9386' d='M1 2h1'/%3E%3Cpath stroke='%23ffaea5' d='M2 2h1'/%3E%3Cpath stroke='%23ffb2a9' d='M3 2h1'/%3E%3Cpath stroke='%23ffa99f' d='M4 2h1'/%3E%3Cpath stroke='%23ff9e93' d='M5 2h1m0 1h1M5 4h1'/%3E%3Cpath stroke='%23ff998d' d='M6 2h1M4 6h1'/%3E%3Cpath stroke='%23ff9488' d='M7 2h1m0 1h1'/%3E%3Cpath stroke='%23ff9083' d='M8 2h1M3 8h1'/%3E%3Cpath stroke='%23ff8e80' d='M9 2h1'/%3E%3Cpath stroke='%23ff8b7d' d='M10 2h1M5 8h1M3 9h1'/%3E%3Cpath stroke='%23ff887a' d='M11 2h1m0 1h1M5 9h1'/%3E%3Cpath stroke='%23ff8475' d='M12 2h1M8 5h1'/%3E%3Cpath stroke='%23ff8172' d='M13 2h1M7 9h1m-3 3h1'/%3E%3Cpath stroke='%23ff7c6d' d='M14 2h1'/%3E%3Cpath stroke='%23ff7666' d='M15 2h1M1 7h1m1 6h1m0 1h1'/%3E%3Cpath stroke='%23fc6352' d='M17 2h1'/%3E%3Cpath stroke='%23e54' d='M18 2h1'/%3E%3Cpath stroke='%23d3594b' d='M19 2h1'/%3E%3Cpath stroke='%23ff8d80' d='M1 3h1'/%3E%3Cpath stroke='%23ffb3ab' d='M2 3h1'/%3E%3Cpath stroke='%23ffb8b0' d='M3 3h1'/%3E%3Cpath stroke='%23ffb0a6' d='M4 3h1M3 4h1'/%3E%3Cpath stroke='%23ffa49a' d='M5 3h1'/%3E%3Cpath stroke='%23ff988d' d='M7 3h1M6 4h1'/%3E%3Cpath stroke='%23ff9184' d='M9 3h1'/%3E%3Cpath stroke='%23ff8e81' d='M10 3h1M4 8h1'/%3E%3Cpath stroke='%23ff8c7e' d='M11 3h1M2 8h1'/%3E%3Cpath stroke='%23ff8576' d='M13 3h1M6 9h1m-4 1h1'/%3E%3Cpath stroke='%23ff7f70' d='M14 3h1M1 5h1m0 5h1m1 2h1'/%3E%3Cpath stroke='%23ff796a' d='M15 3h1M2 11h1'/%3E%3Cpath stroke='%23ff7161' d='M16 3h1M3 14h1'/%3E%3Cpath stroke='%23fc6857' d='M17 3h1'/%3E%3Cpath stroke='%23ed5948' d='M18 3h1M6 18h1'/%3E%3Cpath stroke='%23cb4233' d='M19 3h1'/%3E%3Cpath stroke='%23ff8577' d='M1 4h1m0 5h1'/%3E%3Cpath stroke='%23ffaaa0' d='M2 4h1'/%3E%3Cpath stroke='%23ffa89e' d='M4 4h1'/%3E%3Cpath stroke='%23ff8d7f' d='M8 4h1'/%3E%3Cpath stroke='%23ff8879' d='M9 4h1'/%3E%3Cpath stroke='%23ff8071' d='M11 4h1M8 6h1'/%3E%3Cpath stroke='%23ff7a6b' d='M12 4h1M1 6h1m7 0h1m-6 7h1'/%3E%3Cpath stroke='%23ff7969' d='M13 4h1'/%3E%3Cpath stroke='%23ff7464' d='M14 4h1m-5 2h1'/%3E%3Cpath stroke='%23ff7060' d='M15 4h1'/%3E%3Cpath stroke='%23ff6c5b' d='M16 4h1m-4 1h1'/%3E%3Cpath stroke='%23fc6655' d='M17 4h1'/%3E%3Cpath stroke='%23ef5c4b' d='M18 4h1'/%3E%3Cpath stroke='%23cc4636' d='M19 4h1'/%3E%3Cpath stroke='%23ffa095' d='M2 5h1'/%3E%3Cpath stroke='%23ffa59b' d='M3 5h1'/%3E%3Cpath stroke='%23ff9f94' d='M4 5h1'/%3E%3Cpath stroke='%23ffd5d1' d='M5 5h1'/%3E%3Cpath stroke='%23ff8a7c' d='M7 5h1'/%3E%3Cpath stroke='%23ff7e6f' d='M9 5h1'/%3E%3Cpath stroke='%23ffc2bb' d='M15 5h1'/%3E%3Cpath stroke='%23ff6554' d='M16 5h1'/%3E%3Cpath stroke='%23fc6453' d='M17 5h1'/%3E%3Cpath stroke='%23ee5d4d' d='M18 5h1'/%3E%3Cpath stroke='%23cd4939' d='M19 5h1'/%3E%3Cpath stroke='%23ff998e' d='M2 6h1'/%3E%3Cpath stroke='%23ff9d92' d='M3 6h1'/%3E%3Cpath stroke='%23ff6f5e' d='M11 6h1'/%3E%3Cpath stroke='%23ff6a58' d='M12 6h1'/%3E%3Cpath stroke='%23ff6451' d='M16 6h1'/%3E%3Cpath stroke='%23fd6451' d='M17 6h1'/%3E%3Cpath stroke='%23ee5e4d' d='M18 6h1'/%3E%3Cpath stroke='%23ce4a3a' d='M19 6h1'/%3E%3Cpath stroke='%23ff968a' d='M3 7h1'/%3E%3Cpath stroke='%23ff9487' d='M4 7h1'/%3E%3Cpath stroke='%23ff8f82' d='M5 7h1'/%3E%3Cpath stroke='%23ff7968' d='M9 7h1m-3 8h1'/%3E%3Cpath stroke='%23ff7463' d='M10 7h1'/%3E%3Cpath stroke='%23ff6f5d' d='M11 7h1'/%3E%3Cpath stroke='%23ff6450' d='M15 7h1'/%3E%3Cpath stroke='%23ff6552' d='M16 7h1'/%3E%3Cpath stroke='%23fd6653' d='M17 7h1'/%3E%3Cpath stroke='%23f0604e' d='M18 7h1'/%3E%3Cpath stroke='%23ce4a3b' d='M19 7h1'/%3E%3Cpath stroke='%23ff7565' d='M1 8h1'/%3E%3Cpath stroke='%23ff8677' d='M6 8h1m-2 2h1'/%3E%3Cpath stroke='%23ff7664' d='M10 8h1'/%3E%3Cpath stroke='%23ff6a53' d='M14 8h1'/%3E%3Cpath stroke='%23ff6953' d='M15 8h1'/%3E%3Cpath stroke='%23ff6b55' d='M16 8h1'/%3E%3Cpath stroke='%23fd6b56' d='M17 8h1'/%3E%3Cpath stroke='%23f06350' d='M18 8h1'/%3E%3Cpath stroke='%23cf4c3b' d='M19 8h1'/%3E%3Cpath stroke='%23ff6d5d' d='M1 9h1'/%3E%3Cpath stroke='%23ff8b7c' d='M4 9h1'/%3E%3Cpath stroke='%23ff775d' d='M13 9h1'/%3E%3Cpath stroke='%23ff745a' d='M14 9h1'/%3E%3Cpath stroke='%23ff7359' d='M15 9h1'/%3E%3Cpath stroke='%23ff735a' d='M16 9h1'/%3E%3Cpath stroke='%23fd715a' d='M17 9h1'/%3E%3Cpath stroke='%23f16752' d='M18 9h1'/%3E%3Cpath stroke='%23d24e3c' d='M19 9h1'/%3E%3Cpath stroke='%23ff6a59' d='M1 10h1m2 6h1'/%3E%3Cpath stroke='%23ff8778' d='M4 10h1'/%3E%3Cpath stroke='%23ff8374' d='M6 10h1m-3 1h2'/%3E%3Cpath stroke='%23ff8171' d='M7 10h1m-5 1h1'/%3E%3Cpath stroke='%23ff8271' d='M8 10h1m-2 1h1'/%3E%3Cpath stroke='%23ff8369' d='M12 10h1'/%3E%3Cpath stroke='%23ff8165' d='M13 10h1'/%3E%3Cpath stroke='%23ff7e61' d='M14 10h1'/%3E%3Cpath stroke='%23ff7d5f' d='M15 10h1'/%3E%3Cpath stroke='%23ff7b5f' d='M16 10h1'/%3E%3Cpath stroke='%23fd775d' d='M17 10h1'/%3E%3Cpath stroke='%23f36a53' d='M18 10h1'/%3E%3Cpath stroke='%23d34e3c' d='M19 10h1'/%3E%3Cpath stroke='%23ff6553' d='M1 11h1'/%3E%3Cpath stroke='%23ff8273' d='M6 11h1'/%3E%3Cpath stroke='%23ff8c6c' d='M13 11h1'/%3E%3Cpath stroke='%23ff8969' d='M14 11h1'/%3E%3Cpath stroke='%23ff8665' d='M15 11h1'/%3E%3Cpath stroke='%23ff8262' d='M16 11h1'/%3E%3Cpath stroke='%23fd7c5e' d='M17 11h1'/%3E%3Cpath stroke='%23f46d54' d='M18 11h1'/%3E%3Cpath stroke='%23d64f3b' d='M19 11h1'/%3E%3Cpath stroke='%23ff5f4d' d='M1 12h1'/%3E%3Cpath stroke='%23ff8070' d='M6 12h1'/%3E%3Cpath stroke='%23ff9279' d='M10 12h1'/%3E%3Cpath stroke='%23fff8f6' d='M13 12h1'/%3E%3Cpath stroke='%23ff936f' d='M14 12h1'/%3E%3Cpath stroke='%23ff906c' d='M15 12h1'/%3E%3Cpath stroke='%23ff8967' d='M16 12h1'/%3E%3Cpath stroke='%23fe7f5f' d='M17 12h1'/%3E%3Cpath stroke='%23f56e52' d='M18 12h1'/%3E%3Cpath stroke='%23d84f39' d='M19 12h1'/%3E%3Cpath stroke='%23ff5c4a' d='M1 13h1'/%3E%3Cpath stroke='%23ff7d6e' d='M5 13h1'/%3E%3Cpath stroke='%23ff907a' d='M9 13h1'/%3E%3Cpath stroke='%23ff957c' d='M10 13h1'/%3E%3Cpath stroke='%23ff9a7e' d='M11 13h1'/%3E%3Cpath stroke='%23ff9670' d='M15 13h1'/%3E%3Cpath stroke='%23ff8e68' d='M16 13h1'/%3E%3Cpath stroke='%23fe815e' d='M17 13h1'/%3E%3Cpath stroke='%23f66c4f' d='M18 13h1'/%3E%3Cpath stroke='%23da4d36' d='M19 13h1'/%3E%3Cpath stroke='%23ff5744' d='M1 14h1'/%3E%3Cpath stroke='%23ff6857' d='M2 14h1'/%3E%3Cpath stroke='%23ff8672' d='M8 14h1'/%3E%3Cpath stroke='%23ff8f78' d='M9 14h1'/%3E%3Cpath stroke='%23ff967c' d='M10 14h1'/%3E%3Cpath stroke='%23ff9c7e' d='M11 14h1'/%3E%3Cpath stroke='%23ffa07e' d='M12 14h1'/%3E%3Cpath stroke='%23ff8e66' d='M16 14h1'/%3E%3Cpath stroke='%23fe7f5a' d='M17 14h1m-3 3h1'/%3E%3Cpath stroke='%23f76a4b' d='M18 14h1'/%3E%3Cpath stroke='%23da4a33' d='M19 14h1'/%3E%3Cpath stroke='%23ff523f' d='M1 15h1'/%3E%3Cpath stroke='%23ff7160' d='M4 15h1'/%3E%3Cpath stroke='%23ffc7c1' d='M5 15h1'/%3E%3Cpath stroke='%23ff836f' d='M8 15h1'/%3E%3Cpath stroke='%23ff8b74' d='M9 15h1'/%3E%3Cpath stroke='%23ff9379' d='M10 15h1'/%3E%3Cpath stroke='%23ff9a7c' d='M11 15h1'/%3E%3Cpath stroke='%23ff9e7c' d='M12 15h1'/%3E%3Cpath stroke='%23ffa07a' d='M13 15h1'/%3E%3Cpath stroke='%23ffd5c5' d='M15 15h1'/%3E%3Cpath stroke='%23ff8b62' d='M16 15h1'/%3E%3Cpath stroke='%23fe7c56' d='M17 15h1'/%3E%3Cpath stroke='%23f76545' d='M18 15h1'/%3E%3Cpath stroke='%23db4931' d='M19 15h1'/%3E%3Cpath stroke='%23ff4f3a' d='M1 16h1'/%3E%3Cpath stroke='%23ff5c49' d='M2 16h1'/%3E%3Cpath stroke='%23ff6452' d='M3 16h1'/%3E%3Cpath stroke='%23ff6e5e' d='M5 16h1'/%3E%3Cpath stroke='%23ff7462' d='M7 16h1'/%3E%3Cpath stroke='%23ff7c68' d='M8 16h1'/%3E%3Cpath stroke='%23ff846d' d='M9 16h1'/%3E%3Cpath stroke='%23ff8b71' d='M10 16h1'/%3E%3Cpath stroke='%23ff9174' d='M11 16h1'/%3E%3Cpath stroke='%23ff9674' d='M12 16h1'/%3E%3Cpath stroke='%23ff9571' d='M13 16h1'/%3E%3Cpath stroke='%23ff946d' d='M14 16h1'/%3E%3Cpath stroke='%23ff8d66' d='M15 16h1'/%3E%3Cpath stroke='%23ff855c' d='M16 16h1'/%3E%3Cpath stroke='%23fe7650' d='M17 16h1'/%3E%3Cpath stroke='%23f66141' d='M18 16h1'/%3E%3Cpath stroke='%23da462f' d='M19 16h1'/%3E%3Cpath stroke='%23fa4935' d='M1 17h1'/%3E%3Cpath stroke='%23fb5441' d='M2 17h1'/%3E%3Cpath stroke='%23fc5c4a' d='M3 17h1'/%3E%3Cpath stroke='%23fb6150' d='M4 17h1'/%3E%3Cpath stroke='%23fc6554' d='M5 17h1'/%3E%3Cpath stroke='%23fc6756' d='M6 17h1'/%3E%3Cpath stroke='%23fc6a58' d='M7 17h1'/%3E%3Cpath stroke='%23fc715c' d='M8 17h1'/%3E%3Cpath stroke='%23fc7761' d='M9 17h1'/%3E%3Cpath stroke='%23fd7e64' d='M10 17h1'/%3E%3Cpath stroke='%23fd8367' d='M11 17h1'/%3E%3Cpath stroke='%23fe8566' d='M12 17h1'/%3E%3Cpath stroke='%23fe8664' d='M13 17h1'/%3E%3Cpath stroke='%23fe8460' d='M14 17h1'/%3E%3Cpath stroke='%23fe7651' d='M16 17h1'/%3E%3Cpath stroke='%23fc6b47' d='M17 17h1'/%3E%3Cpath stroke='%23f2573a' d='M18 17h1'/%3E%3Cpath stroke='%23d4402a' d='M19 17h1'/%3E%3Cpath stroke='%23e85848' d='M1 18h1'/%3E%3Cpath stroke='%23ed4a37' d='M2 18h1'/%3E%3Cpath stroke='%23ec4f3d' d='M3 18h1'/%3E%3Cpath stroke='%23ee5443' d='M4 18h1'/%3E%3Cpath stroke='%23ed5746' d='M5 18h1'/%3E%3Cpath stroke='%23ee5a48' d='M7 18h1'/%3E%3Cpath stroke='%23ef5e4b' d='M8 18h1'/%3E%3Cpath stroke='%23f0644e' d='M9 18h1'/%3E%3Cpath stroke='%23f16750' d='M10 18h1'/%3E%3Cpath stroke='%23f46c52' d='M11 18h1'/%3E%3Cpath stroke='%23f66d51' d='M12 18h1'/%3E%3Cpath stroke='%23f66e51' d='M13 18h1'/%3E%3Cpath stroke='%23f66c4e' d='M14 18h1'/%3E%3Cpath stroke='%23f86a4a' d='M15 18h1'/%3E%3Cpath stroke='%23f76343' d='M16 18h1'/%3E%3Cpath stroke='%23f3583a' d='M17 18h1'/%3E%3Cpath stroke='%23e54930' d='M18 18h1'/%3E%3Cpath stroke='%23cd5140' d='M19 18h1'/%3E%3Cpath stroke='%23f6d9d6' d='M1 19h1'/%3E%3Cpath stroke='%23d25344' d='M2 19h1'/%3E%3Cpath stroke='%23c93c2b' d='M3 19h1'/%3E%3Cpath stroke='%23ca3f2f' d='M4 19h1'/%3E%3Cpath stroke='%23ca4131' d='M5 19h1'/%3E%3Cpath stroke='%23ca4333' d='M6 19h1'/%3E%3Cpath stroke='%23cc4332' d='M7 19h1'/%3E%3Cpath stroke='%23cf4434' d='M8 19h1'/%3E%3Cpath stroke='%23d24936' d='M9 19h1'/%3E%3Cpath stroke='%23d34936' d='M10 19h1'/%3E%3Cpath stroke='%23d84b37' d='M11 19h1'/%3E%3Cpath stroke='%23da4c36' d='M12 19h1'/%3E%3Cpath stroke='%23dc4d36' d='M13 19h1'/%3E%3Cpath stroke='%23d94933' d='M14 19h1'/%3E%3Cpath stroke='%23de4a32' d='M15 19h1'/%3E%3Cpath stroke='%23dd482f' d='M16 19h1'/%3E%3Cpath stroke='%23d6402a' d='M17 19h1'/%3E%3Cpath stroke='%23cf5140' d='M18 19h1'/%3E%3Cpath stroke='%23f1d8d5' d='M19 19h1'/%3E%3Cpath stroke='%23fefefe' d='M6 20h1m3 0h1'/%3E%3Cpath stroke='%23fdfdfd' d='M7 20h1m1 0h1'/%3E%3Cpath stroke='%23fcfcfc' d='M8 20h1'/%3E%3C/svg%3E\")}.title-bar-controls button[aria-label=Close]:not(:disabled):active{background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 -0.5 21 21' shape-rendering='crispEdges'%3E%3Cpath stroke='%23a7bced' d='M1 0h1M0 1h1'/%3E%3Cpath stroke='%23f4f6fd' d='M2 0h1m15 0h1M0 2h1m19 0h1M0 18h1m19 0h1M2 20h1m15 0h1'/%3E%3Cpath stroke='%23fff' d='M3 0h15M0 3h1m19 0h1M0 4h1m19 0h1M0 5h1m19 0h1M0 6h1m19 0h1M0 7h1m19 0h1M0 8h1m19 0h1M0 9h1m19 0h1M0 10h1m19 0h1M0 11h1m19 0h1M0 12h1m19 0h1M0 13h1m19 0h1M0 14h1m19 0h1M0 15h1m19 0h1M0 16h1m19 0h1M0 17h1m19 0h1M3 20h15'/%3E%3Cpath stroke='%23a7baec' d='M19 0h1m0 1h1'/%3E%3Cpath stroke='%23dad2d0' d='M1 1h1'/%3E%3Cpath stroke='%23643529' d='M2 1h1M1 2h1'/%3E%3Cpath stroke='%235a1d0d' d='M3 1h1'/%3E%3Cpath stroke='%235d1e0d' d='M4 1h1'/%3E%3Cpath stroke='%235f1e0e' d='M5 1h1'/%3E%3Cpath stroke='%2363200e' d='M6 1h1'/%3E%3Cpath stroke='%2368210f' d='M7 1h1'/%3E%3Cpath stroke='%236f2310' d='M8 1h1'/%3E%3Cpath stroke='%23732511' d='M9 1h1'/%3E%3Cpath stroke='%23752511' d='M10 1h1M1 10h1'/%3E%3Cpath stroke='%237c2712' d='M11 1h1'/%3E%3Cpath stroke='%23822912' d='M12 1h1M5 2h1'/%3E%3Cpath stroke='%23852a13' d='M13 1h1M2 5h1m-2 8h1'/%3E%3Cpath stroke='%23892b13' d='M14 1h1'/%3E%3Cpath stroke='%238a2b14' d='M15 1h1M6 2h1'/%3E%3Cpath stroke='%238e2d14' d='M16 1h1M7 2h1'/%3E%3Cpath stroke='%238c2c14' d='M17 1h1M2 6h1'/%3E%3Cpath stroke='%239d4732' d='M18 1h1M1 18h1'/%3E%3Cpath stroke='%23ebd8d3' d='M19 1h1'/%3E%3Cpath stroke='%2369220f' d='M2 2h1'/%3E%3Cpath stroke='%23782611' d='M3 2h1'/%3E%3Cpath stroke='%237e2812' d='M4 2h1'/%3E%3Cpath stroke='%23932e15' d='M8 2h1'/%3E%3Cpath stroke='%239a3016' d='M9 2h1'/%3E%3Cpath stroke='%239c3116' d='M10 2h1'/%3E%3Cpath stroke='%23a03217' d='M11 2h1'/%3E%3Cpath stroke='%23a43418' d='M12 2h1'/%3E%3Cpath stroke='%23a73518' d='M13 2h1'/%3E%3Cpath stroke='%23aa3618' d='M14 2h1M2 14h1'/%3E%3Cpath stroke='%23ab3618' d='M15 2h1'/%3E%3Cpath stroke='%23ad3719' d='M16 2h1m1 0h1M2 16h1m-1 1h1'/%3E%3Cpath stroke='%23ac3618' d='M17 2h1'/%3E%3Cpath stroke='%23b24e35' d='M19 2h1'/%3E%3Cpath stroke='%23591c0d' d='M1 3h1M1 4h1'/%3E%3Cpath stroke='%23792711' d='M2 3h1'/%3E%3Cpath stroke='%238d2c14' d='M3 3h1'/%3E%3Cpath stroke='%23962e15' d='M4 3h1'/%3E%3Cpath stroke='%239a2f16' d='M5 3h1'/%3E%3Cpath stroke='%23a13117' d='M6 3h1'/%3E%3Cpath stroke='%23a63317' d='M7 3h1'/%3E%3Cpath stroke='%23aa3418' d='M8 3h1'/%3E%3Cpath stroke='%23af3619' d='M9 3h1'/%3E%3Cpath stroke='%23b23719' d='M10 3h1M8 4h1M4 8h1'/%3E%3Cpath stroke='%23b5391a' d='M11 3h1'/%3E%3Cpath stroke='%23b73a1b' d='M12 3h1'/%3E%3Cpath stroke='%23b93b1b' d='M13 3h1'/%3E%3Cpath stroke='%23ba3b1b' d='M14 3h2m3 0h1M3 13h1m-1 1h1m-1 5h1'/%3E%3Cpath stroke='%23bb3b1b' d='M16 3h3M3 15h1'/%3E%3Cpath stroke='%23802812' d='M2 4h1m-2 8h1'/%3E%3Cpath stroke='%23962f15' d='M3 4h1'/%3E%3Cpath stroke='%239e3016' d='M4 4h1'/%3E%3Cpath stroke='%23a43216' d='M5 4h1'/%3E%3Cpath stroke='%23aa3317' d='M6 4h1M4 6h1'/%3E%3Cpath stroke='%23ae3518' d='M7 4h1'/%3E%3Cpath stroke='%23b5381a' d='M9 4h1M4 9h1'/%3E%3Cpath stroke='%23b8391a' d='M10 4h1m-7 6h1'/%3E%3Cpath stroke='%23ba3a1b' d='M11 4h1m-8 7h2'/%3E%3Cpath stroke='%23bc3b1c' d='M12 4h1m-9 8h1'/%3E%3Cpath stroke='%23bd3c1c' d='M13 4h1m-1 1h1m-2 1h1m-7 6h1m-3 1h2'/%3E%3Cpath stroke='%23be3d1c' d='M14 4h3m-1 1h1m-1 1h1M4 14h1m-1 1h1m-1 1h2'/%3E%3Cpath stroke='%23bf3d1c' d='M17 4h3m-3 1h3m-2 1h2m-1 1h1M4 17h2m-2 1h4m-4 1h4'/%3E%3Cpath stroke='%235b1d0d' d='M1 5h1'/%3E%3Cpath stroke='%239c3016' d='M3 5h1'/%3E%3Cpath stroke='%23a43217' d='M4 5h1'/%3E%3Cpath stroke='%23b8553e' d='M5 5h1'/%3E%3Cpath stroke='%23d59485' d='M6 5h1M5 6h1'/%3E%3Cpath stroke='%23b33619' d='M7 5h1'/%3E%3Cpath stroke='%23b53719' d='M8 5h1'/%3E%3Cpath stroke='%23b8381a' d='M9 5h1M6 8h1'/%3E%3Cpath stroke='%23b9391b' d='M10 5h1'/%3E%3Cpath stroke='%23ba391b' d='M11 5h1M6 9h1m-2 1h1'/%3E%3Cpath stroke='%23bc3b1b' d='M12 5h1m-2 1h1m-6 5h1m-2 1h1'/%3E%3Cpath stroke='%23dc9887' d='M14 5h1'/%3E%3Cpath stroke='%23c85d42' d='M15 5h1M5 15h1'/%3E%3Cpath stroke='%23611f0e' d='M1 6h1'/%3E%3Cpath stroke='%23a23217' d='M3 6h1'/%3E%3Cpath stroke='%23d79585' d='M6 6h1'/%3E%3Cpath stroke='%23d89585' d='M7 6h1'/%3E%3Cpath stroke='%23b8371a' d='M8 6h1'/%3E%3Cpath stroke='%23ba391a' d='M9 6h1'/%3E%3Cpath stroke='%23bb3a1b' d='M10 6h1m-5 4h1'/%3E%3Cpath stroke='%23dd9887' d='M13 6h3m-4 1h1m-2 1h1M9 9h1m-2 2h1m-2 1h1m-2 1h1m-2 1h2'/%3E%3Cpath stroke='%23c03e1d' d='M17 6h1m-2 1h3m0 1h1m-1 1h1M7 16h1m-2 1h2m0 1h1'/%3E%3Cpath stroke='%2365200e' d='M1 7h1'/%3E%3Cpath stroke='%23902d15' d='M2 7h1'/%3E%3Cpath stroke='%23a73418' d='M3 7h1'/%3E%3Cpath stroke='%23af3518' d='M4 7h1'/%3E%3Cpath stroke='%23b43619' d='M5 7h1'/%3E%3Cpath stroke='%23d99585' d='M6 7h1'/%3E%3Cpath stroke='%23da9686' d='M7 7h1'/%3E%3Cpath stroke='%23db9686' d='M8 7h1M7 8h1'/%3E%3Cpath stroke='%23bc3a1b' d='M9 7h1M7 9h1'/%3E%3Cpath stroke='%23bd3b1b' d='M10 7h1m-4 3h1'/%3E%3Cpath stroke='%23be3c1c' d='M11 7h1m-2 1h1m-3 2h1m-2 1h1'/%3E%3Cpath stroke='%23de9987' d='M13 7h2m-3 1h2m-4 1h2m-3 1h1m-2 2h1m-2 2h1'/%3E%3Cpath stroke='%23c03f1d' d='M15 7h1m-9 8h1'/%3E%3Cpath stroke='%236a220f' d='M1 8h1'/%3E%3Cpath stroke='%23952f15' d='M2 8h1'/%3E%3Cpath stroke='%23ac3518' d='M3 8h1'/%3E%3Cpath stroke='%23b63719' d='M5 8h1'/%3E%3Cpath stroke='%23dc9786' d='M8 8h2M8 9h1'/%3E%3Cpath stroke='%23c2401d' d='M14 8h1m2 0h1m1 3h1M8 14h1m-1 2h1m-1 1h1m0 1h1m1 1h1'/%3E%3Cpath stroke='%23c2401e' d='M15 8h2m1 1h1M8 15h1'/%3E%3Cpath stroke='%23c13f1d' d='M18 8h1m0 2h1M9 19h2'/%3E%3Cpath stroke='%23702410' d='M1 9h1'/%3E%3Cpath stroke='%239b3016' d='M2 9h1'/%3E%3Cpath stroke='%23b03619' d='M3 9h1'/%3E%3Cpath stroke='%23b9381a' d='M5 9h1'/%3E%3Cpath stroke='%23df9a88' d='M12 9h1m-2 1h1m-2 1h1m-2 1h1'/%3E%3Cpath stroke='%23c4421e' d='M13 9h1m2 0h2m0 1h1M9 13h1m9 1h1m-1 1h1M9 16h1m9 0h1M9 17h1m0 1h1m3 1h3'/%3E%3Cpath stroke='%23c5431e' d='M14 9h1'/%3E%3Cpath stroke='%23c5431f' d='M15 9h1m-4 1h1m5 1h1m-9 1h1m-2 2h1m-1 1h1m0 2h1m0 1h1m6 0h1'/%3E%3Cpath stroke='%239e3217' d='M2 10h1'/%3E%3Cpath stroke='%23b4381a' d='M3 10h1'/%3E%3Cpath stroke='%23df9a87' d='M10 10h1m-2 1h1m-2 2h1'/%3E%3Cpath stroke='%23c6441f' d='M13 10h1m3 0h1m-8 3h1m-1 3h1'/%3E%3Cpath stroke='%23c74520' d='M14 10h2m-6 4h1m-1 1h1m7 2h1m-7 1h1m4 0h1'/%3E%3Cpath stroke='%23c7451f' d='M16 10h1m1 2h1'/%3E%3Cpath stroke='%237b2711' d='M1 11h1'/%3E%3Cpath stroke='%23a13217' d='M2 11h1'/%3E%3Cpath stroke='%23b7391a' d='M3 11h1'/%3E%3Cpath stroke='%23e09b88' d='M11 11h1'/%3E%3Cpath stroke='%23e29d89' d='M12 11h1'/%3E%3Cpath stroke='%23c94621' d='M13 11h1m-3 2h1'/%3E%3Cpath stroke='%23ca4721' d='M14 11h1m2 1h1m-7 2h1m-1 1h1m0 2h1m2 1h1'/%3E%3Cpath stroke='%23ca4821' d='M15 11h1m1 6h1'/%3E%3Cpath stroke='%23c94620' d='M16 11h1m1 3h1m-8 2h1m6 0h1'/%3E%3Cpath stroke='%23c84620' d='M17 11h1m0 2h1'/%3E%3Cpath stroke='%23a53418' d='M2 12h1'/%3E%3Cpath stroke='%23b83a1b' d='M3 12h1'/%3E%3Cpath stroke='%23e19d89' d='M11 12h1'/%3E%3Cpath stroke='%23e39e89' d='M12 12h1'/%3E%3Cpath stroke='%23e0947c' d='M13 12h1'/%3E%3Cpath stroke='%23cc4a22' d='M14 12h1m-3 2h1m4 0h1m-6 1h1'/%3E%3Cpath stroke='%23cd4a22' d='M15 12h1m0 1h1m0 2h1m-5 1h1m1 1h1'/%3E%3Cpath stroke='%23cb4922' d='M16 12h1m0 1h1m-5 4h1'/%3E%3Cpath stroke='%23c3411e' d='M19 12h1m-1 1h1m-1 4h1m-8 2h2m3 0h1'/%3E%3Cpath stroke='%23a93618' d='M2 13h1'/%3E%3Cpath stroke='%23dd9987' d='M7 13h1m-2 2h1'/%3E%3Cpath stroke='%23e39f8a' d='M12 13h1'/%3E%3Cpath stroke='%23e59f8b' d='M13 13h1'/%3E%3Cpath stroke='%23e5a08b' d='M14 13h1m-2 1h1'/%3E%3Cpath stroke='%23ce4c23' d='M15 13h1m0 3h1'/%3E%3Cpath stroke='%23882b13' d='M1 14h1'/%3E%3Cpath stroke='%23e6a08b' d='M14 14h1'/%3E%3Cpath stroke='%23e6a18b' d='M15 14h1m-2 1h1'/%3E%3Cpath stroke='%23ce4b23' d='M16 14h1m-4 1h1'/%3E%3Cpath stroke='%238b2c14' d='M1 15h1m-1 1h1'/%3E%3Cpath stroke='%23ac3619' d='M2 15h1'/%3E%3Cpath stroke='%23d76b48' d='M15 15h1'/%3E%3Cpath stroke='%23cf4c23' d='M16 15h1m-2 1h1'/%3E%3Cpath stroke='%23c94721' d='M18 15h1m-3 3h1'/%3E%3Cpath stroke='%23bb3c1b' d='M3 16h1'/%3E%3Cpath stroke='%23bf3e1d' d='M6 16h1'/%3E%3Cpath stroke='%23cb4821' d='M12 16h1'/%3E%3Cpath stroke='%23cd4b23' d='M14 16h1'/%3E%3Cpath stroke='%23cc4922' d='M17 16h1m-4 1h1m1 0h1'/%3E%3Cpath stroke='%238d2d14' d='M1 17h1'/%3E%3Cpath stroke='%23bc3c1b' d='M3 17h1m-1 1h1'/%3E%3Cpath stroke='%23c84520' d='M11 17h1m1 1h1'/%3E%3Cpath stroke='%23ae3719' d='M2 18h1'/%3E%3Cpath stroke='%23c94720' d='M14 18h1'/%3E%3Cpath stroke='%23c95839' d='M19 18h1'/%3E%3Cpath stroke='%23a7bdf0' d='M0 19h1m0 1h1'/%3E%3Cpath stroke='%23ead7d3' d='M1 19h1'/%3E%3Cpath stroke='%23b34e35' d='M2 19h1'/%3E%3Cpath stroke='%23c03e1c' d='M8 19h1'/%3E%3Cpath stroke='%23c9583a' d='M18 19h1'/%3E%3Cpath stroke='%23f3dbd4' d='M19 19h1'/%3E%3Cpath stroke='%23a7bcef' d='M20 19h1m-2 1h1'/%3E%3C/svg%3E\")}.status-bar{margin:0 3px;box-shadow:inset 0 1px 2px grey;padding:2px 1px;gap:0}.status-bar-field{font-family:\"Pixelated MS Sans Serif\",Arial;-webkit-font-smoothing:antialiased;box-shadow:none;padding:1px 2px;border-right:1px solid rgba(208,206,191,.75);border-left:1px solid hsla(0,0%,100%,.75)}.status-bar-field:first-of-type{border-left:none}.status-bar-field:last-of-type{border-right:none}button{font-family:\"Pixelated MS Sans Serif\",Arial;-webkit-font-smoothing:antialiased;font-size:11px;box-sizing:border-box;border:1px solid #003c74;background:linear-gradient(180deg,#fff,#ecebe5 86%,#d8d0c4);box-shadow:none;border-radius:3px}button:not(:disabled).active,button:not(:disabled):active{box-shadow:none;background:linear-gradient(180deg,#cdcac3,#e3e3db 8%,#e5e5de 94%,#f2f2f1)}button:not(:disabled):hover{box-shadow:inset -1px 1px #fff0cf,inset 1px 2px #fdd889,inset -2px 2px #fbc761,inset 2px -2px #e5a01a}button.focused,button:focus{box-shadow:inset -1px 1px #cee7ff,inset 1px 2px #98b8ea,inset -2px 2px #bcd4f6,inset 1px -1px #89ade4,inset 2px -2px #89ade4}button::-moz-focus-inner{border:0}input,label,option,select,textarea{-webkit-font-smoothing:antialiased}select{border:1px solid #7f9db9;background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 -0.5 15 17' shape-rendering='crispEdges'%3E%3Cpath stroke='%23e6eefc' d='M0 0h1'/%3E%3Cpath stroke='%23d1e0fd' d='M1 0h1M0 1h1m3 0h2M2 3h1M2 4h1'/%3E%3Cpath stroke='%23cad8f9' d='M2 0h1M0 2h1'/%3E%3Cpath stroke='%23c4d3f7' d='M3 0h1M0 3h1M0 4h1'/%3E%3Cpath stroke='%23bfd0f8' d='M4 0h2M0 5h1'/%3E%3Cpath stroke='%23bdcef7' d='M6 0h1M0 6h1'/%3E%3Cpath stroke='%23baccf4' d='M7 0h1m6 2h1m-1 5h1m-1 1h1'/%3E%3Cpath stroke='%23b8cbf6' d='M8 0h1M0 7h1M0 8h1'/%3E%3Cpath stroke='%23b7caf5' d='M9 0h2M0 9h1'/%3E%3Cpath stroke='%23b5c8f7' d='M11 0h1'/%3E%3Cpath stroke='%23b3c7f5' d='M12 0h1'/%3E%3Cpath stroke='%23afc5f4' d='M13 0h1'/%3E%3Cpath stroke='%23dce6f9' d='M14 0h1'/%3E%3Cpath stroke='%23e1eafe' d='M1 1h1'/%3E%3Cpath stroke='%23dae6fe' d='M2 1h1M1 2h1'/%3E%3Cpath stroke='%23d4e1fc' d='M3 1h1M1 3h1M1 4h1'/%3E%3Cpath stroke='%23d0ddfc' d='M6 1h1M1 5h1'/%3E%3Cpath stroke='%23cedbfd' d='M7 1h1M4 2h2'/%3E%3Cpath stroke='%23cad9fd' d='M8 1h1M6 2h1M3 5h1'/%3E%3Cpath stroke='%23c8d8fb' d='M9 1h2'/%3E%3Cpath stroke='%23c5d6fc' d='M11 1h1M2 11h4'/%3E%3Cpath stroke='%23c2d3fc' d='M12 1h1m-2 1h1M1 11h1m0 1h2m-2 1h2'/%3E%3Cpath stroke='%23bccefa' d='M13 1h1m-1 1h1m-1 1h1m-1 1h1M3 15h4'/%3E%3Cpath stroke='%23b9c9f3' d='M14 1h1M3 16h4'/%3E%3Cpath stroke='%23d8e3fc' d='M2 2h1'/%3E%3Cpath stroke='%23d1defd' d='M3 2h1'/%3E%3Cpath stroke='%23c9d8fc' d='M7 2h1M4 3h3M4 4h3M3 6h1m1 0h2M1 7h1M1 8h1'/%3E%3Cpath stroke='%23c5d5fc' d='M8 2h1m-8 8h5'/%3E%3Cpath stroke='%23c5d3fc' d='M9 2h2'/%3E%3Cpath stroke='%23bed0fc' d='M12 2h1M8 3h1M8 4h1m-8 8h1m-1 1h1m0 1h1m1 0h3'/%3E%3Cpath stroke='%23cddbfc' d='M3 3h1M3 4h1M1 6h2'/%3E%3Cpath stroke='%23c8d5fb' d='M7 3h1M7 4h1'/%3E%3Cpath stroke='%23bbcefd' d='M9 3h4M9 4h4M8 5h1M7 6h1'/%3E%3Cpath stroke='%23bcccf3' d='M14 3h1m-1 1h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23ceddfd' d='M2 5h1'/%3E%3Cpath stroke='%23c8d6fb' d='M4 5h4M1 9h3'/%3E%3Cpath stroke='%23bacdfc' d='M9 5h2m1 0h2M1 14h1'/%3E%3Cpath stroke='%23b9cdfb' d='M11 5h1M8 6h2m2 0h2m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%234d6185' d='M4 6h1m5 0h1M3 7h3m3 0h3M4 8h3m1 0h3M5 9h5m-4 1h3m-2 1h1'/%3E%3Cpath stroke='%23b7cdfc' d='M11 6h1m0 1h1m-1 1h1'/%3E%3Cpath stroke='%23cad8fd' d='M2 7h1M2 8h2'/%3E%3Cpath stroke='%23c1d3fb' d='M6 7h2M7 8h1M4 9h1'/%3E%3Cpath stroke='%23b6cefb' d='M8 7h1m2 1h1m-2 1h3m-2 1h2'/%3E%3Cpath stroke='%23b6cdfb' d='M13 9h1m-6 6h1'/%3E%3Cpath stroke='%23b9cbf3' d='M14 9h1'/%3E%3Cpath stroke='%23b4c8f6' d='M0 10h1'/%3E%3Cpath stroke='%23bdd3fb' d='M9 10h2m-4 4h1'/%3E%3Cpath stroke='%23b5cdfa' d='M13 10h1'/%3E%3Cpath stroke='%23b5c9f3' d='M14 10h1'/%3E%3Cpath stroke='%23b1c7f6' d='M0 11h1'/%3E%3Cpath stroke='%23c3d5fd' d='M6 11h1'/%3E%3Cpath stroke='%23bad4fc' d='M8 11h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23b2cffb' d='M9 11h4m-2 3h1'/%3E%3Cpath stroke='%23b1cbfa' d='M13 11h1m-3 4h1'/%3E%3Cpath stroke='%23b3c8f5' d='M14 11h1m-7 5h3'/%3E%3Cpath stroke='%23adc3f6' d='M0 12h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23c2d5fc' d='M4 12h4m-4 1h4'/%3E%3Cpath stroke='%23b7d3fc' d='M9 12h2m-2 1h2m-3 1h1'/%3E%3Cpath stroke='%23b3d1fc' d='M11 12h1m-1 1h1'/%3E%3Cpath stroke='%23afcdfb' d='M12 12h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23afcbfa' d='M13 12h1m-1 1h1'/%3E%3Cpath stroke='%23b2c8f4' d='M14 12h1m-1 1h1m-4 3h1'/%3E%3Cpath stroke='%23c1d2fb' d='M3 14h1'/%3E%3Cpath stroke='%23b6d1fb' d='M9 14h2'/%3E%3Cpath stroke='%23adc9f9' d='M13 14h1m-2 1h1'/%3E%3Cpath stroke='%23b1c6f3' d='M14 14h1m-3 2h1'/%3E%3Cpath stroke='%23abc1f4' d='M0 15h1'/%3E%3Cpath stroke='%23b7cbf9' d='M1 15h1'/%3E%3Cpath stroke='%23b9cefb' d='M2 15h1'/%3E%3Cpath stroke='%23b9cffb' d='M7 15h1'/%3E%3Cpath stroke='%23b2cdfb' d='M9 15h2'/%3E%3Cpath stroke='%23aec8f7' d='M13 15h1'/%3E%3Cpath stroke='%23b0c5f2' d='M14 15h1m-2 1h1'/%3E%3Cpath stroke='%23dbe3f8' d='M0 16h1'/%3E%3Cpath stroke='%23b7c6f1' d='M1 16h1'/%3E%3Cpath stroke='%23b8c9f2' d='M2 16h1m4 0h1'/%3E%3Cpath stroke='%23d9e3f6' d='M14 16h1'/%3E%3C/svg%3E\");background-size:15px}select:focus{box-shadow:inset 2px 2px #fff,inset -2px -2px #fff}select:hover{background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 -0.5 15 17' shape-rendering='crispEdges'%3E%3Cpath stroke='%23cfdbec' d='M0 0h1'/%3E%3Cpath stroke='%23b4c7eb' d='M1 0h1M0 1h1'/%3E%3Cpath stroke='%23abbee9' d='M2 0h1M0 2h1'/%3E%3Cpath stroke='%23a5b9e7' d='M3 0h1M0 3h1M0 4h1'/%3E%3Cpath stroke='%23a0b4e5' d='M4 0h2'/%3E%3Cpath stroke='%239cb2e5' d='M6 0h1M0 6h1'/%3E%3Cpath stroke='%2398b1e4' d='M7 0h1'/%3E%3Cpath stroke='%2397aee2' d='M8 0h1M0 7h1M0 8h1'/%3E%3Cpath stroke='%2395ade5' d='M9 0h2M0 9h1'/%3E%3Cpath stroke='%2394abe4' d='M11 0h1'/%3E%3Cpath stroke='%2390aae4' d='M12 0h1M0 10h1'/%3E%3Cpath stroke='%238aa6e3' d='M13 0h1M0 12h1m-1 1h1'/%3E%3Cpath stroke='%23c2cfea' d='M14 0h1'/%3E%3Cpath stroke='%23fdffff' d='M1 1h1'/%3E%3Cpath stroke='%23f6ffff' d='M2 1h1M1 2h1'/%3E%3Cpath stroke='%23eefcff' d='M3 1h1M1 3h1M1 4h1'/%3E%3Cpath stroke='%23e9faff' d='M4 1h2M3 2h1M2 3h1M2 4h1'/%3E%3Cpath stroke='%23e5f6ff' d='M6 1h1M1 5h2'/%3E%3Cpath stroke='%23e3f4ff' d='M7 1h1M4 2h2'/%3E%3Cpath stroke='%23dff2fe' d='M8 1h1M6 2h1M3 5h1'/%3E%3Cpath stroke='%23dbf1ff' d='M9 1h2'/%3E%3Cpath stroke='%23d8edfc' d='M11 1h1M2 11h4'/%3E%3Cpath stroke='%23d4e9fc' d='M12 1h1m-2 1h1M1 11h1m0 1h1m-1 1h1'/%3E%3Cpath stroke='%23cde2fc' d='M13 1h1m-1 1h1m-1 1h1m-1 1h1M3 15h3'/%3E%3Cpath stroke='%2397ace0' d='M14 1h1M3 16h4'/%3E%3Cpath stroke='%23f2feff' d='M2 2h1'/%3E%3Cpath stroke='%23ddf0fe' d='M7 2h1M4 3h3M4 4h3M3 6h1m1 0h2M1 7h1M1 8h1'/%3E%3Cpath stroke='%23d8ebfc' d='M8 2h1m-8 8h5'/%3E%3Cpath stroke='%23d8eafc' d='M9 2h2'/%3E%3Cpath stroke='%23d0e5fc' d='M12 2h1M8 3h1M8 4h1m-8 8h1m-1 1h1m0 1h1'/%3E%3Cpath stroke='%2398aee1' d='M14 2h1m-1 5h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23e2f3fd' d='M3 3h1M3 4h1M1 6h1'/%3E%3Cpath stroke='%23dcecfd' d='M7 3h1M7 4h1'/%3E%3Cpath stroke='%23cde4fc' d='M9 3h4M9 4h4M8 5h1'/%3E%3Cpath stroke='%239bafe0' d='M14 3h1m-1 1h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%239db3e6' d='M0 5h1'/%3E%3Cpath stroke='%23dcedfd' d='M4 5h4M1 9h3'/%3E%3Cpath stroke='%23cae2fd' d='M9 5h2m1 0h2M1 14h1m4 1h1'/%3E%3Cpath stroke='%23c8e1fb' d='M11 5h1M8 6h2m2 0h2m-1 1h1m-1 1h1M2 15h1'/%3E%3Cpath stroke='%23e1f2fe' d='M2 6h1'/%3E%3Cpath stroke='%234d6185' d='M4 6h1m5 0h1M3 7h3m3 0h3M4 8h3m1 0h3M5 9h5m-4 1h3m-2 1h1'/%3E%3Cpath stroke='%23d2eafe' d='M7 6h1M6 7h2M7 8h1M4 9h1'/%3E%3Cpath stroke='%23c6e2fd' d='M11 6h1m0 1h1m-1 1h1'/%3E%3Cpath stroke='%23dff0fc' d='M2 7h1M2 8h2'/%3E%3Cpath stroke='%23c5e3fc' d='M8 7h1m2 1h1m-2 1h4m-3 1h2m-5 5h1'/%3E%3Cpath stroke='%23ceeafd' d='M9 10h2m-4 4h1'/%3E%3Cpath stroke='%23c2e2fc' d='M13 10h1m-5 5h2'/%3E%3Cpath stroke='%2392ade2' d='M14 10h1'/%3E%3Cpath stroke='%238da9e4' d='M0 11h1'/%3E%3Cpath stroke='%23d5ebfd' d='M6 11h1'/%3E%3Cpath stroke='%23cbedff' d='M8 11h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23c0e4fc' d='M9 11h4m-2 3h1'/%3E%3Cpath stroke='%23bee0fc' d='M13 11h1'/%3E%3Cpath stroke='%2391abe1' d='M14 11h1m-7 5h3'/%3E%3Cpath stroke='%23d5eafd' d='M3 12h1m-1 1h1'/%3E%3Cpath stroke='%23d4ebfc' d='M4 12h3m-3 1h3'/%3E%3Cpath stroke='%23d2ecfe' d='M7 12h1m-1 1h1'/%3E%3Cpath stroke='%23c6ebff' d='M9 12h2m-2 1h2m-3 1h1'/%3E%3Cpath stroke='%23c2eaff' d='M11 12h1m-1 1h1'/%3E%3Cpath stroke='%23bde3fd' d='M12 12h1m-1 1h1'/%3E%3Cpath stroke='%23b9ddfb' d='M13 12h1m-1 1h1m-1 1h1m-2 1h1'/%3E%3Cpath stroke='%238da9e1' d='M14 12h1m-1 1h1m-3 3h1'/%3E%3Cpath stroke='%2388a3e3' d='M0 14h1'/%3E%3Cpath stroke='%23d2e9fe' d='M3 14h1'/%3E%3Cpath stroke='%23d0e8fc' d='M4 14h3'/%3E%3Cpath stroke='%23c4e8fe' d='M9 14h2'/%3E%3Cpath stroke='%23bde0fb' d='M12 14h1m-2 1h1'/%3E%3Cpath stroke='%238fa7df' d='M14 14h1'/%3E%3Cpath stroke='%2387a0de' d='M0 15h1'/%3E%3Cpath stroke='%23c5ddfc' d='M1 15h1'/%3E%3Cpath stroke='%23c8e3fb' d='M7 15h1'/%3E%3Cpath stroke='%23b9dafb' d='M13 15h1'/%3E%3Cpath stroke='%238ca7de' d='M14 15h1m-2 1h1'/%3E%3Cpath stroke='%23c3cde7' d='M0 16h1'/%3E%3Cpath stroke='%2393a7df' d='M1 16h1'/%3E%3Cpath stroke='%2396acdf' d='M2 16h1m4 0h1'/%3E%3Cpath stroke='%2390abe0' d='M11 16h1'/%3E%3Cpath stroke='%23becbe6' d='M14 16h1'/%3E%3C/svg%3E\")}select:active{background-image:url(\"data:image/svg+xml;charset=utf-8,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 -0.5 15 17' shape-rendering='crispEdges'%3E%3Cpath stroke='%23bbc2dc' d='M0 0h1'/%3E%3Cpath stroke='%23a2acdc' d='M1 0h1M0 1h1'/%3E%3Cpath stroke='%2398a1dd' d='M2 0h1M0 2h1'/%3E%3Cpath stroke='%23919bda' d='M3 0h1M0 3h1M0 4h1'/%3E%3Cpath stroke='%238c97dc' d='M4 0h2M0 5h1'/%3E%3Cpath stroke='%238793db' d='M6 0h1M0 6h1'/%3E%3Cpath stroke='%238592d9' d='M7 0h1'/%3E%3Cpath stroke='%23838fda' d='M8 0h1M0 7h1M0 8h1'/%3E%3Cpath stroke='%23828dd9' d='M9 0h2M0 9h1'/%3E%3Cpath stroke='%237f8ada' d='M11 0h1'/%3E%3Cpath stroke='%237b88da' d='M12 0h1'/%3E%3Cpath stroke='%237786d9' d='M13 0h1'/%3E%3Cpath stroke='%23afb6db' d='M14 0h1'/%3E%3Cpath stroke='%236e8ef1' d='M1 1h1'/%3E%3Cpath stroke='%236e94f1' d='M2 1h1M1 2h1'/%3E%3Cpath stroke='%237598f1' d='M3 1h1'/%3E%3Cpath stroke='%237d9df1' d='M4 1h2M1 6h1'/%3E%3Cpath stroke='%23809df1' d='M6 1h1M1 7h1M1 8h1'/%3E%3Cpath stroke='%2383a0f0' d='M7 1h1'/%3E%3Cpath stroke='%23899fef' d='M8 1h1m4 1h1'/%3E%3Cpath stroke='%238d9eef' d='M9 1h3M1 12h1m-1 1h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23869df0' d='M12 1h1'/%3E%3Cpath stroke='%238096f1' d='M13 1h1'/%3E%3Cpath stroke='%23b9c9f3' d='M14 1h1M3 16h4'/%3E%3Cpath stroke='%23759cf1' d='M2 2h1'/%3E%3Cpath stroke='%237ca4f0' d='M3 2h1M3 3h1M3 4h1M2 5h2'/%3E%3Cpath stroke='%2380a6f1' d='M4 2h2'/%3E%3Cpath stroke='%2383a7f0' d='M6 2h1M4 3h2M4 4h2'/%3E%3Cpath stroke='%238daaef' d='M7 2h1'/%3E%3Cpath stroke='%2390a7ef' d='M8 2h3'/%3E%3Cpath stroke='%2394a8ef' d='M11 2h1'/%3E%3Cpath stroke='%2390a4ef' d='M12 2h1m0 1h1m-1 1h1m-8 8h1m-1 1h1m-5 1h1'/%3E%3Cpath stroke='%23baccf4' d='M14 2h1m-1 5h1m-1 1h1'/%3E%3Cpath stroke='%237396f2' d='M1 3h1M1 4h1'/%3E%3Cpath stroke='%23759ef1' d='M2 3h1M2 4h1'/%3E%3Cpath stroke='%2390b1f2' d='M6 3h1M6 4h1M6 5h1'/%3E%3Cpath stroke='%2393aff2' d='M7 3h1M7 4h1'/%3E%3Cpath stroke='%2397afee' d='M8 3h3M8 4h3'/%3E%3Cpath stroke='%239aaaee' d='M11 3h1m-1 1h1'/%3E%3Cpath stroke='%2397abee' d='M12 3h1m-1 1h1m0 1h1M3 14h1m-2 1h1'/%3E%3Cpath stroke='%23bcccf3' d='M14 3h1m-1 1h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%237997f1' d='M1 5h1'/%3E%3Cpath stroke='%2383a6f4' d='M4 5h2'/%3E%3Cpath stroke='%239aaff1' d='M7 5h1'/%3E%3Cpath stroke='%239eb1f2' d='M8 5h1m0 6h2'/%3E%3Cpath stroke='%239bb0ef' d='M9 5h4M3 15h1'/%3E%3Cpath stroke='%23809ff1' d='M2 6h1'/%3E%3Cpath stroke='%2387a3f4' d='M3 6h1M2 7h1M2 8h2'/%3E%3Cpath stroke='%234d6185' d='M4 6h1m5 0h1M3 7h3m3 0h3M4 8h3m1 0h3M5 9h5m-4 1h3m-2 1h1'/%3E%3Cpath stroke='%2393acf2' d='M5 6h3M6 7h2M7 8h1M4 9h1'/%3E%3Cpath stroke='%239eaef2' d='M8 6h2'/%3E%3Cpath stroke='%239baeef' d='M11 6h3m-8 8h1'/%3E%3Cpath stroke='%2395abf4' d='M8 7h1'/%3E%3Cpath stroke='%23a2b2ee' d='M12 7h2m-2 1h2m-7 6h1'/%3E%3Cpath stroke='%23a3b4f3' d='M11 8h1m-4 3h1'/%3E%3Cpath stroke='%238da3f3' d='M1 9h1m-1 1h5m-3 1h1'/%3E%3Cpath stroke='%2389a3f3' d='M2 9h2'/%3E%3Cpath stroke='%2397adf2' d='M10 9h1'/%3E%3Cpath stroke='%23a6b6f2' d='M11 9h2m-2 2h1'/%3E%3Cpath stroke='%23a4b7ed' d='M13 9h1m-6 3h3m-3 1h3'/%3E%3Cpath stroke='%23b9cbf3' d='M14 9h1'/%3E%3Cpath stroke='%237c89db' d='M0 10h1'/%3E%3Cpath stroke='%23a5b6f1' d='M9 10h2'/%3E%3Cpath stroke='%23a8b8f1' d='M11 10h2'/%3E%3Cpath stroke='%23aabcef' d='M13 10h1m-3 2h1m-1 1h1m-5 2h1'/%3E%3Cpath stroke='%23b5c9f3' d='M14 10h1'/%3E%3Cpath stroke='%237988db' d='M0 11h1'/%3E%3Cpath stroke='%2390a4f2' d='M1 11h2m1 0h2'/%3E%3Cpath stroke='%238da1ef' d='M6 11h1m-5 1h4m-4 1h4'/%3E%3Cpath stroke='%23a9bcf2' d='M12 11h1'/%3E%3Cpath stroke='%23afbfed' d='M13 11h1m-6 4h1'/%3E%3Cpath stroke='%23b3c8f5' d='M14 11h1m-7 5h3'/%3E%3Cpath stroke='%237583db' d='M0 12h1m-1 1h1'/%3E%3Cpath stroke='%2397a7ee' d='M7 12h1m-1 1h1'/%3E%3Cpath stroke='%23b4c7ee' d='M12 12h1m-1 1h1m-4 2h2'/%3E%3Cpath stroke='%23b7c9ee' d='M13 12h1m-1 1h1m-3 2h1'/%3E%3Cpath stroke='%23b2c8f4' d='M14 12h1m-1 1h1m-4 3h1'/%3E%3Cpath stroke='%237381d9' d='M0 14h1m-1 1h1'/%3E%3Cpath stroke='%239baaef' d='M4 14h2'/%3E%3Cpath stroke='%23a9baee' d='M8 14h1m-3 1h1'/%3E%3Cpath stroke='%23adbcef' d='M9 14h2'/%3E%3Cpath stroke='%23b4c2ee' d='M11 14h1'/%3E%3Cpath stroke='%23becfed' d='M12 14h1'/%3E%3Cpath stroke='%23c4d4ec' d='M13 14h1m-2 1h1'/%3E%3Cpath stroke='%23b1c6f3' d='M14 14h1m-3 2h1'/%3E%3Cpath stroke='%23a2b6ee' d='M4 15h2'/%3E%3Cpath stroke='%23d2deeb' d='M13 15h1'/%3E%3Cpath stroke='%23b0c5f2' d='M14 15h1m-2 1h1'/%3E%3Cpath stroke='%23dbe3f8' d='M0 16h1'/%3E%3Cpath stroke='%23b7c6f1' d='M1 16h1'/%3E%3Cpath stroke='%23b8c9f2' d='M2 16h1m4 0h1'/%3E%3Cpath stroke='%23d9e3f6' d='M14 16h1'/%3E%3C/svg%3E\")}input[type=radio]{appearance:none;-webkit-appearance:none;-moz-appearance:none;margin:0;background:0;position:fixed;opacity:0;border:none}input[type=radio]+label{line-height:16px}input[type=radio]+label:before{background:linear-gradient(135deg,#dcdcd7,#fff);border-radius:50%;border:1px solid #1d5281}input[type=radio]:not([disabled]):not(:active)+label:hover:before{box-shadow:inset -2px -2px #f8b636,inset 2px 2px #fedf9c}input[type=radio]:active+label:before{background:linear-gradient(135deg,#b0b0a7,#e3e1d2)}input[type=radio]:checked+label:after{background:url(\"data:image/svg+xml;charset=utf-8,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 -0.5 5 5' shape-rendering='crispEdges'%3E%3Cpath stroke='%23a9dca6' d='M1 0h1M0 1h1'/%3E%3Cpath stroke='%234dbf4a' d='M2 0h1M0 2h1'/%3E%3Cpath stroke='%23a0d29e' d='M3 0h1M0 3h1'/%3E%3Cpath stroke='%2355d551' d='M1 1h1'/%3E%3Cpath stroke='%2343c33f' d='M2 1h1'/%3E%3Cpath stroke='%2329a826' d='M3 1h1'/%3E%3Cpath stroke='%239acc98' d='M4 1h1M1 4h1'/%3E%3Cpath stroke='%2342c33f' d='M1 2h1'/%3E%3Cpath stroke='%2338b935' d='M2 2h1'/%3E%3Cpath stroke='%2321a121' d='M3 2h1'/%3E%3Cpath stroke='%23269623' d='M4 2h1'/%3E%3Cpath stroke='%232aa827' d='M1 3h1'/%3E%3Cpath stroke='%2322a220' d='M2 3h1'/%3E%3Cpath stroke='%23139210' d='M3 3h1'/%3E%3Cpath stroke='%2398c897' d='M4 3h1'/%3E%3Cpath stroke='%23249624' d='M2 4h1'/%3E%3Cpath stroke='%2398c997' d='M3 4h1'/%3E%3C/svg%3E\")}input[type=radio]:focus+label{outline:1px dotted #000}input[type=radio][disabled]+label:before{border:1px solid #cac8bb;background:#fff}input[type=radio][disabled]:checked+label:after{background:url(\"data:image/svg+xml;charset=utf-8,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 -0.5 5 5' shape-rendering='crispEdges'%3E%3Cpath stroke='%23e8e6da' d='M1 0h1M0 1h1'/%3E%3Cpath stroke='%23d2ceb5' d='M2 0h1M0 2h1'/%3E%3Cpath stroke='%23e5e3d4' d='M3 0h1M0 3h1'/%3E%3Cpath stroke='%23d7d3bd' d='M1 1h1'/%3E%3Cpath stroke='%23d0ccb2' d='M2 1h1M1 2h1'/%3E%3Cpath stroke='%23c7c2a2' d='M3 1h1M1 3h1'/%3E%3Cpath stroke='%23e2dfd0' d='M4 1h1M1 4h1'/%3E%3Cpath stroke='%23cdc8ac' d='M2 2h1'/%3E%3Cpath stroke='%23c5bf9f' d='M3 2h1M2 3h1'/%3E%3Cpath stroke='%23c3bd9c' d='M4 2h1'/%3E%3Cpath stroke='%23bfb995' d='M3 3h1'/%3E%3Cpath stroke='%23e2dfcf' d='M4 3h1M3 4h1'/%3E%3Cpath stroke='%23c4be9d' d='M2 4h1'/%3E%3C/svg%3E\")}input[type=checkbox]+label:before{box-shadow:none;border:1px solid #1d5281;background:linear-gradient(135deg,#dcdcd7,#fff)}input[type=checkbox]:checked+label:after{background:url(\"data:image/svg+xml;charset=utf-8,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 -0.5 11 11' shape-rendering='crispEdges'%3E%3Cpath stroke='%2322a122' d='M8 2h1M7 3h2M2 4h1m3 0h3M2 5h2m1 0h3M2 6h5M3 7h3M4 8h1'/%3E%3C/svg%3E\");left:-17px}input[type=checkbox]:not([disabled]):not(:active)+label:hover:before{box-shadow:inset -2px -2px #f8b636,inset 2px 2px #fedf9c}input[type=checkbox]:active+label:before{background:linear-gradient(135deg,#b0b0a7,#e3e1d2)}input[type=checkbox][disabled]+label:before{background:#fff;border:1px solid #cac8bb}input[type=checkbox][disabled]:checked+label:after{background:url(\"data:image/svg+xml;charset=utf-8,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 -0.5 11 11' shape-rendering='crispEdges'%3E%3Cpath stroke='%23cac8bb' d='M8 2h1M7 3h2M2 4h1m3 0h3M2 5h2m1 0h3M2 6h5M3 7h3M4 8h1'/%3E%3C/svg%3E\")}input[type=email],input[type=password],input[type=text],select{height:23px}input::selection{background:#2267cb;color:#fff}textarea::selection{background:#2267cb;color:#fff}input[type=range]::-webkit-slider-thumb{height:21px;width:11px;background:url(\"data:image/svg+xml;charset=utf-8,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 -0.5 11 21' shape-rendering='crispEdges'%3E%3Cpath stroke='%23becbd3' d='M1 0h1M0 1h1'/%3E%3Cpath stroke='%23b6c5cd' d='M2 0h1M0 2h1'/%3E%3Cpath stroke='%23b5c4cd' d='M3 0h5M0 3h1M0 4h1M0 5h1M0 6h1M0 7h1M0 8h1M0 9h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23afbfc8' d='M8 0h1M0 14h1'/%3E%3Cpath stroke='%239fb2be' d='M9 0h1M0 15h1'/%3E%3Cpath stroke='%23a6d1b1' d='M1 1h1'/%3E%3Cpath stroke='%236fd16e' d='M2 1h1M1 2h1'/%3E%3Cpath stroke='%2367ce65' d='M3 1h1M1 3h1'/%3E%3Cpath stroke='%2366ce64' d='M4 1h3'/%3E%3Cpath stroke='%2362cd61' d='M7 1h1'/%3E%3Cpath stroke='%2345c343' d='M8 1h1M7 2h1'/%3E%3Cpath stroke='%2363ac76' d='M9 1h1M2 16h1m0 1h1m0 1h1'/%3E%3Cpath stroke='%23879aa6' d='M10 1h1'/%3E%3Cpath stroke='%2363cd62' d='M2 2h1'/%3E%3Cpath stroke='%2349c547' d='M3 2h1M2 3h1'/%3E%3Cpath stroke='%2347c446' d='M4 2h3'/%3E%3Cpath stroke='%2321b71f' d='M8 2h1'/%3E%3Cpath stroke='%231da41c' d='M9 2h1'/%3E%3Cpath stroke='%237d8e99' d='M10 2h1'/%3E%3Cpath stroke='%2325b923' d='M3 3h1'/%3E%3Cpath stroke='%2321b81f' d='M4 3h4M2 15h1'/%3E%3Cpath stroke='%231ea71c' d='M8 3h1'/%3E%3Cpath stroke='%231b9619' d='M9 3h1'/%3E%3Cpath stroke='%23778892' d='M10 3h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23f7f7f4' d='M1 4h1M1 5h1M1 6h1M1 7h1M1 8h1M1 9h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23f5f5f2' d='M2 4h1M2 5h1M2 6h1M2 7h1M2 8h1M2 9h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23f3f3ef' d='M3 4h5M3 5h5M3 6h5M3 7h5M3 8h5M3 9h5m-5 1h5m-5 1h5m-5 1h5m-5 1h4m-4 1h3m-2 1h1'/%3E%3Cpath stroke='%23dcdcd9' d='M8 4h1M8 5h1M8 6h1M8 7h1M8 8h1M8 9h1m-1 1h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23c3c3c0' d='M9 4h1M9 5h1M9 6h1M9 7h1M9 8h1M9 9h1m-1 1h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23f1f1ed' d='M7 13h1m-2 1h1m-2 1h1'/%3E%3Cpath stroke='%23dbdbd8' d='M8 13h1'/%3E%3Cpath stroke='%23c4c4c1' d='M9 13h1'/%3E%3Cpath stroke='%234bc549' d='M1 14h1'/%3E%3Cpath stroke='%23f4f4f1' d='M2 14h1'/%3E%3Cpath stroke='%23e6e6e2' d='M7 14h1m-2 1h1'/%3E%3Cpath stroke='%23cececa' d='M8 14h1'/%3E%3Cpath stroke='%231a9319' d='M9 14h1'/%3E%3Cpath stroke='%23788993' d='M10 14h1'/%3E%3Cpath stroke='%2369b17b' d='M1 15h1'/%3E%3Cpath stroke='%23f2f2ee' d='M3 15h1m0 1h1'/%3E%3Cpath stroke='%23d0d0cc' d='M7 15h1m-2 1h1'/%3E%3Cpath stroke='%231a9118' d='M8 15h1m-2 1h1m-2 1h1'/%3E%3Cpath stroke='%234c845a' d='M9 15h1'/%3E%3Cpath stroke='%2372838d' d='M10 15h1'/%3E%3Cpath stroke='%2391a6b2' d='M1 16h1m0 1h1m0 1h1m0 1h1'/%3E%3Cpath stroke='%2321b61f' d='M3 16h1m0 1h1'/%3E%3Cpath stroke='%23e7e7e3' d='M5 16h1'/%3E%3Cpath stroke='%234b8259' d='M8 16h1m-2 1h1m-2 1h1'/%3E%3Cpath stroke='%236e7e88' d='M9 16h1m-2 1h1m-2 1h1m-2 1h1'/%3E%3Cpath stroke='%23d7d7d4' d='M5 17h1'/%3E%3Cpath stroke='%231da21b' d='M5 18h1'/%3E%3Cpath stroke='%23589868' d='M5 19h1'/%3E%3Cpath stroke='%2380929e' d='M5 20h1'/%3E%3C/svg%3E\");transform:translateY(-8px)}input[type=range]::-moz-range-thumb{height:21px;width:11px;border:0;border-radius:0;background:url(\"data:image/svg+xml;charset=utf-8,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 -0.5 11 21' shape-rendering='crispEdges'%3E%3Cpath stroke='%23becbd3' d='M1 0h1M0 1h1'/%3E%3Cpath stroke='%23b6c5cd' d='M2 0h1M0 2h1'/%3E%3Cpath stroke='%23b5c4cd' d='M3 0h5M0 3h1M0 4h1M0 5h1M0 6h1M0 7h1M0 8h1M0 9h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23afbfc8' d='M8 0h1M0 14h1'/%3E%3Cpath stroke='%239fb2be' d='M9 0h1M0 15h1'/%3E%3Cpath stroke='%23a6d1b1' d='M1 1h1'/%3E%3Cpath stroke='%236fd16e' d='M2 1h1M1 2h1'/%3E%3Cpath stroke='%2367ce65' d='M3 1h1M1 3h1'/%3E%3Cpath stroke='%2366ce64' d='M4 1h3'/%3E%3Cpath stroke='%2362cd61' d='M7 1h1'/%3E%3Cpath stroke='%2345c343' d='M8 1h1M7 2h1'/%3E%3Cpath stroke='%2363ac76' d='M9 1h1M2 16h1m0 1h1m0 1h1'/%3E%3Cpath stroke='%23879aa6' d='M10 1h1'/%3E%3Cpath stroke='%2363cd62' d='M2 2h1'/%3E%3Cpath stroke='%2349c547' d='M3 2h1M2 3h1'/%3E%3Cpath stroke='%2347c446' d='M4 2h3'/%3E%3Cpath stroke='%2321b71f' d='M8 2h1'/%3E%3Cpath stroke='%231da41c' d='M9 2h1'/%3E%3Cpath stroke='%237d8e99' d='M10 2h1'/%3E%3Cpath stroke='%2325b923' d='M3 3h1'/%3E%3Cpath stroke='%2321b81f' d='M4 3h4M2 15h1'/%3E%3Cpath stroke='%231ea71c' d='M8 3h1'/%3E%3Cpath stroke='%231b9619' d='M9 3h1'/%3E%3Cpath stroke='%23778892' d='M10 3h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23f7f7f4' d='M1 4h1M1 5h1M1 6h1M1 7h1M1 8h1M1 9h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23f5f5f2' d='M2 4h1M2 5h1M2 6h1M2 7h1M2 8h1M2 9h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23f3f3ef' d='M3 4h5M3 5h5M3 6h5M3 7h5M3 8h5M3 9h5m-5 1h5m-5 1h5m-5 1h5m-5 1h4m-4 1h3m-2 1h1'/%3E%3Cpath stroke='%23dcdcd9' d='M8 4h1M8 5h1M8 6h1M8 7h1M8 8h1M8 9h1m-1 1h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23c3c3c0' d='M9 4h1M9 5h1M9 6h1M9 7h1M9 8h1M9 9h1m-1 1h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23f1f1ed' d='M7 13h1m-2 1h1m-2 1h1'/%3E%3Cpath stroke='%23dbdbd8' d='M8 13h1'/%3E%3Cpath stroke='%23c4c4c1' d='M9 13h1'/%3E%3Cpath stroke='%234bc549' d='M1 14h1'/%3E%3Cpath stroke='%23f4f4f1' d='M2 14h1'/%3E%3Cpath stroke='%23e6e6e2' d='M7 14h1m-2 1h1'/%3E%3Cpath stroke='%23cececa' d='M8 14h1'/%3E%3Cpath stroke='%231a9319' d='M9 14h1'/%3E%3Cpath stroke='%23788993' d='M10 14h1'/%3E%3Cpath stroke='%2369b17b' d='M1 15h1'/%3E%3Cpath stroke='%23f2f2ee' d='M3 15h1m0 1h1'/%3E%3Cpath stroke='%23d0d0cc' d='M7 15h1m-2 1h1'/%3E%3Cpath stroke='%231a9118' d='M8 15h1m-2 1h1m-2 1h1'/%3E%3Cpath stroke='%234c845a' d='M9 15h1'/%3E%3Cpath stroke='%2372838d' d='M10 15h1'/%3E%3Cpath stroke='%2391a6b2' d='M1 16h1m0 1h1m0 1h1m0 1h1'/%3E%3Cpath stroke='%2321b61f' d='M3 16h1m0 1h1'/%3E%3Cpath stroke='%23e7e7e3' d='M5 16h1'/%3E%3Cpath stroke='%234b8259' d='M8 16h1m-2 1h1m-2 1h1'/%3E%3Cpath stroke='%236e7e88' d='M9 16h1m-2 1h1m-2 1h1m-2 1h1'/%3E%3Cpath stroke='%23d7d7d4' d='M5 17h1'/%3E%3Cpath stroke='%231da21b' d='M5 18h1'/%3E%3Cpath stroke='%23589868' d='M5 19h1'/%3E%3Cpath stroke='%2380929e' d='M5 20h1'/%3E%3C/svg%3E\");transform:translateY(2px)}input[type=range]::-webkit-slider-runnable-track{width:100%;height:2px;box-sizing:border-box;background:#ecebe4;border-right:1px solid #f3f2ea;border-bottom:1px solid #f3f2ea;border-radius:2px;box-shadow:1px 0 0 #fff,1px 1px 0 #fff,0 1px 0 #fff,-1px 0 0 #9d9c99,-1px -1px 0 #9d9c99,0 -1px 0 #9d9c99,-1px 1px 0 #fff,1px -1px #9d9c99}input[type=range]::-moz-range-track{width:100%;height:2px;box-sizing:border-box;background:#ecebe4;border-right:1px solid #f3f2ea;border-bottom:1px solid #f3f2ea;border-radius:2px;box-shadow:1px 0 0 #fff,1px 1px 0 #fff,0 1px 0 #fff,-1px 0 0 #9d9c99,-1px -1px 0 #9d9c99,0 -1px 0 #9d9c99,-1px 1px 0 #fff,1px -1px #9d9c99}input[type=range].has-box-indicator::-webkit-slider-thumb{background:url(\"data:image/svg+xml;charset=utf-8,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 -0.5 11 22' shape-rendering='crispEdges'%3E%3Cpath stroke='%23f2f1e7' d='M0 0h1m9 0h1M0 21h1m9 0h1'/%3E%3Cpath stroke='%23879aa6' d='M1 0h1m8 20h1'/%3E%3Cpath stroke='%237d8e99' d='M2 0h1m7 19h1'/%3E%3Cpath stroke='%23778892' d='M3 0h5m2 3h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23788993' d='M8 0h1m1 2h1'/%3E%3Cpath stroke='%2372838d' d='M9 0h1m0 1h1'/%3E%3Cpath stroke='%239fb2be' d='M0 1h1m8 20h1'/%3E%3Cpath stroke='%2363af76' d='M1 1h1m7 19h1'/%3E%3Cpath stroke='%231eab1c' d='M2 1h1m6 18h1'/%3E%3Cpath stroke='%231c9d1a' d='M3 1h1'/%3E%3Cpath stroke='%231b9a1a' d='M4 1h3m1 0h1m0 1h1'/%3E%3Cpath stroke='%231b9b1a' d='M7 1h1'/%3E%3Cpath stroke='%234d875b' d='M9 1h1'/%3E%3Cpath stroke='%23afbfc8' d='M0 2h1m7 19h1'/%3E%3Cpath stroke='%2346ca44' d='M1 2h1m5 17h1m0 1h1'/%3E%3Cpath stroke='%2322be20' d='M2 2h1m5 17h1'/%3E%3Cpath stroke='%231faf1d' d='M3 2h1'/%3E%3Cpath stroke='%231fae1d' d='M4 2h3'/%3E%3Cpath stroke='%231fad1d' d='M7 2h1'/%3E%3Cpath stroke='%231da11b' d='M8 2h1'/%3E%3Cpath stroke='%23b5c4cd' d='M0 3h1M0 4h1M0 5h1M0 6h1M0 7h1M0 8h1M0 9h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m2 3h5'/%3E%3Cpath stroke='%23f7f7f4' d='M1 3h1M1 4h1M1 5h1M1 6h1M1 7h1M1 8h1M1 9h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23f5f5f2' d='M2 3h1M2 4h1M2 5h1M2 6h1M2 7h1M2 8h1M2 9h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23f3f3ef' d='M3 3h4M3 4h5M3 5h5M3 6h5M3 7h5M3 8h5M3 9h5m-5 1h5m-5 1h5m-5 1h5m-5 1h5m-5 1h5m-5 1h5m-5 1h5m-5 1h5m-5 1h5'/%3E%3Cpath stroke='%23f1f1ed' d='M7 3h1'/%3E%3Cpath stroke='%23dbdbd8' d='M8 3h1'/%3E%3Cpath stroke='%23c4c4c1' d='M9 3h1'/%3E%3Cpath stroke='%23ddddd9' d='M8 4h1M8 18h1'/%3E%3Cpath stroke='%23c6c6c3' d='M9 4h1M9 18h1'/%3E%3Cpath stroke='%23dcdcd9' d='M8 5h1M8 6h1M8 7h1M8 8h1M8 9h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23c3c3c0' d='M9 5h1M9 6h1M9 7h1M9 8h1M9 9h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23b6c5cd' d='M0 19h1m1 2h1'/%3E%3Cpath stroke='%2370d66f' d='M1 19h1m0 1h1'/%3E%3Cpath stroke='%2364d362' d='M2 19h1'/%3E%3Cpath stroke='%234acb48' d='M3 19h1'/%3E%3Cpath stroke='%2348cb46' d='M4 19h3'/%3E%3Cpath stroke='%23becbd3' d='M0 20h1m0 1h1'/%3E%3Cpath stroke='%23a6d2b1' d='M1 20h1'/%3E%3Cpath stroke='%2367d466' d='M3 20h1'/%3E%3Cpath stroke='%2366d465' d='M4 20h3'/%3E%3Cpath stroke='%2363d362' d='M7 20h1'/%3E%3C/svg%3E\");transform:translateY(-10px)}input[type=range].has-box-indicator::-moz-range-thumb{background:url(\"data:image/svg+xml;charset=utf-8,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 -0.5 11 22' shape-rendering='crispEdges'%3E%3Cpath stroke='%23f2f1e7' d='M0 0h1m9 0h1M0 21h1m9 0h1'/%3E%3Cpath stroke='%23879aa6' d='M1 0h1m8 20h1'/%3E%3Cpath stroke='%237d8e99' d='M2 0h1m7 19h1'/%3E%3Cpath stroke='%23778892' d='M3 0h5m2 3h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23788993' d='M8 0h1m1 2h1'/%3E%3Cpath stroke='%2372838d' d='M9 0h1m0 1h1'/%3E%3Cpath stroke='%239fb2be' d='M0 1h1m8 20h1'/%3E%3Cpath stroke='%2363af76' d='M1 1h1m7 19h1'/%3E%3Cpath stroke='%231eab1c' d='M2 1h1m6 18h1'/%3E%3Cpath stroke='%231c9d1a' d='M3 1h1'/%3E%3Cpath stroke='%231b9a1a' d='M4 1h3m1 0h1m0 1h1'/%3E%3Cpath stroke='%231b9b1a' d='M7 1h1'/%3E%3Cpath stroke='%234d875b' d='M9 1h1'/%3E%3Cpath stroke='%23afbfc8' d='M0 2h1m7 19h1'/%3E%3Cpath stroke='%2346ca44' d='M1 2h1m5 17h1m0 1h1'/%3E%3Cpath stroke='%2322be20' d='M2 2h1m5 17h1'/%3E%3Cpath stroke='%231faf1d' d='M3 2h1'/%3E%3Cpath stroke='%231fae1d' d='M4 2h3'/%3E%3Cpath stroke='%231fad1d' d='M7 2h1'/%3E%3Cpath stroke='%231da11b' d='M8 2h1'/%3E%3Cpath stroke='%23b5c4cd' d='M0 3h1M0 4h1M0 5h1M0 6h1M0 7h1M0 8h1M0 9h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m2 3h5'/%3E%3Cpath stroke='%23f7f7f4' d='M1 3h1M1 4h1M1 5h1M1 6h1M1 7h1M1 8h1M1 9h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23f5f5f2' d='M2 3h1M2 4h1M2 5h1M2 6h1M2 7h1M2 8h1M2 9h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23f3f3ef' d='M3 3h4M3 4h5M3 5h5M3 6h5M3 7h5M3 8h5M3 9h5m-5 1h5m-5 1h5m-5 1h5m-5 1h5m-5 1h5m-5 1h5m-5 1h5m-5 1h5m-5 1h5'/%3E%3Cpath stroke='%23f1f1ed' d='M7 3h1'/%3E%3Cpath stroke='%23dbdbd8' d='M8 3h1'/%3E%3Cpath stroke='%23c4c4c1' d='M9 3h1'/%3E%3Cpath stroke='%23ddddd9' d='M8 4h1M8 18h1'/%3E%3Cpath stroke='%23c6c6c3' d='M9 4h1M9 18h1'/%3E%3Cpath stroke='%23dcdcd9' d='M8 5h1M8 6h1M8 7h1M8 8h1M8 9h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23c3c3c0' d='M9 5h1M9 6h1M9 7h1M9 8h1M9 9h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1m-1 1h1'/%3E%3Cpath stroke='%23b6c5cd' d='M0 19h1m1 2h1'/%3E%3Cpath stroke='%2370d66f' d='M1 19h1m0 1h1'/%3E%3Cpath stroke='%2364d362' d='M2 19h1'/%3E%3Cpath stroke='%234acb48' d='M3 19h1'/%3E%3Cpath stroke='%2348cb46' d='M4 19h3'/%3E%3Cpath stroke='%23becbd3' d='M0 20h1m0 1h1'/%3E%3Cpath stroke='%23a6d2b1' d='M1 20h1'/%3E%3Cpath stroke='%2367d466' d='M3 20h1'/%3E%3Cpath stroke='%2366d465' d='M4 20h3'/%3E%3Cpath stroke='%2363d362' d='M7 20h1'/%3E%3C/svg%3E\");transform:translateY(0)}.is-vertical>input[type=range]::-webkit-slider-runnable-track{border-left:1px solid #f3f2ea;border-right:0;border-bottom:1px solid #f3f2ea;box-shadow:-1px 0 0 #fff,-1px 1px 0 #fff,0 1px 0 #fff,1px 0 0 #9d9c99,1px -1px 0 #9d9c99,0 -1px 0 #9d9c99,1px 1px 0 #fff,-1px -1px #9d9c99}.is-vertical>input[type=range]::-moz-range-track{border-left:1px solid #f3f2ea;border-right:0;border-bottom:1px solid #f3f2ea;box-shadow:-1px 0 0 #fff,-1px 1px 0 #fff,0 1px 0 #fff,1px 0 0 #9d9c99,1px -1px 0 #9d9c99,0 -1px 0 #9d9c99,1px 1px 0 #fff,-1px -1px #9d9c99}fieldset{box-shadow:none;background:#fff;border:1px solid #d0d0bf;border-radius:4px;padding-top:10px}legend{background:transparent;color:#0046d5}.field-row{display:flex;align-items:center}.field-row>*+*{margin-left:6px}[class^=field-row]+[class^=field-row]{margin-top:6px}.field-row-stacked{display:flex;flex-direction:column}.field-row-stacked *+*{margin-top:6px}menu[role=tablist] button{background:linear-gradient(180deg,#fff,#fafaf9 26%,#f0f0ea 95%,#ecebe5);margin-left:-1px;margin-right:2px;border-radius:0;border-color:#91a7b4;border-top-right-radius:3px;border-top-left-radius:3px;padding:0 12px 3px}menu[role=tablist] button:hover{box-shadow:unset;border-top:1px solid #e68b2c;box-shadow:inset 0 2px #ffc73c}menu[role=tablist] button[aria-selected=true]{background:#fcfcfe;border-color:#919b9c;margin-right:-1px;border-bottom:1px solid transparent;border-top:1px solid #e68b2c;box-shadow:inset 0 2px #ffc73c}menu[role=tablist] button[aria-selected=true]:first-of-type:before{content:\"\";display:block;position:absolute;z-index:-1;top:100%;left:-1px;height:2px;width:0;border-left:1px solid #919b9c}[role=tabpanel]{box-shadow:inset 1px 1px #fcfcfe,inset -1px -1px #fcfcfe,1px 2px 2px 0 rgba(208,206,191,.75)}ul.tree-view{-webkit-font-smoothing:auto;border:1px solid #7f9db9;padding:2px 5px}@keyframes sliding{0%{transform:translateX(-30px)}to{transform:translateX(100%)}}progress{box-sizing:border-box;appearance:none;-webkit-appearance:none;-moz-appearance:none;height:14px;border:1px solid #686868;border-radius:4px;padding:1px 2px 1px 0;overflow:hidden;background-color:#fff;-webkit-box-shadow:inset 0 0 1px 0 #686868;-moz-box-shadow:inset 0 0 1px 0 #686868}progress,progress:not([value]){box-shadow:inset 0 0 1px 0 #686868}progress:not([value]){-moz-box-shadow:inset 0 0 1px 0 #686868;-webkit-box-shadow:inset 0 0 1px 0 #686868;height:14px}progress[value]::-webkit-progress-bar{background-color:transparent}progress[value]::-webkit-progress-value{border-radius:2px;background:repeating-linear-gradient(90deg,#fff 0,#fff 2px,transparent 0,transparent 10px),linear-gradient(180deg,#acedad 0,#7be47d 14%,#4cda50 28%,#2ed330 42%,#42d845 57%,#76e275 71%,#8fe791 85%,#fff)}progress[value]::-moz-progress-bar{border-radius:2px;background:repeating-linear-gradient(90deg,#fff 0,#fff 2px,transparent 0,transparent 10px),linear-gradient(180deg,#acedad 0,#7be47d 14%,#4cda50 28%,#2ed330 42%,#42d845 57%,#76e275 71%,#8fe791 85%,#fff)}progress:not([value])::-webkit-progress-bar{width:100%;background:repeating-linear-gradient(90deg,transparent 0,transparent 8px,#fff 0,#fff 10px,transparent 0,transparent 18px,#fff 0,#fff 20px,transparent 0,transparent 28px,#fff 0,#fff),linear-gradient(180deg,#acedad 0,#7be47d 14%,#4cda50 28%,#2ed330 42%,#42d845 57%,#76e275 71%,#8fe791 85%,#fff);animation:sliding 2s linear 0s infinite}progress:not([value])::-webkit-progress-bar:not([value]){animation:sliding 2s linear 0s infinite;background:repeating-linear-gradient(90deg,transparent 0,transparent 8px,#fff 0,#fff 10px,transparent 0,transparent 18px,#fff 0,#fff 20px,transparent 0,transparent 28px,#fff 0,#fff),linear-gradient(180deg,#acedad 0,#7be47d 14%,#4cda50 28%,#2ed330 42%,#42d845 57%,#76e275 71%,#8fe791 85%,#fff)}progress:not([value]){position:relative}progress:not([value]):before{box-sizing:border-box;content:\"\";position:absolute;top:0;left:0;width:100%;height:100%;background-color:#fff;-webkit-box-shadow:inset 0 0 1px 0 #686868;-moz-box-shadow:inset 0 0 1px 0 #686868}progress:not([value]):before,progress:not([value]):before:not([value]){box-shadow:inset 0 0 1px 0 #686868}progress:not([value]):before:not([value]){-moz-box-shadow:inset 0 0 1px 0 #686868;-webkit-box-shadow:inset 0 0 1px 0 #686868}progress:not([value]):after{box-sizing:border-box;content:\"\";position:absolute;top:1px;left:2px;width:100%;height:calc(100% - 2px);padding:1px 2px;border-radius:2px;background:repeating-linear-gradient(90deg,transparent 0,transparent 8px,#fff 0,#fff 10px,transparent 0,transparent 18px,#fff 0,#fff 20px,transparent 0,transparent 28px,#fff 0,#fff),linear-gradient(180deg,#acedad 0,#7be47d 14%,#4cda50 28%,#2ed330 42%,#42d845 57%,#76e275 71%,#8fe791 85%,#fff)}progress:not([value]):after,progress:not([value]):after:not([value]){animation:sliding 2s linear 0s infinite}progress:not([value]):after:not([value]){background:repeating-linear-gradient(90deg,transparent 0,transparent 8px,#fff 0,#fff 10px,transparent 0,transparent 18px,#fff 0,#fff 20px,transparent 0,transparent 28px,#fff 0,#fff),linear-gradient(180deg,#acedad 0,#7be47d 14%,#4cda50 28%,#2ed330 42%,#42d845 57%,#76e275 71%,#8fe791 85%,#fff)}progress:not([value])::-moz-progress-bar{width:100%;background:repeating-linear-gradient(90deg,transparent 0,transparent 8px,#fff 0,#fff 10px,transparent 0,transparent 18px,#fff 0,#fff 20px,transparent 0,transparent 28px,#fff 0,#fff),linear-gradient(180deg,#acedad 0,#7be47d 14%,#4cda50 28%,#2ed330 42%,#42d845 57%,#76e275 71%,#8fe791 85%,#fff);animation:sliding 2s linear 0s infinite}progress:not([value])::-moz-progress-bar:not([value]){animation:sliding 2s linear 0s infinite;background:repeating-linear-gradient(90deg,transparent 0,transparent 8px,#fff 0,#fff 10px,transparent 0,transparent 18px,#fff 0,#fff 20px,transparent 0,transparent 28px,#fff 0,#fff),linear-gradient(180deg,#acedad 0,#7be47d 14%,#4cda50 28%,#2ed330 42%,#42d845 57%,#76e275 71%,#8fe791 85%,#fff)}\n/*# sourceMappingURL=XP.css.map */"
RAW_FILES['7'] = RawFiles({"credit": "<a href='http://khang-nd.github.io/7.css/' target='_blank'>7.css</a> by <code>khang-nd</code>"}, {}, {}, {})
RAW_FILES['7'].styles['7.css'] = "/*! 7.css v0.17.0 - https://khang-nd.github.io/7.css */:root{--font:9pt \"Segoe UI\",\"SegoeUI\",\"Noto Sans\",sans-serif;--surface:#f0f0f0;--button-highlight:#fff;--button-face:#f2f2f2;--button-face-hover:#eaf6fd;--button-face-active:#c4e5f6;--button-face-disabled:#f4f4f4;--button-shade-light:#ebebeb;--button-shade-light-default:#c3dcea;--button-shade-light-hovered:#bee6fd;--button-shade-light-active:#98d1ef;--button-shade-dark:#cfcfcf;--button-shadow:inset 0 0 0 1px #fffc;--button-shadow-active:inset 1px 1px 0 #0003,inset -1px 1px 0 #0001;--button-border:1px solid;--button-border-color:#8e8f8f;--button-border-color-default:#5586a3;--button-border-color-hovered:#3c7fb1;--button-border-color-active:#6d91ab;--button-border-color-disabled:#aeb2b5;--button-text-color-disabled:#838383;--button-gradient:linear-gradient(to bottom,var(--button-face) 45%,var(--button-shade-light) 45%,var(--button-shade-dark));--button-gradient-hovered:linear-gradient(to bottom,var(--button-face-hover) 45%,var(--button-shade-light-hovered) 45%,#a7d9f5);--button-gradient-active:linear-gradient(to bottom,#e5f4fc,var(--button-face-active) 30% 50%,var(--button-shade-light-active) 50%,#68b3db);--element-spacing:8px;--grouped-element-spacing:6px;--border-radius:3px;--border-width:1px}body{color:#222;font-family:Segoe UI,SegoeUI,Noto Sans,sans-serif;font-size:9pt}.surface{background:var(--surface)}h1{font-size:5rem}h2{font-size:2.5rem}h3{font-size:2rem}h4{font-size:1.5rem}u{border-bottom:.5px solid #222;text-decoration:none}code,code *{font-family:monospace}pre{background:#fff;border:1px solid var(--button-border-color);display:block;margin:0;padding:12px 8px}summary:focus-visible{outline:1px dotted #000}:root{--balloon-border-color:rgba(0,0,0,.4);--balloon-tail-top:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='20' height='20' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath d='M.5 1.5v18h18l-18-18Z' fill='%23fff' stroke='%23fff'/%3E%3Cpath d='M1 19.5H.5v-18l18 18H18' stroke='%23939393'/%3E%3C/svg%3E\");--balloon-tail-bottom:url(\"data:image/svg+xml;charset=utf-8,%3Csvg width='20' height='20' fill='none' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath d='M.5 1.5v18h18l-18-18Z' fill='%23ddd' stroke='%23ddd'/%3E%3Cpath d='M1 19.5H.5v-18l18 18H18' stroke='%23939393'/%3E%3C/svg%3E\");--balloon-tail-size:18px;--balloon-tail-offset:1em}[role=tooltip]{background:linear-gradient(to bottom,#fff,var(--button-shade-light));border:1px solid var(--balloon-border-color);border-radius:var(--border-radius);box-shadow:5px 5px 3px -3px var(--balloon-border-color);padding:1em 1em 1em 2.5em;position:relative}[role=tooltip]:before{background:var(--balloon-tail-top);content:\"\";height:var(--balloon-tail-size);left:var(--balloon-tail-offset);position:absolute;top:calc(var(--balloon-tail-size)*-1);width:var(--balloon-tail-size)}[role=tooltip][id]{position:absolute}[role=tooltip].is-top:before{background:var(--balloon-tail-bottom);bottom:calc(var(--balloon-tail-size)*-1);top:unset;transform:scale(-1)}[role=tooltip].is-top.is-right:before{transform:scaleY(-1)}[role=tooltip].is-left:before{left:unset;right:var(--balloon-tail-offset)}[role=tooltip].is-left.is-bottom:before{transform:scaleX(-1)}[role=button],button{background:var(--button-gradient);border:var(--button-border);border-color:var(--button-border-color);border-radius:var(--border-radius);box-shadow:var(--button-shadow);box-sizing:border-box;color:#222;font:var(--font);min-height:23px;min-width:75px;padding:0 12px;position:relative;text-align:center;z-index:0}[role=button]:before,button:before{background:var(--button-gradient-hovered);border-radius:var(--border-radius);box-shadow:var(--button-shadow)}[role=button]:after,[role=button]:before,button:after,button:before{content:\"\";height:100%;left:0;margin:0;opacity:0;padding:0;position:absolute;top:0;transition:opacity .3s;width:100%;z-index:-1}[role=button]:after,button:after{background:var(--button-gradient-active);border-radius:2px;box-shadow:var(--button-shadow-active)}[role=button]:disabled,button:disabled{background:var(--button-face-disabled);border-color:var(--button-border-color-disabled);color:var(--button-text-color-disabled)}[role=button]:not(:disabled):hover,button:not(:disabled):hover{border-color:var(--button-border-color-hovered);transition:border-color .3s}[role=button]:not(:disabled):hover:before,button:not(:disabled):hover:before{opacity:1;transition:opacity .3s}[role=button]:not(:disabled):not(:hover),button:not(:disabled):not(:hover){border-color:var(--button-border-color);transition:border-color 1s linear}[role=button]:not(:disabled):not(:hover):before,button:not(:disabled):not(:hover):before{opacity:0;transition:opacity 1s linear}[role=button]:not(:disabled).active,[role=button]:not(:disabled):active,button:not(:disabled).active,button:not(:disabled):active{border-color:var(--button-border-color-active);transition:border-color .3s}[role=button]:not(:disabled).active:after,[role=button]:not(:disabled):active:after,button:not(:disabled).active:after,button:not(:disabled):active:after{opacity:1;transition:opacity .3s}[role=button].focused,[role=button]:focus-visible,button.focused,button:focus-visible{box-shadow:inset 0 0 0 2px var(--button-shade-light-active);outline:1px dotted #000;outline-offset:-4px}[role=button].default,[role=button].focused,[role=button]:focus,button.default,button.focused,button:focus{-webkit-animation:pulse-anim 1s ease infinite alternate;animation:pulse-anim 1s ease infinite alternate;background-image:var(--button-gradient-hovered);border-color:var(--button-border-color-default)}@-webkit-keyframes pulse-anim{0%{box-shadow:inset 0 0 3px 1px #34deffdd}to{box-shadow:inset 0 0 1px 1px #34deffdd}}@keyframes pulse-anim{0%{box-shadow:inset 0 0 3px 1px #34deffdd}to{box-shadow:inset 0 0 1px 1px #34deffdd}}:root{--checkbox-width:14px;--checkbox-label-spacing:6px;--checkbox-total-width-precalc:0;--checkbox-total-width:calc(var(--checkbox-total-width-precalc));--checkmark-top:0;--checkmark-left:2px}input[type=checkbox]{appearance:none;-webkit-appearance:none;-moz-appearance:none;background:none;border:none;font:var(--font);margin:0;opacity:0}input[type=checkbox]+label{align-items:center;display:inline-flex;font:var(--font);margin-left:var(--checkbox-total-width);position:relative}input[type=checkbox]+label:before{background:#f6f6f6;border:1px solid;border-color:var(--button-border-color);box-shadow:inset 0 0 0 1px #f4f4f4,inset 1px 1px 0 1px #aeaeae,inset -1px -1px 0 1px #ddd,inset 3px 3px 6px #ccc;box-sizing:border-box;content:\"\";display:inline-block;height:var(--checkbox-width);margin-right:var(--checkbox-label-spacing);transition:.4s;width:var(--checkbox-width)}input[type=checkbox]+label:hover:before{background:#e9f7fe;border-color:var(--button-border-color-hovered);box-shadow:inset 0 0 0 1px #def9fa,inset 1px 1px 0 1px #79c6f9,inset -1px -1px 0 1px #c6e9fc,inset 3px 3px 6px #b1dffd}input[type=checkbox]:focus-visible+label{outline:1px dotted #000}input[type=checkbox]:checked+label:after{color:#4a5f97;content:\"\\2714\";display:block;font-weight:700;left:var(--checkmark-left);position:absolute;top:var(--checkmark-top)}input[type=checkbox]:disabled+label{opacity:.6}:root{--chevron-size:5px}details{margin-top:0}details>summary{cursor:pointer;display:inline;margin-bottom:0;position:relative}details>summary:before{border:var(--chevron-size) solid transparent;border-left-color:#000;border-radius:3px;content:\"\";position:absolute;right:100%;top:calc(50% - var(--chevron-size))}details[open]>summary:before{top:calc(50% - var(--chevron-size)/2);transform:rotate(45deg)}:root{--combobox-chevron-size:4px}.combobox{display:inline-block;position:relative}.combobox input[type=text]{padding-right:20px;width:100%}.combobox button{background:url(\"data:image/svg+xml;base64,PHN2ZyB3aWR0aD0iMTYiIGhlaWdodD0iMTciIGZpbGw9Im5vbmUiIHhtbG5zPSJodHRwOi8vd3d3LnczLm9yZy8yMDAwL3N2ZyI+PGRlZnM+PGxpbmVhckdyYWRpZW50IGlkPSJhIiB4MT0iMCUiIHkxPSIwJSIgeDI9IjEwMCUiIHkyPSIwJSI+PHN0b3Agb2Zmc2V0PSIwJSIgc3R5bGU9InN0b3AtY29sb3I6IzMzMztzdG9wLW9wYWNpdHk6MSIvPjxzdG9wIG9mZnNldD0iMTAwJSIgc3R5bGU9InN0b3AtY29sb3I6I2FhYTtzdG9wLW9wYWNpdHk6MSIvPjwvbGluZWFyR3JhZGllbnQ+PC9kZWZzPjxwYXRoIGQ9Ik0xMSA2SDR2MWgxdjFoMXYxaDF2MWgxVjloMVY4aDFWN2gxVjZaIiBmaWxsPSJ1cmwoI2EpIi8+PC9zdmc+\") 50% no-repeat,var(--button-gradient);min-width:16px;padding:0;position:absolute;right:0}.combobox button,.combobox button:before{border-bottom-left-radius:0;border-top-left-radius:0}.combobox button:before{background:url(\"data:image/svg+xml;base64,PHN2ZyB3aWR0aD0iMTYiIGhlaWdodD0iMTciIGZpbGw9Im5vbmUiIHhtbG5zPSJodHRwOi8vd3d3LnczLm9yZy8yMDAwL3N2ZyI+PGRlZnM+PGxpbmVhckdyYWRpZW50IGlkPSJhIiB4MT0iMCUiIHkxPSIwJSIgeDI9IjEwMCUiIHkyPSIwJSI+PHN0b3Agb2Zmc2V0PSIwJSIgc3R5bGU9InN0b3AtY29sb3I6IzMzMztzdG9wLW9wYWNpdHk6MSIvPjxzdG9wIG9mZnNldD0iMTAwJSIgc3R5bGU9InN0b3AtY29sb3I6I2FhYTtzdG9wLW9wYWNpdHk6MSIvPjwvbGluZWFyR3JhZGllbnQ+PC9kZWZzPjxwYXRoIGQ9Ik0xMSA2SDR2MWgxdjFoMXYxaDF2MWgxVjloMVY4aDFWN2gxVjZaIiBmaWxsPSJ1cmwoI2EpIi8+PC9zdmc+\") 50% no-repeat,var(--button-gradient-hovered)}.combobox button:after{background:url(\"data:image/svg+xml;base64,PHN2ZyB3aWR0aD0iMTYiIGhlaWdodD0iMTciIGZpbGw9Im5vbmUiIHhtbG5zPSJodHRwOi8vd3d3LnczLm9yZy8yMDAwL3N2ZyI+PGRlZnM+PGxpbmVhckdyYWRpZW50IGlkPSJhIiB4MT0iMCUiIHkxPSIwJSIgeDI9IjEwMCUiIHkyPSIwJSI+PHN0b3Agb2Zmc2V0PSIwJSIgc3R5bGU9InN0b3AtY29sb3I6IzMzMztzdG9wLW9wYWNpdHk6MSIvPjxzdG9wIG9mZnNldD0iMTAwJSIgc3R5bGU9InN0b3AtY29sb3I6I2FhYTtzdG9wLW9wYWNpdHk6MSIvPjwvbGluZWFyR3JhZGllbnQ+PC9kZWZzPjxwYXRoIGQ9Ik0xMSA2SDR2MWgxdjFoMXYxaDF2MWgxVjloMVY4aDFWN2gxVjZaIiBmaWxsPSJ1cmwoI2EpIi8+PC9zdmc+\") 50% no-repeat,var(--button-gradient-active);border-bottom-left-radius:0;border-top-left-radius:0}.combobox button:focus{box-shadow:none;outline:none}select:not([multiple]){-webkit-appearance:none;-moz-appearance:none;appearance:none;background:url(\"data:image/svg+xml;base64,PHN2ZyB3aWR0aD0iMTYiIGhlaWdodD0iMTciIGZpbGw9Im5vbmUiIHhtbG5zPSJodHRwOi8vd3d3LnczLm9yZy8yMDAwL3N2ZyI+PGRlZnM+PGxpbmVhckdyYWRpZW50IGlkPSJhIiB4MT0iMCUiIHkxPSIwJSIgeDI9IjEwMCUiIHkyPSIwJSI+PHN0b3Agb2Zmc2V0PSIwJSIgc3R5bGU9InN0b3AtY29sb3I6IzMzMztzdG9wLW9wYWNpdHk6MSIvPjxzdG9wIG9mZnNldD0iMTAwJSIgc3R5bGU9InN0b3AtY29sb3I6I2FhYTtzdG9wLW9wYWNpdHk6MSIvPjwvbGluZWFyR3JhZGllbnQ+PC9kZWZzPjxwYXRoIGQ9Ik0xMSA2SDR2MWgxdjFoMXYxaDF2MWgxVjloMVY4aDFWN2gxVjZaIiBmaWxsPSJ1cmwoI2EpIi8+PC9zdmc+\"),var(--button-gradient);background-position:100%;background-repeat:no-repeat;border:var(--button-border);border-color:var(--button-border-color);border-radius:var(--border-radius);box-shadow:var(--button-shadow);box-sizing:border-box;color:#222;font:var(--font);padding:2px 30px 2px 3px;position:relative}select:not([multiple]):hover{background-image:url(\"data:image/svg+xml;base64,PHN2ZyB3aWR0aD0iMTYiIGhlaWdodD0iMTciIGZpbGw9Im5vbmUiIHhtbG5zPSJodHRwOi8vd3d3LnczLm9yZy8yMDAwL3N2ZyI+PGRlZnM+PGxpbmVhckdyYWRpZW50IGlkPSJhIiB4MT0iMCUiIHkxPSIwJSIgeDI9IjEwMCUiIHkyPSIwJSI+PHN0b3Agb2Zmc2V0PSIwJSIgc3R5bGU9InN0b3AtY29sb3I6IzMzMztzdG9wLW9wYWNpdHk6MSIvPjxzdG9wIG9mZnNldD0iMTAwJSIgc3R5bGU9InN0b3AtY29sb3I6I2FhYTtzdG9wLW9wYWNpdHk6MSIvPjwvbGluZWFyR3JhZGllbnQ+PC9kZWZzPjxwYXRoIGQ9Ik0xMSA2SDR2MWgxdjFoMXYxaDF2MWgxVjloMVY4aDFWN2gxVjZaIiBmaWxsPSJ1cmwoI2EpIi8+PC9zdmc+\"),var(--button-gradient-hovered);border-color:var(--button-border-color-hovered)}select:not([multiple]):active{background-image:url(\"data:image/svg+xml;base64,PHN2ZyB3aWR0aD0iMTYiIGhlaWdodD0iMTciIGZpbGw9Im5vbmUiIHhtbG5zPSJodHRwOi8vd3d3LnczLm9yZy8yMDAwL3N2ZyI+PGRlZnM+PGxpbmVhckdyYWRpZW50IGlkPSJhIiB4MT0iMCUiIHkxPSIwJSIgeDI9IjEwMCUiIHkyPSIwJSI+PHN0b3Agb2Zmc2V0PSIwJSIgc3R5bGU9InN0b3AtY29sb3I6IzMzMztzdG9wLW9wYWNpdHk6MSIvPjxzdG9wIG9mZnNldD0iMTAwJSIgc3R5bGU9InN0b3AtY29sb3I6I2FhYTtzdG9wLW9wYWNpdHk6MSIvPjwvbGluZWFyR3JhZGllbnQ+PC9kZWZzPjxwYXRoIGQ9Ik0xMSA2SDR2MWgxdjFoMXYxaDF2MWgxVjloMVY4aDFWN2gxVjZaIiBmaWxsPSJ1cmwoI2EpIi8+PC9zdmc+\"),var(--button-gradient-active);border-color:var(--button-border-color-active);box-shadow:var(--button-shadow-active);outline:none}select:not([multiple]):focus{box-shadow:inset 0 0 0 2px var(--button-shade-light-active);outline:1px dotted #000;outline-offset:-4px}fieldset{border:1px solid #cdd7db;border-radius:var(--border-radius);box-shadow:inset 0 0 0 1px #fff;margin:0;padding:calc(var(--border-width)*2 + var(--element-spacing));-webkit-padding-before:var(--element-spacing);padding-block-start:var(--element-spacing)}fieldset legend{font:var(--font)}.field-row{align-items:center;display:flex}.field-row>*+*{margin-left:var(--grouped-element-spacing)}.field-row-stacked{display:flex;flex-direction:column}.field-row-stacked *+*,[class^=field-row]+[class^=field-row]{margin-top:var(--grouped-element-spacing)}[role=listbox],select[multiple]{background:#fff;border:1px solid #c0c1cd;display:block;font:var(--font);overflow-y:scroll}[role=listbox].has-shadow,select[multiple].has-shadow{box-shadow:4px 4px 3px -2px #999}[role=listbox].has-hover li:hover,select[multiple].has-hover li:hover{background-color:#2a90ff;color:#fff}[role=listbox]:focus,select[multiple]:focus{outline:none}[role=listbox] [role=option],[role=listbox] option,select[multiple] [role=option],select[multiple] option{padding:2px}[role=listbox] [role=option]:focus,[role=listbox] [role=option][aria-selected],[role=listbox] option:focus,[role=listbox] option[aria-selected],select[multiple] [role=option]:focus,select[multiple] [role=option][aria-selected],select[multiple] option:focus,select[multiple] option[aria-selected]{background-color:#2a90ff;color:#fff}:root{--menu-padding:2px;--item-offset-left:30px;--item-hover-background:linear-gradient(180deg,#fff9,#e6ecf5cc 90%,#fffc)}ul[role]{cursor:default;list-style:none;margin:0;padding:0}ul[role=menubar]{background:linear-gradient(#fff 20%,#f1f4fa 25%,#f1f4fa 43%,#d4dbee 48%,#e6eaf6);display:flex}ul[role=menubar]>[role=menuitem]{padding:6px 10px;position:relative}ul[role=menubar]>[role=menuitem]:focus,ul[role=menubar]>[role=menuitem]:focus-within,ul[role=menubar]>[role=menuitem]:hover{background:#39f;color:#fff;outline:none}ul[role=menubar]>[role=menuitem]:focus-within~[role=menuitem]:focus,ul[role=menubar]>[role=menuitem]:focus-within~[role=menuitem]:focus-within,ul[role=menubar]>[role=menuitem]:focus~[role=menuitem]:focus,ul[role=menubar]>[role=menuitem]:focus~[role=menuitem]:focus-within,ul[role=menubar]>[role=menuitem]:hover~[role=menuitem]:focus,ul[role=menubar]>[role=menuitem]:hover~[role=menuitem]:focus-within{background:transparent;color:inherit}ul[role=menubar]>[role=menuitem]:focus-within:has(~[role=menuitem]:hover),ul[role=menubar]>[role=menuitem]:focus:has(~[role=menuitem]:hover),ul[role=menubar]>[role=menuitem]:hover:has(~[role=menuitem]:hover){background:transparent;color:inherit}ul[role=menu]{background:var(--surface);border:1px solid rgba(0,0,0,.4);box-shadow:4px 4px 3px -2px rgba(0,0,0,.5);color:initial;min-width:150px;padding:var(--menu-padding);position:relative}ul[role=menu]:before{box-shadow:inset 1px 0 rgba(0,0,0,.15),inset -1px 0 #fff;content:\"\";height:calc(100% - 4px);left:var(--item-offset-left);pointer-events:none;position:absolute;width:2px}[role=menuitem] ul[role=menu]{display:none;left:0;position:absolute;top:100%;z-index:99}ul[role=menu] [role=menuitem]>[role=menu]{left:100%;top:-4px}ul[role=menu]>[role=menuitem]>a,ul[role=menu]>[role=menuitem]>button,ul[role=menu]>[role=menuitem]>label,ul[role=menu]>[role=menuitem][aria-haspopup=true]{all:unset;border:1px solid transparent;border-radius:3px;box-sizing:border-box;display:flex;justify-content:space-between;padding:4px 10px 4px 32px;position:relative;white-space:nowrap;width:100%}ul[role=menu]>[role=menuitem]>a:focus-visible,ul[role=menu]>[role=menuitem]>a:hover,ul[role=menu]>[role=menuitem]>button:focus-visible,ul[role=menu]>[role=menuitem]>button:hover,ul[role=menu]>[role=menuitem]>label:focus-visible,ul[role=menu]>[role=menuitem]>label:hover,ul[role=menu]>[role=menuitem][aria-haspopup=true]:focus-visible,ul[role=menu]>[role=menuitem][aria-haspopup=true]:hover{background:var(--item-hover-background);border-color:#b8d6fb}ul[role=menu]>[role=menuitem]>button:after,ul[role=menu]>[role=menuitem]>button:hover:before{content:none}ul[role=menu]>[role=menuitem][aria-haspopup=true]:after{border:4px solid transparent;border-left-color:currentcolor;content:\"\";position:absolute;right:2px;top:50%;transform:translateY(-50%)}ul [role=menuitem]{position:relative}ul [role=menuitem]>input[type]{display:none}ul [role=menuitem]>input[type]+label{display:block;position:relative}ul [role=menuitem]>input[type]+label:before{all:unset;background:var(--item-hover-background);border-radius:inherit;box-shadow:0 0 0 1px #b3d3f9;box-sizing:border-box;height:22px;left:0;position:absolute;top:0;width:22px}ul [role=menuitem]>input[type]:checked+label:before{content:\"\"}ul [role=menuitem]>input[type][type=radio]:checked+label:after{background:radial-gradient(circle at 75% 25%,#d5d4ea,#333583);border:1px solid #1a1490;box-shadow:none;left:8px;top:50%;transform:translateY(-50%)}ul [role=menuitem]>input[type][type=checkbox]:checked+label:after{color:#0c12a1;font-size:10pt;left:6px;top:50%;transform:translateY(-52%)}ul [role=menuitem]:focus-within>[role=menu],ul [role=menuitem]:focus>[role=menu]{display:block}ul [role=menuitem]:focus-within:has(~[role=menuitem]:hover)>[role=menu],ul [role=menuitem]:focus:has(~[role=menuitem]:hover)>[role=menu]{display:none}ul [role=menuitem][aria-disabled]{opacity:.5;pointer-events:none}ul [role=menuitem].has-divider:after{box-shadow:inset 0 1px #00000026,inset 0 -1px #fff;content:\"\";display:block;height:2px;margin:3px 0 2px;margin-left:var(--item-offset-left);pointer-events:none}ul [role=menuitem] img{left:2px;pointer-events:none;position:absolute;top:50%;transform:translateY(-50%);z-index:1}ul [role=menuitem] span{margin-left:32px}ul.can-hover [role=menuitem]:hover>[role=menu]{display:block}ul.can-hover [role=menuitem]:hover~[role=menuitem]:focus,ul.can-hover [role=menuitem]:hover~[role=menuitem]:focus-within{background:transparent;color:inherit}ul.can-hover [role=menuitem]:hover~[role=menuitem]:focus-within>[role=menu],ul.can-hover [role=menuitem]:hover~[role=menuitem]:focus>[role=menu]{display:none}:root{--progress-color-normal:#0bd82c;--progress-color-paused:#e6df1b;--progress-color-error:#ef0000}[role=progressbar]{background:radial-gradient(circle at 0 50%,#0000001f 10px,transparent 30px),radial-gradient(circle at 100% 50%,#0000001f 10px,transparent 30px),linear-gradient(180deg,#f3f3f3af,#fcfcfcaf 3px,#dbdbdbaf 6px,#cacacaaf 0,#d5d5d5af),#ddd;border:1px solid var(--button-border-color);border-radius:var(--border-radius);box-shadow:inset 0 0 0 1px #f3f3f388,0 0 0 1px #eaeaea88;height:15px;margin:2px 0;overflow:hidden}[role=progressbar]>div{background-color:var(--progress-color-normal);background-image:linear-gradient(180deg,#f3f3f3af,#fcfcfcaf 3px,#dbdbdbaf 6px,transparent 0),radial-gradient(circle at 0 50%,#0000002f 10px,transparent 30px),radial-gradient(circle at 100% 50%,#0000002f 10px,transparent 30px),linear-gradient(180deg,transparent 65%,#ffffff55),linear-gradient(180deg,transparent 6px,#cacaca33 0,#d5d5d533);box-shadow:inset 0 0 0 1px #ffffff1f;height:100%;overflow:hidden}[role=progressbar].paused>div{background-color:var(--progress-color-paused)}[role=progressbar].error>div{background-color:var(--progress-color-error)}[role=progressbar].animate>div:before,[role=progressbar].marquee:before{-webkit-animation:progressbar 3s linear infinite;animation:progressbar 3s linear infinite;background:linear-gradient(90deg,transparent,#ffffff80,transparent 40%);content:\"\";display:block;height:100%}[role=progressbar].marquee:before{background:linear-gradient(to right,transparent,var(--progress-color-normal),transparent 35%);opacity:.5}@-webkit-keyframes progressbar{0%{transform:translateX(-40%)}60%{transform:translateX(100%)}to{transform:translateX(100%)}}@keyframes progressbar{0%{transform:translateX(-40%)}60%{transform:translateX(100%)}to{transform:translateX(100%)}}:root{--radio-width:14px;--radio-label-spacing:6px;--radio-total-width-precalc:var(--radio-width) + var(--radio-label-spacing);--radio-total-width:calc(var(--radio-total-width-precalc));--radio-dot-width:8px;--radio-dot-offset:calc(var(--radio-width)/2);--radio-dot-top:calc(var(--radio-dot-offset) - var(--radio-dot-width)/2);--radio-dot-left:calc(var(--radio-total-width)*-1 + var(--radio-width)/2 - var(\n        --radio-dot-width\n      )/2)}input[type=radio]{appearance:none;-webkit-appearance:none;-moz-appearance:none;background:0;border:none;margin:0;opacity:0;position:fixed}input[type=radio]+label{align-items:center;display:inline-flex;font:var(--font);margin-left:var(--radio-total-width);position:relative}input[type=radio]+label:before{background:#f6f6f6;border:1px solid;border-color:var(--button-border-color);border-radius:50%;box-shadow:inset 0 0 0 1.5px #f4f4f4,inset 1px 1px 0 1.5px #aeaeae,inset -1px 0 0 1.5px #ddd,inset 3px 3px 6px #ccc;box-sizing:border-box;content:\"\";display:inline-block;height:var(--radio-width);left:calc(var(--radio-total-width)*-1);margin-right:var(--radio-label-spacing);position:absolute;top:0;transition:.4s;width:var(--radio-width)}input[type=radio]+label:hover:before{border-color:var(--button-border-color-hovered);box-shadow:inset 0 0 0 1.5px #def9fa,inset 1px 1px 0 1.5px #79c6f9,inset -1px -1px 0 1.5px #c6e9fc,inset 3px 3px 6px #b1dffd}input[type=radio]:checked+label:after{background:#7cd3eb;border:1.5px solid #27506d;border-radius:50%;box-shadow:inset -1px -1px 0 .5px #16638f,inset -1px -1px 0 1px #1985c0;box-sizing:border-box;content:\"\";display:block;height:var(--radio-dot-width);left:var(--radio-dot-left);position:absolute;top:var(--radio-dot-top);width:var(--radio-dot-width)}input[type=radio]:focus-visible+label{outline:1px dotted #000}input[type=radio]:disabled+label{opacity:.6}:root{--scrollbar-y:linear-gradient(to right,#e5e5e5,var(--surface) 20%);--scrollbar-x:linear-gradient(to bottom,#e5e5e5,var(--surface) 20%)}.has-scrollbar::-webkit-scrollbar{width:16px}.has-scrollbar::-webkit-scrollbar:horizontal{height:17px}.has-scrollbar::-webkit-scrollbar-corner{background:var(--button-face)}.has-scrollbar::-webkit-scrollbar-track:vertical{background:var(--scrollbar-y)}.has-scrollbar::-webkit-scrollbar-track:horizontal{background:var(--scrollbar-x)}.has-scrollbar::-webkit-scrollbar-thumb{background-color:var(--button-face);border:var(--button-border);border-color:var(--button-border-color);border-radius:var(--border-radius);box-shadow:var(--button-shadow)}.has-scrollbar::-webkit-scrollbar-thumb:vertical{background:url(\"data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAkAAAAKCAIAAADpZ+PpAAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAAAJcEhZcwAAEnQAABJ0Ad5mH3gAAADrSURBVChTTc5LboJQGAXguyoCu4ERCzAGlRk7UOwGWIDh0s4M4kxb06RSq/jAB6AxJkJ4lTDrue3AnvyzP+fLId+/yfM8juP7PQmCCOf7B3e+ZD+O40RRVFW12VQUpd3r9U3T2m4OpKoqWZYNwzBZLEqfh0N7NnvfrPcEWlEUWZb9mWF4Ph6D0ylcLbfM5HkeJrhGA2hb15/QXnv+w7RYXsDatjOdvnmrHSnLEizMNE2v11sUXQBCnn98kbquBUGQJAlmq9WB2e3qg4HJdqKkaRql1HGc0WgMcDJ5dd0F24kediZJ8t/ELT69H+8py0CYSIO5AAAAAElFTkSuQmCC\") no-repeat 50%,linear-gradient(to right,var(--button-face) 45%,var(--button-shade-light) 45%,var(--button-shade-dark))}.has-scrollbar::-webkit-scrollbar-thumb:horizontal{background:url(\"data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAoAAAAJCAYAAAALpr0TAAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAAAJcEhZcwAAEnQAABJ0Ad5mH3gAAADcSURBVChTNZBLqoUwEEQrURQUxZGCvy24ACfiityJi7tv8GauQoPxk5tquA2RQ9vVVYk6z9NZaxFFEe77htYazjk8z4MwDIVZ+rourOuKaZrwvi+WZcE8z1BKCbPPCjk4DAO2bRP1OI7wLiL6Mbd7J408z1GWpQwWRYGqqiQG+03TgMu0MacfUN4qANmn8UOv9MjW3sKaSm7iIdOSlziOQ3LScd93aPonSYK6riVLlmVo21aYfVqzND9pmqLrOlGT+76XbcxLZkb19/l3fEP+oF0cx8KMEASBsDEGX2/CgZCHkg+8AAAAAElFTkSuQmCC\") no-repeat 50%,linear-gradient(to bottom,var(--button-face) 45%,var(--button-shade-light) 45%,var(--button-shade-dark))}.has-scrollbar::-webkit-scrollbar-thumb:hover:vertical{background:url(\"data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAkAAAAKCAIAAADpZ+PpAAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAAAJcEhZcwAAEnQAABJ0Ad5mH3gAAADrSURBVChTTc5LboJQGAXguyoCu4ERCzAGlRk7UOwGWIDh0s4M4kxb06RSq/jAB6AxJkJ4lTDrue3AnvyzP+fLId+/yfM8juP7PQmCCOf7B3e+ZD+O40RRVFW12VQUpd3r9U3T2m4OpKoqWZYNwzBZLEqfh0N7NnvfrPcEWlEUWZb9mWF4Ph6D0ylcLbfM5HkeJrhGA2hb15/QXnv+w7RYXsDatjOdvnmrHSnLEizMNE2v11sUXQBCnn98kbquBUGQJAlmq9WB2e3qg4HJdqKkaRql1HGc0WgMcDJ5dd0F24kediZJ8t/ELT69H+8py0CYSIO5AAAAAElFTkSuQmCC\") no-repeat 50%,linear-gradient(to right,var(--button-face-hover) 45%,var(--button-shade-light-hovered) 45%);border-color:var(--button-border-color-hovered)}.has-scrollbar::-webkit-scrollbar-thumb:hover:horizontal{background:url(\"data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAoAAAAJCAYAAAALpr0TAAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAAAJcEhZcwAAEnQAABJ0Ad5mH3gAAADcSURBVChTNZBLqoUwEEQrURQUxZGCvy24ACfiityJi7tv8GauQoPxk5tquA2RQ9vVVYk6z9NZaxFFEe77htYazjk8z4MwDIVZ+rourOuKaZrwvi+WZcE8z1BKCbPPCjk4DAO2bRP1OI7wLiL6Mbd7J408z1GWpQwWRYGqqiQG+03TgMu0MacfUN4qANmn8UOv9MjW3sKaSm7iIdOSlziOQ3LScd93aPonSYK6riVLlmVo21aYfVqzND9pmqLrOlGT+76XbcxLZkb19/l3fEP+oF0cx8KMEASBsDEGX2/CgZCHkg+8AAAAAElFTkSuQmCC\") no-repeat 50%,linear-gradient(to bottom,var(--button-face-hover) 45%,var(--button-shade-light-hovered) 45%);border-color:var(--button-border-color-hovered)}.has-scrollbar::-webkit-scrollbar-thumb:active:vertical{background:url(\"data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAkAAAAKCAIAAADpZ+PpAAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAAAJcEhZcwAAEnQAABJ0Ad5mH3gAAADrSURBVChTTc5LboJQGAXguyoCu4ERCzAGlRk7UOwGWIDh0s4M4kxb06RSq/jAB6AxJkJ4lTDrue3AnvyzP+fLId+/yfM8juP7PQmCCOf7B3e+ZD+O40RRVFW12VQUpd3r9U3T2m4OpKoqWZYNwzBZLEqfh0N7NnvfrPcEWlEUWZb9mWF4Ph6D0ylcLbfM5HkeJrhGA2hb15/QXnv+w7RYXsDatjOdvnmrHSnLEizMNE2v11sUXQBCnn98kbquBUGQJAlmq9WB2e3qg4HJdqKkaRql1HGc0WgMcDJ5dd0F24kediZJ8t/ELT69H+8py0CYSIO5AAAAAElFTkSuQmCC\") no-repeat 50%,linear-gradient(to right,var(--button-face-active) 45%,var(--button-shade-light-active) 45%);border-color:var(--button-border-color-active)}.has-scrollbar::-webkit-scrollbar-thumb:active:horizontal{background:url(\"data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAoAAAAJCAYAAAALpr0TAAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAAAJcEhZcwAAEnQAABJ0Ad5mH3gAAADcSURBVChTNZBLqoUwEEQrURQUxZGCvy24ACfiityJi7tv8GauQoPxk5tquA2RQ9vVVYk6z9NZaxFFEe77htYazjk8z4MwDIVZ+rourOuKaZrwvi+WZcE8z1BKCbPPCjk4DAO2bRP1OI7wLiL6Mbd7J408z1GWpQwWRYGqqiQG+03TgMu0MacfUN4qANmn8UOv9MjW3sKaSm7iIdOSlziOQ3LScd93aPonSYK6riVLlmVo21aYfVqzND9pmqLrOlGT+76XbcxLZkb19/l3fEP+oF0cx8KMEASBsDEGX2/CgZCHkg+8AAAAAElFTkSuQmCC\") no-repeat 50%,linear-gradient(to bottom,var(--button-face-active) 45%,var(--button-shade-light-active) 45%);border-color:var(--button-border-color-active)}.has-scrollbar::-webkit-scrollbar-button:horizontal:end:decrement,.has-scrollbar::-webkit-scrollbar-button:horizontal:start:increment,.has-scrollbar::-webkit-scrollbar-button:vertical:end:decrement,.has-scrollbar::-webkit-scrollbar-button:vertical:start:increment{display:none}.has-scrollbar::-webkit-scrollbar-button{border:var(--button-border) transparent}.has-scrollbar::-webkit-scrollbar-button:vertical{height:17px}.has-scrollbar::-webkit-scrollbar-button:vertical:start{background:url(\"data:image/svg+xml;base64,PHN2ZyB3aWR0aD0iMTYiIGhlaWdodD0iMTciIGZpbGw9Im5vbmUiIHhtbG5zPSJodHRwOi8vd3d3LnczLm9yZy8yMDAwL3N2ZyI+PGRlZnM+PGxpbmVhckdyYWRpZW50IGlkPSJhIiB4MT0iMCUiIHkxPSIwJSIgeDI9IjEwMCUiIHkyPSIwJSI+PHN0b3Agb2Zmc2V0PSIwJSIgc3R5bGU9InN0b3AtY29sb3I6IzMzMztzdG9wLW9wYWNpdHk6MSIvPjxzdG9wIG9mZnNldD0iMTAwJSIgc3R5bGU9InN0b3AtY29sb3I6I2FhYTtzdG9wLW9wYWNpdHk6MSIvPjwvbGluZWFyR3JhZGllbnQ+PC9kZWZzPjxwYXRoIGQ9Ik04IDZIN3YxSDZ2MUg1djFINHYxaDdWOWgtMVY4SDlWN0g4VjZaIiBmaWxsPSJ1cmwoI2EpIi8+PC9zdmc+\"),var(--scrollbar-y)}.has-scrollbar::-webkit-scrollbar-button:vertical:end{background:url(\"data:image/svg+xml;base64,PHN2ZyB3aWR0aD0iMTYiIGhlaWdodD0iMTciIGZpbGw9Im5vbmUiIHhtbG5zPSJodHRwOi8vd3d3LnczLm9yZy8yMDAwL3N2ZyI+PGRlZnM+PGxpbmVhckdyYWRpZW50IGlkPSJhIiB4MT0iMCUiIHkxPSIwJSIgeDI9IjEwMCUiIHkyPSIwJSI+PHN0b3Agb2Zmc2V0PSIwJSIgc3R5bGU9InN0b3AtY29sb3I6IzMzMztzdG9wLW9wYWNpdHk6MSIvPjxzdG9wIG9mZnNldD0iMTAwJSIgc3R5bGU9InN0b3AtY29sb3I6I2FhYTtzdG9wLW9wYWNpdHk6MSIvPjwvbGluZWFyR3JhZGllbnQ+PC9kZWZzPjxwYXRoIGQ9Ik0xMSA2SDR2MWgxdjFoMXYxaDF2MWgxVjloMVY4aDFWN2gxVjZaIiBmaWxsPSJ1cmwoI2EpIi8+PC9zdmc+\"),var(--scrollbar-y)}.has-scrollbar::-webkit-scrollbar-button:horizontal{width:16px}.has-scrollbar::-webkit-scrollbar-button:horizontal:start{background:url(\"data:image/svg+xml;base64,PHN2ZyB3aWR0aD0iMTYiIGhlaWdodD0iMTciIGZpbGw9Im5vbmUiIHhtbG5zPSJodHRwOi8vd3d3LnczLm9yZy8yMDAwL3N2ZyI+PGRlZnM+PGxpbmVhckdyYWRpZW50IGlkPSJhIiB4MT0iMCUiIHkxPSIxMDAlIiB4Mj0iMCUiIHkyPSIwJSI+PHN0b3Agb2Zmc2V0PSIwJSIgc3R5bGU9InN0b3AtY29sb3I6IzMzMztzdG9wLW9wYWNpdHk6MSIvPjxzdG9wIG9mZnNldD0iMTAwJSIgc3R5bGU9InN0b3AtY29sb3I6I2FhYTtzdG9wLW9wYWNpdHk6MSIvPjwvbGluZWFyR3JhZGllbnQ+PC9kZWZzPjxwYXRoIGQ9Ik05IDRIOHYxSDd2MUg2djFINXYxaDF2MWgxdjFoMXYxaDFWNFoiIGZpbGw9InVybCgjYSkiLz48L3N2Zz4=\"),var(--scrollbar-x)}.has-scrollbar::-webkit-scrollbar-button:horizontal:end{background:url(\"data:image/svg+xml;base64,PHN2ZyB3aWR0aD0iMTYiIGhlaWdodD0iMTciIGZpbGw9Im5vbmUiIHhtbG5zPSJodHRwOi8vd3d3LnczLm9yZy8yMDAwL3N2ZyI+PGRlZnM+PGxpbmVhckdyYWRpZW50IGlkPSJhIiB4MT0iMCUiIHkxPSIxMDAlIiB4Mj0iMCUiIHkyPSIwJSI+PHN0b3Agb2Zmc2V0PSIwJSIgc3R5bGU9InN0b3AtY29sb3I6IzMzMztzdG9wLW9wYWNpdHk6MSIvPjxzdG9wIG9mZnNldD0iMTAwJSIgc3R5bGU9InN0b3AtY29sb3I6I2FhYTtzdG9wLW9wYWNpdHk6MSIvPjwvbGluZWFyR3JhZGllbnQ+PC9kZWZzPjxwYXRoIGQ9Ik03IDRINnY3aDF2LTFoMVY5aDFWOGgxVjdIOVY2SDhWNUg3VjRaIiBmaWxsPSJ1cmwoI2EpIi8+PC9zdmc+\"),var(--scrollbar-x)}.has-scrollbar::-webkit-scrollbar-button:hover{background-color:var(--button-face);border-color:var(--button-border-color);border-radius:var(--border-radius);box-shadow:var(--button-shadow)}.has-scrollbar::-webkit-scrollbar-button:hover:vertical{height:17px}.has-scrollbar::-webkit-scrollbar-button:hover:vertical:start{background:url(\"data:image/svg+xml;base64,PHN2ZyB3aWR0aD0iMTYiIGhlaWdodD0iMTciIGZpbGw9Im5vbmUiIHhtbG5zPSJodHRwOi8vd3d3LnczLm9yZy8yMDAwL3N2ZyI+PGRlZnM+PGxpbmVhckdyYWRpZW50IGlkPSJhIiB4MT0iMCUiIHkxPSIwJSIgeDI9IjEwMCUiIHkyPSIwJSI+PHN0b3Agb2Zmc2V0PSIwJSIgc3R5bGU9InN0b3AtY29sb3I6IzMzMztzdG9wLW9wYWNpdHk6MSIvPjxzdG9wIG9mZnNldD0iMTAwJSIgc3R5bGU9InN0b3AtY29sb3I6I2FhYTtzdG9wLW9wYWNpdHk6MSIvPjwvbGluZWFyR3JhZGllbnQ+PC9kZWZzPjxwYXRoIGQ9Ik04IDZIN3YxSDZ2MUg1djFINHYxaDdWOWgtMVY4SDlWN0g4VjZaIiBmaWxsPSJ1cmwoI2EpIi8+PC9zdmc+\"),linear-gradient(to right,var(--button-face-hover) 45%,var(--button-shade-light-hovered) 45%);border-color:var(--button-border-color-hovered)}.has-scrollbar::-webkit-scrollbar-button:hover:vertical:end{background:url(\"data:image/svg+xml;base64,PHN2ZyB3aWR0aD0iMTYiIGhlaWdodD0iMTciIGZpbGw9Im5vbmUiIHhtbG5zPSJodHRwOi8vd3d3LnczLm9yZy8yMDAwL3N2ZyI+PGRlZnM+PGxpbmVhckdyYWRpZW50IGlkPSJhIiB4MT0iMCUiIHkxPSIwJSIgeDI9IjEwMCUiIHkyPSIwJSI+PHN0b3Agb2Zmc2V0PSIwJSIgc3R5bGU9InN0b3AtY29sb3I6IzMzMztzdG9wLW9wYWNpdHk6MSIvPjxzdG9wIG9mZnNldD0iMTAwJSIgc3R5bGU9InN0b3AtY29sb3I6I2FhYTtzdG9wLW9wYWNpdHk6MSIvPjwvbGluZWFyR3JhZGllbnQ+PC9kZWZzPjxwYXRoIGQ9Ik0xMSA2SDR2MWgxdjFoMXYxaDF2MWgxVjloMVY4aDFWN2gxVjZaIiBmaWxsPSJ1cmwoI2EpIi8+PC9zdmc+\"),linear-gradient(to right,var(--button-face-hover) 45%,var(--button-shade-light-hovered) 45%);border-color:var(--button-border-color-hovered)}.has-scrollbar::-webkit-scrollbar-button:hover:horizontal{width:16px}.has-scrollbar::-webkit-scrollbar-button:hover:horizontal:start{background:url(\"data:image/svg+xml;base64,PHN2ZyB3aWR0aD0iMTYiIGhlaWdodD0iMTciIGZpbGw9Im5vbmUiIHhtbG5zPSJodHRwOi8vd3d3LnczLm9yZy8yMDAwL3N2ZyI+PGRlZnM+PGxpbmVhckdyYWRpZW50IGlkPSJhIiB4MT0iMCUiIHkxPSIxMDAlIiB4Mj0iMCUiIHkyPSIwJSI+PHN0b3Agb2Zmc2V0PSIwJSIgc3R5bGU9InN0b3AtY29sb3I6IzMzMztzdG9wLW9wYWNpdHk6MSIvPjxzdG9wIG9mZnNldD0iMTAwJSIgc3R5bGU9InN0b3AtY29sb3I6I2FhYTtzdG9wLW9wYWNpdHk6MSIvPjwvbGluZWFyR3JhZGllbnQ+PC9kZWZzPjxwYXRoIGQ9Ik05IDRIOHYxSDd2MUg2djFINXYxaDF2MWgxdjFoMXYxaDFWNFoiIGZpbGw9InVybCgjYSkiLz48L3N2Zz4=\"),linear-gradient(to bottom,var(--button-face-hover) 45%,var(--button-shade-light-hovered) 45%);border-color:var(--button-border-color-hovered)}.has-scrollbar::-webkit-scrollbar-button:hover:horizontal:end{background:url(\"data:image/svg+xml;base64,PHN2ZyB3aWR0aD0iMTYiIGhlaWdodD0iMTciIGZpbGw9Im5vbmUiIHhtbG5zPSJodHRwOi8vd3d3LnczLm9yZy8yMDAwL3N2ZyI+PGRlZnM+PGxpbmVhckdyYWRpZW50IGlkPSJhIiB4MT0iMCUiIHkxPSIxMDAlIiB4Mj0iMCUiIHkyPSIwJSI+PHN0b3Agb2Zmc2V0PSIwJSIgc3R5bGU9InN0b3AtY29sb3I6IzMzMztzdG9wLW9wYWNpdHk6MSIvPjxzdG9wIG9mZnNldD0iMTAwJSIgc3R5bGU9InN0b3AtY29sb3I6I2FhYTtzdG9wLW9wYWNpdHk6MSIvPjwvbGluZWFyR3JhZGllbnQ+PC9kZWZzPjxwYXRoIGQ9Ik03IDRINnY3aDF2LTFoMVY5aDFWOGgxVjdIOVY2SDhWNUg3VjRaIiBmaWxsPSJ1cmwoI2EpIi8+PC9zdmc+\"),linear-gradient(to bottom,var(--button-face-hover) 45%,var(--button-shade-light-hovered) 45%);border-color:var(--button-border-color-hovered)}.has-scrollbar::-webkit-scrollbar-button:active{background-color:var(--button-face);border-color:var(--button-border-color);border-radius:var(--border-radius);box-shadow:var(--button-shadow)}.has-scrollbar::-webkit-scrollbar-button:active:vertical{height:17px}.has-scrollbar::-webkit-scrollbar-button:active:vertical:start{background:url(\"data:image/svg+xml;base64,PHN2ZyB3aWR0aD0iMTYiIGhlaWdodD0iMTciIGZpbGw9Im5vbmUiIHhtbG5zPSJodHRwOi8vd3d3LnczLm9yZy8yMDAwL3N2ZyI+PGRlZnM+PGxpbmVhckdyYWRpZW50IGlkPSJhIiB4MT0iMCUiIHkxPSIwJSIgeDI9IjEwMCUiIHkyPSIwJSI+PHN0b3Agb2Zmc2V0PSIwJSIgc3R5bGU9InN0b3AtY29sb3I6IzMzMztzdG9wLW9wYWNpdHk6MSIvPjxzdG9wIG9mZnNldD0iMTAwJSIgc3R5bGU9InN0b3AtY29sb3I6I2FhYTtzdG9wLW9wYWNpdHk6MSIvPjwvbGluZWFyR3JhZGllbnQ+PC9kZWZzPjxwYXRoIGQ9Ik04IDZIN3YxSDZ2MUg1djFINHYxaDdWOWgtMVY4SDlWN0g4VjZaIiBmaWxsPSJ1cmwoI2EpIi8+PC9zdmc+\"),linear-gradient(to right,var(--button-face-active) 45%,var(--button-shade-light-active) 45%);border-color:var(--button-border-color-active)}.has-scrollbar::-webkit-scrollbar-button:active:vertical:end{background:url(\"data:image/svg+xml;base64,PHN2ZyB3aWR0aD0iMTYiIGhlaWdodD0iMTciIGZpbGw9Im5vbmUiIHhtbG5zPSJodHRwOi8vd3d3LnczLm9yZy8yMDAwL3N2ZyI+PGRlZnM+PGxpbmVhckdyYWRpZW50IGlkPSJhIiB4MT0iMCUiIHkxPSIwJSIgeDI9IjEwMCUiIHkyPSIwJSI+PHN0b3Agb2Zmc2V0PSIwJSIgc3R5bGU9InN0b3AtY29sb3I6IzMzMztzdG9wLW9wYWNpdHk6MSIvPjxzdG9wIG9mZnNldD0iMTAwJSIgc3R5bGU9InN0b3AtY29sb3I6I2FhYTtzdG9wLW9wYWNpdHk6MSIvPjwvbGluZWFyR3JhZGllbnQ+PC9kZWZzPjxwYXRoIGQ9Ik0xMSA2SDR2MWgxdjFoMXYxaDF2MWgxVjloMVY4aDFWN2gxVjZaIiBmaWxsPSJ1cmwoI2EpIi8+PC9zdmc+\"),linear-gradient(to right,var(--button-face-active) 45%,var(--button-shade-light-active) 45%);border-color:var(--button-border-color-active)}.has-scrollbar::-webkit-scrollbar-button:active:horizontal{width:16px}.has-scrollbar::-webkit-scrollbar-button:active:horizontal:start{background:url(\"data:image/svg+xml;base64,PHN2ZyB3aWR0aD0iMTYiIGhlaWdodD0iMTciIGZpbGw9Im5vbmUiIHhtbG5zPSJodHRwOi8vd3d3LnczLm9yZy8yMDAwL3N2ZyI+PGRlZnM+PGxpbmVhckdyYWRpZW50IGlkPSJhIiB4MT0iMCUiIHkxPSIxMDAlIiB4Mj0iMCUiIHkyPSIwJSI+PHN0b3Agb2Zmc2V0PSIwJSIgc3R5bGU9InN0b3AtY29sb3I6IzMzMztzdG9wLW9wYWNpdHk6MSIvPjxzdG9wIG9mZnNldD0iMTAwJSIgc3R5bGU9InN0b3AtY29sb3I6I2FhYTtzdG9wLW9wYWNpdHk6MSIvPjwvbGluZWFyR3JhZGllbnQ+PC9kZWZzPjxwYXRoIGQ9Ik05IDRIOHYxSDd2MUg2djFINXYxaDF2MWgxdjFoMXYxaDFWNFoiIGZpbGw9InVybCgjYSkiLz48L3N2Zz4=\"),linear-gradient(to bottom,var(--button-face-active) 45%,var(--button-shade-light-active) 45%);border-color:var(--button-border-color-active)}.has-scrollbar::-webkit-scrollbar-button:active:horizontal:end{background:url(\"data:image/svg+xml;base64,PHN2ZyB3aWR0aD0iMTYiIGhlaWdodD0iMTciIGZpbGw9Im5vbmUiIHhtbG5zPSJodHRwOi8vd3d3LnczLm9yZy8yMDAwL3N2ZyI+PGRlZnM+PGxpbmVhckdyYWRpZW50IGlkPSJhIiB4MT0iMCUiIHkxPSIxMDAlIiB4Mj0iMCUiIHkyPSIwJSI+PHN0b3Agb2Zmc2V0PSIwJSIgc3R5bGU9InN0b3AtY29sb3I6IzMzMztzdG9wLW9wYWNpdHk6MSIvPjxzdG9wIG9mZnNldD0iMTAwJSIgc3R5bGU9InN0b3AtY29sb3I6I2FhYTtzdG9wLW9wYWNpdHk6MSIvPjwvbGluZWFyR3JhZGllbnQ+PC9kZWZzPjxwYXRoIGQ9Ik03IDRINnY3aDF2LTFoMVY5aDFWOGgxVjdIOVY2SDhWNUg3VjRaIiBmaWxsPSJ1cmwoI2EpIi8+PC9zdmc+\"),linear-gradient(to bottom,var(--button-face-active) 45%,var(--button-shade-light-active) 45%);border-color:var(--button-border-color-active)}:root{--search-icon:url(\"data:image/svg+xml;base64,PHN2ZyB4bWxucz0iaHR0cDovL3d3dy53My5vcmcvMjAwMC9zdmciIHdpZHRoPSIxNiIgaGVpZ2h0PSIxNiI+PHBhdGggc3Ryb2tlPSIjMjA3MGI5IiBkPSJNMTAuNSAxQzguMDIgMSA2IDMuMDIgNiA1LjVhNC40NSA0LjQ1IDAgMCAwIDEgMi43OTNMMi4wMjMgMTMuMjdsLjcwNC43MUw3LjcwNyA5Yy43Ny42MTcgMS43MzQgMSAyLjc5MyAxIDIuNDggMCA0LjUtMi4wMiA0LjUtNC41UzEyLjk4IDEgMTAuNSAxWm0wIDFDMTIuNDM4IDIgMTQgMy41NjMgMTQgNS41IDE0IDcuNDM4IDEyLjQzNyA5IDEwLjUgOUEzLjQ5NCAzLjQ5NCAwIDAgMSA3IDUuNUM3IDMuNTYyIDguNTYzIDIgMTAuNSAyWiIvPjwvc3ZnPg==\");--search-button:var(--search-icon) no-repeat center}[type=search]{background-color:var(--button-highlight);border:1px solid transparent;border-radius:2px;box-shadow:inset 1px 1px 0 var(--button-border-color),inset -1px -1px 0 #ccc;box-sizing:border-box;font:var(--font);height:24px;min-width:187px;padding:3px 6px}[type=search]:-moz-placeholder-shown{background-image:var(--search-icon);background-position:calc(100% - 8px);background-repeat:no-repeat;background-size:14px}[type=search]:placeholder-shown{background-image:var(--search-icon);background-position:calc(100% - 8px);background-repeat:no-repeat;background-size:14px}[type=search]:focus{outline:none}[type=search]::-moz-placeholder{font-style:italic}[type=search]::placeholder{font-style:italic}.searchbox [type=search]{padding-right:26px}.searchbox [type=search]+[aria-label=search]{background:var(--search-button),var(--button-gradient);background-size:14px;border-radius:0;min-height:22px;min-width:26px;padding:0;position:absolute;right:1px;top:1px}.searchbox [type=search]+[aria-label=search]:before{background:var(--search-button),var(--button-gradient-hovered);background-size:14px;border-radius:0}.searchbox [type=search]+[aria-label=search]:after{background:var(--search-button),var(--button-gradient-active);background-size:14px;border-radius:0}.searchbox [type=search]+[aria-label=search]:focus-visible{outline:1px dotted #000;outline-offset:-4px}.searchbox{display:inline-block;position:relative}:root{--thumb-w:10px;--thumb-h:18px;--thumb-icon:url(\"data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAoAAAASCAYAAABit09LAAAACXBIWXMAAAsTAAALEwEAmpwYAAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAAC7SURBVHgBlZLBCYQwEEVj9CDYgRfbswF7sRAbsArPdhBICDmEzPoDE9CdXeKH0eHP48fINOrSPM+k/mhd16YDtCyLmqZJhM7zxIs6PAF570WQAzJIRLkksd89DUl939eB1Ym3b0wpiQBmIYSXiTFGZYwRgWEY6o8uIPQLZGlu2rYtP54L3g3c912N45gHSEahh4dZERZj2zZyztFxHLnQw/vaLIattbmeULkMdg6XxLFaa3WB7MlCirTIHxVUkxicbwSEAAAAAElFTkSuQmCC\");--thumb-icon-box:url(\"data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAoAAAASCAYAAABit09LAAAACXBIWXMAAAsTAAALEwEAmpwYAAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAABrSURBVHgB7dLNCcAgDAXgWASjm9TJ3Egns8N4SGv6c2gbyQA+kED40IPPwJGUEsEgpRRjO8o5i6i11gfZe7HVCojIJ4QA6D3vnXM8F1DmBUkL1TdOOIRGC2X7hcIvPu1ZY/wFV83OhzTF3QGRrxiDB3GCSQAAAABJRU5ErkJggg==\");--thumb-icon-active:url(\"data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAoAAAASCAYAAABit09LAAAACXBIWXMAAAsTAAALEwEAmpwYAAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAACsSURBVHgB3ZKxCoMwEIYvJWOn0r10LHTuVDp1KIU+RN+gT9FZ6AP0LYqDY+nQWXAScXMQdRYh+ouRqEF0cPGHHJfcl7sbfkaFroYlqEefx5lxQK/blu6rjRZ6R34RLcFxAWSkmRZEbUkeLWigZgVymZhuoid264lGjwMvTkhmtYsqvKEGMQQY43jYNwoy//7t0j31b8DPXyBOcVYe5Kr9mDpKdoY6ndoCrDNyDnRZRNbxQWFyAAAAAElFTkSuQmCC\");--thumb-icon-box-active:url(\"data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAoAAAASCAYAAABit09LAAAACXBIWXMAAAsTAAALEwEAmpwYAAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAABmSURBVHgB7dKxCYAwEAXQfyEKYhrBThdwCwsLyRKukFGcRiyyh0tY2kcT0fJIn/zijg+v/IQnerUOTDYzkfToWkbOQcM6+ZX5OMNvVIVeFehqEbopXyIQmQzTgf969qFlIfkTM9wbGH4SS8YdBX4AAAAASUVORK5CYII=\")}input[type=range]{-webkit-appearance:none;-moz-appearance:none;appearance:none;background:transparent;padding:10px 1px;width:100%}input[type=range]:focus-visible{outline:1px dotted #000}input[type=range]::-webkit-slider-thumb{-webkit-appearance:none;background:var(--thumb-icon);height:var(--thumb-h);transform:translateY(-7px);width:var(--thumb-w)}input[type=range]::-webkit-slider-thumb:active{background:var(--thumb-icon-active)}input[type=range]::-moz-range-thumb{background:var(--thumb-icon);border:0;border-radius:0;height:var(--thumb-h);width:var(--thumb-w)}input[type=range]::-moz-range-thumb:active{background:var(--thumb-icon-active)}input[type=range]::-webkit-slider-runnable-track{background:var(--surface);box-shadow:inset 1px 1px 1px #999,inset -1px 0 #999,0 1px #fff;box-sizing:border-box;height:3px;width:100%}input[type=range]::-moz-range-track{background:var(--surface);box-shadow:inset 1px 1px 1px #999,inset -1px 0 #999,0 1px #fff;box-sizing:border-box;height:3px;width:100%}input[type=range].has-box-indicator::-webkit-slider-thumb{background:var(--thumb-icon-box)}input[type=range].has-box-indicator::-webkit-slider-thumb:active{background:var(--thumb-icon-box-active)}input[type=range].has-box-indicator::-moz-range-thumb{background:var(--thumb-icon-box)}input[type=range].has-box-indicator::-moz-range-thumb:active{background:var(--thumb-icon-box-active)}.is-vertical{display:inline-block;height:150px;transform:translateY(50%);width:4px}.is-vertical>input[type=range]{height:4px;margin:0 calc(var(--grouped-element-spacing) + var(--range-spacing)) 0 var(--range-spacing);transform:rotate(270deg) translateX(calc(-50% + var(--element-spacing)));transform-origin:left;width:150px}.is-vertical>input[type=range]::-webkit-slider-thumb{transform:translateY(-8px) scaleX(-1)}.is-vertical>input[type=range]::-moz-range-thumb{transform:translateY(2px) scaleX(-1)}.is-vertical>input[type=range].has-box-indicator::-webkit-slider-thumb{transform:translateY(-9px) scaleX(-1)}.is-vertical>input[type=range].has-box-indicator::-moz-range-thumb{transform:translateY(0) scaleX(-1)}.loader,.spinner{background:url(\"data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAACAAAAAgCAYAAABzenr0AAAABGdBTUEAALGPC/xhBQAAACBjSFJNAAB6JgAAgIQAAPoAAACA6AAAdTAAAOpgAAA6mAAAF3CculE8AAAABmJLR0QA/wD/AP+gvaeTAAAGiElEQVRYw+2UWWxcVxnH/+fcdebeuTNjj+2pM96bOCkle900Dq3dLBVulVBSiUpVQECpoC9ICPpSJChIiBceeEQtTRSQUNOGCqrgxkqaQkpoFodmoa4TO7Yn9nibxbPcmbucew4P5imJhfuGRH4vR/oezvnp/33fAe5zn/93yJ2FX/z0l+CgCKAhICpqnCIcrUO2bMOnMv4xfHmDkOVtfiBaNU1r8X2fuq47pShSOhTWhjdv2zxChA9TJQgLDmLbUDjHz1/93j0F5DsLHAAnyyejgB42kK+U8PHF4e2aZe0+dOjQCy1d7e1dnbEwpZCCAGAMLJOp1MYnJ6ZPnxn6w80bI4O7v/TYJUWVIZPl+1biLoGALj/uUyAAkF/K4/wnVw/07dn3/f69T+00LUvzGMdcxkEQMACA53myZVmRfXu+uKG7u+vHH5w52X/i3eO/PvjMwJ8C371Hzv9FwBMcgixL/Pnk4IFvvPTdV58+sH97Zi5LZqfGIISA4AJCCGSzWYQNA6GCjrFJgaYHkmp//5N9qgTryOu/kV547qt/FJ9HwCMCNSbAAh+/P/bW9qefPfjy2ocf2nYzPUUmpm/DZU6+5rmnajX7L47jfMgYE6qq9hm6ORBStX2STuJhLUQe7Xl0y8T46MuvHz2cfvHb37y0aoGaCOByDp/IUA2zv/+pvb0eZ3RyahrZbHZEONVfmYb626Rlwkg2QFVU2FX7aKlUOVotFV68frXyw5aWlu6oadCdu3p3DJ0afLLKvBUFpDsLu54ZQEAp3n3v5Pqdj/e90rGuu2ssncZ8vrAoHO+1hxJNbzaAIOwxKNUqZLsKkwnUyxrqjejlXKFk11y/t1DIheuTCdVhjnXsd0c+Qj6bXVUCN/UoBGXISXSriEY6rk3fwnxuCYGQBpOGeTikSDAFBWcBuMfBmQ9CPMiyBKZ6KLv24ZJT6Usm4l83IiYaGxvblFh8qw98di8BemdhpncA+Z7HwU1j02Rgx/6VmwWPG1hwau9bySbkvRKWvDJc34HgDCoBVCFAGIMf+BhdnEGG2ScypTxODL2PXDZr0YBvWPUM3DYtxBUioatjjUhEQ4uBg5npW6gzEh/ViEfClIlA8sAFIACw/+xYQIGqSrEkPBD4fw35DqauX0NdSNNBad2qBQgvQ1a4Jq3tiKgNdSJpmmQ2kwFUxTIbIwLlNCg8cAq4FAAl8CngyMCSRuGFKIQIxLXxUSRiESiKDNkMKe4KAne1gCsOoDFNNjXe2tbBd/XswBOP9CCs0631rQla1DmKOoetcjgyYKtAIQTMRoA5E3AMBbbwH1GjJqiho/nBdle2wpVVJxAYDjxGvWg0Wuh5eKvYbCqwzRgaYuq36pPxt4YjkudETNRXOAyuwEGABdnHXEyCb5ooahySkAYaE3HIjGHo3NmMzf1PVxK4K4E1lSpiJdtuMM0rc3MLNSaAjpCMzlRD75mLp3/Q9sQO/YYmYco0cA0M00YIWcNETg9jSdFkJRY/pMYiBzc91gNbMFScyoRQ6LmVBO76Bw7ufx6RoouwHrm8kM9/rblpTWMZLjEVIuVpdfPxC58EUurB87mQLkQqReZkGVpzirhckcZuTH1HUcM/6uxoa3dKRSzMzuQqpaW3U22t75TOXVxdC7oWy8gSjniq2f84PfHOzXx7Z2trk2EigJRoSDS2iZ+MzMztASWD3F8Yiqg6RjJzW+Ja+LnudRt3m2bIhOdgfCYdVKTaPxEVp0vWSiN4D4EosaHIMhaL89jYlvzZqavDa5vJFw40pawIEyo2rtsYfqChsLdUKu3NZrPwHReRNSnEIhY62zuQTt/GxNQkW/LKl8re3Jvdz/ecc1FBabUCeuCBCI6AACVVRqQ1+crfpm/MJ1H3UlPEMho8RuPRCBJxC+vbOuHCBRMM1WoVl65fgO96laqdv0LswhvPfnngWLo4KnT6ORKQAwoIwHKBEOOoUm/WXJt87VZm/LMlT/9KUTd2JXhU931fUhRFDofDWFzM+o5Ts8tBftYuFS7EiHN8ixV9L3b9NhIMkIiMK6sVoAKQOUA4R8TjqHcD1MrVcn0s/obHi0c+nby8U9Kx3zDNNtd1k67jEEmSMkG1+veQbFxfn2w+o4+VeKMToNVXECwBuqSuPgFOKASW9zPmcGzKcaSqHphcRc4qMr/FOVswi2cF5iUVgE4lzWd+TZZkkuJ6IJ0fRUvZRNwDaC2ApOgQXOA+9/mf5d+3RhGneJlNNgAAACV0RVh0ZGF0ZTpjcmVhdGUAMjAyMi0xMS0wNFQxMjo1OToxNyswMDowMD1ktU4AAAAldEVYdGRhdGU6bW9kaWZ5ADIwMjItMTEtMDRUMTI6NTk6MTcrMDA6MDBMOQ3yAAAAKHRFWHRkYXRlOnRpbWVzdGFtcAAyMDIyLTExLTA0VDEyOjU5OjE3KzAwOjAwGywsLQAAACV0RVh0d2VicDptdXgtYmxlbmQAQXRvcFByZXZpb3VzQWxwaGFCbGVuZH56rZIAAAAASUVORK5CYII=\") 50%;color:transparent;display:inline-block;height:20px;width:20px}.loader.animate,.spinner.animate{background-image:url(\"data:image/gif;base64, \")}:root{--tab-border:1px solid #888;--tab-border-color:#888;--tab-bg:#fff}menu[role=tablist]{display:flex;list-style-type:none;margin:0 0 -2px;padding-left:3px;position:relative;text-indent:0}menu[role=tablist] button{border-radius:0;color:#222;display:block;min-width:unset;padding:2px 6px;text-decoration:none;z-index:1}menu[role=tablist] button[aria-selected=true]{background:var(--tab-bg);border-bottom:0;box-shadow:none;margin:-2px 0 1px -3px;padding-bottom:4px;position:relative;z-index:8}menu[role=tablist] button[aria-selected=true]:after,menu[role=tablist] button[aria-selected=true]:before{content:none}menu[role=tablist] button[aria-selected=true]:hover{border-color:var(--tab-border-color)}menu[role=tablist] button[aria-selected=true].active,menu[role=tablist] button[aria-selected=true]:active,menu[role=tablist] button[aria-selected=true]:focus{-webkit-animation:none;animation:none;border-color:var(--tab-border-color)}menu[role=tablist] button[aria-selected=true]:focus-visible{outline:1px dotted #222;outline-offset:-4px}menu[role=tablist] button:before{border-radius:0}menu[role=tablist] button:after{content:none}menu[role=tablist] button:disabled{opacity:.6}menu[role=tablist].justified button{flex-grow:1;text-align:center}[role=tabpanel]{background:var(--tab-bg);border:var(--tab-border);clear:both;margin-bottom:9px;padding:14px;position:relative;z-index:2}input[type=email],input[type=number],input[type=password],input[type=text]{height:23px}input[type=email],input[type=number],input[type=password],input[type=text],textarea{background-color:var(--button-highlight);border:1px solid #ccc;border-radius:2px;border-top:1px solid var(--button-border-color);box-sizing:border-box;font:var(--font);padding:3px 4px 5px}input[type=email]:focus,input[type=number]:focus,input[type=password]:focus,input[type=text]:focus,textarea:focus{border-color:var(--button-shade-light-hovered);border-top-color:var(--button-shade-light-active);outline:none}:root{--treeview-square-size:8px;--treeview-spacing:20px}ul.tree-view{display:block;font:var(--font);margin:0;padding:6px 6px 6px var(--treeview-spacing)}ul.tree-view li{list-style-type:none;margin-top:4px;position:relative}ul.tree-view a{color:#000;text-decoration:none}ul.tree-view ul{margin-top:4px;padding-left:var(--treeview-spacing)}ul.tree-view.has-container{background:#fff;border:1px solid var(--button-border-color)}ul.tree-view.has-collapse-button details>summary:before{background:linear-gradient(to bottom,var(--button-face) 45%,var(--button-shade-light));border:1px solid #919191;border-radius:1px;color:#4b63a7;content:\"\\002b\";font-size:8pt;font-weight:700;height:var(--treeview-square-size);left:calc(var(--treeview-square-size)*2*-1);line-height:.5;margin:0;right:unset;text-align:center;top:calc(50% - var(--treeview-square-size)/2);width:var(--treeview-square-size)}ul.tree-view.has-collapse-button details[open]>summary:before{content:\"\\2013\";transform:none}ul.tree-view.has-connector ul{position:relative}ul.tree-view.has-connector ul:before{border-left:1px dotted #000;content:\"\";height:calc(100% - var(--treeview-square-size));left:var(--treeview-square-size);position:absolute;top:0}ul.tree-view.has-connector ul li:before{border-bottom:1px dotted #000;content:\"\";position:absolute;right:calc(100% + var(--treeview-square-size)/4);top:var(--treeview-square-size);width:calc(var(--treeview-spacing)/2)}:root{--link-color:#06c;--link-color-hovered:#39f;--primary-color:#039;--secondary-color:#000}a{color:var(--link-color);text-decoration:none}a:focus-visible{outline:1px dotted var(--link-color)}a:focus,a:hover{color:var(--link-color-hovered);text-decoration:underline}.instruction{color:var(--secondary-color);font:var(--font);font-weight:400;margin:0 0 20px}.instruction-primary{color:var(--primary-color);font-size:12pt}.header{font:var(--font);font-weight:400}.header-document{color:var(--secondary-color);font-family:Calibri,Noto Sans,sans-serif;font-size:17pt}.header-group{color:var(--primary-color);font-size:11pt}:root{--window-spacing:6px;--window-border:1px solid;--window-border-radius:6px;--window-border-color:rgba(0,0,0,.7);--window-background-color:#4580c4;--window-background-glass-stripes:linear-gradient(135deg,#fff5 70px,transparent 100px),linear-gradient(225deg,#fff5 70px,transparent 100px),linear-gradient(54deg,#0002 0 4%,#6661 6% 6%,#0002 8% 10%,#0002 15% 16%,#aaa1 17% 18%,#0002 23% 24%,#bbb2 25% 26%,#0002 31% 33%,#0002 34% 34.5%,#bbb2 36% 40%,#0002 41% 41.5%,#bbb2 44% 45%,#bbb2 46% 47%,#0002 48% 49%,#0002 50% 50.5%,#0002 56% 56.5%,#bbb2 57% 63%,#0002 67% 69%,#bbb2 69.5% 70%,#0002 73.5% 74%,#bbb2 74.5% 79%,#0002 80% 84%,#aaa2 85% 86%,#0002 87%,#bbb1 90%) left center/100vw 100vh no-repeat fixed;--window-background:linear-gradient(to right,#ffffff66,#0000001a,#ffffff33),var(--window-background-color);--control-border-color:rgba(0,0,0,.3);--control-border-radius:5px;--control-inset-shadow:inset 0 0 0 1px #fffa;--control-background:linear-gradient(hsla(0,0%,100%,.5),hsla(0,0%,100%,.3) 45%,rgba(0,0,0,.1) 50%,rgba(0,0,0,.1) 75%,hsla(0,0%,100%,.5));--control-background-hovered:radial-gradient(circle at bottom,#2aceda,transparent 65%),linear-gradient(#b6d9ee 50%,#1a6ca1 0);--control-background-active:radial-gradient(circle at bottom,#0bfdfa,transparent 65%),linear-gradient(#86a7bc 50%,#092747 0);--control-background-close:radial-gradient(circle at -60% 50%,#0007 5% 10%,#0000 50%),radial-gradient(circle at 160% 50%,#0007 5% 10%,#0000 50%),linear-gradient(#e0a197e5,#cf796a 25% 50%,#d54f36 50%);--control-background-close-hovered:radial-gradient(circle at 50% 170%,#f4e676 10% 20%,#0000 60%),radial-gradient(circle at -60% 50%,#000a 5% 10%,#0000 50%),radial-gradient(circle at 160% 50%,#000a 5% 10%,#0000 50%),linear-gradient(#fb9d8b,#ee6d56 25% 50%,#d42809 50%);--control-background-close-active:radial-gradient(circle at 50% 170%,#dcc03f 10% 20%,#0000 60%),radial-gradient(circle at -60% 50%,#000 5% 10%,#0000 50%),radial-gradient(circle at 160% 50%,#000 5% 10%,#0000 50%),linear-gradient(#d1a894,#b67562 25% 50%,#7d0d01 50%)}.window{border:var(--window-border) var(--window-border-color);box-shadow:2px 2px 10px 1px var(--window-border-color),inset 0 0 0 1px #fffa;font:var(--font);position:relative;z-index:0}.window,.window:before{border-radius:var(--window-border-radius)}.window:before{background:linear-gradient(transparent 20%,#ffffffb3 40%,transparent 41%),var(--window-background);background-color:var(--window-background-color);box-shadow:inset 0 0 0 1px #fffd;content:\"\";height:100%;left:0;position:absolute;top:0;width:100%;z-index:-1}.window-body{background:var(--surface);border:var(--window-border) var(--window-border-color);box-shadow:0 0 0 1px #fff9;margin:var(--window-spacing);margin-top:0}.window-body.has-space{padding:var(--window-spacing)}.window-body pre{margin:calc(var(--window-spacing)*-1)}.window footer,.window-footer{background:var(--surface);border:var(--window-border) var(--window-border-color);border-top:0;box-shadow:0 .5px 1px .5px #fff;margin:var(--window-spacing);margin-top:calc(var(--window-spacing)*-1 - 1px);padding:10px;position:relative}.window footer:before,.window-footer:before{box-shadow:inset 0 1px rgba(0,0,0,.3),inset 0 -1px #fff;content:\"\";height:2px;left:0;position:absolute;top:0;width:100%}.window.is-bright .window-body{background:#fff}.window.is-bright .window-footer,.window.is-bright footer{background:#eee;box-shadow:inset 0 1px 3px #ddd,0 1px 0 #fff9,1px 1px 0 #fff9,-1px 1px 0 #fff9}.window.is-bright .window-footer:before,.window.is-bright footer:before{content:none}.window.glass:before{opacity:.6}.window.glass:after{-webkit-backdrop-filter:blur(4px);backdrop-filter:blur(4px);border:none;border-radius:var(--window-border-radius);content:\"\";height:100%;left:0;position:absolute;top:0;width:100%;z-index:-10}.window.glass>.title-bar{background:var(--window-background-glass-stripes)}.window>.title-bar{border:0;box-shadow:inset 0 1px 0 #fffd,inset 1px 0 0 #fffd,inset -1px 0 0 #fffd}.window fieldset{margin-bottom:9px}.window[role=dialog]{left:50%;opacity:0;position:fixed;top:50%;transform:translate(-50%,-50%);transition-duration:.2s;transition-property:visibility,opacity;visibility:hidden;z-index:999}.window[role=dialog]:target{opacity:1;visibility:visible}.title-bar{align-items:center;background:var(--window-background);background-color:var(--window-background-color);border:var(--window-border) var(--window-border-color);border-radius:var(--window-border-radius) var(--window-border-radius) 0 0;box-shadow:inset 0 0 0 1px #fff9;display:flex;font:var(--font);justify-content:space-between;padding:var(--window-spacing);padding-top:0}.title-bar-text{color:#000;letter-spacing:0;line-height:15px;padding-top:var(--window-spacing);text-shadow:0 0 10px #fff,0 0 10px #fff,0 0 10px #fff,0 0 10px #fff,0 0 10px #fff,0 0 10px #fff,0 0 10px #fff,0 0 10px #fff}.title-bar-controls{background:#fff3;border:var(--window-border) var(--control-border-color);border-radius:0 0 var(--control-border-radius) var(--control-border-radius);border-top:0;box-shadow:0 1px 0 #fffa,1px 0 0 #fffa,-1px 0 0 #fffa;display:flex}.title-bar-controls button{background:none;border:0;border-radius:0;border-right:var(--window-border) var(--control-border-color);box-shadow:none;box-sizing:border-box;min-height:19px;min-width:29px;padding:0;position:relative}.title-bar-controls button:after{content:none}.title-bar-controls button:active,.title-bar-controls button:hover{background:none}.title-bar-controls button:disabled:before{opacity:.4}.title-bar-controls button:before{border-radius:0;bottom:0;box-shadow:inset 0 0 0 1px #fff5;content:\"\";left:0;opacity:1;position:absolute;right:0;top:0}.title-bar-controls button:not(:hover):before{opacity:1;transition:none}.title-bar-controls button.is-minimize:before,.title-bar-controls button[aria-label=Minimize]:before{background:url(\"data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAwAAAAFCAYAAABxeg0vAAAACXBIWXMAAAsTAAALEwEAmpwYAAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAAA7SURBVHgBlY+xDQAgDMNaiQ9B8AHcxAlI/NYtKwj6QOohUzxYc+1HAqQ3e03qXNpwwcyE5QsAaEGjDRf8ZAza6Bz6VQAAAABJRU5ErkJggg==\") no-repeat center 10px}.title-bar-controls button.is-maximize:before,.title-bar-controls button[aria-label=Maximize]:before{background:url(\"data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAwAAAAKCAYAAACALL/6AAAACXBIWXMAAAsTAAALEwEAmpwYAAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAABsSURBVHgBlZHBCYAwDEVT6YaKbqArOILgBLqBosP1VkhAaQ6S1lKad+ihPy8kxLT9+IACG57r2KqKu2GCBpTYtEMOOQELzrnv4z53I4vDjjJnwXsPJWTOAiJGHVNB5pGwLjPk+AlEBLUY7eFebCosBHOR7vYAAAAASUVORK5CYII=\") no-repeat 50%}.title-bar-controls button.is-help:before,.title-bar-controls button[aria-label=Help]:before{background:url(\"data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAgAAAAMBAMAAACtsOGuAAAAG1BMVEUAAABTVmb////39/fa2trm5ubj4+Pq6urg4OByvnYHAAAAAXRSTlMAQObYZgAAADtJREFUCNdjYBQUFGAQVFISZBBSUlJkEDYUNmQQFBAvZGBgDA1kYBBLE2RgkGgUYGAQBDIYRBwhBJgLAKBIBQ2X+PZ0AAAAAElFTkSuQmCC\") no-repeat 50%}.title-bar-controls button.is-restore:before,.title-bar-controls button[aria-label=Restore]:before{background:url(\"data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAwAAAAMCAYAAABWdVznAAAACXBIWXMAAAsTAAALEwEAmpwYAAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAACVSURBVHgBlVDLCoMwENwU/7ClvRYK7b2lCIJHEQQ/QEHwqujfBXJIQMnihg34iHPJ7GNm2AhYcL2/JwhAxIuhq3aXb48PiFBna2YFERVHzoQLnAQmSCmxeL5+3rBtSsdpBwVKKTcY+1rY197G+8RRoLVejed94igwxrgB/zXeJ+4Jijz2Erjg+8/WE7ZAt6EgSUsIxQwaJEWv7G6SSgAAAABJRU5ErkJggg==\") no-repeat 50%}.title-bar-controls button.is-close,.title-bar-controls button[aria-label=Close]{min-width:48px}.title-bar-controls button.is-close:before,.title-bar-controls button[aria-label=Close]:before{background:url(\"data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAsAAAAKCAYAAABi8KSDAAAACXBIWXMAAAsTAAALEwEAmpwYAAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAACUSURBVHgBlZFNCoQwDIVbmRvOMHOAAd0K/uDGlYKiCG4VPICi9yu0oCSQoFUXFtKEl6+voZXvr7sKay1TL690FOexY+Hz87i2decOuDKQsNGVe8AGYTRJIh0Y+orBvxszCPllOymlxN06jNE26Qnwg4zdHQLrMhFaawwAqAadDPE1ijwUxhiMKCmxCZk06OMYTz5lAx3xWDeXkLNAAAAAAElFTkSuQmCC\") no-repeat 50%}.title-bar-controls button:first-child,.title-bar-controls button:first-child:before{border-bottom-left-radius:var(--control-border-radius)}.title-bar-controls button:last-child,.title-bar-controls button:last-child:before{border:0;border-bottom-right-radius:var(--control-border-radius)}.title-bar-controls button:focus{-webkit-animation:none;animation:none;outline:none}.title-bar.active .title-bar-controls,.window.active .title-bar .title-bar-controls{border-color:var(--window-border-color)}.title-bar.active .title-bar-controls button,.window.active .title-bar .title-bar-controls button{border-color:var(--window-border-color);box-shadow:var(--control-inset-shadow)}.title-bar.active .title-bar-controls button:after,.window.active .title-bar .title-bar-controls button:after{content:none}.title-bar.active .title-bar-controls button.is-minimize,.title-bar.active .title-bar-controls button[aria-label=Minimize],.window.active .title-bar .title-bar-controls button.is-minimize,.window.active .title-bar .title-bar-controls button[aria-label=Minimize]{background:url(\"data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAwAAAAFCAYAAABxeg0vAAAACXBIWXMAAAsTAAALEwEAmpwYAAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAAA7SURBVHgBlY+xDQAgDMNaiQ9B8AHcxAlI/NYtKwj6QOohUzxYc+1HAqQ3e03qXNpwwcyE5QsAaEGjDRf8ZAza6Bz6VQAAAABJRU5ErkJggg==\") no-repeat center 10px,var(--control-background)}.title-bar.active .title-bar-controls button.is-minimize:before,.title-bar.active .title-bar-controls button[aria-label=Minimize]:before,.window.active .title-bar .title-bar-controls button.is-minimize:before,.window.active .title-bar .title-bar-controls button[aria-label=Minimize]:before{background:url(\"data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAwAAAAFCAYAAABxeg0vAAAACXBIWXMAAAsTAAALEwEAmpwYAAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAAA7SURBVHgBlY+xDQAgDMNaiQ9B8AHcxAlI/NYtKwj6QOohUzxYc+1HAqQ3e03qXNpwwcyE5QsAaEGjDRf8ZAza6Bz6VQAAAABJRU5ErkJggg==\") no-repeat center 10px,var(--control-background-hovered);border-bottom-left-radius:var(--control-border-radius);box-shadow:0 0 7px 3px #5dc4f0,var(--control-inset-shadow);content:\"\";opacity:0;transition:opacity .3s linear}.title-bar.active .title-bar-controls button.is-minimize:focus-visible:before,.title-bar.active .title-bar-controls button.is-minimize:hover:before,.title-bar.active .title-bar-controls button[aria-label=Minimize]:focus-visible:before,.title-bar.active .title-bar-controls button[aria-label=Minimize]:hover:before,.window.active .title-bar .title-bar-controls button.is-minimize:focus-visible:before,.window.active .title-bar .title-bar-controls button.is-minimize:hover:before,.window.active .title-bar .title-bar-controls button[aria-label=Minimize]:focus-visible:before,.window.active .title-bar .title-bar-controls button[aria-label=Minimize]:hover:before{opacity:1;transition:opacity .1s linear}.title-bar.active .title-bar-controls button.is-minimize:active:before,.title-bar.active .title-bar-controls button[aria-label=Minimize]:active:before,.window.active .title-bar .title-bar-controls button.is-minimize:active:before,.window.active .title-bar .title-bar-controls button[aria-label=Minimize]:active:before{background:url(\"data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAwAAAAFCAYAAABxeg0vAAAACXBIWXMAAAsTAAALEwEAmpwYAAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAAA7SURBVHgBlY+xDQAgDMNaiQ9B8AHcxAlI/NYtKwj6QOohUzxYc+1HAqQ3e03qXNpwwcyE5QsAaEGjDRf8ZAza6Bz6VQAAAABJRU5ErkJggg==\") no-repeat center 10px,var(--control-background-active)}.title-bar.active .title-bar-controls button.is-maximize,.title-bar.active .title-bar-controls button[aria-label=Maximize],.window.active .title-bar .title-bar-controls button.is-maximize,.window.active .title-bar .title-bar-controls button[aria-label=Maximize]{background:url(\"data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAwAAAAKCAYAAACALL/6AAAACXBIWXMAAAsTAAALEwEAmpwYAAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAABsSURBVHgBlZHBCYAwDEVT6YaKbqArOILgBLqBosP1VkhAaQ6S1lKad+ihPy8kxLT9+IACG57r2KqKu2GCBpTYtEMOOQELzrnv4z53I4vDjjJnwXsPJWTOAiJGHVNB5pGwLjPk+AlEBLUY7eFebCosBHOR7vYAAAAASUVORK5CYII=\") no-repeat 50%,var(--control-background)}.title-bar.active .title-bar-controls button.is-maximize:before,.title-bar.active .title-bar-controls button[aria-label=Maximize]:before,.window.active .title-bar .title-bar-controls button.is-maximize:before,.window.active .title-bar .title-bar-controls button[aria-label=Maximize]:before{background:url(\"data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAwAAAAKCAYAAACALL/6AAAACXBIWXMAAAsTAAALEwEAmpwYAAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAABsSURBVHgBlZHBCYAwDEVT6YaKbqArOILgBLqBosP1VkhAaQ6S1lKad+ihPy8kxLT9+IACG57r2KqKu2GCBpTYtEMOOQELzrnv4z53I4vDjjJnwXsPJWTOAiJGHVNB5pGwLjPk+AlEBLUY7eFebCosBHOR7vYAAAAASUVORK5CYII=\") no-repeat 50%,var(--control-background-hovered);box-shadow:0 0 7px 3px #5dc4f0,var(--control-inset-shadow);content:\"\";opacity:0;transition:opacity .3s linear}.title-bar.active .title-bar-controls button.is-maximize:focus-visible:before,.title-bar.active .title-bar-controls button.is-maximize:hover:before,.title-bar.active .title-bar-controls button[aria-label=Maximize]:focus-visible:before,.title-bar.active .title-bar-controls button[aria-label=Maximize]:hover:before,.window.active .title-bar .title-bar-controls button.is-maximize:focus-visible:before,.window.active .title-bar .title-bar-controls button.is-maximize:hover:before,.window.active .title-bar .title-bar-controls button[aria-label=Maximize]:focus-visible:before,.window.active .title-bar .title-bar-controls button[aria-label=Maximize]:hover:before{opacity:1;transition:opacity .1s linear}.title-bar.active .title-bar-controls button.is-maximize:active:before,.title-bar.active .title-bar-controls button[aria-label=Maximize]:active:before,.window.active .title-bar .title-bar-controls button.is-maximize:active:before,.window.active .title-bar .title-bar-controls button[aria-label=Maximize]:active:before{background:url(\"data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAwAAAAKCAYAAACALL/6AAAACXBIWXMAAAsTAAALEwEAmpwYAAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAABsSURBVHgBlZHBCYAwDEVT6YaKbqArOILgBLqBosP1VkhAaQ6S1lKad+ihPy8kxLT9+IACG57r2KqKu2GCBpTYtEMOOQELzrnv4z53I4vDjjJnwXsPJWTOAiJGHVNB5pGwLjPk+AlEBLUY7eFebCosBHOR7vYAAAAASUVORK5CYII=\") no-repeat 50%,var(--control-background-active)}.title-bar.active .title-bar-controls button.is-help,.title-bar.active .title-bar-controls button[aria-label=Help],.window.active .title-bar .title-bar-controls button.is-help,.window.active .title-bar .title-bar-controls button[aria-label=Help]{background:url(\"data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAgAAAAMBAMAAACtsOGuAAAAG1BMVEUAAABTVmb////39/fa2trm5ubj4+Pq6urg4OByvnYHAAAAAXRSTlMAQObYZgAAADtJREFUCNdjYBQUFGAQVFISZBBSUlJkEDYUNmQQFBAvZGBgDA1kYBBLE2RgkGgUYGAQBDIYRBwhBJgLAKBIBQ2X+PZ0AAAAAElFTkSuQmCC\") no-repeat 50%,var(--control-background)}.title-bar.active .title-bar-controls button.is-help:before,.title-bar.active .title-bar-controls button[aria-label=Help]:before,.window.active .title-bar .title-bar-controls button.is-help:before,.window.active .title-bar .title-bar-controls button[aria-label=Help]:before{background:url(\"data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAgAAAAMBAMAAACtsOGuAAAAG1BMVEUAAABTVmb////39/fa2trm5ubj4+Pq6urg4OByvnYHAAAAAXRSTlMAQObYZgAAADtJREFUCNdjYBQUFGAQVFISZBBSUlJkEDYUNmQQFBAvZGBgDA1kYBBLE2RgkGgUYGAQBDIYRBwhBJgLAKBIBQ2X+PZ0AAAAAElFTkSuQmCC\") no-repeat 50%,var(--control-background-hovered);box-shadow:0 0 7px 3px #5dc4f0,var(--control-inset-shadow);content:\"\";opacity:0;transition:opacity .3s linear}.title-bar.active .title-bar-controls button.is-help:focus-visible:before,.title-bar.active .title-bar-controls button.is-help:hover:before,.title-bar.active .title-bar-controls button[aria-label=Help]:focus-visible:before,.title-bar.active .title-bar-controls button[aria-label=Help]:hover:before,.window.active .title-bar .title-bar-controls button.is-help:focus-visible:before,.window.active .title-bar .title-bar-controls button.is-help:hover:before,.window.active .title-bar .title-bar-controls button[aria-label=Help]:focus-visible:before,.window.active .title-bar .title-bar-controls button[aria-label=Help]:hover:before{opacity:1;transition:opacity .1s linear}.title-bar.active .title-bar-controls button.is-help:active:before,.title-bar.active .title-bar-controls button[aria-label=Help]:active:before,.window.active .title-bar .title-bar-controls button.is-help:active:before,.window.active .title-bar .title-bar-controls button[aria-label=Help]:active:before{background:url(\"data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAgAAAAMBAMAAACtsOGuAAAAG1BMVEUAAABTVmb////39/fa2trm5ubj4+Pq6urg4OByvnYHAAAAAXRSTlMAQObYZgAAADtJREFUCNdjYBQUFGAQVFISZBBSUlJkEDYUNmQQFBAvZGBgDA1kYBBLE2RgkGgUYGAQBDIYRBwhBJgLAKBIBQ2X+PZ0AAAAAElFTkSuQmCC\") no-repeat 50%,var(--control-background-active)}.title-bar.active .title-bar-controls button.is-restore,.title-bar.active .title-bar-controls button[aria-label=Restore],.window.active .title-bar .title-bar-controls button.is-restore,.window.active .title-bar .title-bar-controls button[aria-label=Restore]{background:url(\"data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAwAAAAMCAYAAABWdVznAAAACXBIWXMAAAsTAAALEwEAmpwYAAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAACVSURBVHgBlVDLCoMwENwU/7ClvRYK7b2lCIJHEQQ/QEHwqujfBXJIQMnihg34iHPJ7GNm2AhYcL2/JwhAxIuhq3aXb48PiFBna2YFERVHzoQLnAQmSCmxeL5+3rBtSsdpBwVKKTcY+1rY197G+8RRoLVejed94igwxrgB/zXeJ+4Jijz2Erjg+8/WE7ZAt6EgSUsIxQwaJEWv7G6SSgAAAABJRU5ErkJggg==\") no-repeat 50%,var(--control-background)}.title-bar.active .title-bar-controls button.is-restore:before,.title-bar.active .title-bar-controls button[aria-label=Restore]:before,.window.active .title-bar .title-bar-controls button.is-restore:before,.window.active .title-bar .title-bar-controls button[aria-label=Restore]:before{background:url(\"data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAwAAAAMCAYAAABWdVznAAAACXBIWXMAAAsTAAALEwEAmpwYAAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAACVSURBVHgBlVDLCoMwENwU/7ClvRYK7b2lCIJHEQQ/QEHwqujfBXJIQMnihg34iHPJ7GNm2AhYcL2/JwhAxIuhq3aXb48PiFBna2YFERVHzoQLnAQmSCmxeL5+3rBtSsdpBwVKKTcY+1rY197G+8RRoLVejed94igwxrgB/zXeJ+4Jijz2Erjg+8/WE7ZAt6EgSUsIxQwaJEWv7G6SSgAAAABJRU5ErkJggg==\") no-repeat 50%,var(--control-background-hovered);box-shadow:0 0 7px 3px #5dc4f0,var(--control-inset-shadow);content:\"\";opacity:0;transition:opacity .3s linear}.title-bar.active .title-bar-controls button.is-restore:focus-visible:before,.title-bar.active .title-bar-controls button.is-restore:hover:before,.title-bar.active .title-bar-controls button[aria-label=Restore]:focus-visible:before,.title-bar.active .title-bar-controls button[aria-label=Restore]:hover:before,.window.active .title-bar .title-bar-controls button.is-restore:focus-visible:before,.window.active .title-bar .title-bar-controls button.is-restore:hover:before,.window.active .title-bar .title-bar-controls button[aria-label=Restore]:focus-visible:before,.window.active .title-bar .title-bar-controls button[aria-label=Restore]:hover:before{opacity:1;transition:opacity .1s linear}.title-bar.active .title-bar-controls button.is-restore:active:before,.title-bar.active .title-bar-controls button[aria-label=Restore]:active:before,.window.active .title-bar .title-bar-controls button.is-restore:active:before,.window.active .title-bar .title-bar-controls button[aria-label=Restore]:active:before{background:url(\"data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAwAAAAMCAYAAABWdVznAAAACXBIWXMAAAsTAAALEwEAmpwYAAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAACVSURBVHgBlVDLCoMwENwU/7ClvRYK7b2lCIJHEQQ/QEHwqujfBXJIQMnihg34iHPJ7GNm2AhYcL2/JwhAxIuhq3aXb48PiFBna2YFERVHzoQLnAQmSCmxeL5+3rBtSsdpBwVKKTcY+1rY197G+8RRoLVejed94igwxrgB/zXeJ+4Jijz2Erjg+8/WE7ZAt6EgSUsIxQwaJEWv7G6SSgAAAABJRU5ErkJggg==\") no-repeat 50%,var(--control-background-active)}.title-bar.active .title-bar-controls button.is-close,.title-bar.active .title-bar-controls button[aria-label=Close],.window.active .title-bar .title-bar-controls button.is-close,.window.active .title-bar .title-bar-controls button[aria-label=Close]{background:url(\"data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAsAAAAKCAYAAABi8KSDAAAACXBIWXMAAAsTAAALEwEAmpwYAAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAACUSURBVHgBlZFNCoQwDIVbmRvOMHOAAd0K/uDGlYKiCG4VPICi9yu0oCSQoFUXFtKEl6+voZXvr7sKay1TL690FOexY+Hz87i2decOuDKQsNGVe8AGYTRJIh0Y+orBvxszCPllOymlxN06jNE26Qnwg4zdHQLrMhFaawwAqAadDPE1ijwUxhiMKCmxCZk06OMYTz5lAx3xWDeXkLNAAAAAAElFTkSuQmCC\") no-repeat 50%,var(--control-background),var(--control-background-close);box-shadow:var(--control-inset-shadow)}.title-bar.active .title-bar-controls button.is-close:before,.title-bar.active .title-bar-controls button[aria-label=Close]:before,.window.active .title-bar .title-bar-controls button.is-close:before,.window.active .title-bar .title-bar-controls button[aria-label=Close]:before{background:url(\"data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAsAAAAKCAYAAABi8KSDAAAACXBIWXMAAAsTAAALEwEAmpwYAAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAACUSURBVHgBlZFNCoQwDIVbmRvOMHOAAd0K/uDGlYKiCG4VPICi9yu0oCSQoFUXFtKEl6+voZXvr7sKay1TL690FOexY+Hz87i2decOuDKQsNGVe8AGYTRJIh0Y+orBvxszCPllOymlxN06jNE26Qnwg4zdHQLrMhFaawwAqAadDPE1ijwUxhiMKCmxCZk06OMYTz5lAx3xWDeXkLNAAAAAAElFTkSuQmCC\") no-repeat 50%,var(--control-background),var(--control-background-close-hovered);border-bottom-right-radius:var(--control-border-radius);box-shadow:0 0 7px 3px #e68e75,var(--control-inset-shadow);content:\"\";opacity:0;transition:opacity .3s linear}.title-bar.active .title-bar-controls button.is-close:focus-visible:before,.title-bar.active .title-bar-controls button.is-close:hover:before,.title-bar.active .title-bar-controls button[aria-label=Close]:focus-visible:before,.title-bar.active .title-bar-controls button[aria-label=Close]:hover:before,.window.active .title-bar .title-bar-controls button.is-close:focus-visible:before,.window.active .title-bar .title-bar-controls button.is-close:hover:before,.window.active .title-bar .title-bar-controls button[aria-label=Close]:focus-visible:before,.window.active .title-bar .title-bar-controls button[aria-label=Close]:hover:before{opacity:1;transition:opacity .1s linear}.title-bar.active .title-bar-controls button.is-close:active:before,.title-bar.active .title-bar-controls button[aria-label=Close]:active:before,.window.active .title-bar .title-bar-controls button.is-close:active:before,.window.active .title-bar .title-bar-controls button[aria-label=Close]:active:before{background:url(\"data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAsAAAAKCAYAAABi8KSDAAAACXBIWXMAAAsTAAALEwEAmpwYAAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAACUSURBVHgBlZFNCoQwDIVbmRvOMHOAAd0K/uDGlYKiCG4VPICi9yu0oCSQoFUXFtKEl6+voZXvr7sKay1TL690FOexY+Hz87i2decOuDKQsNGVe8AGYTRJIh0Y+orBvxszCPllOymlxN06jNE26Qnwg4zdHQLrMhFaawwAqAadDPE1ijwUxhiMKCmxCZk06OMYTz5lAx3xWDeXkLNAAAAAAElFTkSuQmCC\") no-repeat 50%,var(--control-background),var(--control-background-close-active)}.title-bar.active .title-bar-controls button:disabled.is-maximize,.title-bar.active .title-bar-controls button:disabled[aria-label=Maximize],.window.active .title-bar .title-bar-controls button:disabled.is-maximize,.window.active .title-bar .title-bar-controls button:disabled[aria-label=Maximize]{background:var(--control-background)}.title-bar.active .title-bar-controls button:disabled.is-maximize:before,.title-bar.active .title-bar-controls button:disabled[aria-label=Maximize]:before,.window.active .title-bar .title-bar-controls button:disabled.is-maximize:before,.window.active .title-bar .title-bar-controls button:disabled[aria-label=Maximize]:before{background:url(\"data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAwAAAAKCAYAAACALL/6AAAACXBIWXMAAAsTAAALEwEAmpwYAAAAAXNSR0IArs4c6QAAAARnQU1BAACxjwv8YQUAAABsSURBVHgBlZHBCYAwDEVT6YaKbqArOILgBLqBosP1VkhAaQ6S1lKad+ihPy8kxLT9+IACG57r2KqKu2GCBpTYtEMOOQELzrnv4z53I4vDjjJnwXsPJWTOAiJGHVNB5pGwLjPk+AlEBLUY7eFebCosBHOR7vYAAAAASUVORK5CYII=\") no-repeat 50%;content:\"\";opacity:.4}.status-bar{background:var(--surface);border:var(--window-border) var(--window-border-color);border-top:0;box-shadow:0 1px 0 #fff9,1px 0 0 #fff9,-1px 0 0 #fff9;display:flex;margin:var(--window-spacing);margin-top:calc(var(--window-spacing)*-1)}.status-bar-field{border-right:var(--window-border) #cfcfcf;flex-grow:1;margin:0;padding:2px 3px}.status-bar-field:last-child{border-right:0}\n/*# sourceMappingURL=7.css.map */"
RAW_FILES['mvp'] = RawFiles({"credit": "<a href='http://andybrewer.github.io/mvp/' target='_blank'>MVP.css</a> by <code>andybrewer</code>"}, {}, {}, {})
RAW_FILES['mvp'].styles['mvp.css'] = "/* MVP.css v1.17.1 - https://github.com/andybrewer/mvp */\n\n:root {\n  --active-brightness: 0.85;\n  --border-radius: 5px;\n  --box-shadow: 2px 2px 10px;\n  --color-accent: #118bee15;\n  --color-bg: #fff;\n  --color-bg-secondary: #e9e9e9;\n  --color-link: #118bee;\n  --color-secondary: #920de9;\n  --color-secondary-accent: #920de90b;\n  --color-shadow: #f4f4f4;\n  --color-table: #118bee;\n  --color-text: #000;\n  --color-text-secondary: #999;\n  --color-scrollbar: #cacae8;\n  --font-family: -apple-system, BlinkMacSystemFont, \"Segoe UI\", Roboto, Oxygen-Sans, Ubuntu, Cantarell, \"Helvetica Neue\", sans-serif;\n  --hover-brightness: 1.2;\n  --justify-important: center;\n  --justify-normal: left;\n  --line-height: 1.5;\n  --width-card: 285px;\n  --width-card-medium: 460px;\n  --width-card-wide: 800px;\n  --width-content: 1080px;\n}\n\n@media (prefers-color-scheme: dark) {\n  :root[color-mode=\"user\"] {\n    --color-accent: #0097fc4f;\n    --color-bg: #333;\n    --color-bg-secondary: #555;\n    --color-link: #0097fc;\n    --color-secondary: #e20de9;\n    --color-secondary-accent: #e20de94f;\n    --color-shadow: #bbbbbb20;\n    --color-table: #0097fc;\n    --color-text: #f7f7f7;\n    --color-text-secondary: #aaa;\n  }\n}\n\nhtml {\n  scroll-behavior: smooth;\n}\n\n@media (prefers-reduced-motion: reduce) {\n  html {\n    scroll-behavior: auto;\n  }\n}\n\n/* Layout */\narticle aside {\n  background: var(--color-secondary-accent);\n  border-left: 4px solid var(--color-secondary);\n  padding: 0.01rem 0.8rem;\n}\n\nbody {\n  background: var(--color-bg);\n  color: var(--color-text);\n  font-family: var(--font-family);\n  line-height: var(--line-height);\n  margin: 0;\n  overflow-x: hidden;\n  padding: 0;\n}\n\nfooter,\nheader,\nmain {\n  margin: 0 auto;\n  max-width: var(--width-content);\n  padding: 3rem 1rem;\n}\n\nhr {\n  background-color: var(--color-bg-secondary);\n  border: none;\n  height: 1px;\n  margin: 4rem 0;\n  width: 100%;\n}\n\nsection {\n  display: flex;\n  flex-wrap: wrap;\n  justify-content: var(--justify-important);\n}\n\nsection img,\narticle img {\n  max-width: 100%;\n}\n\nsection pre {\n  overflow: auto;\n}\n\nsection aside {\n  border: 1px solid var(--color-bg-secondary);\n  border-radius: var(--border-radius);\n  box-shadow: var(--box-shadow) var(--color-shadow);\n  margin: 1rem;\n  padding: 1.25rem;\n  width: var(--width-card);\n}\n\nsection aside:hover {\n  box-shadow: var(--box-shadow) var(--color-bg-secondary);\n}\n\n[hidden] {\n  display: none;\n}\n\n/* Headers */\narticle header,\ndiv header,\nmain header {\n  padding-top: 0;\n}\n\nheader {\n  text-align: var(--justify-important);\n}\n\nheader a b,\nheader a em,\nheader a i,\nheader a strong {\n  margin-left: 0.5rem;\n  margin-right: 0.5rem;\n}\n\nheader nav img {\n  margin: 1rem 0;\n}\n\nsection header {\n  padding-top: 0;\n  width: 100%;\n}\n\n/* Nav */\nnav {\n  align-items: center;\n  display: flex;\n  font-weight: bold;\n  justify-content: space-between;\n  margin-bottom: 7rem;\n}\n\nnav ul {\n  list-style: none;\n  padding: 0;\n}\n\nnav ul li {\n  display: inline-block;\n  margin: 0 0.5rem;\n  position: relative;\n  text-align: left;\n}\n\n/* Nav Dropdown */\nnav ul li:hover ul {\n  display: block;\n}\n\nnav ul li ul {\n  background: var(--color-bg);\n  border: 1px solid var(--color-bg-secondary);\n  border-radius: var(--border-radius);\n  box-shadow: var(--box-shadow) var(--color-shadow);\n  display: none;\n  height: auto;\n  left: -2px;\n  padding: 0.5rem 1rem;\n  position: absolute;\n  top: 1.7rem;\n  white-space: nowrap;\n  width: auto;\n  z-index: 1;\n}\n\nnav ul li ul::before {\n  /* fill gap above to make mousing over them easier */\n  content: \"\";\n  position: absolute;\n  left: 0;\n  right: 0;\n  top: -0.5rem;\n  height: 0.5rem;\n}\n\nnav ul li ul li,\nnav ul li ul li a {\n  display: block;\n}\n\n/* Nav for Mobile */\n@media (max-width: 768px) {\n  nav {\n    flex-wrap: wrap;\n  }\n\n  nav ul li {\n    width: calc(100% - 1em);\n  }\n\n  nav ul li ul {\n    border: none;\n    box-shadow: none;\n    display: block;\n    position: static;\n  }\n}\n\n/* Typography */\ncode,\nsamp {\n  background-color: var(--color-accent);\n  border-radius: var(--border-radius);\n  color: var(--color-text);\n  display: inline-block;\n  margin: 0 0.1rem;\n  padding: 0 0.5rem;\n}\n\ndetails {\n  margin: 1.3rem 0;\n}\n\ndetails summary {\n  font-weight: bold;\n  cursor: pointer;\n}\n\nh1,\nh2,\nh3,\nh4,\nh5,\nh6 {\n  line-height: var(--line-height);\n  text-wrap: balance;\n}\n\nmark {\n  padding: 0.1rem;\n}\n\nol li,\nul li {\n  padding: 0.2rem 0;\n}\n\np {\n  margin: 0.75rem 0;\n  padding: 0;\n  width: 100%;\n}\n\npre {\n  margin: 1rem 0;\n  max-width: var(--width-card-wide);\n  padding: 1rem 0;\n}\n\npre code,\npre samp {\n  display: block;\n  max-width: var(--width-card-wide);\n  padding: 0.5rem 2rem;\n  white-space: pre-wrap;\n}\n\nsmall {\n  color: var(--color-text-secondary);\n}\n\nsup {\n  background-color: var(--color-secondary);\n  border-radius: var(--border-radius);\n  color: var(--color-bg);\n  font-size: xx-small;\n  font-weight: bold;\n  margin: 0.2rem;\n  padding: 0.2rem 0.3rem;\n  position: relative;\n  top: -2px;\n}\n\n/* Links */\na {\n  color: var(--color-link);\n  display: inline-block;\n  font-weight: bold;\n  text-decoration: underline;\n}\n\na:hover {\n  filter: brightness(var(--hover-brightness));\n}\n\na:active {\n  filter: brightness(var(--active-brightness));\n}\n\na b,\na em,\na i,\na strong,\nbutton,\ninput[type=\"submit\"] {\n  border-radius: var(--border-radius);\n  display: inline-block;\n  font-size: medium;\n  font-weight: bold;\n  line-height: var(--line-height);\n  margin: 0.5rem 0;\n  padding: 1rem 2rem;\n}\n\nbutton,\ninput[type=\"submit\"] {\n  font-family: var(--font-family);\n}\n\nbutton:hover,\ninput[type=\"submit\"]:hover {\n  cursor: pointer;\n  filter: brightness(var(--hover-brightness));\n}\n\nbutton:active,\ninput[type=\"submit\"]:active {\n  filter: brightness(var(--active-brightness));\n}\n\na b,\na strong,\nbutton,\ninput[type=\"submit\"] {\n  background-color: var(--color-link);\n  border: 2px solid var(--color-link);\n  color: var(--color-bg);\n}\n\na em,\na i {\n  border: 2px solid var(--color-link);\n  border-radius: var(--border-radius);\n  color: var(--color-link);\n  display: inline-block;\n  padding: 1rem 2rem;\n}\n\narticle aside a {\n  color: var(--color-secondary);\n}\n\n/* Images */\nfigure {\n  margin: 0;\n  padding: 0;\n}\n\nfigure img {\n  max-width: 100%;\n}\n\nfigure figcaption {\n  color: var(--color-text-secondary);\n}\n\n/* Forms */\nbutton:disabled,\ninput:disabled {\n  background: var(--color-bg-secondary);\n  border-color: var(--color-bg-secondary);\n  color: var(--color-text-secondary);\n  cursor: not-allowed;\n}\n\nbutton[disabled]:hover,\ninput[type=\"submit\"][disabled]:hover {\n  filter: none;\n}\n\nform {\n  border: 1px solid var(--color-bg-secondary);\n  border-radius: var(--border-radius);\n  box-shadow: var(--box-shadow) var(--color-shadow);\n  display: block;\n  max-width: var(--width-card-wide);\n  min-width: var(--width-card);\n  padding: 1.5rem;\n  text-align: var(--justify-normal);\n}\n\nform header {\n  margin: 1.5rem 0;\n  padding: 1.5rem 0;\n}\n\ninput,\nlabel,\nselect,\ntextarea {\n  display: block;\n  font-size: inherit;\n  max-width: var(--width-card-wide);\n}\n\ninput[type=\"checkbox\"],\ninput[type=\"radio\"] {\n  display: inline-block;\n}\n\ninput[type=\"checkbox\"] + label,\ninput[type=\"radio\"] + label {\n  display: inline-block;\n  font-weight: normal;\n  position: relative;\n  top: 1px;\n}\n\ninput[type=\"range\"] {\n  padding: 0.4rem 0;\n}\n\ninput,\nselect,\ntextarea {\n  border: 1px solid var(--color-bg-secondary);\n  border-radius: var(--border-radius);\n  margin-bottom: 1rem;\n  padding: 0.4rem 0.8rem;\n}\n\ninput[type=\"text\"],\ninput[type=\"password\"],\ninput[type=\"email\"],\ntextarea {\n  width: calc(100% - 1.6rem);\n}\n\ninput[readonly],\ntextarea[readonly] {\n  background-color: var(--color-bg-secondary);\n}\n\nlabel {\n  font-weight: bold;\n  margin-bottom: 0.2rem;\n}\n\n/* Popups */\ndialog {\n  max-width: 90%;\n  max-height: 85dvh;\n  margin: auto;\n  padding-block: 0;\n  padding-inline: 20px;\n  border: 1px solid var(--color-bg-secondary);\n  border-radius: 0.5rem;\n  overscroll-behavior: contain;\n  scroll-behavior: smooth;\n  scrollbar-width: none; /* Hide scrollbar for Firefox */\n  -ms-overflow-style: none; /* Hide scrollbar for IE and Edge */\n  scrollbar-color: transparent transparent;\n  animation: bottom-to-top 0.25s ease-in-out forwards;\n}\n\ndialog::-webkit-scrollbar {\n  width: 0;\n  display: none;\n}\n\ndialog::-webkit-scrollbar-track {\n  background: transparent;\n}\n\ndialog::-webkit-scrollbar-thumb {\n  background-color: transparent;\n}\n\n@media (min-width: 650px) {\n  dialog {\n    max-width: 39rem;\n  }\n}\n\ndialog::backdrop {\n  background-color: rgba(0, 0, 0, 0.5);\n}\n\n@keyframes bottom-to-top {\n  0% {\n    opacity: 0;\n    transform: translateY(10%);\n  }\n\n  100% {\n    opacity: 1;\n    transform: translateY(0);\n  }\n}\n\ndialog hr {\n  margin-block: 1rem;\n}\n\n/* Tables */\ntable {\n  border: 1px solid var(--color-bg-secondary);\n  border-radius: var(--border-radius);\n  border-spacing: 0;\n  display: inline-block;\n  max-width: 100%;\n  overflow-x: auto;\n  padding: 0;\n  white-space: nowrap;\n}\n\ntable td,\ntable th,\ntable tr {\n  padding: 0.4rem 0.8rem;\n  text-align: var(--justify-important);\n}\n\ntable thead {\n  background-color: var(--color-table);\n  border-collapse: collapse;\n  border-radius: var(--border-radius);\n  color: var(--color-bg);\n  margin: 0;\n  padding: 0;\n}\n\ntable thead tr:first-child th:first-child {\n  border-top-left-radius: var(--border-radius);\n}\n\ntable thead tr:first-child th:last-child {\n  border-top-right-radius: var(--border-radius);\n}\n\ntable thead th:first-child,\ntable tr td:first-child {\n  text-align: var(--justify-normal);\n}\n\ntable tr:nth-child(even) {\n  background-color: var(--color-accent);\n}\n\n/* Quotes */\nblockquote {\n  display: block;\n  font-size: x-large;\n  line-height: var(--line-height);\n  margin: 1rem auto;\n  max-width: var(--width-card-medium);\n  padding: 1.5rem 1rem;\n  text-align: var(--justify-important);\n}\n\nblockquote footer {\n  color: var(--color-text-secondary);\n  display: block;\n  font-size: small;\n  line-height: var(--line-height);\n  padding: 1.5rem 0;\n}\n\n/* Scrollbars */\n* {\n  scrollbar-width: thin;\n  scrollbar-color: var(--color-scrollbar) transparent;\n}\n\n*::-webkit-scrollbar {\n  width: 5px;\n  height: 5px;\n}\n\n*::-webkit-scrollbar-track {\n  background: transparent;\n}\n\n*::-webkit-scrollbar-thumb {\n  background-color: var(--color-scrollbar);\n  border-radius: 10px;\n}\n"
RAW_FILES['sakura'] = RawFiles({"credit": "<a href='https://oxal.org/projects/sakura/' target='_blank'>Sakura</a> by <code>oxal</code>"}, {}, {}, {})
RAW_FILES['sakura'].styles['sakura.css'] = "/* Sakura.css v1.5.0\n * ================\n * Minimal css theme.\n * Project: https://github.com/oxalorg/sakura/\n */\n/* Body */\nhtml {\n  font-size: 62.5%;\n  font-family: -apple-system, BlinkMacSystemFont, \"Segoe UI\", Roboto, \"Helvetica Neue\", Arial, \"Noto Sans\", sans-serif;\n}\n\nbody {\n  font-size: 1.8rem;\n  line-height: 1.618;\n  max-width: 38em;\n  margin: auto;\n  color: #4a4a4a;\n  background-color: #f9f9f9;\n  padding: 13px;\n}\n\n@media (max-width: 684px) {\n  body {\n    font-size: 1.53rem;\n  }\n}\n@media (max-width: 382px) {\n  body {\n    font-size: 1.35rem;\n  }\n}\nh1, h2, h3, h4, h5, h6 {\n  line-height: 1.1;\n  font-family: -apple-system, BlinkMacSystemFont, \"Segoe UI\", Roboto, \"Helvetica Neue\", Arial, \"Noto Sans\", sans-serif;\n  font-weight: 700;\n  margin-top: 3rem;\n  margin-bottom: 1.5rem;\n  overflow-wrap: break-word;\n  word-wrap: break-word;\n  -ms-word-break: break-all;\n  word-break: break-word;\n}\n\nh1 {\n  font-size: 2.35em;\n}\n\nh2 {\n  font-size: 2em;\n}\n\nh3 {\n  font-size: 1.75em;\n}\n\nh4 {\n  font-size: 1.5em;\n}\n\nh5 {\n  font-size: 1.25em;\n}\n\nh6 {\n  font-size: 1em;\n}\n\np {\n  margin-top: 0px;\n  margin-bottom: 2.5rem;\n}\n\nsmall, sub, sup {\n  font-size: 75%;\n}\n\nhr {\n  border-color: #1d7484;\n}\n\na {\n  text-decoration: none;\n  color: #1d7484;\n}\na:visited {\n  color: #144f5a;\n}\na:hover {\n  color: #982c61;\n  border-bottom: 2px solid #4a4a4a;\n}\n\nul {\n  padding-left: 1.4em;\n  margin-top: 0px;\n  margin-bottom: 2.5rem;\n}\n\nli {\n  margin-bottom: 0.4em;\n}\n\nblockquote {\n  margin-left: 0px;\n  margin-right: 0px;\n  padding-left: 1em;\n  padding-top: 0.8em;\n  padding-bottom: 0.8em;\n  padding-right: 0.8em;\n  border-left: 5px solid #1d7484;\n  margin-bottom: 2.5rem;\n  background-color: #f1f1f1;\n}\n\nblockquote p {\n  margin-bottom: 0;\n}\n\nimg, video {\n  height: auto;\n  max-width: 100%;\n  margin-top: 0px;\n  margin-bottom: 2.5rem;\n}\n\n/* Pre and Code */\npre {\n  background-color: #f1f1f1;\n  display: block;\n  padding: 1em;\n  overflow-x: auto;\n  margin-top: 0px;\n  margin-bottom: 2.5rem;\n  font-size: 0.9em;\n}\n\ncode, kbd, samp {\n  font-size: 0.9em;\n  padding: 0 0.5em;\n  background-color: #f1f1f1;\n  white-space: pre-wrap;\n}\n\npre > code {\n  padding: 0;\n  background-color: transparent;\n  white-space: pre;\n  font-size: 1em;\n}\n\n/* Tables */\ntable {\n  text-align: justify;\n  width: 100%;\n  border-collapse: collapse;\n  margin-bottom: 2rem;\n}\n\ntd, th {\n  padding: 0.5em;\n  border-bottom: 1px solid #f1f1f1;\n}\n\n/* Buttons, forms and input */\ninput, textarea {\n  border: 1px solid #4a4a4a;\n}\ninput:focus, textarea:focus {\n  border: 1px solid #1d7484;\n}\n\ntextarea {\n  width: 100%;\n}\n\n.button, button, input[type=submit], input[type=reset], input[type=button], input[type=file]::file-selector-button {\n  display: inline-block;\n  padding: 5px 10px;\n  text-align: center;\n  text-decoration: none;\n  white-space: nowrap;\n  background-color: #1d7484;\n  color: #f9f9f9;\n  border-radius: 1px;\n  border: 1px solid #1d7484;\n  cursor: pointer;\n  box-sizing: border-box;\n}\n.button[disabled], button[disabled], input[type=submit][disabled], input[type=reset][disabled], input[type=button][disabled], input[type=file]::file-selector-button[disabled] {\n  cursor: default;\n  opacity: 0.5;\n}\n.button:hover, button:hover, input[type=submit]:hover, input[type=reset]:hover, input[type=button]:hover, input[type=file]::file-selector-button:hover {\n  background-color: #982c61;\n  color: #f9f9f9;\n  outline: 0;\n}\n.button:focus-visible, button:focus-visible, input[type=submit]:focus-visible, input[type=reset]:focus-visible, input[type=button]:focus-visible, input[type=file]::file-selector-button:focus-visible {\n  outline-style: solid;\n  outline-width: 2px;\n}\n\ntextarea, select, input {\n  color: #4a4a4a;\n  padding: 6px 10px; /* The 6px vertically centers text on FF, ignored by Webkit */\n  margin-bottom: 10px;\n  background-color: #f1f1f1;\n  border: 1px solid #f1f1f1;\n  border-radius: 4px;\n  box-shadow: none;\n  box-sizing: border-box;\n}\ntextarea:focus, select:focus, input:focus {\n  border: 1px solid #1d7484;\n  outline: 0;\n}\n\ninput[type=checkbox]:focus {\n  outline: 1px dotted #1d7484;\n}\n\nlabel, legend, fieldset {\n  display: block;\n  margin-bottom: 0.5rem;\n  font-weight: 600;\n}\n"
RAW_FILES['simple'] = RawFiles({"credit": "<a href='https://simplecss.org/' target='_blank'>Simple.css</a> by <code>kevquirk</code>"}, {}, {}, {})
RAW_FILES['simple'].styles['simple.min.css'] = ":root{--sans-font:-apple-system,BlinkMacSystemFont,\"Avenir Next\",Avenir,\"Nimbus Sans L\",Roboto,\"Noto Sans\",\"Segoe UI\",Arial,Helvetica,\"Helvetica Neue\",sans-serif;--mono-font:Consolas,Menlo,Monaco,\"Andale Mono\",\"Ubuntu Mono\",monospace;--standard-border-radius:5px;--bg:#fff;--accent-bg:#f5f7ff;--text:#212121;--text-light:#585858;--border:#898ea4;--accent:#0d47a1;--accent-hover:#1266e2;--accent-text:var(--bg);--code:#d81b60;--preformatted:#444;--marked:#fd3;--disabled:#efefef}@media (prefers-color-scheme:dark){:root{color-scheme:dark;--bg:#212121;--accent-bg:#2b2b2b;--text:#dcdcdc;--text-light:#ababab;--accent:#ffb300;--accent-hover:#ffe099;--accent-text:var(--bg);--code:#f06292;--preformatted:#ccc;--disabled:#111}img,video{opacity:.8}}*,:before,:after{box-sizing:border-box}textarea,select,input,progress{-webkit-appearance:none;-moz-appearance:none;appearance:none}html{font-family:var(--sans-font);scroll-behavior:smooth}body{color:var(--text);background-color:var(--bg);grid-template-columns:1fr min(45rem,90%) 1fr;margin:0;font-size:1.15rem;line-height:1.5;display:grid}body>*{grid-column:2}body>header{background-color:var(--accent-bg);border-bottom:1px solid var(--border);text-align:center;grid-column:1/-1;padding:0 .5rem 2rem}body>header>:only-child{margin-block-start:2rem}body>header h1{max-width:1200px;margin:1rem auto}body>header p{max-width:40rem;margin:1rem auto}main{padding-top:1.5rem}body>footer{color:var(--text-light);text-align:center;border-top:1px solid var(--border);margin-top:4rem;padding:2rem 1rem 1.5rem;font-size:.9rem}h1{font-size:3rem}h2{margin-top:3rem;font-size:2.6rem}h3{margin-top:3rem;font-size:2rem}h4{font-size:1.44rem}h5{font-size:1.15rem}h6{font-size:.96rem}p{margin:1.5rem 0}p,h1,h2,h3,h4,h5,h6{overflow-wrap:break-word}h1,h2,h3{line-height:1.1}@media only screen and (width<=720px){h1{font-size:2.5rem}h2{font-size:2.1rem}h3{font-size:1.75rem}h4{font-size:1.25rem}}a,a:visited{color:var(--accent)}a:hover{text-decoration:none}button,.button,a.button,input[type=submit],input[type=reset],input[type=button]{border:1px solid var(--accent);background-color:var(--accent);color:var(--accent-text);padding:.5rem .9rem;line-height:normal;text-decoration:none}.button[aria-disabled=true],input:disabled,textarea:disabled,select:disabled,button[disabled]{cursor:not-allowed;background-color:var(--disabled);border-color:var(--disabled);color:var(--text-light)}input[type=range]{padding:0}abbr[title]{cursor:help;text-decoration-line:underline;text-decoration-style:dotted}button:enabled:hover,.button:not([aria-disabled=true]):hover,input[type=submit]:enabled:hover,input[type=reset]:enabled:hover,input[type=button]:enabled:hover{background-color:var(--accent-hover);border-color:var(--accent-hover);cursor:pointer}.button:focus-visible,button:focus-visible:where(:enabled),input:enabled:focus-visible:where([type=submit],[type=reset],[type=button]){outline:2px solid var(--accent);outline-offset:1px}header>nav{padding:1rem 0 0;font-size:1rem;line-height:2}header>nav ul,header>nav ol{flex-flow:wrap;place-content:space-around center;align-items:center;margin:0;padding:0;list-style-type:none;display:flex}header>nav ul li,header>nav ol li{display:inline-block}header>nav a,header>nav a:visited{border:1px solid var(--border);border-radius:var(--standard-border-radius);color:var(--text);margin:0 .5rem 1rem;padding:.1rem 1rem;text-decoration:none;display:inline-block}header>nav a:hover,header>nav a.current,header>nav a[aria-current=page],header>nav a[aria-current=true]{border-color:var(--accent);color:var(--accent);cursor:pointer}@media only screen and (width<=720px){header>nav a{border:none;padding:0;line-height:1;text-decoration:underline}}aside,details,pre,progress{background-color:var(--accent-bg);border:1px solid var(--border);border-radius:var(--standard-border-radius);margin-bottom:1rem}aside{float:right;width:30%;margin-inline-start:15px;padding:0 15px;font-size:1rem}[dir=rtl] aside{float:left}@media only screen and (width<=720px){aside{float:none;width:100%;margin-inline-start:0}}article,fieldset,dialog{border:1px solid var(--border);border-radius:var(--standard-border-radius);margin-bottom:1rem;padding:1rem}article h2:first-child,section h2:first-child,article h3:first-child,section h3:first-child{margin-top:1rem}section{border-top:1px solid var(--border);border-bottom:1px solid var(--border);margin:3rem 0;padding:2rem 1rem}section+section,section:first-child{border-top:0;padding-top:0}section+section{margin-top:0}section:last-child{border-bottom:0;padding-bottom:0}details{padding:.7rem 1rem}summary{cursor:pointer;word-break:break-all;margin:-.7rem -1rem;padding:.7rem 1rem;font-weight:700}details[open]>summary+*{margin-top:0}details[open]>summary{margin-bottom:.5rem}details[open]>:last-child{margin-bottom:0}table{border-collapse:collapse;margin:1.5rem 0}figure>table{width:max-content;margin:0}td,th{border:1px solid var(--border);text-align:start;padding:.5rem}th{background-color:var(--accent-bg);font-weight:700}tr:nth-child(2n){background-color:var(--accent-bg)}table caption{margin-bottom:.5rem;font-weight:700}textarea,select,input,button,.button{font-size:inherit;border-radius:var(--standard-border-radius);box-shadow:none;max-width:100%;margin-bottom:.5rem;padding:.5rem;font-family:inherit;display:inline-block}textarea,select,input{color:var(--text);background-color:var(--bg);border:1px solid var(--border)}label{display:block}textarea:not([cols]){width:100%}select:not([multiple]){background-image:linear-gradient(45deg,transparent 49%,var(--text)51%),linear-gradient(135deg,var(--text)51%,transparent 49%);background-position:calc(100% - 15px),calc(100% - 10px);background-repeat:no-repeat;background-size:5px 5px,5px 5px;padding-inline-end:25px}[dir=rtl] select:not([multiple]){background-position:10px,15px}input[type=checkbox],input[type=radio]{vertical-align:middle;width:min-content;position:relative}input[type=checkbox]+label,input[type=radio]+label{display:inline-block}input[type=radio]{border-radius:100%}input[type=checkbox]:checked,input[type=radio]:checked{background-color:var(--accent)}input[type=checkbox]:checked:after{content:\" \";border-right:solid var(--bg).08em;border-bottom:solid var(--bg).08em;background-color:#0000;border-radius:0;width:.18em;height:.32em;font-size:1.8em;position:absolute;top:.05em;left:.17em;transform:rotate(45deg)}input[type=radio]:checked:after{content:\" \";background-color:var(--bg);border-radius:100%;width:.25em;height:.25em;font-size:32px;position:absolute;top:.125em;left:.125em}@media only screen and (width<=720px){textarea,select,input{width:100%}}input[type=color]{height:2.5rem;padding:.2rem}input[type=file]{border:0}hr{background:var(--border);border:none;height:1px;margin:1rem auto}mark{border-radius:var(--standard-border-radius);background-color:var(--marked);color:#000;padding:2px 5px}mark a{color:#0d47a1}img,video{border-radius:var(--standard-border-radius);max-width:100%;height:auto}figure{margin:0;display:block;overflow-x:auto}figure>img,figure>picture>img{margin-inline:auto;display:block}figcaption{text-align:center;color:var(--text-light);margin-block:1rem;font-size:.9rem}blockquote{border-inline-start:.35rem solid var(--accent);color:var(--text-light);margin-block:2rem;margin-inline:2rem 0;padding:.4rem .8rem;font-style:italic}cite{color:var(--text-light);font-size:.9rem;font-style:normal}dt{color:var(--text-light)}code,pre,pre span,kbd,samp{font-family:var(--mono-font);color:var(--code)}kbd{color:var(--preformatted);border:1px solid var(--preformatted);border-bottom:3px solid var(--preformatted);border-radius:var(--standard-border-radius);padding:.1rem .4rem}pre{max-width:100%;color:var(--preformatted);padding:1rem 1.4rem;overflow:auto}pre code{color:var(--preformatted);background:0 0;margin:0;padding:0}progress{width:100%}progress:indeterminate{background-color:var(--accent-bg)}progress::-webkit-progress-bar{border-radius:var(--standard-border-radius);background-color:var(--accent-bg)}progress::-webkit-progress-value{border-radius:var(--standard-border-radius);background-color:var(--accent)}progress::-moz-progress-bar{border-radius:var(--standard-border-radius);background-color:var(--accent);transition-property:width;transition-duration:.3s}progress:indeterminate::-moz-progress-bar{background-color:var(--accent-bg)}dialog{background-color:var(--bg);max-width:40rem;margin:auto}dialog::backdrop{background-color:var(--bg);opacity:.8}@media only screen and (width<=720px){dialog{max-width:calc(100vw - 2rem)}}sup,sub{vertical-align:baseline;position:relative}sup{top:-.4em}sub{top:.3em}.notice{background:var(--accent-bg);border:2px solid var(--border);border-radius:var(--standard-border-radius);margin:2rem 0;padding:1.5rem}@media print{@page{margin:1cm}body{display:block}body>header{background-color:unset}body>header nav,body>footer{display:none}article{border:none;padding:0}a[href^=http]:after{content:\" <\" attr(href)\">\"}abbr[title]:after{content:\" (\" attr(title)\")\"}a{text-decoration:none}p{widows:3;orphans:3}hr{border-top:1px solid var(--border)}mark{border:1px solid var(--border)}pre,table,figure,img,svg{break-inside:avoid}pre code{white-space:pre-wrap}}"
RAW_FILES['tacit'] = RawFiles({"credit": "<a href='https://yegor256.github.io/tacit/' target='_blank'>Tacit</a> by <code>yegor256</code>"}, {}, {}, {})
RAW_FILES['tacit'].styles['tacit.css'] = "input,textarea,select,button,option,body{font:normal 400 normal 1.125rem/1.85625rem system-ui,\"Helvetica Neue\",Helvetica,Arial,sans-serif}*[hidden]{display:none}th{font-weight:600}td,th{border-bottom:0 solid #595959;padding:0.928125rem 1.125rem;text-align:left;vertical-align:top;word-wrap:break-word}thead th{border-bottom-width:0.135rem;padding-bottom:0.39375rem}table{display:table;table-layout:fixed;width:100%}@media all and (max-width:1024px){table{display:none}}@media all and (max-width:1024px){table thead{display:none}}table tr{border-bottom-width:0.135rem}table tr th{border-bottom-width:0.135rem}table tr td,table tr th{overflow:hidden;padding:0.3375rem 0.225rem}@media all and (max-width:1024px){table tr td,table tr th{border:0;display:inline-block}}@media all and (max-width:1024px){table tr{display:inline-block;margin:0.675rem 0}}@media all and (max-width:1024px){table{display:inline-block}}fieldset label,fieldset legend{display:block}fieldset legend{margin:1.125rem 0}input,textarea,select,button{border-radius:3.6px;display:inline-block;padding:0.61875rem}input+label,input+input[type=\"checkbox\"],input+input[type=\"radio\"],textarea+label,textarea+input[type=\"checkbox\"],textarea+input[type=\"radio\"],select+label,select+input[type=\"checkbox\"],select+input[type=\"radio\"],button+label,button+input[type=\"checkbox\"],button+input[type=\"radio\"]{page-break-before:always}input,select,label{margin-right:0.225rem}textarea{min-height:5.625rem;min-width:22.5rem}label{display:inline-block;margin-bottom:0.7875rem}label+*{page-break-before:always}label>input{margin-bottom:0}input[type=\"submit\"],input[type=\"reset\"],button{background:#f2f2f2;color:#191919;cursor:pointer;display:inline;margin-bottom:1.125rem;margin-right:0.45rem;padding:0.4078125rem 1.4625rem;text-align:center}input[type=\"submit\"]:hover,input[type=\"reset\"]:hover,button:hover{background:#d9d9d9;color:#000}input[type=\"submit\"][disabled],input[type=\"reset\"][disabled],button[disabled]{background:#e6e5e5;color:#403f3f;cursor:not-allowed}input[type=\"submit\"],button[type=\"submit\"]{background:#275a90;color:#fff}input[type=\"submit\"]:hover,button[type=\"submit\"]:hover{background:#173454;color:#bfbfbf}input,select,textarea{margin-bottom:1.125rem}input[type=\"text\"],input[type=\"password\"],input[type=\"email\"],input[type=\"url\"],input[type=\"phone\"],input[type=\"tel\"],input[type=\"number\"],input[type=\"datetime\"],input[type=\"date\"],input[type=\"month\"],input[type=\"week\"],input[type=\"color\"],input[type=\"time\"],input[type=\"search\"],input[type=\"range\"],input[type=\"file\"],input[type=\"datetime-local\"],select,textarea{border:1px solid #595959;padding:0.3375rem 0.39375rem}input[type=\"checkbox\"],input[type=\"radio\"]{flex-grow:0;height:1.85625rem;margin-left:0;margin-right:9px;vertical-align:middle}input[type=\"checkbox\"]+label,input[type=\"radio\"]+label{page-break-before:avoid}select[multiple]{min-width:270px}pre,code,kbd,samp,var,output{font:0.9rem Menlo,Monaco,Consolas,\"Courier New\",monospace}pre{border-left:0 solid #59c072;line-height:1.575rem;overflow:auto;padding-left:18px}pre code{background:none;border:0;line-height:1.85625rem;padding:0}code,kbd{background:#daf1e0;border-radius:3.6px;color:#2a6f3b;display:inline-block;line-height:1.125rem;padding:0.225rem 0.39375rem 0.16875rem}kbd{background:#2a6f3b;color:#fff}mark{background:#ffc;padding:0 0.225rem}h1,h2,h3,h4,h5,h6{color:#000;margin-bottom:1.125rem}h1{font-size:2.25rem;font-weight:500;line-height:2.7rem;margin-top:4.5rem}h2{font-size:1.575rem;font-weight:400;line-height:2.1375rem;margin-top:3.375rem}h3{font-size:1.35rem;line-height:1.6875rem;margin-top:2.25rem}h4{font-size:1.125rem;line-height:1.4625rem;margin-top:1.125rem}h5{font-size:0.9rem;font-weight:bold;line-height:1.35rem;text-transform:uppercase}h6{color:#595959;font-size:0.9rem;font-weight:bold;line-height:1.125rem;text-transform:uppercase}a{color:#275a90;text-decoration:none}a:hover{text-decoration:underline}hr{border-bottom:1px solid #595959}figcaption,small{font-size:0.95625rem}figcaption{color:#595959}var,em,i{font-style:italic}dt,strong,b{font-weight:600}del,s{text-decoration:line-through}ins,u{text-decoration:underline}sub,sup{font-size:75%;line-height:0;position:relative;vertical-align:baseline}sup{top:-.5em}sub{bottom:-.25em}*{border:0;border-collapse:separate;border-spacing:0;box-sizing:border-box;margin:0;max-width:100%;padding:0;vertical-align:baseline}html,body{width:100%}html{height:100%}body{background:#fff;color:#1a1919;padding:36px}p,ul,ol,dl,blockquote,hr,pre,table,form,fieldset,figure,address{margin-bottom:1.85625rem}section{margin-left:auto;margin-right:auto;width:900px}aside{float:right;width:285px}article,header,footer{padding:43.2px}article{background:#fff;border:1px solid #d9d9d9}nav{text-align:center}nav ul{list-style:none;margin-left:0;text-align:center}nav ul li{display:inline-block;margin-left:9px;margin-right:9px;vertical-align:middle}nav ul li:first-child{margin-left:0}nav ul li:last-child{margin-right:0}ol,ul{margin-left:31.5px}li dl,li ol,li ul{margin-bottom:0}dl{display:inline-block}dt{padding:0 1.125rem}dd{padding:0 1.125rem 0.28125rem}dd:last-of-type{border-bottom:0 solid #595959}dd+dt{border-top:0 solid #595959;padding-top:0.5625rem}blockquote{border-left:0 solid #595959;padding:0.28125rem 1.125rem 0.28125rem 0.99rem}blockquote footer{color:#595959;font-size:0.84375rem;margin:0}blockquote p{margin-bottom:0}img{height:auto;margin:0 auto}figure img{display:block}@media (max-width:767px){body{padding:18px 0}article{border:0;padding:18px}header,footer{padding:18px}textarea,input,select{min-width:0}fieldset{min-width:0}fieldset *{flex-grow:1;page-break-before:auto}section{width:auto}x:-moz-any-link{display:table-cell}}"
RAW_FILES['none'] = RawFiles({}, {}, {}, {})



def extract(text):
    return gzip.decompress(b64decode(text)).decode()    
    
CACHED_DECOMPRESSED = {}
def get_raw_files(theme):
    if theme in CACHED_DECOMPRESSED:
        return CACHED_DECOMPRESSED[theme]
    if theme not in RAW_FILES:
        return None
    return RawFiles(
        RAW_FILES[theme].metadata,
        {k: f"<script>{v}</script>" for k, v in RAW_FILES[theme].scripts.items()},
        {k: f"<style>{v}</style>" for k, v in RAW_FILES[theme].styles.items()},
        {k: f"<script>{v}</script>" for k, v in RAW_FILES[theme].deploy.items()},
    )

def get_themes():
    return list(RAW_FILES.keys())
    